/*********************************************************************/
/**
    @file apex.h

    Header file for Apex Micro

    @author ALP

    @version $Id

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

    @par
    Apex Micro is a non-preemptive stack-per-task RTOS.  It provides
    inter-task communication via messages, synchronization and
    mutual exclusion via semaphores, signals and slots, (dynamic
    linking) and time delays via timers.  It
    also provides linked list support

    @note A "Context" section is provided in the documentation block
        for each function in the API.  Contexts are defined as follows:

        @oo Reset: Reset context is code that executes after hard reset
            and before any system calls to Apex.  It is used to prepare
            the environment for C and to set I/O ports to their default
            values immediately after reset.

        @oo Start: Start context is code that may execute in
            Apex_Target_Start(); in other words, a non-blocking system
            call.

        @oo Task: Task context is code that executes as part of a
            task, but not as part of an interrupt service routine.
            Task context occurs only after the call to Apex_Start().

        @oo Interrupt: Interrupt context is code that executes as part
            of an interrupt service routine.

**********************************************************************/
//                  CONDITIONAL COMPILE OPTIONS
//*********************************************************************

    /// Make Apex print the task number of each task it switches
    /// to from the Reschedule() call, on the debug console.

#define TP_APEX_TASK_SWITCH_ID          0

    /// Create a breakpoint when a context switch is going to
    /// be made into this task ID.  (0 to disable).

#define TP_APEX_TASK_SWITCH_TRAP_ID     0

    /// Define this as 'O_FTP0_STROBE()', for example, to strobe
    /// FTP0 when the Apex system timer ISR starts.  Any such
    /// strobe or output command may be used. Leave blank to
    /// disable.  Put the semicolon after the definition here,
    /// not where the macro is used.

#define TP_APEX_SYSTMR_ISR_START

    /// Define this as 'O_FTP0_STROBE()', for example, to strobe
    /// FTP0 when the Apex system timer ISR ends.  Any such
    /// strobe or output command may be used. Leave blank to
    /// disable.  Put the semicolon after the definition here,
    /// not where the macro is used.

#define TP_APEX_SYSTMR_ISR_END

    /// Raise FTP3 while the Apex system task executes the timer
    /// loop.

#define TP_APEX_SYSTMR_TASK             0

    /// Raise FTP2 when a task enters Apex_Pend(), and lower it
    /// when the task leaves Apex_Pend().  Can be used to measure
    /// task duty cycle on a 'scope, amoung other things.  Set to
    /// the ID number for the task to be monitored.  Set to
    /// zero to turn monitoring off.  ID numbers can be found in
    /// app.h and the task's c file.

#define VT_APEX_PEND_DURATION           0

//*********************************************************************
//                      D e f i n e s
//*********************************************************************

//      *** APEX CONFIGURATION DEFINITIONS ****

#define Z_APEX_MIN_STACK    256 ///< Minimum allowable stack size

    /** Idle Stack Size

        Define the idle stack size.  This is the same as the
        parent stack that launches Apex and is defined in
        startup.s.  This definition must match that one.
    */

#define Z_APEX_STK_IDLE     Z_APEX_MIN_STACK

#define APEX_PRI_LEVELS     8       ///< Number of priority levels

    // Apex Module ID and Assigned Numbers:

#define MID_APEX(x) (((unsigned)0xFF00)|(x))

    // The Idle and System Tick task's Task IDs:

#define TID_SYS_IDLE        UINT8_MAX           ///< Idle task ID
#define TID_SYS_TICK        (TID_SYS_IDLE-1)    ///< Timer task ID

//          *** System Timer Definitions ***

#define APEX_TICK_MS                1       // Apex system tick in mS

    // Value that specifies no timeout in Apex_Pend calls:

#define APEX_NO_TIMEOUT             0

//*********************************************************************
//                  A P E X   S t r u c t u r e s
//*********************************************************************
#include "apex_types.h"

//********************************************************************
//                  Apex Application Program Interface               *
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** Initialize Apex after Reset

    This function initializes Apex after power-up or reset.  It
    initializes internal structures that are used by
    system calls that may be made by the user before Apex_Start().

    @return Nothing.

    @context This function may be called anytime after reset and
        before Apex_Start().  It must be called before any other
        Apex system call. It may only be called once.  It should be
        called as soon as practical after reset (like from main),
        and not from task or interrupt code.  It must be called
        with interrupts disabled.  It returns with interrupts
        disabled, and does not enable interrupts internally.
*/
void Apex_Reset(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** Start Multitasking

    This function causes Apex to begin multitasking.  Apex will
    begin by running the highest priority task that has been
    created prior to this call.  It will run each such task in
    priority order and continue to do so as long as the system is
    allowed to run.  Apex_Start() does not return.

    @param priority An integer greater than or equal to zero, and less
    than APEX_PRI_LEVELS.  This value will set the priority of the
    Apex system task (the task that counts down timers and calls their
    service routines).

    @return Nothing.

    @context This function may only be called once, and must be
        called after calling Apex_Reset(). It should be called
        after all modules have been initialized, and may not be
        called from task or interrupt code.  It must be called
        with interrupts disabled.  It will enable interrupts
        before transferring control to the highest priority
        ready task.
*/
void Apex_Start(uint8_t priority);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** Create a Task.

    This function creates a new task and puts it on the ready
    list.

    @param pTask A pointer to the TCB of the task being created.

    @param id A unique identification number.  This number is not
        used by Apex per se, but it is stored in the task's TCB and
        may be used to identify the task in debug traces, etc.

    @param pStartAddr A pointer to the function that embodies the
        code for the task being created.

    @param pStack A pointer to the task's stack.

    @param stackSize An integer indicating the size of the task's
        stack in bytes.

    @param priority An integer from 0 to 7 indicating the task's
    priority.  Each task must be assigned a unique priority; that
        is, no two tasks may have the same priority.  Larger
        priority numbers indicate higher priority.

    @return Nothing.

    @context Task.  It may be called with interrupts
        enabled or disabled, and it returns with them in the same
        state as when it was called.  Interrupts are disabled within
        critical sections of this function.
*/
void Apex_CreateTask(TCB *pTask, uint16_t id, void (*pStartAddr)(void),
    void *pStack, uint16_t stackSize, uint8_t priority);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** Terminate the Current Task

    This function suspends the current task indefinately.  Control
    is passed to the highest priority task on the ready list, and
    if none, the Idle task.

    @return Nothing.
    @context Task only.  It must be called with interrupts
        enabled.
*/
void Apex_TerminateTask(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** Return a Pointer to the Current Task

    This function returns a pointer to the current task.  It is
    most useful to interrupt service routines, common library
    functions, or drivers that need to obtain a pointer to the
    TCB of the current task.

    @return A pointer to the currently executing task's TCB is
        returned.

    @context Task or Interrupt.  Interrupt enable state is not
        changed inside the function.
*/
TCB *Apex_GetCurrentTask(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** Pend for a Message

    This function allows a task to wait for a message.

    @param pBox A pointer to a MBOX structure.  The task will
        wait for the next message at this mailbox.

    @param pID A pointer to a buffer into which the message's ID will
        be returned.

    @param pData1 A pointer to a buffer into which the message's
        data1 parameter will be returned.  If this parameter is NULL,
        it will be ignored.

    @param pData2 A pointer to a buffer into which the message's
        data2 parameter will be returned.  If this parameter is NULL,
        it will be ignored.

    @param timeout A 31 bit unsigned integer specifying the duration
        of an optional timeout.  If the value of timeout is 0, the
        task will pend for a message with no timeout.  If the value
        of timeout is greater than (2^31)-1, an assert will occur
        (assuming asserts are enabled).

    @return True will be returned if a message was delivered and the
        data was written to the caller's buffers.  False will be
        returned if a timeout occurs before a message became available.

    @context Task only.  It must be called with interrupts
        enabled and will return with interrupts enabled.  Interrupts
        may be temporarily disabled around critical sections of the
        function.

    @see Apex_SendMsg.
*/
bool Apex_PendMsg(MBOX *pBox, unsigned *pID,
            uintptr_t *pData1, uintptr_t *pData2, int32_t timeout);

//lint -esym(534, Apex_PendMsg) It's ok to ignore the return value.

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** Initialize a Mailbox

    This function initializes a mailbox structure prior to use.

    @param pBox: A pointer to the mailbox to be initialized.

    @return Nothing.

    @context Task, or interrupt.  It may be called with
        interrupts enabled or disabled, and it returns with the
        interrupt enable state unchanged.

    @see Apex_SendMsg, Apex_PendMsg.
*/
void Apex_InitMailbox(MBOX *pBox);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** Inquire how many messages are waiting at a mailbox.

    This function tells the caller if there are messages awaiting
    delivery at a mailbox.

    @param pBox: A pointer to the mailbox to be inquired.

    @return true if a message is available at the mailbox, false
        otherwise.

    @context Task or interrupt.

    @see Apex_SendMsg, Apex_PendMsg.
*/
bool Apex_InqMsg(const MBOX *pBox);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Send a message.

    This function may be called from a task or interrupt service
    routine to send a message to a mailbox.

    @param pBox: A pointer to the mailbox that should receive the
        message.

    @param id: A user defined integer identifying the type of message.

    @return Nothing

    @context Task or interrupt.

    @see Apex_InitMsg, Apex_SendMsgExt, Apex_SendMsgAbbr, Apex_PendMsg.
*/
void Apex_SendMsg(const MBOX *pBox, unsigned id);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Send a message with extended parameters.

    This function may be called from a task or interrupt service
    routine to send a message to a mailbox.

    @param pBox: A pointer to the mailbox that should receive the
        message.

    @param id: A user defined integer identifying the type of message.

    @param data1: A pointer sized integer which may contain any
        data the user wishes to send.

    @param data2: A pointer sized integer which may contain any
        data the user wishes to send.

    @return Nothing

    @context Task or interrupt.

    @see Apex_InitMsg, Apex_SendMsg, Apex_PendMsg, Apex_SendMsgAbbr.
*/
void Apex_SendMsgExt(const MBOX *pBox,
                unsigned id, uintptr_t data1, uintptr_t data2);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                 Semaphore Related System Calls
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** Initialize a semaphore before first use.

    This function initializes a user allocated semaphore structure.
    Initialization must be performed before the semaphore is used
    for the first time.

    @param[in] pSema: A pointer to the semaphore to be initialized.
    @param count: An unsigned integer specifying the initial value
        of the semaphore's counter.
    @param limit: An unsigned integer specifying the upper limit
        of the semaphore's counter.
    @return Nothing.
    @context Task, or Interrupt.
    @see Apex_InqSema, Apex_PostSema, Apex_PostSema_ISR, Apex_Pend,

*/
void Apex_InitSema(SEMA *pSema, int8_t count, int8_t limit);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Inquire of a semaphore's count.

    This function may be called to obtain the current value of a
    semaphore's counter.

    @param pSema: A pointer to the semaphore, the counter of which
        shall be read.

    @return The unsigned integer contents of the semaphore's counter.
        The count is undefined before the semaphore is initialized with
        a call to Apex_InitSema().
    @context Task or Interrupt.
    @see Apex_InitSema, Apex_PostSema, Apex_PostSema_ISR,
        Apex_Pend,

*/
int8_t Apex_InqSema(const SEMA *pSema);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Post a Semaphore

    This function increments the semaphore's counter and frees a task
    (if one is waiting for control of the semaphore).

    @param pSema: A pointer to the semaphore to post.

    @return Nothing

    @context Task or interrupt.

    @note If the counter is already at the limit when this call occurs,
    the count is not incremented but a task is still freed.  This
    enables the use of a count and limit of 0, effectively creating a
    non-counting semaphore.

    @see Apex_InqSema, Apex_InitSema, Apex_PostSema_ISR,
        Apex_Pend,
*/
void Apex_PostSema(SEMA *pSema);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Pend for Semaphore

    This function allows a task to wait for a semaphore.

    @param pSema A pointer to a SEMA structure.

    @param timeout A 31 bit unsigned integer specifying the duration
        of an optional timeout.  If the value of timeout is 0, the
        task will pend for a message with no timeout.  If the value
        of timeout is greater than (2^31)-1, an assert will occur
        (assuming asserts are enabled).

    @return 'true' is returned if the semaphore is available.  'false'
        is returned if a timeout occured.

    @context Task only.  It must be called with interrupts
        enabled.
*/
bool Apex_PendSema(SEMA *pSema, int32_t timeout);
//lint -esym(534, Apex_PendSema) Don't always need to check return value

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Delay the current task

    This function blocks the current task for some fixed period of
    time, or performs a priority inversion check if the duration
    of the delay is specified as zero.

    @param duration: A signed integer number of Apex_Tick units for
        which the current task shall be blocked.  If duration is
        zero, Apex_Delay() simply checks to see if there are other
        tasks of equal or greater priority, and if so, reschedules
        the current task so the other tasks can execute first.
        If no other tasks of equal or greater priority are waiting,
        it returns back to the current task immediately.  This is
        in effect a priority inversion check.

    @return Nothing

    @context Task only.

    @see Apex_GetSystemTime, Apex_StartTmr, Apex_StopTmr, Apex_Tick
    Apex_InitCallbackTimer
*/
void Apex_Delay(int32_t duration);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Initialize a timer to send a message on timeout.

    This function initializes a user allocated TMR structure before
    its first use.  The timer will send a message to a mailbox upon
    timeout.  This function does not start the timer, it only initializes
    the timer's structure.

    @param pTmr: A pointer to the timer to be initialized.

    @param pBox: Pointer to the mailbox to which the timeout message
        should be sent upon timeout.

    @param id: The id of the message to send when the timer times out.

    @param data1: A pointer sized integer which may contain any
        data the user wishes to send.

    @param data2: A pointer sized integer which may contain any
        data the user wishes to send.

    @return Nothing

    @context Task, or Interrupt.

    @see Apex_GetSystemTime, Apex_StartTmr, Apex_StopTmr, Apex_Tick
    Apex_InitCallbackTimer
*/
void Apex_InitMsgTimer(TMR *pTmr, MBOX *pBox,
                    unsigned id, uintptr_t data1, uintptr_t data2);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Initialize a timer to call a user function on timeout.

    This function initializes a user allocated TMR structure before
    its first use.  The timer will call a user function when it times
    out.  This function does not start the timer, it only initializes
    the timer's structure.

    @param pTmr: A pointer to the timer to be initialized.

    @param pCallbackServiceFunc: Pointer to the user function to be
        called when the timer times out.

    @return Nothing

    @context Task, or Interrupt (although it's not clear
    why one would want to call this function from an ISR since
    Apex_StartTmr() may only be called from a task).

    @see Apex_GetSystemTime, Apex_StartTmr, Apex_StopTmr, Apex_Tick
    Apex_InitMessageTimer
*/
void Apex_InitCallbackTimer(
                TMR *pTmr, void (*pCallbackServiceFunc)(void));

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Get a timer's state.

    This function returns the state (running, stopped, or not initialized)
    of a timer.

    @return An enum indicating the state of the timer.  See the timer
    structure for a definition of this enum.

    @see Apex_StartTmr, Apex_StopTmr, Apex_InitCallbackTimer,
    Apex_InitMessageTimer
*/
enum TAG_TMR_STATE Apex_GetTimerState(const TMR *pTmr);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Read contents of system clock.

    Read the current value of the system clock.  The system clock is
    a 32 bit unsigned integer that is initialized to 0 by Apex_Start()
    and is incremented each time Apex_Tick() is called.

    @return A 32 bit unsigned integer containing the current value
        of the system timer.  (The value is undefined until after
        Apex_Start() is called).

    @context Task or Interrupt.

    @see Apex_InitMessageTimer, Apex_StartTmr, Apex_StopTmr, Apex_Tick
    Apex_InitCallbackTimer
*/
uint32_t Apex_GetSystemTime(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Start a user defined timer.

    This function starts a user defined timer. Space for the timer must
    be allocated by the user and previously initialized with a call to
    Apex_InitTimer().  Once started, the timer will count
    down once each time Apex_Tick() is called.  When the count reaches
    zero, a message with the ID initialized into the timer by a
    previous call to Apex_InitTimer() will be sent to the task specified
    in the Apex_InitTimer() call, or the user-designated function will
    be called.

    @param pTmr: A pointer to the timer to set.

    @param duration: A signed integer number of Apex_Tick units to
        wait until the event is asserted.

    @return Nothing

    @context Task only.  It must be called with interrupts
        enabled.

    @note The value of the duration parameter is signed, however it
        may not contain a zero or negative number, or Apex_StartTmr()
        will assert.

    @see Apex_InitMessageTimer, Apex_GetSystemTime, Apex_StopTmr,
    Apex_Tick, Apex_InitCallbackTimer
*/
void Apex_StartTmr(TMR *pTmr, int32_t duration);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Stop a user defined timer.

    This function stops a user defined timer.  Once a timer has been
    intialized, it is safe to call Apex_StopTmr() even if the timer
    has not yet been started or is not running.

    If the timer has already timed out and sent a timeout message, the
    message is deleted from the task's queue if it has not already been
    delivered.

    @param pTmr: A pointer to the timer to stop.

    @return Nothing

    @context Task only.  It must be called with interrupts
        enabled.

    @see Apex_InitMessageTimer, Apex_GetSystemTime, Apex_StartTmr,
    Apex_Tick Apex_InitCallbackTimer
*/
void Apex_StopTmr(TMR *pTmr);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Increment System Clock

    This function is called by a user provided function to increment
    the system clock and count down all the timers previously set
    using Apex_StartTmr().  Normally the user would call this function
    from an interrupt service routine associated with a hardware-based
    timer (although it could be called by a user-provided task).

    @return Nothing.

    @context Interrupt or Task.  This function should be called with
        interrupts disabled, and it returns with interrupts disabled.
*/
void Apex_Tick(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//              Functions Used to Create Linked Lists
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Initialize a List Link

    This function is called by the user to initialize a list anchor
    or list item before use.  It initializes the forward and backward
    pointers in the LINK structure, as well as the object pointer.
    It is used to initialize LINK structures to be used in either
    singly or doubly linked lists

    @param pLink: A pointer to the link to be initialized.

    @param pObject: A pointer to the object referenced by the link.

    @return Nothing.

    @context Any.  Does not disable or enable interrupts.

    @note The symbol APEX_INIT_LINK(x,y) is an inline synonym for
        this function.  Neither this function nor its inline synonym
    perform any type of mutual exclusion (such as disable/enable
    interrupts) so the caller should do so around the call if
    required.

    @see Apex_Link, Apex_Unlink, Apex_Push, Apex_Pop
*/

void Apex_InitLink(LINK *pLink, void *pObject);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Link an Item to a Doubly-Linked List

    This function is called to attach an object to a doubly-linked
    list.  The object must have been previously initialized using
    Apex_InitLink.

    @param pAnchor: A pointer to a LINK structure to which the object
        is to be linked.  This may either be a statically allocated
        list anchor or it may be a member of a list itself.

    @param pLink: A pointer to the LINK structure that references the
        object.  Normally this structure is contained in the object
        itself, but it doesn't have to be.

    @return Nothing.

    @context Any.  Does not disable or enable interrupts.

    @note The object referenced by pLink shall always be linked
    into the list after the object pointed to by pAnchor.  This
    means that in order to link an object to the tail of the list
    set pAnchor to the last item in the list.  To link the object to
    the head of the list, set pAnchor to the address of the anchor
    structure itself.

    @par
    To link the object to a peer object inside a doubly linked list,
    make pAnchor point to the LINK structure referencing the peer
    object.

    @par
    The symbol APEX_LINK(x,y) is an inline synonym for
    this function.  Neither this function nor its inline synonym
    perform any type of mutual exclusion (such as disable/enable
    interrupts) so the caller should do so around the call if
    required.

    @warning Operating on a list using Apex_Push/Pop and
    Apex_Link/Unlink is undefined and will probably cause a
    crash. Once a list has been established as either singly or
    doubly linked, it is not possible to operate on it using the
    functions designed for the other list type.

    @see Apex_InitLink, Apex_Unlink, Apex_Push, Apex_Pop
*/

void Apex_Link(LINK *pAnchor, LINK *pLink);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Remove an Item from a Doubly-Linked List

    This function is called to remove an object from a doubly linked
    list.  The object must have been previously linked into a list
    using Apex_Link.

    @param pLink: A pointer to the LINK structure that references the
        object.  Normally this structure is contained in the object
        itself, but it doesn't have to be.

    @return Nothing.

    @context Any.  Does not disable or enable interrupts.

    @note The symbol APEX_UNLINK(x) is an inline synonym for
    this function.  Neither this function nor its inline synonym
    perform any type of mutual exclusion (such as disable/enable
    interrupts) so the caller should do so around the call if
    required.

    @warning Operating on a list using Apex_Push/Pop and
    Apex_Link/Unlink is undefined and will probably cause a
    crash. Once a list has been established as either singly or
    doubly linked, it is not possible to operate on it using the
    functions designed for the other list type.

    @see Apex_InitLink, Apex_Link, Apex_Push, Apex_Pop
*/

void Apex_Unlink(const LINK *pLink);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Push Object to the Top of a Singly-Linked List

    This function is called to link an object to the beginning of
    a singlly linked LIFO list.  The object must have been previously
    initialized using Apex_InitLink.

    @param pAnchor: A pointer to a LINK structure representing the
        list (anchor) to which the object shall be linked.  (List
        anchors are usually a statically allocated but don't
        necessarilly need to be).

    @param pLink: A pointer to the LINK structure that references the
        object to be pushed onto the list.  Normally this structure is
        contained in the object itself, but it doesn't necessarilly
        need to be.

    @return Nothing.

    @context Any.  Does not disable or enable interrupts.

    @note This function is somewhat faster than Apex_Link because
    it only deals with a singly linked LIFO list and there are fewer
    pointers to manipulate when linking and unlinking.

    @par
    The symbol APEX_PUSH(x,y) is an inline synonym for
    this function.  Neither this function nor its inline synonym
    perform any type of mutual exclusion (such as disable/enable
    interrupts) so the caller should do so around the call if
    required.

    @warning Operating on a list using Apex_Push/Pop and
    Apex_Link/Unlink is undefined and will probably cause a
    crash. Once a list has been established as either singly or
    doubly linked, it is not possible to operate on it using the
    functions designed for the other list type.

    @see Apex_InitLink, Apex_Link, Apex_Unlink, Apex_Pop

    @todo Set up the inline version of this in apex.c like it is
    in Apex_Link etc.
*/

void Apex_Push(LINK *pAnchor, LINK *pLink);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Pop an Object from the Top of a Singly-Linked List

    This function is called to unlink an object from the top of a
    a singlly linked LIFO list.  The object must have been previously
    initialized using Apex_InitLink and pushed on to the list using
    Apex_Push.

    @param pAnchor: A pointer to a LINK structure representing the
        list (anchor) from which the object shall be unlinked.  (List
        anchors are usually a statically allocated but don't
        necessarilly need to be).

    @return A pointer to the object that was popped from the list
        is returned.

    @context Any.  Does not disable or enable interrupts.

    @note This function is somewhat faster than Apex_Unlink because
    it only deals with a singly linked LIFO list and there are fewer
    pointers to manipulate when linking and unlinking.

    @par
    The symbol APEX_POP(x) is an inline synonym for
    this function.  Neither this function nor its inline synonym
    perform any type of mutual exclusion (such as disable/enable
    interrupts) so the caller should do so around the call if
    required.

    @warning Operating on a list using Apex_Push/Pop and
    Apex_Link/Unlink is undefined and will probably cause a
    crash. Once a list has been established as either singly or
    doubly linked, it is not possible to operate on it using the
    functions designed for the other list type.

    @see Apex_InitLink, Apex_Link, Apex_Unlink, Apex_Push

    @todo Set up the inline version of this in apex.c like it is
    in Apex_Unlink etc.

*/

void *Apex_Pop(LINK *pAnchor);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                  Extension Hook Routines:
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    //  The typedefs make it a little easier to read.

typedef void APEX_EXT_CSWITCH_FUNC(TCB *);
typedef void APEX_EXT_CREATE_FUNC(TCB *);
typedef void APEX_EXT_IDLE_FUNC(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Register Context-Switch Callback

    The user calls this function to register a user-provided function
    to be called each time Apex performs a context switch.

    When the callback function is called it is passed a pointer
    to the TCB of the task being switched to.

    @param pFunc: A pointer to the user's callback function.

    @return A pointer to the function previously registered by this
        call.  The user should call this function as part of the
        registered callback function passed as pFunc above.

    @context Task, or Interrupt.

    @todo Specify interrupt enable state in force during callback
        execution.
*/

APEX_EXT_CSWITCH_FUNC
    *Apex_Hook_SwitchContexts(APEX_EXT_CSWITCH_FUNC *pFunc);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Register Apex Task Creation Callback

    The user calls this function to register a user-provided function
    to be called immediately before Apex_Target_CreateTask during
    the task creation process.  Since Apex_Target_CreateTask
    sets up the task's stack, the user's callback function shall
    see the original stack pointer and length.  This facilitates
    user designed stack checking, which is one of the many reasons
    to register for this callback.

    When the callback function is called, it will be passed a pointer
    to the TCB of the task being created.

    The callback function shall be called with interrupts disabled
        and must return with them disabled.

    @param pFunc: A pointer to the user's callback function.

    @return A pointer to the function previously registered by this
        call.  The user should call this function as part of the
        registered callback function passed as pFunc above.

    @context Task, or Interrupt.
*/

APEX_EXT_CREATE_FUNC
    *Apex_Hook_CreateTask(APEX_EXT_CREATE_FUNC *pFunc);


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Register Idle Loop Callback

    The user calls this function to register a user-provided function
    to be called from the Idle Task each time the idle loop executes.

    @param pFunc: A pointer to the user's callback function.

    @return A pointer to the function previously registered by this
        call.  The user should call this function as part of the
        registered callback function passed as pFunc above.

    The callback function shall be called with interrupts enabled
        and must return with them enabled.

    @context Task, or Interrupt.
*/

APEX_EXT_IDLE_FUNC
    *Apex_Hook_IdleTask(APEX_EXT_IDLE_FUNC *pFunc);


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                  Apex Calls Supplied by the Target Module:
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/// @todo Specify interrupt enable requirements for these functions.

/** Start the Target

    The user supplies this function to be called by Apex from
    Apex_Start.  It should be used to initialize any portion of the
    user's code that must be initialized prior to the start of
    multitasking.  Global interrupts should be enabled by this
    function before returning.  This call is made immediately before
    the callback registered via Apex_Hook_Start.

    @return Nothing.

    @context No potentially blocking system calls may be made from
        this function.
*/
void Apex_Target_Start(void);


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Initialize a Task's Stack

    The user supplies this function to be called by Apex from
    Apex_CreateTask.  It should be used to initialize the new
    task's stack.

    @param pTask: A pointer to the new task's TCB structure.

    @param pStartAddr: A pointer to the starting address of the new
        task's executable code.

    @param stackSize: An integer containing the number of bytes
        allocated to the new task's stack.

    @return Nothing.
*/
void Apex_Target_CreateTask(
    TCB *pTask, void (*pStartAddr)(void), uint16_t stackSize);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Start the System Timer

    This function shall be called by Apex from the system timer
    task immediately after it starts.  The function should
    initialize and start the interrupt or task that calls
    Apex_Tick.

    @return Nothing.
*/
void Apex_Target_StartSystemTimer(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Enable System Timer Interrupts

    This function shall be called by Apex from the system timer
    task when it is necessary to enable the system
    timer interrupt.

    @return Nothing.
*/
void Apex_Target_SystemTimer_EI(void);  // Unnested

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Disable System Timer Interrupts

    This function shall be called by Apex from the system timer
    task when it is necessary to disable the system
    timer interrupt.

    @return Nothing.
*/
void Apex_Target_SystemTimer_DI(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Switch Task Contexts

    This function is called by Apex when the scheduler determines
    that it is necessary to switch tasks.

    @param pOldContext: A pointer to the saved stack pointer in the
        TCB of the task to switch from.

    @param pNewContext: A pointer to the saved stack pointer in the
        TCB of the task to switch to.

    @return Nothing.
*/
void Apex_Target_SwitchContexts(
    void **pOldContext, void **pNewContext);// __attribute__ ((naked));

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

extern const struct TAG_VARIABLE_TABLE apexVarTab[];

#define APEX_H                 // Say we included apex.h

