#ifndef APEX_TYPES_H
#define APEX_TYPES_H
/*********************************************************************/
/**
    @file apex_types.h

    Header file for Apex Micro type definitions

    @author ALP

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

*/
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Module ID Numbers:

        Each module that defines exceptions, tasks, etc,
        must assign ID numbers to those objects.  A convinent way
        to do this is to assign a "module number" to each module
        and derive the object's ID number from that module number.

        The APEX kernel and all APEX library modules use this
        strategy to obtain unique ID numbers for their objects.
        Each module is assigned a 16 bit module ID number, in
        which the low order 8 bits are 0x00, and are reserved
        for use within the module.  Thus, the APEX kernel has
        module number 0xFF00, and may use numbers between 0xFF00
        and 0xFFFF for object IDs.  The dynamic memory manager
        (apexaloc.c) uses module number 0xFE00, and object ID
        numbers between 0xFE00 and 0xFEFF.

        In all, APEX reserves numbers between 0xC0000 and 0xFFFF
        for its own object ID numbers.  If users choose to adopt
        this system also, they may use numbers between 0x0000
        and 0xBFFF for object IDs without conflicting with APEX
        or it's extension modules.

        APEX module ID numbers begin with the prefix "MID_"  The
        individual modules define their ID numbers in their
        respective "*.h" files.

        The MODULE macro extracs the module id's from
        exception codes, message numbers, etc.
    */

    /**
        The MODULE macro extracs the module id's from
        exception codes, message numbers, etc.
    */

#define MODULE(x)       (((x)<<8) & 0xFF00)

//*********************************************************************
//                  A P E X   S t r u c t u r e s
//*********************************************************************

/** @struct TAG_LINK

    The LINK structure is used to create linked lists of generic
    objects.

    @note This structure is typedef'ed as LINK.

    @see Apex
*/

struct TAG_LINK
{
    struct TAG_LINK *pFLink;    ///< Forward pointer to next link
    struct TAG_LINK *pBLink;    ///< Reverse pointer to prev. link
    void *pObject;              ///< A pointer to the linked object.
};

typedef struct TAG_LINK LINK;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** @struct TAG_SEMA

    The TAG_SEMA structure is used to create semaphores.

    @note This structure is typedef'ed as SEMA.
*/

struct TAG_SEMA
{
    int8_t limit;          ///< Count limit
    int8_t count;          ///< Semaphore's count.
    LINK taskList;         ///< List of tasks waiting on semaphore.
};

typedef struct TAG_SEMA SEMA;

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

typedef struct TAG_SEMA MUTEX;

/** Initialize a Mutex
*/
#define APEX_INIT_MUTEX_UNLOCKED(x) Apex_InitSema((x),1,1)
#define APEX_INIT_MUTEX_LOCKED(x)   Apex_InitSema((x),0,1)

/** Lock a Mutex
 * @param m The address of the mutex to lock.
 * @param t A timeout time in system ticks, or 0 if no timeout desired.
*/
#define APEX_LOCK_MUTEX(m,t)  Apex_PendSema((m), (t))

/** Unlock a Mutex
*/
#define APEX_UNLOCK_MUTEX(m)    Apex_PostSema((m))


    /// A semaphore to perform busy/done logic.
typedef struct TAG_SEMA BUSY_SEMA;

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Initialize busy/done semaphore to 'busy':
#define APEX_INIT_BUSY_SEMA(x) Apex_InitSema((x),0,1)

    /// Make the caller wait until the busy semaphore is
    /// marked 'done'.
    /// @param m The address of the busy/done semaphore.
    /// @param t The amount of time to wait.  A value of
    ///     zero waits forever.

#define APEX_WAIT_BUSY(m,t)   Apex_PendSema((m),(t))

    /// Mark the semaphore as done. If there is a task pending
    /// for the semaphore, make that task ready.

#define APEX_MARK_DONE(x)     Apex_PostSema((x))


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** @struct TAG_MSG

    The MSG structure is used to transmit small amounts of data
    from one task to another.  Messages contain a pointer sized
    data object and may also be used to send arrays or structures
    via use of this pointer field.

    Space for messages must be allocated by the user and
    initialized prior to use if the Apex configuration parameter
    APEX_ALLOCATED_MSGS is false. See Apex_InitMsg.

    @note This structure is typedef'ed as MSG.
*/
struct TAG_MSG
{
    LINK link;          ///< Linked list hook. Links to other messages

    // MSS_ stands for "Message State".  Useful in debug.

    enum TAG_MSG_STATE
    {
        MSS_INVALID = 0,    ///< The value 0 is invalid for state.
        MSS_INACTIVE,       ///< Message not in use.
        MSS_SENT_STATIC,    ///< Statically allocated awaiting delivery
        MSS_SENT_DYNAMIC,   ///< Dynamically allocated awaiting delivery
        MSS_DELIVERED       ///< Delivered message.

    } state;                ///< The message's state, .

    unsigned id;           ///< Message's identifier.

    uintptr_t data1;        ///< Pointer sized user data 1
    uintptr_t data2;        ///< Pointer sized user data 2
};

typedef struct TAG_MSG MSG; ///< Message type

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** @struct TAG_MBOX

    Mailboxes are used as destinations for messages.  A task may
    have as many mailboxes as necessary.  Each mailbox contains
    a list of messages awaiting delivery.  Messages are delivered
    from a mailbox in FIFO order.

    Space for mailboxes must be allocated by the user.  They may
    be allocated statically or dynamically.  Mailboxes must be
    initialized prior to use.  See Apex_InitMailbox.

    @note This structure is typedef'ed as MBOX.
*/

struct TAG_MBOX
{
    LINK msgList;               ///< List of messages awaiting delivery
    LINK taskList;              ///< List of tasks awaiting messages
};

typedef struct TAG_MBOX MBOX;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** @struct TAG_TMR

    The timer structure is used in conjunction with the system
    calls Apex_StartTmr and Apex_StopTmr to create timers that
    assert events upont timeout.

    Space for timers must be allocated by the user.  They may
    be allocated statically or dynamically.  Timers must be
    initialized prior to use.  See Apex_InitTimer.

    @note This structure is typedef'ed as TMR.
*/

struct TAG_TMR
{
    LINK link;                  ///< Linked list hook
    uint32_t expireTime;        ///< Time at which timer expires

    MSG msg;                   ///< Message to send upon timeout.
    void *pObject;            ///< Object pointer to use upon timeout

    // TMT_ prefix stands for Timer Type:

    enum TAG_TMR_TYPE
    {
        TMT_INVALID = 0,    ///< The value 0 is invalid for type.
        TMT_MESSAGE,        ///< Timer sends message upon timeout.
        TMT_TIMEOUT,        ///< Timer is a task timeout.
        TMT_CALLBACK        ///< Timer calls user function upon timeout.
    } type;                 ///< Timer's type.

    // The TMS_ prefix stands for Timer State:

    enum TAG_TMR_STATE
    {
        TMS_INVALID = 0,    ///< The value 0 is invalid for state.
        TMS_INACTIVE,       ///< Timer expired or cancled
        TMS_COUNTING        ///< Timer is actively counting.
    } state;                    ///< Timer's state.

    // User Defined Timer Fields

    void *pXTmr;        ///< Pointer to timer extension (user defined)
};

typedef struct TAG_TMR TMR;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** @struct TAG_TCB

    The TCB structure represents a task that may execute
    under the control of Apex.

    Space for TCBs must be allocated by the user.  They may
    be allocated statically or dynamically.  TCBs must be
    initialized prior to use.  See Apex_CreateTask.

    @note This structure is typedef'ed as TCB.
*/

struct TAG_TCB
{
    LINK activeLink;        ///< Link to list of active tasks.
    LINK readyLink;     ///< Link to list of ready tasks

    void *pStack;           ///< Stack pointer
    void *pBlock;           ///< Pointer to object that blocks the task

    uint16_t stackSize; ///< Length of stack in bytes
    uint16_t id;            ///< Task ID number.

    // The THS_ prefix stands for Task State:

    enum TAG_TCB_STATE
    {
        THS_INVALID = 0,    ///< The value 0 is invalid for state.
        THS_INIT,           ///< Task is new. Has not run first time
        THS_READY,          ///< Task is ready due to message or sema.
        THS_EXEC,           ///< Task is currently executing.
        THS_BLOCKED,        ///< Task has been blocked.
        THS_TERMINATED      ///< Task has been terminated.
    } state;                    ///< Task's state

    uint8_t priority;       ///< Task's priority (larger = higher priority)
    uint8_t priorityBit;    ///< Bit mask of task's priority

    TMR timer;              ///< Pend timeout timer
    void *pXTCB;            ///< Pointer to user extension of TCB
};

typedef struct TAG_TCB TCB;

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Slot Type Definition
    */
typedef void slot_t(uintptr_t slotParm1, uintptr_t slotParm2);

slot_t Slot_Null;

    /** Declare existence of a module's signals.
     *
     * This macro makes a signal visible to other
     * modules.  One declaration should be used for each
     * signal.  The declarations should appear in the h
     * file.  No semicolon is required at the end of
     * the declaration.
     *
     * @param name The name of the signal to declare.
     */

#define DECLARE_SIGNAL(name) /**/ \
    void Subscribe_ ## name(slot_t *pSlot); \
    void Unsubscribe_ ## name(void);

    /** Define amodule's signals.
     *
     * This macro creates space and defines subscribe and
     * unsubscribe functions for each signal.  There should
     * be one definition for each of the module's signals.
     * The definitions should appear in the c file.
     *
     * @param name The name of the signal to declare.
     */
#define DEFINE_SIGNAL(name) /**/ \
    static slot_t *name; \
    void Subscribe_ ## name(slot_t *pSlot){(name = pSlot);} \
    void Unsubscribe_ ## name(void){(name = Slot_Null);}

    /** This function actually issues a signal.
     *
     * @param name The name of the signal to issue.
     * @param parm1 A user defined parameter to be passed to
     * the slot subscribed to this signal.
     * @param parm2 A user defined parameter to be passed to
     * the slot subscribed to this signal.
     */
#define EMIT_SIGNAL(name, parm1, parm2)   /**/ \
    (* name)((parm1), (parm2))

#endif

//***************************************************************
//                        End of File
//***************************************************************
