/*********************************************************************/
/**
	@file apmath.h

	API file for AP Math Library

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

	@defgroup apmath Math Library
*/

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//					PSUDO-RANDOM NUMBER FUNCTIONS
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void APML_SeedPRNG(uint32_t seed);

uint32_t APML_UniformPRNG(void);

int32_t APML_GaussianPRNG(void);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//							POWER FUNCTIONS
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

uint32_t APML_Sqrt(uint32_t num);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//							LOG FUNCTIONS
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** Compute the integer part of the log-base-2 of a 32 bit number.

	This function counts the order of the highest order bit in
	a 32 bit double word.

	@param num: An unsigned 32 bit number to be evaluated.
	@return Returns the order of the highest order bit in num.  For
		example, if num is 0x19F30A72, 28 is returned.
*/
uint8_t APML_IntLog2DWord(unsigned num);


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** Compute the integer part of the log-base-2 of a 16 bit number.

	This function counts the order of the highest order bit in
	a 16 bit word.

	@param num: An unsigned 16 bit number to be evaluated.
	@return Returns the order of the highest order bit in num.  For
		example, if num is 0x19F3, 12 is returned.
*/
uint8_t APML_IntLog2Word(uint16_t num);


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** Compute the integer part of the log-base-2 of an 8 bit number.

	This function counts the order of the highest order bit in
	a 8 bit byte.

	@param num: An unsigned 8 bit number to be evaluated.
	@return Returns the order of the highest order bit in num.  For
		example, if num is 0x19, 4 is returned.
*/
uint8_t APML_IntLog2Byte(uint8_t num);


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** Compute the integer and fractional part of the log-base-2
	of a 32 bit number.

	This function evaluates a 32 bit double-word and returns the
	log-base-2 shifted up by a scaling factor to provide fractional
	precision.

	@param num: An unsigned 32 bit number to be evaluated.

	@return Returns the log-base-2 of num, upshifted by E2_LOG2.
		For example, if num is 0x19F3, %% is returned.  IF zero
		is submitted for num, -MIN_INT is returned.

@todo Fill in the %% above.
*/
int32_t APML_Log2(uint32_t num);


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** Log-base-2 Fractional Scaling Factor

	This definition defines the number of bits that APLog2()
	and DB_2_LOG2() upshift their results by in order to
	retain fractional precision.
*/

#define E2_LOG2		16		/// APLog2 is upshifted this much.


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** Convert power dB constant into Log-base-2 equivilent constant.

	This is an in-line macro designed to convert a pre-processor
	constant expressed in units of power dB into units of log-base-2.
	This is done by dividing the dB constant by 3.01, upshifting
	by E2_LOG2 to retain fractional precision, and then truncating.

	This function only works for pre-processor constants because it
	uses floating point arithmetic.  If a variable or run-time const
	are used, it wouldn't give the correct result.

	@param x: An integer value in units of power dB. x must not be
		larger than 2^(31-E_LOG2+2), or the result of the unit
		conversion will overflow 32 bits.
	@return Returns the number in units of log-base-2 upshifted by
		E2_LOG2.
*/

#define DB_2_LOG2(x)	(int32_t)(((x)*((float)(1<<E2_LOG2))/3.01)+0.5)

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//							TRIGONOMETRIC FUNCTIONS
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

int16_t APML_Sine(uint16_t theta);
int16_t APML_Cosine(uint16_t theta);

/****************************************************************
									END
****************************************************************/
