//**************************************************************
/**

    @file apprint.h

    API for a General Purpose Formatted Printting Function.
    (Developed in the context of the PSE-702 project)

    @author ALP

    @version Version 1.0

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

****************************************************************/

/** APPrint:

    @param pStream: A pointer to a function that accepts a character.
        This function will be called by APPrint to process each
        character created as a result of the format specification
        and the parameters.

    @param pF (Field Specification String): A null terminated string
        containing zero or more field specifications.  See below
        for formatting rules.

    @param ... (Parameter List): zero or more char, signed or unsigned
        int, or pointer parameters, one for each field
        specification contained in the pF string.

    @par Field Specification General Format:

    %{width}{dec}{type}

    @par where:
        @oo % is the literal character "%" and signifies the start
            of a field specification.

        @oo {width} is optional, and consists of a number from 1 to
            12 specifying the width of the field in characters.

        @oo {dec} is optional, and consists of the character "."
            followed by a digit from 1 to 8 specifying the number
            of digits to the right of the decimal point.

        @oo {type} is one of the symbols from the following section.

    @par {type} Specifications:

            @oo "c": Signifies that the field consists of a single
                character, which is contained in the next parameter.
                {width} and {dec} must not be present.

            @oo "s": Signifies that the field consists of a null
                terminated string.  A pointer to the string is
                contained in the next parameter.  {width} and {dec}
                must not be present.

            @oo "u": Signifies that the field consists of an unsigned
                integer, the value of which is contained in the
                parameter list.  {dec} must not be present.
                {width} may optionally be present.  If not present or
                if {width} is smaller than the supplied parameter,
                the field will be as long as necessary to print the
                value contained in the parameter (including its
                sign).

            @oo "d": Signifies that the field consists of a signed
                integer, the value of which is contained in the
                parameter list.  {dec} must not be present.
                {width} may optionally be present.  If not present or
                if {width} is smaller than the supplied parameter,
                the field will be as long as necessary to print the
                value contained in the parameter (including its
                sign).

            @oo "f": Signifies that the field consists of a signed
                decimal number, the value of which is contained in
                the parameter list.  The parameter type must be
                signed integer. {width} may optionally be present.
                If not present or if {width is smaller than the
                supplied parameter, the field will be made as long
                as necessary to print the value contained in the
                parameter (including its decimal point and sign).
                {dec} must be present and must be } 0. (Note:
                While not strictly a floating point function,
                "f" allows the printing of numbers represented
                internally as integers, but with internally
                accounted-for decimal places.)

            @oo "xn": Signifies that the field consists of one
                hexadecimal digit (a nibble), contained in the next
                parameter.  "0x" will print followed by one
                character containing the hex-ascii representation
                of that nibble.

            @oo "xb": Signifies that the field consists of two
                hexadecimal digits (a byte), contained in the next
                parameter.  Two characters containing the hex-ascii
                representation of that byte will print.

            @oo "xw": Signifies that the field consists of four
                hexadecimal digits (a word), contained in the next
                parameter.  "0x" will print followed by four
                characters containing the hex-ascii representation
                of that word.

            @oo "xd": Signifies that the field consists of eight
                hexadecimal digits (a double-word), contained in
                the next  parameter.  "0x" will print followed by
                eight characters containing the hex-ascii
                representation of that double-word.

***************************************************************/

void APPrint(void (*pStream)(char c), char *pF, ...);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

#define APPrint_Char(pStream,c)   (*(pStream))(c)

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void APPrint_HexNibble(void (*pStream)(char c), uint32_t data);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void APPrint_HexByte(void (*pStream)(char c), uint32_t data);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void APPrint_HexWord(void (*pStream)(char c), uint32_t data);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void APPrint_HexDWord(void (*pStream)(char c), uint32_t data);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void APPrint_String(void (*pStream)(char c), const char *pString);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
void APPrint_CRLF(void (*pStream)(char c));

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void APPrint_DecUInt(void (*pStream)(char c), unsigned intData);

//***************************************************************
//                          *** END ***
//***************************************************************
