/******************************************************************//**
   @file app.h

   app.h: Application level definitions for PSE-702

   This file contains definitions that are specific to the
   application and are application wide in scope.

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

*//*******************************************************************/

//              *** GENERAL DEFINITIONS ****

//  *** Application Module Number and Related Definitions ***

//#define MID_RESERVED  0x00    /// Reserved ID Number. Do not use.
#define MID_DTMF        0x01    ///< DTMF Detector Base ID Number
#define MID_MORSE       0x02    ///< Morse Code Encoder Base ID Number
#define MID_CON         0x03    ///< Console Cmd Proc Base ID Number
#define MID_FSYS        0x04    ///< Flash File System Base ID Number
#define MID_CEN         0x05    ///< Central control module
#define MID_SQU         0x06    ///< Squelch module
#define MID_VOI         0x07    ///< Voice record/playback module
#define MID_TXSM        0x08    ///< Transmitter state machine
#define MID_IDSM        0x09    ///< ID state machine
#define MID_AUD         0x0A    ///< Audio state machine
#define MID_SATE        0x0B    ///< Subaudible tone encoder ID
#define MID_SATD        0x0C    ///< Subaudible tone decoder ID
#define MID_DUSI        0x0D    ///< DTMF user interface.
#define MID_SUSI        0x0E    ///< Serial user interface
#define MID_XMODM       0x0F    ///< X-Modem protocol state machine
#define MID_SSM         0x10    ///< Simple state machine

//#define MID_RESERVED  0xFE    /// Apex Reserved ID Number. Do not use.
//#define MID_RESERVED  0xFF    /// Apex Reserved ID Number. Do not use.

    // Task ID Numbers.  The task ID is formed by the module
    // number in the high order 8 bits followed by 01 in the
    // low order 8 bits.  Here are a few examples:

    // 0xFF01       Apex Idle Task
    // 0xFE01       Apex Timer task
    // 0x0101       DTMF Detector
    // 0x0201       Morse Code
    // 0x0301       Console Command Processor


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Task Priorities:

    // Each task must be given a priority between 0 and 7.  Larger
    // numbers correspond to higher priority.

    // Apex creates two tasks, the timer task and the idle task.
    // Timer task priority is max (31).  The idle task is the
    // lowest priority, but priority number 0 is assignable.

    // These numbers are defined here instead of in the individual
    // task's h file because it's easier to see and adjust the
    // relative priorities this way.

#define TPRI_APEX       7  ///< System task -> maximum priority.
#define TPRI_DTMF       6  ///< DTMF Detector task priority.
//#define TPRI_SATD       5  ///< Subaudible tone decoder task priority.
#define TPRI_VOI        5  ///< Voice record/playback task priority.
//#define TPRI_SATE       4  ///< Subaudible Tone Encoder priority.
#define TPRI_AUD        3  ///< Audio task priority.
//#define TPRI_IDSM       3  ///< ID task priority.
//#define TPRI_TXSM       3  ///< Transmitter task priority.
//#define TPRI_SQU        3  ///< Squelch task priority
//#define TPRI_MORSE      3  ///< Morse task priority.
#define TPRI_SSM        3  ///< Simple state machine priority.
#define TPRI_CEN        2  ///< Central task priority
#define TPRI_DUSI       1  ///< DTMF user interface priority
#define TPRI_CON        1  ///< Console task priority.

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//          *** COMMON DATA STRUCTURE DEFINITIONS ***

    // Physical Receiver Channel Designators:

enum TAG_RX_CHAN
{
    RX_CHAN_1 = 0,      // Channel 1
    RX_CHAN_2,          // Channel 2
    MAX_RXCHAN          // Number of channels
};

    // Convert chanel number 1 or 2 to enum:
#define ORDINAL_2_RX_CHAN(x)     ((enum TAG_RX_CHAN)((x)-1))

    // The definition below is used with the ASSERT macro to
    // qualify channel numbers:

#define VALID_RX_CHAN(x)    ((enum TAG_RX_CHAN)(x) < MAX_RXCHAN)


    // Physical Transmitter Channel Designators:
    // This enum starts at zero because it is used to index
    // into several arrays.  MAX_TXCHAN gives the size of
    // the array.

enum TAG_TX_CHAN
{
    TX_CHAN_1 = 0,      // Channel 1
    TX_CHAN_2,          // Channel 2
    MAX_TXCHAN          // Number of channels
};

#define TXCHAN_INVALID  ((enum TAG_TX_CHAN)-1)

    // Convert chanel number 1 or 2 to enum:
#define ORDINAL_2_TX_CHAN(x)     ((enum TAG_TX_CHAN)((x)-1))
#define TX_CHAN_2_ORDINAL(x)     ((enum TAG_TX_CHAN)((x)+1))

    // The definition below is used with the ASSERT macro to
    // qualify channel numbers:

#define VALID_TX_CHAN(x)    ((enum TAG_TX_CHAN)(x) < MAX_TXCHAN)

        // Tone Generator Assignments
#define FGEN_TX1_ENCODER_CHAN   0       ///< Tx1 Subaudible Tone
#define FGEN_TX2_ENCODER_CHAN   1       ///< Tx2 Subaudible Tone
#define FGEN_TX1_TONE_CHAN      2       ///< Tx1 Morse Tone
#define FGEN_TX2_TONE_CHAN      3       ///< Tx2 Morse Tone

    /// Number of files reserved for user.
#define NUM_USER_FILES       ((unsigned)8)
#define VALID_USER_FILENUM(x)  ((x)<NUM_USER_FILES)
#define ORDINAL_2_FILENUM(x)    ((uint8_t)((x)-1))
#define FILENUM_2_ORDINAL(x)    ((x)+1)

//*********************************************************************
//                     End of File
//*********************************************************************
