/**************************************************************//**
    @file astream.h

    Audio Processing Stream Interrupt for PSE-702 - API

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

******************************************************************/
//               *** DATA STRUCTURE DEFINITIONS ***
//*****************************************************************

#define MAX_DEEMPH_FILTERS     2   ///< Number of deemphasis filter
                                    ///< profiles to choose from.
#define MAX_CTCSS_REJECT_FILTERS    3  ///< Number of tone reject
                                        ///< filters to choose from.

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Audio Stream Source Signals

typedef enum TAG_ASTREAM_SRCSIG
{
    ASTREAM_SSIG_RX1,       ///< Receiver channel 1
    ASTREAM_SSIG_RX2,       ///< Receiver channel 2
    ASTREAM_SSIG_VOICE1,    ///< Voice Channel 1
    ASTREAM_SSIG_VOICE2,    ///< Voice Channel 2

    ASTREAM_MAX_SSIG        ///< Max number of source signals

} astream_srcsig_t;

#define ASTREAM_NULL_SOURCE     ((astream_srcsig_t)-1)
#define ASTREAM_SOURCE_SIG_IS_RX(x)  \
                    ((astream_srcsig_t)(x) <= ASTREAM_SSIG_RX2)

#define ASTREAM_RX_CHAN_2_SOURCE(x)     ((astream_srcsig_t)(x))

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Audio Stream Destination Signals

typedef enum TAG_ASTREAM_DESTSIG
{
    ASTREAM_DSIG_TX1,       ///< Transmitter channel 1
    ASTREAM_DSIG_TX2,       ///< Transmitter channel 2
    ASTREAM_DSIG_VOICE1,    ///< Voice Channel 1
    ASTREAM_DSIG_VOICE2,    ///< Voice Channel 2

    ASTREAM_MAX_DSIG        ///< Max number of dest signals

} astream_destsig_t;

#define ASTREAM_NULL_DEST     ((astream_destsig_t)-1)

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Path Structure

typedef struct TAG_ASTREAM_PATH
{
    LINK link;                  ///< Linked list hook
    uint16_t id;                ///< Number identifying the path
    astream_srcsig_t srcSig;    ///< Source signal
    astream_destsig_t destSig;  ///< Destination signal
//    bool deemphasis;            ///< Should the path be deemphasized?
    bool mute;                  ///< Is the path muted?
    char priority;              ///< H M or L for destination priority

} astream_path_t;

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Message Structure used with ASTREAM_MessageDestDevices().

typedef struct TAG_ASTREAM_MSG
{
    unsigned event;    /// Code for the event
    uintptr_t data1;    /// General data or pointer variable
    uintptr_t data2;    /// General data or pointer variable

} astream_msg_t;

//*****************************************************************
//                    *** Signals ***
//*****************************************************************

    /** Send signal to transmitter 1 message slot.
     *
     * @oo slotParm1 ASTREAM destination device code identifying
     *      the device to receive the message.
     * @oo slotParm2 Message ID to send to destination device.
     */
DECLARE_SIGNAL(ASTREAM_DestMessage)

    /** Send signal to a source when a higher priority path
     * preempts it's audio.
     *
     * @oo slotParm1 ASTREAM source device code identifying
     *      the device that was preempted.
     * @oo slotParm2 The ID of the path that was preempted.
     */
DECLARE_SIGNAL(ASTREAM_SrcPreempted)

    /** Send signal informing the un-assignment of a destination
     * device to the audio stream.
     *
     * @oo slotParm1 ASTREAM destination signal code identifying
     *      the signal that was unassigned.
     * @oo slotParm2 Unused.
     */
DECLARE_SIGNAL(ASTREAM_DestUnassigned)

    /** Send signal informing a source device that a destination
     *  assigned to it is now ready to accept audio.
     *
     * @oo slotParm1 ASTREAM source device code identifying
     *      the device that can enable audio.
     * @oo slotParm2 ASTREAM destination device code identifying
     *      the device that has become ready to receive audio.
     */
//DECLARE_SIGNAL(ASTREAM_DestReady)

//*****************************************************************
//                *** API Functions ***
//*****************************************************************

void ASTREAM_Reset(void);


void ASTREAM_Start(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void ASTREAM_UnsubscribeAll(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void ASTREAM_AddPath(
        uint16_t pathID, astream_srcsig_t srcSig,
        astream_destsig_t destSig, char priority);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

uint16_t ASTREAM_GetPathID(void);

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *//* * *
    Get a running average that approximates the sum of the
    last 256 squared samples.  The square root of this is
    the approximate RMS of the last 22 ms of the signal.

        @param srcSig Indicates which channel's squared RMS to
            read. Currently only RX1 and RX2 are supported.
    */
uint32_t ASTREAM_GetRmsSquared(astream_srcsig_t srcSig);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void ASTREAM_RemovePathByID(uint16_t pathID);


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *//* * *
    Remove all paths assigned to a particular destination
    device.

    This function is used by the transmitter timeout function
    to remove the audio path that has timed the transmitter out.
    */
void ASTREAM_RemovePathByDestDev(astream_destsig_t destSig);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void ASTREAM_RemoveAllPaths(void);

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    Send a message to all destination devices associated
    with a given path ID.
*/
void ASTREAM_MessageDestDevicesByPathID(
                        uint16_t pathID, astream_msg_t *pMsg);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /* Signal all destination devices currently active due
     * to the current path priorities.
     */
void ASTREAM_MessageDestDevicesByActivePath(
                        uint16_t pathID, astream_msg_t *pMsg);

#if 0
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Notify source devices that destination device is ready.
     *
     * @param destDev Destination device that is now ready to accept
     * audio
     */
void ASTREAM_NotifySourcesDestIsReady(astream_destdev_t destDev);
#endif
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Mute a destination device
     *
     * @param destSig Indicates which destination device to mute.
     */
void ASTREAM_MuteDestDev(astream_destsig_t destSig);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Unmute a destination device
     *
     * @param destSig Indicates which destination device to unmute.
     */
void ASTREAM_UnmuteDestDev(astream_destsig_t destSig);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void ASTREAM_Sample_ISR(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void ASTREAM_SetDeemphasisFilterIndex(
                            enum TAG_RX_CHAN chan, unsigned index);

void ASTREAM_SetCtcssRejectionFilterIndex(
                            enum TAG_RX_CHAN chan, unsigned index);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Get a pointer to a destination's path list.
     *
     * This function is used by SUSI to print a list of active
     * paths.
     *
     * @param destSig The destination signal code for which the
     *  path list shall be returned.
     * @return Returns a pointer to a path list anchor (LINK).
     */
LINK *ASTREAM_GetDestPathList(astream_destsig_t destSig);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Set transmitter audio level.  Levels are 0 to 100.
    // Level 0 corresponds to the minimum gain, and the gain
    // increases logarithmically.

void ASTREAM_SetTx1Level(uint8_t index);
void ASTREAM_SetTx2Level(uint8_t index);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                          END
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
