/******************************************************************//**

    @file audiosm.c

    Audio State Machine for PSE-702

    @defgroup audiosm Audio State Machine

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

**********************************************************************/

//      *** INCLUDE FILES ***

#include <stdint.h>
#include <stdbool.h>        //lint -esym(766, stdbool.h)

#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"
#include "apex.h"
#include "apalloc.h"
#include "app.h"
#include "astream.h"
#include "eep.h"
#include "audiosm.h"
#include "fgen.h"

#include "io_debug.h"       //lint -esym(766, io_debug.h)
#include "apprint.h"        //lint -esym(766, apprint.h)
#include "io_main.h"        //lint -esym(766, io_main.h)

//***************************************************************

//      *** MACROS ****

//lint -esym(750, PRINT_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, SET_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, STATE_CHAN) (Macro not always referenced)

#define PRINT_STATE_CHAR(x)  if(STATE_CHAN == pChan->txChanID) \
        IO_DBPrintChar(x)

#if ((TP_AUD_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
  || (TP_AUD_STATE_FLAGS == STATE_FLAGS_LOWER_CH1))
#   define STATE_CHAN  TX_CHAN_1
#else
#   define STATE_CHAN  TX_CHAN_2
#endif

#if (TP_AUD_STATE_FLAGS == STATE_FLAGS_OFF)
#   define SET_STATE(X,Y)   pChan->pState = (X)

#elif ((TP_AUD_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
    || (TP_AUD_STATE_FLAGS == STATE_FLAGS_UPPER_CH2))
#   define SET_STATE(X,Y)   pChan->pState = (X); PRINT_STATE_CHAR(Y)

#elif ((TP_AUD_STATE_FLAGS == STATE_FLAGS_LOWER_CH1) \
    || (TP_AUD_STATE_FLAGS == STATE_FLAGS_LOWER_CH2))
#   define SET_STATE(X,Y) pChan->pState = (X); \
                                            PRINT_STATE_CHAR((Y)|0x20)
#else
#error "TP_AUD_STATE_FLAGS improperly defined."
#endif


//***************************************************************
    // Stack Size: Measured use of 288 bytes on 3/13/12.

#define Z_STK_AUD       384     // Stack size in bytes


    // Timeout Values:

#define AUD_TIME_TX_STARTUP 500     ///< Tx cold start delay in ms.

    // State Machine Event IDs:

enum TAG_AUD_EVIDS
{
    EV_AUD_START = AUD_ID(0),   ///< Play audio job list.
    EV_AUD_RXAON,               ///< Receiver audio turned on.
    EV_AUD_RXAOFF,              ///< Receiver audio turned off.
    EV_AUD_ID,                  ///< Do an ID at next opportunity.
    EV_AUD_MORSECP,             ///< Morse generation complete.
    EV_AUD_GPTIMER,             ///< General purpose timer timeout.
    EV_AUD_FLASHCP,             ///< Flash audio finished.
    EV_AUD_PREEMPTED,           ///< Generated audio has been preempted.
    EV_AUD_ANNO,                ///< Announcement timed out.
    EV_AUD_TXTO_CLEARED,        ///< Transmitter timeout was cleared.
    EV_AUD_PLAY_CB,             ///< Play a courtesy beep.
    EV_AUD_TIMEOUT,             ///< A transmitter has timed out.
    EV_AUD_DISABLE,             ///< A transmitter is being turned off.
    EV_AUD_PLAYFILE,            ///< Play a file.
    EV_AUD_MAX
};

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//             DEDICATED FLASH FILE NUMBERS
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//#define AUD_TX1_ID_FILE     0   // File contains ID for chan 1
//#define AUD_TX2_ID_FILE     1   // File contains ID for chan 2

//***************************************************************
//               *** DATA STRUCTURE DEFINITIONS ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//***************************************************************
//                *** VARIABLES ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static TCB audioTCB;
static uint32_t audioStack[Z_STK_AUD/sizeof(uint32_t)];
static MBOX mailbox;
//static LINK announcements;      ///< List of announcements.

static struct TAG_AUD_CHAN
{
    enum TAG_TX_CHAN txChanID;        ///< Transmitter channel
    astream_srcsig_t srcID;           ///< Voice ID channel source
    astream_destsig_t destID;         ///< Voice ID channel destination

    // Task state:
    bool (*pState)(enum TAG_AUD_EVIDS event,
                struct TAG_AUD_CHAN *pChan, uintptr_t data);

    /// General purpose timer.  Times Tx startup delay,
    /// courtesy beep delay, courtesy beep duration, etc.
    TMR gpTimer;

    /// The low order 16 flags are usable for general audio jobs.
    /// The high order 16 flags are used for announcements.
    uint32_t flags;
#    define FLAG_CWID          0x0001 ///< Schedule a CW ID.
#    define FLAG_DTMFOK        0x0002 ///< DTMF "OK" message.
#    define FLAG_DTMFERR       0x0004 ///< DTMF "ERROR" message.
#    define FLAG_VID           0x0008 ///< Schedule a voice ID.
//#    define FLAG_JOBS_PENDING  0x0010 ///< Audio jobs are pending.
#    define FLAG_PLAYFILE      0x0020 ///< Play an audio file

#    define FLAG_ANNO(x)   (BV((x)+16)) ///< Announcement flag.
#    define FLAG_ANNO_MASK  0xFFFF0000  ///< Gives all anno flags.

    eep_courtsey_beep_t *pBeep; ///< Courtesy beep parameters

    uint16_t pathID;         ///< ID of path set by audio job.
    uint8_t fgenChan;        ///< Function generator channel
    uint8_t currentAnno;     ///< Index of current announcement.
    uint8_t playFileNum;     ///< Number of file to play.

    bool userAudio;         ///< True if user audio is on.

} txChan[MAX_TXCHAN];

static TMR *pAnnoTimers[MAX_ANNO];  ///< Announcement timers

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//              Signals
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */ /**

DEFINE_SIGNAL(AUD_GeneratedAudioOff)
DEFINE_SIGNAL(AUD_StartMorse)

DEFINE_SIGNAL(AUD_IDAudioStarted)
DEFINE_SIGNAL(AUD_IDAudioComplete)

DEFINE_SIGNAL(AUD_MessageAudioStarted)
DEFINE_SIGNAL(AUD_MessageAudioComplete)

DEFINE_SIGNAL(AUD_VoiceID)
DEFINE_SIGNAL(AUD_CancelPlayback)

DEFINE_SIGNAL(AUD_Announce)

//*********************************************************************
//                  *** CONSTANTS ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Debugging Table for Variable Display

const struct TAG_VARIABLE_TABLE audVarTab[] =
{
    {"stack", audioStack, VAR_TYPE_NONE, sizeof(audioStack)},
    {"tcb", &audioTCB, VAR_TYPE_NONE, sizeof(audioTCB)},
    {"f1", &txChan[TX_CHAN_1].flags, VAR_TYPE_HEXINT, 4},
    {"f2", &txChan[TX_CHAN_2].flags, VAR_TYPE_HEXINT, 4},
    {"state1", &txChan[TX_CHAN_1].pState,
                                VAR_TYPE_HEXINT, sizeof(void *)},
    {"state2", &txChan[TX_CHAN_2].pState,
                                VAR_TYPE_HEXINT, sizeof(void *)},
    {NULL, NULL, VAR_TYPE_NONE, 0}
};


//***************************************************************
//              *** STATIC FUNCTION PROTOTYPES ****
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void AudioTask(void);
static void DoAudioJobList(struct TAG_AUD_CHAN *pChan);

static void Transition_Idle(struct TAG_AUD_CHAN *pChan);    // A
static bool State_Idle(enum TAG_AUD_EVIDS event,
                struct TAG_AUD_CHAN *pChan, uintptr_t data);

static void Transition_TxStartupDelay(
                                struct TAG_AUD_CHAN *pChan);    // B
static bool State_TxStartupDelay(enum TAG_AUD_EVIDS event,
                struct TAG_AUD_CHAN *pChan, uintptr_t data);

static void Transition_MorseAudio(
                                struct TAG_AUD_CHAN *pChan);    // C
static bool State_MorseAudio(enum TAG_AUD_EVIDS event,
                struct TAG_AUD_CHAN *pChan, uintptr_t data);

static void Transition_VoiceID(struct TAG_AUD_CHAN *pChan); // D
static bool State_VoiceID(enum TAG_AUD_EVIDS event,
                struct TAG_AUD_CHAN *pChan, uintptr_t data);

static void Transition_Announce(
                                struct TAG_AUD_CHAN *pChan);    // E
static bool State_Announce(enum TAG_AUD_EVIDS event,
                struct TAG_AUD_CHAN *pChan, uintptr_t data);

static void Transition_CBDelay(struct TAG_AUD_CHAN *pChan); // F
static bool State_CBDelay(enum TAG_AUD_EVIDS event,
                struct TAG_AUD_CHAN *pChan, uintptr_t data);

static void Transition_FirstCourtesyBeep(
                                struct TAG_AUD_CHAN *pChan);    // G
static bool State_FirstCourtesyBeep(enum TAG_AUD_EVIDS event,
                struct TAG_AUD_CHAN *pChan, uintptr_t data);

static void Transition_CWID(struct TAG_AUD_CHAN *pChan);    // H
static bool State_CWID(enum TAG_AUD_EVIDS event,
                struct TAG_AUD_CHAN *pChan, uintptr_t data);

static void Transition_Rxon(struct TAG_AUD_CHAN *pChan);    // I
static bool State_Rxon(enum TAG_AUD_EVIDS event,
                struct TAG_AUD_CHAN *pChan, uintptr_t data);

static void Transition_UserAudioOverID(
                                struct TAG_AUD_CHAN *pChan);    // J
static bool State_UserAudioOverID(enum TAG_AUD_EVIDS event,
                struct TAG_AUD_CHAN *pChan, uintptr_t data);

static void Transition_SecondCourtesyBeep(
                                struct TAG_AUD_CHAN *pChan);    // K
static bool State_SecondCourtesyBeep(enum TAG_AUD_EVIDS event,
                struct TAG_AUD_CHAN *pChan, uintptr_t data);

static void Transition_PlayFile(
                                struct TAG_AUD_CHAN *pChan);    // L
static bool State_PlayFile(enum TAG_AUD_EVIDS event,
                struct TAG_AUD_CHAN *pChan, uintptr_t data);

//***************************************************************
//                     *** API ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void AUD_Reset(void)
{
    unsigned i;

    Apex_InitMailbox(&mailbox);

    Apex_InitMsgTimer(&txChan[TX_CHAN_1].gpTimer,
        &mailbox, EV_AUD_GPTIMER, TX_CHAN_1, 0);
    Apex_InitMsgTimer(&txChan[TX_CHAN_2].gpTimer,
        &mailbox, EV_AUD_GPTIMER, TX_CHAN_2, 0);

    txChan[TX_CHAN_1].txChanID = TX_CHAN_1;
    txChan[TX_CHAN_2].txChanID = TX_CHAN_2;

    txChan[TX_CHAN_1].fgenChan = FGEN_TX1_TONE_CHAN;
    txChan[TX_CHAN_2].fgenChan = FGEN_TX2_TONE_CHAN;

    txChan[TX_CHAN_1].srcID = ASTREAM_SSIG_VOICE1;
    txChan[TX_CHAN_2].srcID = ASTREAM_SSIG_VOICE2;

    txChan[TX_CHAN_1].destID = ASTREAM_DSIG_TX1;
    txChan[TX_CHAN_2].destID = ASTREAM_DSIG_TX2;

    txChan[TX_CHAN_1].flags = 0;
    txChan[TX_CHAN_2].flags = 0;

    txChan[TX_CHAN_1].userAudio = false;
    txChan[TX_CHAN_2].userAudio = false;

    for(i=0; i<MAX_ANNO; i++)
        pAnnoTimers[i] = NULL;


    AUD_UnsubscribeAll();
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void AUD_Start(void)
{
    Apex_CreateTask(&audioTCB, TID_AUD, AudioTask,
        audioStack, (sizeof(audioStack)), TPRI_AUD);

    AUD_AddAllAnno();
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void AUD_AddAnno(unsigned index)
{
    eep_anno_t *pEEPAnno;

    // Is this a new announcement, or a modification of an
    // existing one?
    pEEPAnno = EEP_GetAnno(index);

    // Now, see if we have an active announcement with the
    // index that was submitted.
    if(pAnnoTimers[index] != NULL)
    {
        // The announcement must be enabled if it's in our
        // list here, and SUSI doesn't disable it when a
        // change is made.
        Apex_StopTmr(pAnnoTimers[index]);
        Apex_InitMsgTimer(pAnnoTimers[index], &mailbox,
                            EV_AUD_ANNO, pEEPAnno->chan, index);
        Apex_StartTmr(pAnnoTimers[index],
                        (int32_t)(pEEPAnno->period*TU_1SEC*60));
    }

    // Here, the announcement is not active.
    else
    {
        // Check to see if the announcement is enabled:
        if(pEEPAnno->enabled)
        {
            // Allocate a place to put the announcement timer:
            pAnnoTimers[index] = malloc(sizeof(TMR));
            ASSERT(pAnnoTimers[index] != NULL);
            Apex_InitMsgTimer(pAnnoTimers[index], &mailbox,
                                EV_AUD_ANNO, pEEPAnno->chan, index);
            Apex_StartTmr(pAnnoTimers[index],
                            (int32_t)(pEEPAnno->period*TU_1SEC*60));
        }
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void AUD_DeleteAnno(unsigned index)
{
    if(pAnnoTimers[index] != NULL)
    {
        Apex_StopTmr(pAnnoTimers[index]);
        free(pAnnoTimers[index]);
        pAnnoTimers[index] = NULL;
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void AUD_AddAllAnno(void)
{
    unsigned i;
    eep_anno_t *pEEPAnno;

    // Make a list of the active announcement timers
    for(i=0; i<MAX_ANNO; i++)
    {
        pEEPAnno = EEP_GetAnno(i);     // Get pointer to an announcement.

        // Is it empty?  Yes if invalid Tx.  Check to make
        // sure it's enabled too.
        if((pEEPAnno->chan != TXCHAN_INVALID) && pEEPAnno->enabled)
        {
            AUD_AddAnno(i);
        }
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void AUD_DeleteAllAnno(void)
{
    unsigned i;

    // Now, go through the list and delete each announcement.
    for(i=0; i<MAX_ANNO; i++)
    {
        AUD_DeleteAnno(i);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Unsubscribe to all audio task signals

void AUD_UnsubscribeAll(void)
{
    Unsubscribe_AUD_GeneratedAudioOff();
    Unsubscribe_AUD_StartMorse();

    Unsubscribe_AUD_IDAudioStarted();
    Unsubscribe_AUD_IDAudioComplete();

    Unsubscribe_AUD_MessageAudioStarted();
    Unsubscribe_AUD_MessageAudioComplete();

    Unsubscribe_AUD_VoiceID();
    Unsubscribe_AUD_CancelPlayback();
    Unsubscribe_AUD_Announce();

    AUD_DeleteAllAnno();
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell AudioSM to start generating audio.

void AUD_StartAudioJobs(enum TAG_TX_CHAN chan)
{
    Apex_SendMsgExt(&mailbox, EV_AUD_START, (uintptr_t)chan, 0);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell AudioSM to play a courtesy beep.

void AUD_PlayCB(
            enum TAG_TX_CHAN chan, eep_courtsey_beep_t *pBeep)
{
    Apex_SendMsgExt(&mailbox,
            EV_AUD_PLAY_CB, (uintptr_t)chan, (uintptr_t)pBeep);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell AudioSM to play a particular announcement.

void AUD_PlayAnno(unsigned short index)
{
    Apex_SendMsgExt(&mailbox, EV_AUD_ANNO,
                        EEP_GetAnno(index)->chan, index);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell AudioSM to play a file.

void AUD_PlayFile(enum TAG_TX_CHAN chan, unsigned short fileNum)
{
    txChan[chan].flags |= FLAG_PLAYFILE;
    txChan[chan].playFileNum = (uint8_t)fileNum;
    Apex_SendMsgExt(&mailbox, EV_AUD_PLAYFILE, chan, fileNum);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell AudioSM that a transmitter has timed out.
    // Schedule "TO" and ID messages and unschedule
    // everything else.

void AUD_NotifyTxTimeout(enum TAG_TX_CHAN chan)
{
    Apex_SendMsgExt(&mailbox, EV_AUD_TIMEOUT, chan, 0);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell AudioSM that a transmitter is being turned off.
    // Schedule a CW ID message.

void AUD_NotifyTxDisable(enum TAG_TX_CHAN chan)
{
    Apex_SendMsgExt(&mailbox, EV_AUD_DISABLE, chan, 0);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell AudioSM that there is user audio on the channel:

void AUD_NotifyUserAudioOn(enum TAG_TX_CHAN chan)
{
    txChan[chan].userAudio = true;
    Apex_SendMsgExt(&mailbox, EV_AUD_RXAON, (uintptr_t)chan, 0);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell AudioSM that user audio is off:

void AUD_NotifyUserAudioOff(enum TAG_TX_CHAN chan)
{
    txChan[chan].userAudio = false;
    Apex_SendMsgExt(&mailbox, EV_AUD_RXAOFF, (uintptr_t)chan, 0);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell AudioSM that the audio it has been putting out
    // has been preempted.

void AUD_NotifySourcePreempted(enum TAG_TX_CHAN chan)
{
    Apex_SendMsgExt(
                &mailbox, EV_AUD_PREEMPTED, (uintptr_t)chan, 0);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell AudioSM that morse code generation has completed.

void AUD_NotifyMorseComplete(enum TAG_TX_CHAN chan)
{
    Apex_SendMsgExt(&mailbox, EV_AUD_MORSECP, (uintptr_t)chan, 0);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell AudioSM that the flash file has completed playing.

void AUD_NotifyFlashComplete(enum TAG_TX_CHAN chan)
{
    Apex_SendMsgExt(&mailbox, EV_AUD_FLASHCP, (uintptr_t)chan, 0);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell AudioSM that a transmitter that had previously
    // timed out is now back in service.

void AUD_NotifyTxTimeoutCleared(enum TAG_TX_CHAN chan)
{
    Apex_SendMsgExt(
        &mailbox, EV_AUD_TXTO_CLEARED, (uintptr_t)chan, 0);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell AudioSM to schedule an ID as soon as possible.  We
    // send a message in case the audio state machine is idle.
    // If so it needs to give the transmitter time to start up.

void AUD_ScheduleID(enum TAG_TX_CHAN chan)
{
    if(EEP_GetTxFlags(chan, EEP_TX_FLAG_VID))
        txChan[chan].flags |= FLAG_VID;   // Schedule voice ID
    else
        txChan[chan].flags |= FLAG_CWID;  // Schedule CW ID

    Apex_SendMsgExt(&mailbox, EV_AUD_ID, (uintptr_t)chan, 0);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell AudioSM to schedule a DTMF OK message.

void AUD_ScheduleOK(enum TAG_TX_CHAN chan)
{
    txChan[chan].flags |= FLAG_DTMFOK;  //
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell AudioSM to schedule a DTMF ERR message.

void AUD_ScheduleERR(enum TAG_TX_CHAN chan)
{
    txChan[chan].flags |= FLAG_DTMFERR; //

}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                 *** STATIC FUNCTIONS ***
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void AudioTask(void)
{
    unsigned event;
    uintptr_t data1, data2;
    struct TAG_AUD_CHAN *pChan;
    aud_morse_parms_t morseParms;

    Transition_Idle(&txChan[TX_CHAN_1]);
    Transition_Idle(&txChan[TX_CHAN_2]);

    FOREVER
    {
        // Wait for a message.

        Apex_PendMsg(&mailbox, &event, &data1, &data2, 0);

        ASSERT(data1 < MAX_TXCHAN);
        pChan = &txChan[data1];

        // States return true if the event was processed, or
        // false to cause the default processing in the
        // switch statement below.

        if(!(*pChan->pState)(
                (enum TAG_AUD_EVIDS)event, pChan, data2))
        {
            switch(event)
            {
                // Here, it's time to play an announcement.
                case EV_AUD_ANNO:
                    pChan->flags |= FLAG_ANNO(data2);

                    // Restart the announcement timer for this
                    // announcement (if it's enabled).  It might
                    // not be enabled if we were manually playing
                    // back an announcement.
                    if(EEP_GetAnno(data2)->enabled)
                    {
                        Apex_StartTmr(pAnnoTimers[data2],
                            (int32_t)(EEP_GetAnno(data2)->period
                            *TU_1SEC*60));
                    }
                    break;

                // Here we had a transmitter timeout.
                case EV_AUD_TIMEOUT:
                    pChan->flags = FLAG_CWID;  // Schedule ID
                    FGen_ChannelOff(pChan->fgenChan);
                    Apex_StopTmr(&pChan->gpTimer);
                    morseParms.level =
                        EEP_GetMorseLevel(pChan->txChanID);
                    morseParms.pString = " TO ";

                    EMIT_SIGNAL(AUD_StartMorse,
                        pChan->txChanID, (uintptr_t)&morseParms);

                    Transition_MorseAudio(pChan);
                    break;

                // Here we are disabling a transmitter.  Turn off
                // all flags except DTMFOK and set the CWID flag.
                case EV_AUD_DISABLE:
                    pChan->flags &= FLAG_DTMFOK;
                    pChan->flags |= FLAG_CWID;  // Schedule ID
                    FGen_ChannelOff(pChan->fgenChan);
                    Apex_StopTmr(&pChan->gpTimer);

                    // Tell Central we are putting out ID audio.
                    EMIT_SIGNAL(
                            AUD_IDAudioStarted, pChan->txChanID, 0);
                    Transition_TxStartupDelay(pChan);
                    break;

                default:
                    break;
            }
        }
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // State A: Idle

static void Transition_Idle(struct TAG_AUD_CHAN *pChan)
{
//    pChan->sentAudioStart = false;
    EMIT_SIGNAL(AUD_GeneratedAudioOff, pChan->txChanID, 0);
    SET_STATE(State_Idle, 'A');
}

static bool State_Idle(enum TAG_AUD_EVIDS event,
                struct TAG_AUD_CHAN *pChan, uintptr_t unused)
{
    switch(event)
    {
        case EV_AUD_RXAON:
            Transition_Rxon(pChan);     // Wait for start.
            return(true);

        // Play an announcement or file
        case EV_AUD_PLAYFILE:
            EMIT_SIGNAL(
                AUD_MessageAudioStarted, pChan->txChanID, 0);
            Transition_TxStartupDelay(pChan);
            return(true);   // Indicate event processing complete.

        case EV_AUD_ANNO:
            EMIT_SIGNAL(
                AUD_MessageAudioStarted, pChan->txChanID, 0);
            Transition_TxStartupDelay(pChan);
            return(false);  // Do default processing.

        // Do an ID from cold
        case EV_AUD_ID:
            // Tell Central we are putting out ID audio.
            EMIT_SIGNAL(AUD_IDAudioStarted, pChan->txChanID, 0);
            Transition_TxStartupDelay(pChan);
            return(true);   // Indicate event processing complete.

        default:
            return(false);  // Do default event processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tx Startup delay:
    // This delay allows the transmitter to start up and allows
    // the user's receivers to lock onto a tone before sending
    // the ID audio.

static void Transition_TxStartupDelay(
                                        struct TAG_AUD_CHAN *pChan)
{
    Apex_StartTmr(&pChan->gpTimer, AUD_TIME_TX_STARTUP);
    SET_STATE(State_TxStartupDelay, 'B');
}

static bool State_TxStartupDelay(enum TAG_AUD_EVIDS event,
                    struct TAG_AUD_CHAN *pChan, uintptr_t unused)
{
    switch(event)
    {
        // Start the audio jobs after the timer times out.
        case EV_AUD_GPTIMER:
            DoAudioJobList(pChan);
            return(true);   // Indicate event processing complete.

        default:
            return(false);  // Do default processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // User Audio is on.  Wait until the user audio stops and
    // then we can re-start the audio jobs.

static void Transition_Rxon(struct TAG_AUD_CHAN *pChan)
{
    EMIT_SIGNAL(AUD_GeneratedAudioOff, pChan->txChanID, 0);
    SET_STATE(State_Rxon, 'I');
}

static bool State_Rxon(enum TAG_AUD_EVIDS event,
                struct TAG_AUD_CHAN *pChan, uintptr_t data)
{
    eep_courtsey_beep_t *pBeep = (eep_courtsey_beep_t *) data;

    switch(event)
    {
        // Do the audio job list.
        case EV_AUD_START:
            DoAudioJobList(pChan);  // Will set a new state.
            return(true);   // Indicate event processing complete.

        // Play a courtesy beep.  If not enabled, simply ignore.
        case EV_AUD_PLAY_CB:
            if(pBeep->mode != CB_MODE_DISABLED)
            {
                pChan->pBeep = pBeep;
                Transition_CBDelay(pChan);
            }
            return(true);   // Indicate event processing complete.

        default:
            return(false);  // Do default processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Generate Morse Audio for Timeouts and DTMF Responses

static void Transition_MorseAudio(struct TAG_AUD_CHAN *pChan)
{
    // Tell Central we are putting out audio that must
    // be ID'ed.
    EMIT_SIGNAL(AUD_MessageAudioStarted, pChan->txChanID, 0);
    SET_STATE(State_MorseAudio, 'C');
}

    // Wait in this state for the morse audio to complete.

static bool State_MorseAudio(enum TAG_AUD_EVIDS event,
                    struct TAG_AUD_CHAN *pChan, uintptr_t unused)
{
    switch(event)
    {
        case EV_AUD_MORSECP:
            DoAudioJobList(pChan);  // Will set a new state.
            return(true);   // Indicate event processing complete.

        default:
            return(false);  // Do default processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Generate Voice ID Audio from Flash.

static void Transition_VoiceID(struct TAG_AUD_CHAN *pChan)
{
    pChan->pathID = ASTREAM_GetPathID();

    ASTREAM_AddPath(
        pChan->pathID, pChan->srcID, pChan->destID, 'I');

    EMIT_SIGNAL(AUD_VoiceID, pChan->txChanID, 0);

    SET_STATE(State_VoiceID, 'D');
}

static bool State_VoiceID(enum TAG_AUD_EVIDS event,
                struct TAG_AUD_CHAN *pChan, uintptr_t unused)
{
    aud_morse_parms_t morseParms;

    switch(event)
    {
        // If the ID is complete, check for more jobs to do.

        // The job-list check must come before telling
        // Central that the ID is complete, because the
        // audio-complete signal is sent by the job list
        // when there is no more audio to send, and this
        // must occur before we say the ID was performed,
        // otherwise the ID'er will give continuous IDs
        // every 10 minutes, thinking that there was
        // more audio after the ID completed.
//@todo is this really true?
        case EV_AUD_FLASHCP:
            ASTREAM_RemovePathByID(pChan->pathID);
            DoAudioJobList(pChan);  // Must come before EMIT.
            EMIT_SIGNAL(AUD_IDAudioComplete, pChan->txChanID, 0);
            return(true);   // Indicate event processing complete.

        // If we change state here, we need to switch to CW ID
        // at 1/2 level.
        case EV_AUD_PREEMPTED:
            ASTREAM_RemovePathByID(pChan->pathID);
            EMIT_SIGNAL(AUD_CancelPlayback, pChan->txChanID, 0);

            morseParms.pString =
                EEP_GetIDString(pChan->txChanID);
            morseParms.level =
                EEP_GetMorseLevel(pChan->txChanID);
            EMIT_SIGNAL(AUD_StartMorse,
                pChan->txChanID, (uintptr_t)&morseParms);

            Transition_UserAudioOverID(pChan);
            return(true);   // Indicate event processing complete.

        // Here we had a transmitter timeout.
        case EV_AUD_TIMEOUT:
            ASTREAM_RemovePathByID(pChan->pathID);
            EMIT_SIGNAL(AUD_CancelPlayback, pChan->txChanID, 0);
            return(false);  // Do default processing.

        default:
            return(false);  // Do default processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Play an audio file

static void Transition_PlayFile(
                                struct TAG_AUD_CHAN *pChan)
{
    pChan->pathID = ASTREAM_GetPathID();

    ASTREAM_AddPath(
        pChan->pathID, pChan->srcID, pChan->destID, 'H');
    EMIT_SIGNAL(AUD_Announce, pChan->txChanID, pChan->playFileNum);

    SET_STATE(State_PlayFile, 'L');
}


static bool State_PlayFile(enum TAG_AUD_EVIDS event,
                struct TAG_AUD_CHAN *pChan, uintptr_t unused)
{
    switch(event)
    {
        case EV_AUD_FLASHCP:
            ASTREAM_RemovePathByID(pChan->pathID);
            DoAudioJobList(pChan);  // Must come before EMIT.
            return(true);   // Indicate event processing complete.

        case EV_AUD_RXAON:
            ASTREAM_RemovePathByID(pChan->pathID);
            EMIT_SIGNAL(AUD_CancelPlayback, pChan->txChanID, 0);
            Transition_Rxon(pChan);     // Wait for start.
            return(true);   // Indicate event processing complete.

        case EV_AUD_TIMEOUT:
            ASTREAM_RemovePathByID(pChan->pathID);
            EMIT_SIGNAL(AUD_CancelPlayback, pChan->txChanID, 0);
            return(false);  // Do default processing.

        default:
            return(false);  // Do default processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Generate Announcement Audio from Flash.

static void Transition_Announce(struct TAG_AUD_CHAN *pChan)
{
    eep_anno_t *pEEPAnno;

    pEEPAnno = EEP_GetAnno(pChan->currentAnno);

    pChan->pathID = ASTREAM_GetPathID();

    ASTREAM_AddPath(pChan->pathID, pChan->srcID,
                            pChan->destID, pEEPAnno->priority);
    // This signal starts the file playing.
    EMIT_SIGNAL(AUD_Announce, pChan->txChanID, pEEPAnno->fileNum);
    SET_STATE(State_Announce, 'E');
}

static bool State_Announce(enum TAG_AUD_EVIDS event,
                    struct TAG_AUD_CHAN *pChan, uintptr_t unused)
{
    switch(event)
    {
        case EV_AUD_FLASHCP:
            ASTREAM_RemovePathByID(pChan->pathID);

            // We may want to transfer to Rxon if this announcement
            // was higher priority than the user audio and was not
            // preempted.
            if(pChan->userAudio)
                Transition_Rxon(pChan);
            else
                DoAudioJobList(pChan);  // Will set a new state.

            return(true);   // Indicate event processing complete.

        // If we get this event, a user trumped our announcement.
        // Restart the timer for 1 minute (if it's an enabled
        // announcement).
        case EV_AUD_PREEMPTED:
            ASTREAM_RemovePathByID(pChan->pathID);
            EMIT_SIGNAL(AUD_CancelPlayback, pChan->txChanID, 0);
            if(EEP_GetAnno(pChan->currentAnno)->enabled)
            {
                Apex_StopTmr(pAnnoTimers[pChan->currentAnno]);
                Apex_StartTmr(
                        pAnnoTimers[pChan->currentAnno],
                        (TU_1SEC*60));
            }
            Transition_Rxon(pChan);
            return(true);   // Indicate event processing complete.

        // Here we had a transmitter timeout.
        case EV_AUD_TIMEOUT:
            ASTREAM_RemovePathByID(pChan->pathID);
            EMIT_SIGNAL(AUD_CancelPlayback, pChan->txChanID, 0);
            return(false);  // Do default processing.

        // If we get EV_AUD_RXAON, ignore it
        // because if the announcement was lower priority we would
        // have received the "preempt" event instead.  Just wait
        // until the announcement is complete.

        default:
            return(false);  // Do default processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Pre-courtesy beep delay

static void Transition_CBDelay(struct TAG_AUD_CHAN *pChan)
{
    Apex_StartTmr(&pChan->gpTimer, pChan->pBeep->delay);
    SET_STATE(State_CBDelay, 'F');
}

static bool State_CBDelay(enum TAG_AUD_EVIDS event,
                    struct TAG_AUD_CHAN *pChan, uintptr_t unused)
{
    aud_morse_parms_t morseParms;

    switch(event)
    {
        // Start courtesy beep audio w/ function generator
        // Start courtesy beep timer.
        case EV_AUD_GPTIMER:
            switch(pChan->pBeep->mode)
            {
                case CB_MODE_SINGLE:
                case CB_MODE_DOUBLE:
                    FGen_SetChannelFreq(pChan->fgenChan,
                        (uint16_t)(pChan->pBeep->freq1 * 10));

                    FGen_SetChannelGain(pChan->fgenChan,
                            EEP_GetCourtesyBeepLevel(pChan->txChanID));

                    FGen_ChannelOn(pChan->fgenChan);
                    Apex_StartTmr(&pChan->gpTimer,
                                    pChan->pBeep->duration1);
                    Transition_FirstCourtesyBeep(pChan);
                    break;

                case CB_MODE_MORSE:
                    morseParms.level =
                            EEP_GetCourtesyBeepLevel(pChan->txChanID);
                    morseParms.pString =  pChan->pBeep->morse;

                    EMIT_SIGNAL(AUD_StartMorse,
                            pChan->txChanID, (uintptr_t)&morseParms);
                    Transition_FirstCourtesyBeep(pChan);
                    break;

                // If it's the disabled case, there must be a crash.
                default:
                    FATAL();
                    break;
            }
            return(true);   // Indicate event processing complete.

        // We are receiving user audio
        case EV_AUD_RXAON:
            FGen_ChannelOff(pChan->fgenChan);
            Apex_StopTmr(&pChan->gpTimer);
            Transition_Rxon(pChan);     // Wait for start.
            return(true);   // Indicate event processing complete.

        default:
            return(false);  // Do default processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Wait for first courtesy beep to complete.

static void Transition_FirstCourtesyBeep(
                                    struct TAG_AUD_CHAN *pChan)
{
    SET_STATE(State_FirstCourtesyBeep, 'G');
}

static bool State_FirstCourtesyBeep(enum TAG_AUD_EVIDS event,
                    struct TAG_AUD_CHAN *pChan, uintptr_t unused)
{
    switch(event)
    {
        // Here, courtesy beep is complete.
        case EV_AUD_GPTIMER:
            switch(pChan->pBeep->mode)
            {
                case CB_MODE_SINGLE:
                    FGen_ChannelOff(pChan->fgenChan);
                    DoAudioJobList(pChan);
                    break;

                // Here, we are doing a double courtesy beep.
                case CB_MODE_DOUBLE:
                    FGen_SetChannelFreq(pChan->fgenChan,
                        (uint16_t)(pChan->pBeep->freq2 * 10));

                    FGen_SetChannelGain(pChan->fgenChan,
                            EEP_GetCourtesyBeepLevel(pChan->txChanID));

                    FGen_ChannelOn(pChan->fgenChan);
                    Apex_StartTmr(&pChan->gpTimer,
                                    pChan->pBeep->duration2);
                    Transition_SecondCourtesyBeep(pChan);
                    break;

                // This shouldn't occur because the timer wasn't
                // set if we are doing a morse courtesy beep.
                case CB_MODE_MORSE:
                    break;

                // If it's the disabled case, there must be a crash.
                default:
                    FATAL();
                    break;
            }
            return(true);   // Indicate event processing complete.

        // The morse unit is signaling that the courtesy beep
        // is complete.
        case EV_AUD_MORSECP:
            DoAudioJobList(pChan);
            return(true);   // Indicate event processing complete.

        // We are receiving user audio
        case EV_AUD_RXAON:
            FGen_ChannelOff(pChan->fgenChan);
            Apex_StopTmr(&pChan->gpTimer);
            Transition_Rxon(pChan);     // Wait for start.
            return(true);   // Indicate event processing complete.

        default:
            return(false);  // Do default processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Wait for second courtesy beep to complete.

static void Transition_SecondCourtesyBeep(
                                    struct TAG_AUD_CHAN *pChan)
{
    SET_STATE(State_SecondCourtesyBeep, 'K');
}

static bool State_SecondCourtesyBeep(enum TAG_AUD_EVIDS event,
                    struct TAG_AUD_CHAN *pChan, uintptr_t unused)
{
    switch(event)
    {
        // Here, courtesy beep is complete.
        case EV_AUD_GPTIMER:
            FGen_ChannelOff(pChan->fgenChan);
            DoAudioJobList(pChan);
            return(true);   // Indicate event processing complete.

        // We are receiving user audio
        case EV_AUD_RXAON:
            FGen_ChannelOff(pChan->fgenChan);
            Apex_StopTmr(&pChan->gpTimer);
            Transition_Rxon(pChan);     // Wait for start.
            return(true);   // Indicate event processing complete.

        default:
            return(false);  // Do default processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Generate CW ID Audio

static void Transition_CWID(struct TAG_AUD_CHAN *pChan)
{
    SET_STATE(State_CWID, 'H');
}

static bool State_CWID(enum TAG_AUD_EVIDS event,
                struct TAG_AUD_CHAN *pChan, uintptr_t unused)
{
    switch(event)
    {
        // If the ID is complete, check for more jobs to do.
        // The job-list check must come before telling
        // Central that the ID is complete, because the
        // audio-complete signal is sent by the job list
        // when there is no more audio to send, and this
        // must occur before we say the ID was performed,
        // otherwise the ID'er will give continuous IDs
        // every 10 minutes, thinking that there was
        // more audio after the ID completed.

        case EV_AUD_MORSECP:
            DoAudioJobList(pChan);      // Will set a new state.
            EMIT_SIGNAL(AUD_IDAudioComplete, pChan->txChanID, 0);
            return(true);   // Indicate event processing complete.

        // Here, go to a special state while repeating user audio
        // and ID at the same time.
        case EV_AUD_RXAON:
            Transition_UserAudioOverID(pChan);
            return(true);   // Indicate event processing complete.

        default:
            return(false);  // Do default processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Here, we already started doing an ID and then the
    // user keyed up.

static void Transition_UserAudioOverID(
                                struct TAG_AUD_CHAN *pChan)
{
    SET_STATE(State_UserAudioOverID, 'J');
}

static bool State_UserAudioOverID(enum TAG_AUD_EVIDS event,
                    struct TAG_AUD_CHAN *pChan, uintptr_t unused)
{
    switch(event)
    {
        // If the ID ends and the user is still talking,
        // go to the Rxon state and wait for the user
        // to unkey.
        case EV_AUD_MORSECP:
            EMIT_SIGNAL(
                AUD_MessageAudioComplete, pChan->txChanID, 0);
            EMIT_SIGNAL(AUD_IDAudioComplete, pChan->txChanID, 0);
            Transition_Rxon(pChan);
            return(true);   // Indicate event processing complete.

        // Go back to the normal CWID state if the user
        // audio stops.  Can increase amplitude to normal
        // too.
        case EV_AUD_START:
            Transition_CWID(pChan);
            return(true);   // Indicate event processing complete.

        default:
            return(false);  // Do default processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Check the audio job list and dispatch to the proper
    // state.

static void DoAudioJobList(struct TAG_AUD_CHAN *pChan)
{
    unsigned i;
    aud_morse_parms_t morseParms;

    // If an OK is scheduled, play it.
    if((pChan->flags & FLAG_DTMFOK) != 0)
    {
        pChan->flags &= ~FLAG_DTMFOK;       // Reset the flag.
        morseParms.pString = " OK ";
        morseParms.level = EEP_GetMorseLevel(pChan->txChanID);
        EMIT_SIGNAL(AUD_StartMorse,
            pChan->txChanID, (uintptr_t)&morseParms);

        Transition_MorseAudio(pChan);
    }

    // If an ERR is scheduled, play it.
    else if((pChan->flags & FLAG_DTMFERR) != 0)
    {
        pChan->flags &= ~FLAG_DTMFERR;  // Reset the flag.
        morseParms.pString = " ERR ";
        morseParms.level = EEP_GetMorseLevel(pChan->txChanID);
        EMIT_SIGNAL(AUD_StartMorse,
            pChan->txChanID, (uintptr_t)&morseParms);

        Transition_MorseAudio(pChan);
    }

    else if((pChan->flags & FLAG_PLAYFILE) != 0)
    {
        pChan->flags &= ~FLAG_PLAYFILE;
        Transition_PlayFile(pChan);
    }

    // Do an announcement
    // ALP: I'm changing this so that we try to play the
    // announcement even if timed out.  The transmitter
    // is muted, so it won't transmit anyway.  That way
    // the announcement system will continue to work
    // normally.  If the transmitter is timed out, the
    // announcement will simply appear to be skipped.

//    else if(((pChan->flags & FLAG_ANNO_MASK) != 0)
//        && !pChan->timedOut)
    else if((pChan->flags & FLAG_ANNO_MASK) != 0)
    {
        for(i=0; i<MAX_ANNO; i++)
        {
            if((pChan->flags & FLAG_ANNO(i)) != 0)
            {
                pChan->flags &= ~FLAG_ANNO(i);
                pChan->currentAnno = (uint8_t)i;
                Transition_Announce(pChan);
                break;
            }
        }
    }

    // Do a voice ID.

    else if((pChan->flags & FLAG_VID) != 0)
    {
        pChan->flags &= ~FLAG_VID;       // Reset the voice flag.
        Transition_VoiceID(pChan);

        // Tell Central we are putting out ID audio.
        EMIT_SIGNAL(AUD_IDAudioStarted, pChan->txChanID, 0);
    }

    // Do a CW ID.

    else if((pChan->flags & FLAG_CWID) != 0)
    {
        pChan->flags &= ~FLAG_CWID;       // Reset the flag.

        morseParms.pString =
            EEP_GetIDString(pChan->txChanID);
        morseParms.level =
            EEP_GetMorseLevel(pChan->txChanID);
        EMIT_SIGNAL(AUD_StartMorse,
            pChan->txChanID, (uintptr_t)&morseParms);
        Transition_CWID(pChan);

        // Tell Central we are putting out ID audio.
        EMIT_SIGNAL(AUD_IDAudioStarted, pChan->txChanID, 0);
    }

    // If we get here, all audio jobs are done.
    else
    {
        EMIT_SIGNAL(
            AUD_MessageAudioComplete, pChan->txChanID, 0);

        if(pChan->userAudio)
            Transition_Rxon(pChan);
        else
            Transition_Idle(pChan);
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                      END
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
