/******************************************************************//**
    @file central.c

    PSE-702 central control module.  Integrates all other modules
    and maintains the top level state.

    @author ALP

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

    @defgroup central Central Control Module

**********************************************************************/
///@{

//      *** INCLUDE FILES ***

#include <stdint.h>
#include <stdbool.h>

#include "lpc213x.h"
#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"

#include "apex.h"
#include "app.h"
#include "console.h"
#include "io_map.h"
#include "io_codec.h"
#include "astream.h"
#include "io_misc.h"
#include "filesys.h"
#include "dtmfdet.h"
#include "fgen.h"
#include "eep.h"
#include "audiosm.h"
#include "ssm_squelch.h"
#include "ssm_morse.h"
#include "ssm_txsm.h"
#include "ssm_idsm.h"
#include "ssm_satd.h"
#include "ssm_sate.h"
#include "central.h"
#include "new_dusi.h"
#include "xmodem.h"
#include "susi.h"
#include "voice.h"
#include "filter.h"

#undef APEX_H               // This prevents the error msg
#include "simplesm.h"

#include "io_debug.h"   //lint -esym(766, io_debug.h)
#include "apprint.h"    //lint -esym(766, apprint.h)
#include "io_main.h"    //lint -esym(766, io_main.h)

//***************************************************************
//                       *** MACROS ****
//***************************************************************
//      *** MACROS ****

//lint -esym(750, PRINT_RX_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, PRINT_TX_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, PRINT_SQ_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, RX_STATE_CHAN) (Macro not always referenced)
//lint -esym(750, TX_STATE_CHAN) (Macro not always referenced)
//lint -esym(750, SQ_STATE_CHAN) (Macro not always referenced)

#define PRINT_RX_STATE_CHAR(x)  \
        if(RX_STATE_CHAN == pChan->chanID) IO_DBPrintChar(x)

#if ((TP_CEN_RX_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
  || (TP_CEN_RX_STATE_FLAGS == STATE_FLAGS_LOWER_CH1))
#   define RX_STATE_CHAN  RX_CHAN_1
#else
#   define RX_STATE_CHAN  RX_CHAN_2
#endif

#if (TP_CEN_RX_STATE_FLAGS == STATE_FLAGS_OFF)
#   define SET_RX_STATE(X,Y) pChan->pRxState = (X)

#elif ((TP_CEN_RX_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
    || (TP_CEN_RX_STATE_FLAGS == STATE_FLAGS_UPPER_CH2))
#   define SET_RX_STATE(X,Y)  pChan->pRxState = (X); \
        PRINT_RX_STATE_CHAR((Y)+(TP_CEN_RX_STATE_OFFSET))

#elif ((TP_CEN_RX_STATE_FLAGS == STATE_FLAGS_LOWER_CH1) \
    || (TP_CEN_RX_STATE_FLAGS == STATE_FLAGS_LOWER_CH2))
#   define SET_RX_STATE(X,Y)  pChan->pRxState = (X);
        PRINT_RX_STATE_CHAR(((Y)|0x20)+(TP_CEN_RX_STATE_OFFSET))
#else
#error "TP_CEN_RX_STATE_FLAGS improperly defined."
#endif

#define PRINT_SQ_STATE_CHAR(x)  \
        if(SQ_STATE_CHAN == pChan->chanID) IO_DBPrintChar(x)

#if ((TP_CEN_SQ_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
  || (TP_CEN_SQ_STATE_FLAGS == STATE_FLAGS_LOWER_CH1))
#   define SQ_STATE_CHAN  RX_CHAN_1
#else
#   define SQ_STATE_CHAN  RX_CHAN_2
#endif

#if (TP_CEN_SQ_STATE_FLAGS == STATE_FLAGS_OFF)
#   define SET_SQ_STATE(X,Y)   pChan->pSqState = (X);

#elif ((TP_CEN_SQ_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
    || (TP_CEN_SQ_STATE_FLAGS == STATE_FLAGS_UPPER_CH2))
#   define SET_SQ_STATE(X,Y)  pChan->pSqState = (X); \
        PRINT_SQ_STATE_CHAR((Y)+(TP_CEN_SQ_STATE_OFFSET))


#elif ((TP_CEN_SQ_STATE_FLAGS == STATE_FLAGS_LOWER_CH1) \
    || (TP_CEN_SQ_STATE_FLAGS == STATE_FLAGS_LOWER_CH2))
#   define SET_SQ_STATE(X,Y)  pChan->pSqState = (X); \
        PRINT_SQ_STATE_CHAR(((Y)|0x20)+(TP_CEN_SQ_STATE_OFFSET))
#else
#error "TP_CEN_SQ_STATE_FLAGS improperly defined."
#endif

#define PRINT_TX_STATE_CHAR(x)  \
        if(TX_STATE_CHAN == pChan->chanID) IO_DBPrintChar(x)

#if ((TP_CEN_TX_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
  || (TP_CEN_TX_STATE_FLAGS == STATE_FLAGS_LOWER_CH1))
#   define TX_STATE_CHAN  TX_CHAN_1
#else
#   define TX_STATE_CHAN  TX_CHAN_2
#endif

#if (TP_CEN_TX_STATE_FLAGS == STATE_FLAGS_OFF)
#   define SET_TX_STATE(X,Y)   pChan->pTxState = (X);

#elif ((TP_CEN_TX_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
    || (TP_CEN_TX_STATE_FLAGS == STATE_FLAGS_UPPER_CH2))
#   define SET_TX_STATE(X,Y) pChan->pTxState = (X); \
        PRINT_TX_STATE_CHAR((Y)+(TP_CEN_TX_STATE_OFFSET))

#elif ((TP_CEN_TX_STATE_FLAGS == STATE_FLAGS_LOWER_CH1) \
    || (TP_CEN_TX_STATE_FLAGS == STATE_FLAGS_LOWER_CH2))
#   define SET_TX_STATE(X,Y) pChan->pTxState = (X); \
        PRINT_TX_STATE_CHAR(((Y)|0x20)+(TP_CEN_TX_STATE_OFFSET))
#else
#error "TP_CEN_TX_STATE_FLAGS improperly defined."
#endif

//***************************************************************
//           *** MODULE ID DEFINITIONS ****
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Stack Size: Measured use of 336 bytes on 3/13/12.

#define Z_STK_CEN       420     ///< Stack size in bytes

    // State Machine Event IDs:

enum TAG_CEN_EVIDS
{
    EV_CEN_NULL = CEN_ID(0),    ///< Used to return a null event.

    // The following events are associated with a receiver channel
    // number:

    EV_CEN_CARRIER_SQU_CLOSED,  ///< A channel's carrier squelch closed.
    EV_CEN_CARRIER_SQU_OPENED,  ///< A channel's carrier squelch opened.

    EV_CEN_TONE_SQU_OPENED,     ///< We locked onto a CTCSS tone
    EV_CEN_TONE_SQU_CLOSED,     ///< We lost tone lock.

    EV_CEN_RX_SQUELCHED,        ///< Receiver is fully squelched.
    EV_CEN_RX_UNSQUELCHED,      ///< Receiver is fully unsquelched.

    EV_CEN_RX_ENABLE,           ///< Unmute a receiver.
    EV_CEN_RX_DISABLE,          ///< Mute a receiver.

    EV_CEN_DTMF_PRESS,          ///< DTMF digit pressed.

    // The following events are associated with a transmitter channel
    // number:

    EV_CEN_SATE_ON,             ///< Say CTCSS encoder is on
    EV_CEN_SATE_OFF,            ///< Say CTCSS encoder is off

    EV_CEN_DOID,                ///< Schedule an ID.

    EV_CEN_RECORD,              ///< Command to record a voice file
    EV_CEN_RECORD_TIMER,        ///< The arm-record timer timed out.
    EV_CEN_CANCEL_RECORD,       ///< Cancel a recording in progress.
    EV_CEN_PLAYBACK,            ///< Command to playback a voice file
    EV_CEN_RECORD_FINI,         ///< Signal voice record complete.
    EV_CEN_PLAYBACK_FINI,       ///< Signal voice playback complete.

    EV_CEN_DUSI_ERR,            ///< A DUSI command entered in error.
    EV_CEN_DUSI_OK,             ///< A DUSI command was entered OK.

    EV_CEN_AUDIO_CP,            ///< Say AUD completed its tasks.

    EV_CEN_SET_USER_CONTEXT,    ///< Set operation as repeater
    EV_CEN_SET_TEST_CONTEXT,    ///< Set operation for testing

    // Transmitter related events:

    EV_CEN_TX_GENAUDIO_STARTED, ///< Transmitting generated audio
    EV_CEN_TX_MSGAUDIO_STARTED, ///< Transmitting generated audio
    EV_CEN_TX_UAUDIO_STARTED,   ///< Transmitting user audio.
    EV_CEN_TX_UAUDIO_STOPPED,   ///< Stopped transmitting user audio.

    EV_CEN_TX_PATH_UNASSIGNED,  ///< Tx was unassigned an audio stream.
    EV_CEN_TX_DO_ID,            ///< Time to do an ID.
    EV_CEN_TX_ID_COMPLETE,      ///< AUD signaled the ID is complete.
    EV_CEN_TX_DISABLE,          ///< Disable the transmitter.
    EV_CEN_TX_ENABLE,           ///< Enable the transmitter.
    EV_CEN_TX_DTMF,             ///< Tx will be transmitting DTMF.
    EV_CEN_TX_DTMF_OK,          ///< Tell the transmitter to send OK.
    EV_CEN_TX_DTMF_ERR,         ///< Tell the transmitter to send ERR.

    EV_CEN_TX_ON,               ///< Ask CEN to turn transmitter on
    EV_CEN_TX_OFF,              ///< Say it's ok to turn tx off.
    EV_CEN_TX_TIMEOUT,          ///< A transmitter timed out.

    EV_CEN_TEST_TONE_REQ,       ///< Request a test tone
    EV_CEN_TEST_TONE_OFF,       ///< Finished test tone.

    EV_CEN_WDTIME,              ///< Time to refresh watchdog.

    EV_CEN_MAX_EVENT            ///< Largest event number.
};


    /// @name Timeout Definitions:
///@{
    /// Max time to start recording
#define CEN_TIME_ARM_RECORD   (TU_1SEC*10)

    /// Delay when starting up a Tx before starting tone
    /// audio or announcement.
//#define CEN_TIME_TX_STARTUP   (TU_1MS*500)

#if WATCHDOG_ENABLE != 0
    /// Watchdog refresh rate.  The watchdog timer times
    /// out in 100 ms so we must refresh in less than that.
    /// I tried 90 ms but got a couple resets, presumably
    /// because it took longer than 10 ms to service the
    /// higher priority tasks.
    /// (ALP 11-24-2012 Ver 2.3 code)
#define CEN_TIME_WDOG_REFRESH   (TU_1MS*80) // Timeout in 100 ms.
#endif
///@}
//***************************************************************
//              *** DATA STRUCTURE DEFINITIONS ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


//***************************************************************
//                     *** VARIABLES ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static TCB cenTask;
static uint32_t cenStack[Z_STK_CEN/sizeof(uint32_t)];
static MBOX mailbox;

#if WATCHDOG_ENABLE != 0
static TMR wdRefreshTmr;    // Watchdog refresh timer
#endif
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Receiver channels: Each channel is represented by
    /// a structure:

static struct TAG_CEN_RX_CHAN
{
    enum TAG_RX_CHAN chanID;    ///< Must be initialized at reset.
    astream_srcsig_t srcDev;     ///< Initialized at reset.

    /// Squelch state:
    void (*pSqState)(enum TAG_CEN_EVIDS event,
        struct TAG_CEN_RX_CHAN *pRxChan, uint32_t data);

    /// Receiver state:
    void (*pRxState)(enum TAG_CEN_EVIDS event,
        struct TAG_CEN_RX_CHAN *pRxChan, uint32_t data);

    uint16_t pathID;    ///< ID of paths currently in use for this Rx

} rxChans[MAX_RXCHAN];


    /// Transmitter Channel Structure

static struct TAG_CEN_TX_CHAN
{
    enum TAG_TX_CHAN chanID;    ///< Must be initialized at reset.
    enum TAG_VOI_CHAN vidChan;  ///< Voice ID channel assigned this Tx

    astream_destsig_t destDev;  ///< Initialized at reset.

    /// Transmitter state:
    void (*pTxState)(enum TAG_CEN_EVIDS event,
        struct TAG_CEN_TX_CHAN *pTxChan, uint32_t data);

//    TMR txDelayTimer;   ///< Times Tx startup delay.

} txChans[MAX_TXCHAN];

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static enum TAG_RX_CHAN dusiChan;

    // Pointer to the channel on which we are recording a
    // voice file, or NULL if we are not recording.
static struct TAG_CEN_RX_CHAN *pRecordingChan;
static unsigned recordFileNum;  ///< File number to which we are
                                  ///< recording.

static TMR recordTimer;    ///< If this times out, cancel the recording.

//***************************************************************
//                  *** CONSTANTS ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Debugging Table for Variable Display

const struct TAG_VARIABLE_TABLE cenVarTab[] =
{
    {"stack", cenStack, VAR_TYPE_NONE, sizeof(cenStack)},
    {"tcb", &cenTask, VAR_TYPE_NONE, sizeof(cenTask)},
    {"rx1state", &rxChans[RX_CHAN_1].pRxState,
                                VAR_TYPE_HEXINT, sizeof(void *)},
    {"rx2state", &rxChans[RX_CHAN_2].pRxState,
                                VAR_TYPE_HEXINT, sizeof(void *)},
    {"sq1state", &rxChans[RX_CHAN_1].pSqState,
                                VAR_TYPE_HEXINT, sizeof(void *)},
    {"sq2state", &rxChans[RX_CHAN_2].pSqState,
                                VAR_TYPE_HEXINT, sizeof(void *)},
    {"tx1state", &txChans[TX_CHAN_1].pTxState,
                                VAR_TYPE_HEXINT, sizeof(void *)},
    {"tx2state", &txChans[TX_CHAN_2].pTxState,
                                VAR_TYPE_HEXINT, sizeof(void *)},
    {NULL, NULL, VAR_TYPE_NONE, 0}
};

//***************************************************************
//                      *** SLOTS ****
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static slot_t Slot_DTMF_Digit;      // DTMF digit pressed signal
static slot_t Slot_DTMF_NotDigit;   // DTMF digit released signal
static slot_t Slot_DTMF_RMS2;       // DTMF signal mag squared.

static slot_t Slot_SATD_Locked;
static slot_t Slot_SATD_Unlocked;

static slot_t Slot_Voice_RecordComplete;    // Voice recording complete
static slot_t Slot_Voice_PlaybackComplete;  // Voice playback complete

static slot_t Slot_SquelchClosed;
static slot_t Slot_SquelchOpened;

static slot_t Slot_AUD_MessageAudioStarted;
static slot_t Slot_AUD_MessageAudioComplete;
static slot_t Slot_AUD_IDAudioStarted;
static slot_t Slot_AUD_IDAudioComplete;
static slot_t Slot_AUD_StartMorse;
//static slot_t Slot_AUD_GeneratedAudioOff;
static slot_t Slot_AUD_VoiceID;
static slot_t Slot_AUD_CancelPlayback;
static slot_t Slot_AUD_Announce;

static slot_t Slot_ASTREAM_DestMessage;
static slot_t Slot_ASTREAM_DestUnassigned;
static slot_t Slot_ASTREAM_SrcPreempted;

static slot_t Slot_DUSI_OK;
static slot_t Slot_DUSI_ERR;
static slot_t Slot_DUSI_ArmRecorder;

static slot_t Slot_SATE_TurnedOn;
static slot_t Slot_SATE_TurnedOff;

static slot_t Slot_MORSE_Complete;

static slot_t Slot_IDSM_ScheduleID;
static slot_t Slot_IDSM_ScheduleOpeningID;

static slot_t Slot_TXSM_TxOn;
static slot_t Slot_TXSM_TxOff;
static slot_t Slot_TXSM_TxTimeout;
static slot_t Slot_TXSM_StartGeneratedAudio;
//static slot_t Slot_TXSM_ScheduleCB;
static slot_t Slot_TXSM_TxComplete;

//***************************************************************
//             *** STATIC FUNCTION PROTOTYPES ****
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void CentralTask(void);

//static void NotifyTransmitterOn(enum TAG_TX_CHAN chan);

//static bool SetCrosspointForRxChan(enum TAG_RX_CHAN rxChan);
//static void ResetCrosspoint(enum TAG_XPOINT_CODE xpoint);

//* * * * * * * * * Top Level Event Decoder * * * * * * * * * * * *

static void DoTopStateMachine(
    enum TAG_CEN_EVIDS event, uintptr_t chan, uintptr_t data);


//* * * * * * *  Transmitter State Machine  * * * * * * * * * * * *

static void TxTransition_Idle(struct TAG_CEN_TX_CHAN *pChan); // A
static void TxState_Idle(enum TAG_CEN_EVIDS event,
                struct TAG_CEN_TX_CHAN *pChan, uint32_t unused);
static void
        TxTransition_Repeating(struct TAG_CEN_TX_CHAN *pChan); // B
static void TxState_Repeating(enum TAG_CEN_EVIDS event,
                struct TAG_CEN_TX_CHAN *pChan, uint32_t unused);
static void
        TxTransition_Dtmf(struct TAG_CEN_TX_CHAN *pChan);       // C
static void TxState_Dtmf(enum TAG_CEN_EVIDS event,
                struct TAG_CEN_TX_CHAN *pChan, uint32_t unused);
static void
        TxTransition_TestTone(struct TAG_CEN_TX_CHAN *pChan);  // D
static void TxState_TestTone(enum TAG_CEN_EVIDS event,
                struct TAG_CEN_TX_CHAN *pChan, uint32_t unused);

static void TxTransition_GenAudio(
                                struct TAG_CEN_TX_CHAN *pChan);   // E
static void TxState_GenAudio(enum TAG_CEN_EVIDS event,
                struct TAG_CEN_TX_CHAN *pChan, uint32_t unused);

static void TxTransition_Disabled(
                                struct TAG_CEN_TX_CHAN *pChan);   // F
static void TxState_Disabled(enum TAG_CEN_EVIDS event,
                    struct TAG_CEN_TX_CHAN *pChan, uint32_t unused);

static void TxTransition_Timeout(
                                struct TAG_CEN_TX_CHAN *pChan);   // G
static void TxState_Timeout(enum TAG_CEN_EVIDS event,
                    struct TAG_CEN_TX_CHAN *pChan, uint32_t unused);

static void TxTransition_Wait4TxOff(
                                struct TAG_CEN_TX_CHAN *pChan);   // H
static void TxState_Wait4TxOff(enum TAG_CEN_EVIDS event,
                    struct TAG_CEN_TX_CHAN *pChan, uint32_t unused);

static void TxTransition_DisableWaitForID(
                                struct TAG_CEN_TX_CHAN *pChan);   // I
static void
    TxState_DisableWaitForID(enum TAG_CEN_EVIDS event,
                struct TAG_CEN_TX_CHAN *pChan, uint32_t unused);

//* * * * * * * * * Squelch State Machines  * * * * * * * * * * * *

static void SquelchTransition_Closed(
                            struct TAG_CEN_RX_CHAN *pChan);  // A
static void SquelchState_Closed(enum TAG_CEN_EVIDS event,
    struct TAG_CEN_RX_CHAN *pChan, uint32_t unused);

static void SquelchTransition_CarrierUnsquelched(
                              struct TAG_CEN_RX_CHAN *pChan); // B
static void
    SquelchState_CarrierUnsquelched(enum TAG_CEN_EVIDS event,
                    struct TAG_CEN_RX_CHAN *pChan, uint32_t unused);

static void SquelchTransition_Opened(
                              struct TAG_CEN_RX_CHAN *pChan); // C
static void
    SquelchState_Opened(enum TAG_CEN_EVIDS event,
                      struct TAG_CEN_RX_CHAN *pChan, uint32_t unused);

//* * * * * * * * * Receiver State Machines * * * * * * * * * * * *

static void RxTransition_Squelched(
                            struct TAG_CEN_RX_CHAN *pRxChan); // A
static void RxState_Squelched(enum TAG_CEN_EVIDS event,
    struct TAG_CEN_RX_CHAN *pRxChan, uint32_t data);

static void RxTransition_Unsquelched(
                        struct TAG_CEN_RX_CHAN *pChan);     // B
static void RxState_Unsquelched(enum TAG_CEN_EVIDS event,
    struct TAG_CEN_RX_CHAN *pChan, uint32_t data);

static void
    RxTransition_DUSI_Complete(
                            struct TAG_CEN_RX_CHAN *pChan);  // D
static void RxState_DUSI_Complete(enum TAG_CEN_EVIDS event,
    struct TAG_CEN_RX_CHAN *pChan, uint32_t data);

static void RxTransition_DTMF(
                            struct TAG_CEN_RX_CHAN *pRxChan); // C
static void RxState_DTMF(enum TAG_CEN_EVIDS event,
    struct TAG_CEN_RX_CHAN *pRxChan, uint32_t data);

static void RxTransition_RecorderArmed(
                            struct TAG_CEN_RX_CHAN *pChan); // E

static void RxState_RecorderArmed(enum TAG_CEN_EVIDS event,
    struct TAG_CEN_RX_CHAN *pChan, uint32_t unused);

static void
    RxTransition_Recording(struct TAG_CEN_RX_CHAN *pChan);  // F

static void RxState_Recording(enum TAG_CEN_EVIDS event,
    struct TAG_CEN_RX_CHAN *pChan, uint32_t unused);

static void
    RxTransition_Disabled(struct TAG_CEN_RX_CHAN *pChan);      // G
static void RxState_Disabled(enum TAG_CEN_EVIDS event,
            struct TAG_CEN_RX_CHAN *pChan, uint32_t unused);

//***************************************************************
//                       *** API ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Reset Initialization
    */

void CEN_Reset(void)
{
    Apex_CreateTask(&cenTask, TID_CEN, CentralTask,
        cenStack, (sizeof(cenStack)), TPRI_CEN);

    Apex_InitMailbox(&mailbox);

    Apex_InitMsgTimer(
            &recordTimer, &mailbox, EV_CEN_RECORD_TIMER, 0, 0);

#if WATCHDOG_ENABLE != 0
    // This timer refreshes the watchdog.
    Apex_InitMsgTimer(
            &wdRefreshTmr, &mailbox, EV_CEN_WDTIME, 0, 0);
#endif

    // Init the receiver channel structures:

    rxChans[RX_CHAN_1].chanID = RX_CHAN_1;
    rxChans[RX_CHAN_2].chanID = RX_CHAN_2;

    rxChans[RX_CHAN_1].srcDev = ASTREAM_SSIG_RX1;
    rxChans[RX_CHAN_2].srcDev = ASTREAM_SSIG_RX2;

    txChans[TX_CHAN_1].chanID = TX_CHAN_1;
    txChans[TX_CHAN_2].chanID = TX_CHAN_2;

    txChans[TX_CHAN_1].destDev = ASTREAM_DSIG_TX1;
    txChans[TX_CHAN_2].destDev = ASTREAM_DSIG_TX2;

    txChans[TX_CHAN_1].vidChan = VOI_CHAN_1;
    txChans[TX_CHAN_2].vidChan = VOI_CHAN_2;

    pRecordingChan = NULL;

}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void CEN_SetConsoleToUserContext(void)
{
    Apex_SendMsg(&mailbox, EV_CEN_SET_USER_CONTEXT);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void CEN_SetConsoleToTestContext(void)
{
    Apex_SendMsg(&mailbox, EV_CEN_SET_TEST_CONTEXT);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void CEN_RequestTestTone(void)
{
    Apex_SendMsg(&mailbox, EV_CEN_TEST_TONE_REQ);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void CEN_TestToneOff(void)
{
    Apex_SendMsg(&mailbox, EV_CEN_TEST_TONE_OFF);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void CEN_PlayID(enum TAG_TX_CHAN chan)
{
    AUD_ScheduleID(chan);
}

void CEN_PlayAnno(unsigned short index)
{
    AUD_PlayAnno(index);
}

void CEN_PlayFile(enum TAG_TX_CHAN chan, unsigned short fileNum)
{
    AUD_PlayFile(chan, fileNum);
}

    /** SUSI and DUSI call this to tell central to disable a
        transmitter.

        @param txChan: Transmitter channel number.
    */
void CEN_TxDisable(enum TAG_TX_CHAN txChan)
{
    Apex_SendMsgExt(&mailbox, EV_CEN_TX_DISABLE, txChan, 0);
}

    /** SUSI and DUSI call this to tell central to enable a
        transmitter.

        @param txChan: Transmitter channel number.
    */
void CEN_TxEnable(enum TAG_TX_CHAN txChan)
{
    Apex_SendMsgExt(&mailbox, EV_CEN_TX_ENABLE, txChan, 0);
}

/** SUSI and DUSI call this to tell central to mute or
    unmute a receiver.

    @param rxChan: Receiver channel number.
*/
void CEN_RxEnable(enum TAG_RX_CHAN rxChan, bool flag)
{
    if(flag)
        Apex_SendMsgExt(&mailbox, EV_CEN_RX_ENABLE, rxChan, 0);
    else
        Apex_SendMsgExt(&mailbox, EV_CEN_RX_DISABLE, rxChan, 0);
}

//***************************************************************
//                  *** SLOTS ***
//***************************************************************
//               DTMF Decoder Slots
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** A DTMF digit was pressed

        Tell CENTRAL that a DTMF digit was either pressed or
        released.  CENTRAL controls whether the DTMF decoder
        is listening for just a command prefix or all 16
        digits on either or both channels.

        @param chan This parameter contains the channel
        number of the DTMF channel on which the digit was
        detected.

        @param data This parameter contains the character
        code that was detected.
    */
static void Slot_DTMF_Digit(
                    uintptr_t chan, uintptr_t data)
{
    // The DTMF channel is the same as the Rx channel.  This
    // is enforced by SUSI/DUSI when the paths for DTMF are
    // entered.
    Apex_SendMsgExt(&mailbox, EV_CEN_DTMF_PRESS,
                    (uintptr_t)DTMF_CHAN_2_RX_CHAN(chan), data);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** A DTMF digit was released
    */
static void Slot_DTMF_NotDigit(
                    uintptr_t unused1, uintptr_t unused2)
{}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** DTMF system is reporting signal power.

        Param 1 is Rx channel.
        Param 2 is signal magnitude squared.
    */
static void Slot_DTMF_RMS2(
                    uintptr_t unused1, uintptr_t unused2)
{
//    IO_DBPrintDWord(magSquared);
//    IO_DB_CRLF();
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                   API for Squelch Task
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** The squelch state machine detected a channel on which
        squelch closed.
    */
static void Slot_SquelchClosed(
                            uintptr_t rxChan, uintptr_t unused2)
{
    Apex_SendMsgExt(
        &mailbox, EV_CEN_CARRIER_SQU_CLOSED, (uintptr_t)rxChan, 0);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** The squelch state machine detected a channel on which
        squelch opened.
    */
static void Slot_SquelchOpened(
                            uintptr_t rxChan, uintptr_t unused2)
{
    Apex_SendMsgExt(
        &mailbox, EV_CEN_CARRIER_SQU_OPENED, (uintptr_t)rxChan, 0);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                  CTCSS Decoder Slots
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** CTCSS decoder has locked onto a signal.
    */
static void Slot_SATD_Locked(
                            uintptr_t rxChan, uintptr_t unused2)
{
    Apex_SendMsgExt(
        &mailbox, EV_CEN_TONE_SQU_OPENED, rxChan, 0);
}

    /** CTCSS decoder has unlocked from a signal.
    */
static void Slot_SATD_Unlocked(
                            uintptr_t rxChan, uintptr_t unused2)
{
    Apex_SendMsgExt(
        &mailbox, EV_CEN_TONE_SQU_CLOSED, rxChan, 0);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                  CTCSS Encoder Interface
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // SATE informs central that the CTCSS encoder has
    // completed it's output.

static void Slot_SATE_TurnedOff(
                        uintptr_t txChan, uintptr_t unused2)
{
    TXSM_NotifySATE_TurnedOff((enum TAG_TX_CHAN)txChan);
}

// SATE informs central that the CTCSS encoder has
// completed it's output.

static void Slot_SATE_TurnedOn(
                        uintptr_t txChan, uintptr_t unused2)
{
    TXSM_NotifySATE_TurnedOn((enum TAG_TX_CHAN)txChan);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//              API for Transmitter State Machine
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // TXSM calls this to Tell central that a transmitter is
    // turning on.

static void Slot_TXSM_TxOn(uintptr_t txChan, uintptr_t unused2)
{
    // Turn CTCSS encoder on if necessary.
    if(EEP_GetTxFlags((enum TAG_TX_CHAN)txChan, EEP_TX_FLAG_SATE) &&
      (EEP_GetCTCSSEncodeMode(
                        (enum TAG_TX_CHAN)txChan) == SATE_MODE_PTT))
    {
        SATE_On((enum TAG_TX_CHAN)txChan);
    }
}

    // Tell central that TXSM is turning a transmitter off

static void Slot_TXSM_TxOff(uintptr_t txChan, uintptr_t unused2)
{
    Apex_SendMsgExt(&mailbox, EV_CEN_TX_OFF, txChan, 0);
}


static void Slot_TXSM_TxTimeout(
                        uintptr_t txChan, uintptr_t unused2)
{
    Apex_SendMsgExt(&mailbox, EV_CEN_TX_TIMEOUT, txChan, 0);
}

    // Txsm calls this to tell Central to start playing
    // generated audio on transmitter channel.

static void Slot_TXSM_StartGeneratedAudio(
                    uintptr_t txChan, uintptr_t unused2)
{
    AUD_StartAudioJobs((enum TAG_TX_CHAN)txChan);
}
#if 0
    // Txsm calls this to tell Central to schedule a courtesy
    // beep on a transmitter channel.

static void Slot_TXSM_ScheduleCB(
                        uintptr_t txChan, uintptr_t unused2)
{
    // If a courtesy beep is in progress, the audio task will
    // cancel it automatically.  Tell it to schedule a new one.
    AUD_ScheduleCB((enum TAG_TX_CHAN)txChan);
}
#endif

    // Txsm signals that all transmit jobs are complete and it
    // wants to turn the transmitter off, but it will wait for
    // SATE complete if there is a tone being output in case of
    // reverse burst.
static void Slot_TXSM_TxComplete(
                            uintptr_t txChan, uintptr_t unused2)
{
    // Turn CTCSS encoder off if necessary.
    if(EEP_GetTxFlags(
       (enum TAG_TX_CHAN)txChan, EEP_TX_FLAG_SATE) &&
       (EEP_GetCTCSSEncodeMode(
       (enum TAG_TX_CHAN)txChan) == SATE_MODE_PTT))
    {
        SATE_Off((enum TAG_TX_CHAN)txChan);
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                   API for ID Module
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Called by the ID state machine to schedule an ID,

static void Slot_IDSM_ScheduleID(
                            uintptr_t txChan, uintptr_t unused2)
{
    Apex_SendMsgExt(&mailbox, EV_CEN_TX_DO_ID, txChan, 0);
}

    // Called by ID state machine to schedule an opening ID.

static void Slot_IDSM_ScheduleOpeningID(
                            uintptr_t txChan, uintptr_t unused2)
{
    Apex_SendMsgExt(&mailbox, EV_CEN_TX_DO_ID, txChan, 0);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//              API to Morse Encoder Module
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Morse sends this to inform Central that morse code
    /// transmission has been completed.

static void Slot_MORSE_Complete(
                            uintptr_t txChan, uintptr_t unused2)

{
    AUD_NotifyMorseComplete((enum TAG_TX_CHAN)txChan);
}

#if 0
void CEN_MorseWhisper(enum TAG_TX_CHAN chan)
{
    Morse_Whisper(chan);
}

void CEN_MorseShout(enum TAG_TX_CHAN chan)
{
    Morse_Shout(chan);
}
#endif

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//              Audio State Machine Slots
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Audiosm emits this signal to tell Central that the audio
// state machine is generating audio that must be ID'ed.
// (That is, anything but an ID itself or a courtesy beep).

static void Slot_AUD_MessageAudioStarted(
                                uintptr_t txChan, uintptr_t unused2)
{
    Apex_SendMsgExt(&mailbox, EV_CEN_TX_MSGAUDIO_STARTED, txChan, 0);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Audiosm emits this signal to tell Central that all scheduled
    // ID'able audio jobs are complete.

static void Slot_AUD_MessageAudioComplete(
                                uintptr_t chan, uintptr_t unused2)
{
    TXSM_NotifyGeneratedAudioOff((enum TAG_TX_CHAN)chan);
    IDSM_NotifyAudioOff((enum TAG_TX_CHAN)chan);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Audiosm emits this signal to tell Central to start morse
    // output on the selected channel.

static void Slot_AUD_StartMorse(uintptr_t chan, uintptr_t data)
{
    aud_morse_parms_t *pMorseParms = (aud_morse_parms_t *)data;

    Morse_Encode((enum TAG_TX_CHAN)chan,
                pMorseParms->level, pMorseParms->pString);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Audiosm emits this signal to tell Central that it
    // has started an ID.

static void Slot_AUD_IDAudioStarted(
                            uintptr_t chan, uintptr_t unused2)
{
    TXSM_NotifyGeneratedAudioOn((enum TAG_TX_CHAN)chan);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Audiosm calls this to tell Central that it
    // has completed an ID.

static void Slot_AUD_IDAudioComplete(
                            uintptr_t chan, uintptr_t unused2)
{
    IDSM_NotifyDoneID((enum TAG_TX_CHAN)chan);
    Apex_SendMsgExt(&mailbox, EV_CEN_TX_ID_COMPLETE, chan, 0);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Audiosm emits this signal to tell Central that it wants the
    // voice playback unit to start.

    // chan is an enum TAG_TX_CHAN.

static void Slot_AUD_VoiceID(uintptr_t chan, uintptr_t unused)
{
    VOI_PlaybackFile(txChans[chan].vidChan,
                    EEP_GetVidFileNum((enum TAG_TX_CHAN)chan));
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Audiosm emits this signal to tell Central that it wants the
    // voice playback unit to stop.

static void Slot_AUD_CancelPlayback(
                                uintptr_t chan, uintptr_t unused)
{
    VOI_StopPlayback(txChans[chan].vidChan);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Audiosm emits this signal to tell Central that it wants the
    // voice playback unit to play an announcement.

static void Slot_AUD_Announce(
                                uintptr_t chan, uintptr_t file)
{
    VOI_PlaybackFile(txChans[chan].vidChan, file);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//              ASTREAM Slots
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // ASTREAM is telling us that there is a message for
    // one of the transmitters.

static void Slot_ASTREAM_DestMessage(
                            uintptr_t dest, uintptr_t msgPtr)
{
    astream_msg_t *pMsg = (astream_msg_t *) msgPtr;

    switch((astream_destsig_t)dest)
    {
        case ASTREAM_DSIG_TX1:
            Apex_SendMsgExt(&mailbox,
                    pMsg->event, (uintptr_t)TX_CHAN_1, pMsg->data1);
            break;

        case ASTREAM_DSIG_TX2:
            Apex_SendMsgExt(&mailbox,
                    pMsg->event, (uintptr_t)TX_CHAN_2, pMsg->data1);
            break;

        case ASTREAM_DSIG_VOICE1:
        case ASTREAM_DSIG_VOICE2:
            break;

        default:
            FATAL();
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // ASTREAM is telling us that a transmitter is no longer
    // being used.

static void Slot_ASTREAM_DestUnassigned(
                            uintptr_t dest, uintptr_t unused2)
{
    switch((astream_destsig_t)dest)
    {
        case ASTREAM_DSIG_TX1:
            Apex_SendMsgExt(&mailbox,
                EV_CEN_TX_PATH_UNASSIGNED, (uintptr_t)TX_CHAN_1, 0);
            break;

        case ASTREAM_DSIG_TX2:
            Apex_SendMsgExt(&mailbox,
                EV_CEN_TX_PATH_UNASSIGNED, (uintptr_t)TX_CHAN_2, 0);
            break;

        case ASTREAM_DSIG_VOICE1:
        case ASTREAM_DSIG_VOICE2:
            break;

        default:
            FATAL();
    }
}


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // ASTREAM is telling us that a signal source has been
    // preempted.

static void Slot_ASTREAM_SrcPreempted(
                            uintptr_t src, uintptr_t unused)
{
    switch((astream_srcsig_t)src)
    {
        case ASTREAM_SSIG_RX1:
        case ASTREAM_SSIG_RX2:
            break;

        case ASTREAM_SSIG_VOICE1:
            AUD_NotifySourcePreempted(TX_CHAN_1);
            break;

        case ASTREAM_SSIG_VOICE2:
            AUD_NotifySourcePreempted(TX_CHAN_2);
            break;

        default:
            FATAL();
    }
}



//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//              DUSI Module Slots
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // DUSI calls this to tell Central to schedule an OK
    // response upon completion of a DTMF command.

static void Slot_DUSI_OK(uintptr_t unused1, uintptr_t unused2)
{
    Apex_SendMsgExt(&mailbox, EV_CEN_DUSI_OK, dusiChan, 0);
}

    // DUSI calls this to tell Central to schedule an ERR
    // response upon completion of a DTMF command.

static void Slot_DUSI_ERR(uintptr_t unused1, uintptr_t unused2)
{
    Apex_SendMsgExt(&mailbox, EV_CEN_DUSI_ERR, dusiChan, 0);
}

// DUSI calls this to tell Central to schedule an OK
// response upon completion of a DTMF command.

static void Slot_DUSI_ArmRecorder(
                        uintptr_t unused1, uintptr_t file)
{
    Apex_SendMsgExt(&mailbox, EV_CEN_RECORD, dusiChan, file);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//          Voice Store-Playback API (Called from SUSI)
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Inerface functions called externally to preset Central
    // to capture an audio file, or to play one back.

void CEN_RecordAudioFile(
                        enum TAG_RX_CHAN rxChan, unsigned file)
{
    Apex_SendMsgExt(&mailbox, EV_CEN_RECORD,
                        (uintptr_t)rxChan, (uintptr_t)file);
}

void CEN_CancelRecording(void)
{
    Apex_SendMsg(&mailbox, EV_CEN_CANCEL_RECORD);
}

#if 0

void CEN_PlaybackAudioFile(
                        enum TAG_TX_CHAN txChan, unsigned file)
{
    Apex_SendMsgExt(&mailbox, EV_CEN_PLAYBACK,
        (uintptr_t)txChan, (uintptr_t)file);
}

void CEN_CancelPlayback(void)
{
    VOI_StopPlayback();
}
#endif

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Voice module completed record command.

        @param slotParm1 This parameter contains the channel
        number of the channel that just completed voice
        recording.
        @param slotParm2 This parameter contains the status
        of the record operation:  True means the operation
        completed OK.  False means there was an overflow because
        we ran out of flash during the recording.
    */
static void Slot_Voice_RecordComplete(
        uintptr_t slotParm1, uintptr_t slotParm2)
{
    Apex_SendMsgExt(
            &mailbox, EV_CEN_RECORD_FINI, slotParm1, slotParm2);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Voice module completed playback command.

        @param chan This parameter contains the channel
        number of the channel that just completed voice
        playback.
        @param unused2 Parameter not used.
    */
static void Slot_Voice_PlaybackComplete(
                            uintptr_t chan, uintptr_t unused2)
{
    AUD_NotifyFlashComplete((enum TAG_TX_CHAN)chan);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Issue all subscriptions for Central.
     *
     */
void CEN_SubscribeAll(void)
{
    // Subscribe to DTMF detector module signals:
    Subscribe_DTMF_Signal_Digit(Slot_DTMF_Digit);
    Subscribe_DTMF_Signal_NotDigit(Slot_DTMF_NotDigit);
    Subscribe_DTMF_Signal_RMS2(Slot_DTMF_RMS2);

    // Subscribe to Voice module signals
    Subscribe_VOI_Signal_RecordComplete(Slot_Voice_RecordComplete);
    Subscribe_VOI_Signal_PlaybackComplete(Slot_Voice_PlaybackComplete);

    // Subscribe to Audio Task signals
    Subscribe_AUD_StartMorse(Slot_AUD_StartMorse);
    Subscribe_AUD_IDAudioStarted(Slot_AUD_IDAudioStarted);
    Subscribe_AUD_IDAudioComplete(Slot_AUD_IDAudioComplete);
    Subscribe_AUD_MessageAudioStarted(Slot_AUD_MessageAudioStarted);
    Subscribe_AUD_MessageAudioComplete(Slot_AUD_MessageAudioComplete);
    Subscribe_AUD_VoiceID(Slot_AUD_VoiceID);
    Subscribe_AUD_CancelPlayback(Slot_AUD_CancelPlayback);
    Subscribe_AUD_Announce(Slot_AUD_Announce);

    // Subscribe to Audio Stream signals

    Subscribe_ASTREAM_DestMessage(Slot_ASTREAM_DestMessage);
    Subscribe_ASTREAM_DestUnassigned(Slot_ASTREAM_DestUnassigned);
    Subscribe_ASTREAM_SrcPreempted(Slot_ASTREAM_SrcPreempted);

    // Set the de-emphasis selection
    ASTREAM_SetDeemphasisFilterIndex(RX_CHAN_1,
                            EEP_GetDeemphasisFilterIndex(RX_CHAN_1));
    ASTREAM_SetDeemphasisFilterIndex(RX_CHAN_2,
                            EEP_GetDeemphasisFilterIndex(RX_CHAN_2));

    ASTREAM_SetCtcssRejectionFilterIndex(RX_CHAN_1,
                        EEP_GetCtcssRejectionFilterIndex(RX_CHAN_1));
    ASTREAM_SetCtcssRejectionFilterIndex(RX_CHAN_2,
                        EEP_GetCtcssRejectionFilterIndex(RX_CHAN_2));

    // Subscribe to Squelch Task signals
    Subscribe_SQU_SquelchClosed(Slot_SquelchClosed);
    Subscribe_SQU_SquelchOpened(Slot_SquelchOpened);

    // Subscribe to Subaudible Tone Decoder Signals:
    Subscribe_SATD_Locked(Slot_SATD_Locked);
    Subscribe_SATD_Unlocked(Slot_SATD_Unlocked);

    // Subscribe to DUSI Signals
    Subscribe_DUSI_CommandOK(Slot_DUSI_OK);
    Subscribe_DUSI_CommandERR(Slot_DUSI_ERR);
    Subscribe_DUSI_ArmRecorder(Slot_DUSI_ArmRecorder);

    // Subscribe to SATE Signals
    Subscribe_SATE_TurnedOff(Slot_SATE_TurnedOff);
    Subscribe_SATE_TurnedOn(Slot_SATE_TurnedOn);

    // Subscribe to Morse Generator Module
    Subscribe_MORSE_Complete(Slot_MORSE_Complete);

    /// Subscribe to ID State Machine Signals
    Subscribe_IDSM_ScheduleID(Slot_IDSM_ScheduleID);
    Subscribe_IDSM_ScheduleOpeningID(Slot_IDSM_ScheduleOpeningID);

    /// Subscribe to Transmit State Machine Signals
    Subscribe_TXSM_TxOn(Slot_TXSM_TxOn);
    Subscribe_TXSM_TxOff(Slot_TXSM_TxOff);
    Subscribe_TXSM_TxTimeout(Slot_TXSM_TxTimeout);
    Subscribe_TXSM_StartGeneratedAudio(Slot_TXSM_StartGeneratedAudio);
//    Subscribe_TXSM_ScheduleCB(Slot_TXSM_ScheduleCB);
    Subscribe_TXSM_TxComplete(Slot_TXSM_TxComplete);

    // Subscribe Tone Generator to its hardware:
    IO_PWM_Subscribe(PWM_CHAN_1, FGen_SourceChan0_ISR);
    IO_PWM_Subscribe(PWM_CHAN_2, FGen_SourceChan1_ISR);

    CDX_Subscribe(ASTREAM_Sample_ISR);
}


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void CEN_UnsubscribeAll(void)
{
    IO_PWM_Unsubscribe(PWM_CHAN_1);
    IO_PWM_Unsubscribe(PWM_CHAN_2);

    CDX_Unsubscribe();

    ASTREAM_RemoveAllPaths();
    ASTREAM_UnsubscribeAll();

    XMODM_UnsubscribeAllSignals();
    VOI_UnsubscribeAllSignals();
    DTMF_UnsubscribeAllSignals();
    DUSI_UnsubscribeAll();
    SQU_UnsubscribeAll();
    AUD_UnsubscribeAll();
    SATD_UnsubscribeAll();
    SATE_UnsubscribeAll();
    MORSE_UnsubscribeAll();
    IDSM_UnsubscribeAll();
    TXSM_UnsubscribeAll();
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                      STATIC FUNCTIONS
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Central Task
     */
static void CentralTask(void)
{
    unsigned event;
    uintptr_t chan, data;

    // Disable codec ISR to prevent it from giving an overflow
    // assert when debugging.  See makefile for definition.

    FILT_Start();       // Init the deemphasis filter.

#ifndef CODEC_ISR_DISABLE
    CDX_Start();        // Startup the codec.
#endif

    IO_PWM_Start();     // Startup PWM hardware.

    FSys_Start();       // Get file system running.

    EEP_Start();        // Read user parameters from flash.

    SSM_Start();

    // Start the console task before enabling UART interrupts
    // because the UART receiver interrupt could send a message
    // to the console.
    Con_Start();
    IO_EnableIntDrivenRxUART0();
    IO_EnableIntDrivenTxUART0();

    // Start the function generator.  These functions
    // must be called before SATD_Start();

    FGen_Channel_Init(0, (uint16_t)(PWM_SAMPLE_RATE_HZ+0.5));
    FGen_Channel_Init(1, (uint16_t)(PWM_SAMPLE_RATE_HZ+0.5));
    FGen_Channel_Init(2, (uint16_t)(CDX_SAMPLE_RATE_HZ+0.5));
    FGen_Channel_Init(3, (uint16_t)(CDX_SAMPLE_RATE_HZ+0.5));

    SUSI_Start();       // Susi makes central subscribe.

    DTMF_Start();
    DUSI_Start();
    VOI_Start();
    AUD_Start();
    SATD_Start();
    IDSM_Start();
    Morse_Start();
    SQU_Start();
    TXSM_Start();
    SATE_Start();

    // Set the Receiver and Transmitter Potentiometers

    IO_SetRx1Level(EEP_GetRxLevel(RX_CHAN_1));
    IO_SetRx2Level(EEP_GetRxLevel(RX_CHAN_2));
    ASTREAM_SetTx1Level(EEP_GetTxLevel(TX_CHAN_1));
    ASTREAM_SetTx2Level(EEP_GetTxLevel(TX_CHAN_2));

    CEN_SubscribeAll();     // Subscribe to other module's signals.

    // Turn off audio to the transmitters:

    SquelchTransition_Closed(&rxChans[RX_CHAN_1]);
    SquelchTransition_Closed(&rxChans[RX_CHAN_2]);

    RxTransition_Squelched(&rxChans[RX_CHAN_1]);
    RxTransition_Squelched(&rxChans[RX_CHAN_2]);

    TxTransition_Idle(&txChans[TX_CHAN_1]);
    TxTransition_Idle(&txChans[TX_CHAN_2]);

//    CEN_SetConsoleToUserContext();      // This also sets up
                                        // module subscriptions.
#if WATCHDOG_ENABLE != 0
    Apex_StartTmr(&wdRefreshTmr, CEN_TIME_WDOG_REFRESH);
#endif
    FOREVER
    {
        // If there is a message, read it.

        // Read the message
        Apex_PendMsg(&mailbox, &event, &chan, &data, 0);
        DoTopStateMachine((enum TAG_CEN_EVIDS)event, chan, data);
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** The top-level state machine evaluates events and decides
     * whether to process the event statelessly or to pass the
     * event to a state machine.
    */
static void DoTopStateMachine(
    enum TAG_CEN_EVIDS event, uintptr_t chan, uintptr_t data)
{
    struct TAG_CEN_RX_CHAN *pRxChan;
    struct TAG_CEN_TX_CHAN *pTxChan;

   /* Execute the receiver state machine for any of the events
     * below. (Do it this way because the 'chan' parameter
     * is only a valid Rx channel number if it's passed with one of
     * the following events):
    */
    switch(event)
    {
        // Refresh the watchdog timer.
#if WATCHDOG_ENABLE != 0
        case EV_CEN_WDTIME:
            DI();
            WDFEED = 0xAA;
            WDFEED = 0x55;
            EI();
            Apex_StartTmr(&wdRefreshTmr, CEN_TIME_WDOG_REFRESH);
            break;
#endif

        // Squelch state machine events:
        case EV_CEN_CARRIER_SQU_OPENED:
        case EV_CEN_TONE_SQU_OPENED:
        case EV_CEN_CARRIER_SQU_CLOSED:
        case EV_CEN_TONE_SQU_CLOSED:
            // Select the channel and execute the squelch
            // state machine associated with that channel.

            ASSERT(VALID_RX_CHAN(chan));
            pRxChan = &rxChans[(enum TAG_RX_CHAN)chan];

            // Run the squelch state machine.
            ASSERT(pRxChan->pSqState != NULL);
            (*pRxChan->pSqState)(event, pRxChan, data);
            break;

        // Receiver state machine events:
        case EV_CEN_RX_SQUELCHED:
        case EV_CEN_RX_UNSQUELCHED:
        case EV_CEN_DTMF_PRESS:
        case EV_CEN_DUSI_ERR:
        case EV_CEN_DUSI_OK:
        case EV_CEN_RECORD:
        case EV_CEN_RX_ENABLE:
        case EV_CEN_RX_DISABLE:

            // Select the channel and execute the receiver
            // state machine associated with that channel.

            ASSERT(VALID_RX_CHAN(chan));
            pRxChan = &rxChans[(enum TAG_RX_CHAN)chan];

            // Run the receiver state machine.
            ASSERT(pRxChan->pRxState != NULL);
            (*pRxChan->pRxState)(event, pRxChan, data);
            break;

        case EV_CEN_CANCEL_RECORD:
        case EV_CEN_RECORD_TIMER:
            ASSERT(pRecordingChan != NULL);
            (*pRecordingChan->pRxState)(event, pRecordingChan, data);
            break;

        // Transmitter state machine events
        // The following events are transmitter related.
        // The channel number is a transmitter channel.
        case EV_CEN_TX_TIMEOUT:
        case EV_CEN_TEST_TONE_REQ:
        case EV_CEN_TEST_TONE_OFF:
        case EV_CEN_TX_ON:
        case EV_CEN_TX_OFF:
        case EV_CEN_TX_DTMF:
        case EV_CEN_TX_MSGAUDIO_STARTED:
        case EV_CEN_TX_UAUDIO_STARTED:
        case EV_CEN_TX_UAUDIO_STOPPED:
        case EV_CEN_TX_PATH_UNASSIGNED:
        case EV_CEN_TX_DTMF_OK:
        case EV_CEN_TX_DTMF_ERR:
        case EV_CEN_TX_DISABLE:
        case EV_CEN_TX_ENABLE:
        case EV_CEN_TX_ID_COMPLETE:
            ASSERT(VALID_TX_CHAN(chan));
            pTxChan = &txChans[(enum TAG_TX_CHAN)chan];
            (*pTxChan->pTxState)(event, pTxChan, data);
            break;

        // The ID state machine says to do an ID.
        case EV_CEN_TX_DO_ID:
            AUD_ScheduleID((enum TAG_TX_CHAN)chan);
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Central Transmitter Control State Machine

    @dot
    digraph Central_Tx_State_Machine
    {
        ratio=1;
        size="10,10";
        node [shape = circle];

        State_Idle [label="State_Idle"];
        State_Repeating [label="State_Repeating"];
        State_DtmfMute [label="State_DtmfMute"];
        State_TestTone [label="State_TestTone"];
        State_GenAudio [label="State_GenAudio"];
        State_StartupDelay [label="State_StartupDelay"];
        State_Timeout [label="State_Timeout"];

        edge [fontsize=10];
        node [shape = point]; Reset

        Reset -> State_Idle [label="Reset (R)"];

        State_Idle -> State_Repeating [label="RUS"];
        State_Idle -> State_TestTone [label="Test Tone Req"];
        State_Idle -> State_StartupDelay [label="ID or Announcement"];

        State_Repeating -> State_Idle [label="~RUS" ];
        State_Repeating -> State_DtmfMute [label="DTMF Key Decoded" ];
        State_Repeating -> State_Timeout [label="Tx Timeout" ];

        State_DtmfMute -> State_StartupDelay [label="~RUS"];

        State_TestTone -> State_Idle [label="Test Tone Fini"];

        State_GenAudio -> State_Idle [label="Audio Fini"];
        State_GenAudio -> State_Repeating [label="RUS"];

        State_StartupDelay -> State_GenAudio [label="Delay Timeout"];

        State_Timeout -> State_Idle [label="~RUS" ];
    }
    @enddot
*/
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** (A) Transmitter Idle State
     */
static void TxTransition_Idle(struct TAG_CEN_TX_CHAN *pChan)
{
    // Set new state.
    SET_TX_STATE(TxState_Idle, 'A');
}

    /**
     */
static void TxState_Idle(enum TAG_CEN_EVIDS event,
                struct TAG_CEN_TX_CHAN *pChan, uint32_t unused)
{
    switch(event)
    {
        // Event from audiosm saying that it is generating audio
        // that must be ID'ed.
        case EV_CEN_TX_MSGAUDIO_STARTED:
            TXSM_NotifyGeneratedAudioOn(pChan->chanID);
            IDSM_NotifyAudioOn(pChan->chanID);
            TxTransition_GenAudio(pChan);
            break;

        // ID state machine says it's time to do an ID.  The
        // transmitter is not currently active so we need to
        // delay before starting the audio.
        case EV_CEN_TX_DO_ID:
            AUD_ScheduleID(pChan->chanID);
            TXSM_NotifyGeneratedAudioOn(pChan->chanID);
            TxTransition_GenAudio(pChan);
            break;

        // A receiver went unsquelched, and this transmitter has
        // been assigned.
        case EV_CEN_TX_UAUDIO_STARTED:
            TXSM_NotifyUserAudioOn(pChan->chanID);
            AUD_NotifyUserAudioOn(pChan->chanID);
            IDSM_NotifyAudioOn(pChan->chanID);

            // Turn CTCSS encoder on if necessary.
            if(EEP_GetTxFlags(pChan->chanID, EEP_TX_FLAG_SATE) &&
              (EEP_GetCTCSSEncodeMode(pChan->chanID) == SATE_MODE_RUS))
            {
                SATE_On(pChan->chanID);
            }

            TxTransition_Repeating(pChan);
            break;

        // User is requesting a test tone:
        case EV_CEN_TEST_TONE_REQ:
            SUSI_GrantTestTone();
            TxTransition_TestTone(pChan);
            break;

        // If timeout event received, go to timeout state.
        case EV_CEN_TX_TIMEOUT:
            TxTransition_Timeout(pChan);
            break;

        // Received a command to disable the transmitter.
        case EV_CEN_TX_DISABLE:
            Morse_Stop(pChan->chanID);
            TxTransition_DisableWaitForID(pChan);
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** (B) In this state the transmitter is repeating user audio.
     */
static void
        TxTransition_Repeating(struct TAG_CEN_TX_CHAN *pChan)
{
    // Set new state.
    SET_TX_STATE(TxState_Repeating, 'B');
}

    /**
     */
static void TxState_Repeating(enum TAG_CEN_EVIDS event,
                struct TAG_CEN_TX_CHAN *pChan, uint32_t data)
{
    switch(event)
    {
        // Here, user audio has stopped. Emit a courtesy
        // beep.
        case EV_CEN_TX_UAUDIO_STOPPED:
            // Only emit beep if it is enabled for this tx.
            if(EEP_TX_CB_ENABLED(pChan->chanID))
            {
                AUD_PlayCB(pChan->chanID,
                                (eep_courtsey_beep_t *)data);
            }
            break;

        // Here, all paths assigned to this transmitter
        // have been removed so shut down the transmitter.
        case EV_CEN_TX_PATH_UNASSIGNED:
            AUD_NotifyUserAudioOff(pChan->chanID);
            AUD_StartAudioJobs(pChan->chanID);
            TXSM_NotifyUserAudioOff(pChan->chanID);
            IDSM_NotifyAudioOff(pChan->chanID);

            // Turn CTCSS encoder off if necessary.
            if(EEP_GetTxFlags(pChan->chanID, EEP_TX_FLAG_SATE) &&
              (EEP_GetCTCSSEncodeMode(pChan->chanID) == SATE_MODE_RUS))
            {
                SATE_Off(pChan->chanID);
            }

            TxTransition_Wait4TxOff(pChan);
            break;

        // Here, the source device is telling us that it is receiving
        // DTMF codes, and we should mute if that is selected by
        // the user configuration.
        case EV_CEN_TX_DTMF:
            TxTransition_Dtmf(pChan);
            break;

        // If timeout event received, go to timeout state.
        case EV_CEN_TX_TIMEOUT:
            TxTransition_Timeout(pChan);
            break;

        // Received a command to disable the transmitter.
        case EV_CEN_TX_DISABLE:
            TxTransition_DisableWaitForID(pChan);
            break;

        // User is requesting a test tone:
        case EV_CEN_TEST_TONE_REQ:
            SUSI_DenyTestTone();
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** (C) In this state the transmitter is on but the audio may
     * be muted because the audio source receiver is receiving
     * DTMF tones.
     */
static void
        TxTransition_Dtmf(struct TAG_CEN_TX_CHAN *pChan)
{
    if(EEP_GetTxFlags(pChan->chanID, EEP_TX_FLAG_DMUTE))
    {
        ASTREAM_MuteDestDev(pChan->destDev);
    }

    // Set new state.
    SET_TX_STATE(TxState_Dtmf, 'C');
}

    /**
     */
static void TxState_Dtmf(enum TAG_CEN_EVIDS event,
                struct TAG_CEN_TX_CHAN *pChan, uint32_t unused)
{
    switch(event)
    {
        // Here, the user unkeyed.
        case EV_CEN_TX_DTMF_OK:
            TXSM_NotifyUserAudioOff(pChan->chanID);
            IDSM_NotifyAudioOff(pChan->chanID);
            AUD_NotifyUserAudioOff(pChan->chanID);
            AUD_ScheduleOK(pChan->chanID);
            AUD_StartAudioJobs(pChan->chanID);

            // If transmitter is being disabled, go to that
            // state.  Otherwise go to a different state.
            if(!EEP_GetTxFlags(pChan->chanID, EEP_TX_FLAG_TXENA))
                TxTransition_DisableWaitForID(pChan);
            else
            {
                ASTREAM_UnmuteDestDev(pChan->destDev);

                // Turn CTCSS encoder off if necessary.
                if(EEP_GetTxFlags(
                    pChan->chanID, EEP_TX_FLAG_SATE) &&
                    (EEP_GetCTCSSEncodeMode(pChan->chanID)
                    == SATE_MODE_RUS))
                {
                    SATE_Off(pChan->chanID);
                }
                TxTransition_Wait4TxOff(pChan);
            }
            break;

        // Here, the user unkeyed.
        case EV_CEN_TX_DTMF_ERR:
            TXSM_NotifyUserAudioOff(pChan->chanID);
            IDSM_NotifyAudioOff(pChan->chanID);
            AUD_NotifyUserAudioOff(pChan->chanID);
            AUD_StartAudioJobs(pChan->chanID);
            AUD_ScheduleERR(pChan->chanID);

            // Turn CTCSS encoder off if necessary.
            if(EEP_GetTxFlags(pChan->chanID, EEP_TX_FLAG_SATE) &&
              (EEP_GetCTCSSEncodeMode(pChan->chanID) == SATE_MODE_RUS))
            {
                SATE_Off(pChan->chanID);
            }

            ASTREAM_UnmuteDestDev(pChan->destDev);
            TxTransition_Wait4TxOff(pChan);
            break;

        // If timeout event received, go to timeout state.
        case EV_CEN_TX_TIMEOUT:
            TxTransition_Timeout(pChan);
            break;

        // Here, received a command to disable the transmitter.
        // This could happen in this state if someone was entering
        // DTMF while someone else with access to the repeater is
        // disabling it via the terminal.  Note that if someone
        // is disabling the transmitter via touch-tones, it will
        // be handled in the EV_CEN_TX_DTMF_OK event handler
        // above.

        case EV_CEN_TX_DISABLE:
            TxTransition_DisableWaitForID(pChan);
            break;

        // User is requesting a test tone:
        case EV_CEN_TEST_TONE_REQ:
            SUSI_DenyTestTone();
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Come into this state if the user requested and was granted
     * permission to output a test tone on this transmitter.
     */
static void
        TxTransition_TestTone(struct TAG_CEN_TX_CHAN *pChan)
{
    // Turn CTCSS encoder on if enabled.
    if(EEP_GetTxFlags(pChan->chanID, EEP_TX_FLAG_SATE))
    {
        SATE_On(pChan->chanID);
    }

    // Set new state.
    SET_TX_STATE(TxState_TestTone, 'D');
}

    /**
     */
static void TxState_TestTone(enum TAG_CEN_EVIDS event,
                struct TAG_CEN_TX_CHAN *pChan, uint32_t unused)
{
    switch(event)
    {
        case EV_CEN_TEST_TONE_OFF:
            // Turn CTCSS encoder off if enabled.
            if(EEP_GetTxFlags(pChan->chanID, EEP_TX_FLAG_SATE))
            {
                SATE_Off(pChan->chanID);
            }

            TxTransition_Idle(pChan);
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** The transmitter is transmitting generated audio in this
     * state.
     */
static void TxTransition_GenAudio(
                                struct TAG_CEN_TX_CHAN *pChan)
{
    // Set new state.
    SET_TX_STATE(TxState_GenAudio, 'E');
}

    /**
     */
static void
    TxState_GenAudio(enum TAG_CEN_EVIDS event,
                struct TAG_CEN_TX_CHAN *pChan, uint32_t unused)
{
    switch(event)
    {
        // Here, RUS was asserted to a receiver that is tied to
        // this transmitter.
        case EV_CEN_TX_UAUDIO_STARTED:
            TXSM_NotifyUserAudioOn(pChan->chanID);
            AUD_NotifyUserAudioOn(pChan->chanID);
            IDSM_NotifyAudioOn(pChan->chanID);

            // Turn CTCSS encoder on if necessary.
            if(EEP_GetTxFlags(pChan->chanID, EEP_TX_FLAG_SATE) &&
              (EEP_GetCTCSSEncodeMode(pChan->chanID) == SATE_MODE_RUS))
            {
                SATE_On(pChan->chanID);
            }

            TxTransition_Repeating(pChan);
            break;

        // Here, all paths assigned to this transmitter
        // have been removed so shut down the transmitter.
        // In this case the paths will be from an announcement.
        case EV_CEN_TX_PATH_UNASSIGNED:
            TXSM_NotifyUserAudioOff(pChan->chanID);
            IDSM_NotifyAudioOff(pChan->chanID);
            TxTransition_Wait4TxOff(pChan);
            break;

        // User is requesting a test tone:
        case EV_CEN_TEST_TONE_REQ:
            SUSI_DenyTestTone();
            break;

        // If timeout event received, go to timeout state.
        case EV_CEN_TX_TIMEOUT:
            TxTransition_Timeout(pChan);
            break;

        // Received a command to disable the transmitter.
        case EV_CEN_TX_DISABLE:
            TxTransition_DisableWaitForID(pChan);
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Come here if the transmitter times out.
     */
static void TxTransition_Timeout(
                                struct TAG_CEN_TX_CHAN *pChan)
{
    TXSM_NotifyUserAudioOff(pChan->chanID);

    // Turn off subaudible tone encoder if enabled.
    if(EEP_GetTxFlags(pChan->chanID, EEP_TX_FLAG_SATE) &&
       (EEP_GetCTCSSEncodeMode(pChan->chanID) == SATE_MODE_RUS))
    {
        SATE_Off(pChan->chanID);
    }

    ASTREAM_MuteDestDev(pChan->destDev);

    // Start sending the "TO" message and the final ID.
    Morse_Stop(pChan->chanID);
    IDSM_NotifyTxTimeout(pChan->chanID);
    AUD_NotifyTxTimeout(pChan->chanID);

    // Set new state.
    SET_TX_STATE(TxState_Timeout, 'G');
}

    /**
     */
static void
    TxState_Timeout(enum TAG_CEN_EVIDS event,
                struct TAG_CEN_TX_CHAN *pChan, uint32_t unused)
{
    switch(event)
    {
        // User is requesting a test tone:
        case EV_CEN_TEST_TONE_REQ:
            SUSI_DenyTestTone();
            break;

        case EV_CEN_TX_PATH_UNASSIGNED:
            AUD_NotifyUserAudioOff(pChan->chanID);
            ASTREAM_UnmuteDestDev(pChan->destDev);
            AUD_NotifyTxTimeoutCleared(pChan->chanID);
            TXSM_NotifyTxTimeoutCleared(pChan->chanID);
            TxTransition_Idle(pChan);
            break;

        // Received a command to disable the transmitter.
        case EV_CEN_TX_DISABLE:
            TxTransition_DisableWaitForID(pChan);
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Come here to disable the transmitter.
     */
static void TxTransition_DisableWaitForID(
                                struct TAG_CEN_TX_CHAN *pChan)
{
    TXSM_NotifyTxDisable(pChan->chanID);

    // Turn off subaudible tone encoder if enabled.
    if(EEP_GetTxFlags(pChan->chanID, EEP_TX_FLAG_SATE) &&
       (EEP_GetCTCSSEncodeMode(pChan->chanID) == SATE_MODE_RUS))
    {
        SATE_Off(pChan->chanID);
    }

    ASTREAM_MuteDestDev(pChan->destDev);

    // Start sending the final ID.
    IDSM_NotifyTxDisable(pChan->chanID);
    AUD_NotifyTxDisable(pChan->chanID);

    // Set new state.
    SET_TX_STATE(TxState_DisableWaitForID, 'I');
}

    /**
     */
static void
    TxState_DisableWaitForID(enum TAG_CEN_EVIDS event,
                struct TAG_CEN_TX_CHAN *pChan, uint32_t unused)
{
    switch(event)
    {
        // User is requesting a test tone:
        case EV_CEN_TEST_TONE_REQ:
            SUSI_DenyTestTone();
            break;

        // The ID is complete now, so go to the disabled state.
        case EV_CEN_TX_ID_COMPLETE:
            TxTransition_Disabled(pChan);
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Come here when the transmitter is disabled.
     */
static void TxTransition_Disabled(
                                struct TAG_CEN_TX_CHAN *pChan)
{
    // Set new state.
    SET_TX_STATE(TxState_Disabled, 'F');
}

    /**
     */
static void
    TxState_Disabled(enum TAG_CEN_EVIDS event,
                struct TAG_CEN_TX_CHAN *pChan, uint32_t unused)
{
    switch(event)
    {
        // User is requesting a test tone:
        case EV_CEN_TEST_TONE_REQ:
            SUSI_DenyTestTone();
            break;

        case EV_CEN_TX_ENABLE:
            ASTREAM_UnmuteDestDev(pChan->destDev);
            TXSM_NotifyTxEnable(pChan->chanID);
            TxTransition_Idle(pChan);
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** (H) In this state we have told the transmitter to turn off
     * and we are waiting for an acknowledgement back from
     * the transmitter to signal that it has complied.  It will
     * complete its audio jobs first.
     */
static void TxTransition_Wait4TxOff(
                                struct TAG_CEN_TX_CHAN *pChan)
{
    // Turn off subaudible tone encoder if enabled.
    if(EEP_GetTxFlags(pChan->chanID, EEP_TX_FLAG_SATE) &&
       (EEP_GetCTCSSEncodeMode(pChan->chanID)
       == SATE_MODE_RUS))
    {
        SATE_Off(pChan->chanID);
    }
    SET_TX_STATE(TxState_Wait4TxOff, 'H');
}

    /**
     */
static void
    TxState_Wait4TxOff(enum TAG_CEN_EVIDS event,
                struct TAG_CEN_TX_CHAN *pChan, uint32_t unused)
{
    switch(event)
    {
        // Transmitter says it's off.
        case EV_CEN_TX_OFF:
            TxTransition_Idle(pChan);
            break;

        // A receiver went unsquelched, and this transmitter has
        // been assigned.
        case EV_CEN_TX_UAUDIO_STARTED:
            TXSM_NotifyUserAudioOn(pChan->chanID);
            AUD_NotifyUserAudioOn(pChan->chanID);
            IDSM_NotifyAudioOn(pChan->chanID);

            // Turn CTCSS encoder on if necessary.
            if(EEP_GetTxFlags(pChan->chanID, EEP_TX_FLAG_SATE) &&
              (EEP_GetCTCSSEncodeMode(pChan->chanID) == SATE_MODE_RUS))
            {
                SATE_On(pChan->chanID);
            }

            TxTransition_Repeating(pChan);
            break;

        // User is requesting a test tone:
        case EV_CEN_TEST_TONE_REQ:
            SUSI_DenyTestTone();
            break;

#if 0   // If we timeout in this state we really don't have
        // to do anything because the transmitter is already
        // in the process of turning off.

        // If timeout event received, go to timeout state.
        case EV_CEN_TX_TIMEOUT:
            TxTransition_Timeout(pChan);
            break;
#endif

        // Received a command to disable the transmitter.
        case EV_CEN_TX_DISABLE:
            TxTransition_Disabled(pChan);
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                 Squelch State Machine
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void SquelchTransition_Closed(
                                struct TAG_CEN_RX_CHAN *pChan)
{
    // Turn off squelch-opened LED.
    if(pChan->chanID == RX_CHAN_1)
        IO_ClearRUS1LED();
    else
        IO_ClearRUS2LED();

    Apex_SendMsgExt(
            &mailbox, EV_CEN_RX_SQUELCHED, pChan->chanID, 0);

    SATD_DisableDecoder(pChan->chanID);

    // Set Closed as the new state.
    SET_SQ_STATE(SquelchState_Closed, 'A');
}

    /** In this state, all receivers are squelched.  We are
     * wating for an unsquelch event.
     */
static void SquelchState_Closed(enum TAG_CEN_EVIDS event,
    struct TAG_CEN_RX_CHAN *pChan, uint32_t unused)
{
    switch(event)
    {
        // The carrier squelch for a channel has opened.
        case EV_CEN_CARRIER_SQU_OPENED:

            // Enable tone squelch if tone is required.
            if(EEP_GetRxFlags(pChan->chanID, EEP_RX_FLAG_SATD))
            {
                // Enable subaudible tone decoder:
                SATD_EnableDecoder(pChan->chanID);
                SquelchTransition_CarrierUnsquelched(pChan);
            }

            // Transition to SquelchOpened if tone is not
            // required.
            else
                SquelchTransition_Opened(pChan);
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Transition into the Carrier Unsquelched state.
     *
     * Enter this state when the carrier is unsquelched but we are
     * waiting for CTCSS decoder.
    */
static void SquelchTransition_CarrierUnsquelched(
                                    struct TAG_CEN_RX_CHAN *pChan)
{
    // Set new state.
    SET_SQ_STATE(SquelchState_CarrierUnsquelched, 'B');
}

    /** In this state, carrier is unsquelched.  We are
     * wating for the tone decoder to go unsquelched.
     */
static void
    SquelchState_CarrierUnsquelched(enum TAG_CEN_EVIDS event,
                        struct TAG_CEN_RX_CHAN *pChan, uint32_t unused)
{
    switch(event)
    {
        // The tone squelch for a channel has opened.

        case EV_CEN_TONE_SQU_OPENED:
            SquelchTransition_Opened(pChan);
            break;

        case EV_CEN_CARRIER_SQU_CLOSED:
            SquelchTransition_Closed(pChan);
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Transition into the Squelch Opened state
     *
     * Enter this state when the squelch is fully open.
    */
static void SquelchTransition_Opened(
                                    struct TAG_CEN_RX_CHAN *pChan)
{
    // Turn on the squelch LED for the channel.
    if(pChan->chanID == RX_CHAN_1)
        IO_SetRUS1LED();
    else
        IO_SetRUS2LED();

    /// Tell the receiver it is unsquelched:

    Apex_SendMsgExt(
            &mailbox, EV_CEN_RX_UNSQUELCHED, pChan->chanID, 0);

    /// Set new state.
    SET_SQ_STATE(SquelchState_Opened, 'C');
}

    /** In this state, carrier is unsquelched.  We are
     * wating for the tone decoder to go unsquelched.
     */
static void
    SquelchState_Opened(enum TAG_CEN_EVIDS event,
                        struct TAG_CEN_RX_CHAN *pChan, uint32_t unused)
{
    switch(event)
    {
        // The tone squelch for a channel has closed.
        case EV_CEN_TONE_SQU_CLOSED:
            // Turn off squelch-opened LED.
            if(pChan->chanID == RX_CHAN_1)
                IO_ClearRUS1LED();
            else
                IO_ClearRUS2LED();

            Apex_SendMsgExt(
                &mailbox, EV_CEN_RX_SQUELCHED, pChan->chanID, 0);

            SquelchTransition_CarrierUnsquelched(pChan);
            break;

        // Carrier squelch for the channel has closed.
        case EV_CEN_CARRIER_SQU_CLOSED:
            SquelchTransition_Closed(pChan);
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                   Receiver State Machine
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** (A) Transition into the Idle (receiver squelched) state.
    */
static void RxTransition_Squelched(struct TAG_CEN_RX_CHAN *pChan)
{
    /// Set Idle as the new state.
    SET_RX_STATE(RxState_Squelched, 'A');
}

    /** In this state the receiver is squelched.  We are
     * wating for an unsquelch event.
     */
static void RxState_Squelched(enum TAG_CEN_EVIDS event,
    struct TAG_CEN_RX_CHAN *pChan, uint32_t data)
{
    switch(event)
    {
        // The channel went unsquelched:
        case EV_CEN_RX_UNSQUELCHED:
            // Only open the transmitter if we are not muted.
            if(EEP_RX_IS_ENABLED(pChan->chanID))
                RxTransition_Unsquelched(pChan);
            else
                RxTransition_Disabled(pChan);
            break;

        // There is a request to record a voice file on the
        // selected channel:
        case EV_CEN_RECORD:
            pRecordingChan = pChan;
            recordFileNum = data;
            RxTransition_RecorderArmed(pChan);
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** (B) Transition to the Unsquelched (repeating) state.  This
     * function sets up the system to begin repeating audio from
     * receiver to transmitter(s).
     */
static void RxTransition_Unsquelched(
                            struct TAG_CEN_RX_CHAN *pChan)
{
    struct TAG_EEP_PATH *pEEPath;
    astream_msg_t astream_msg;

    pChan->pathID = ASTREAM_GetPathID();

    // Set the paths.
    EEP_InitFindPathBySrc();   // Initialize the path search
    pChan->pathID = ASTREAM_GetPathID();

    while((pEEPath = EEP_FindPathBySrc(pChan->srcDev)) != NULL)
    {
        ASTREAM_AddPath(pChan->pathID, pEEPath->src,
                        pEEPath->dest, pEEPath->priority);
    }

    // Tell the transmitters we are transmitting user audio.
    astream_msg.event = EV_CEN_TX_UAUDIO_STARTED;
    ASTREAM_MessageDestDevicesByPathID(
                                pChan->pathID, &astream_msg);

    DTMF_On(RX_CHAN_2_DTMF_CHAN(pChan->chanID));

    SET_RX_STATE(RxState_Unsquelched, 'B');
}

    /** In this state, squelch is opened and we are repeating.
     */
static void RxState_Unsquelched(enum TAG_CEN_EVIDS event,
    struct TAG_CEN_RX_CHAN *pChan, uint32_t data)
{
    astream_msg_t astream_msg;

    switch(event)
    {
        // Carrier squelch closed on a channel.

        case EV_CEN_RX_SQUELCHED:
            DTMF_Off(RX_CHAN_2_DTMF_CHAN(pChan->chanID));

            // Start the courtesy beep cycle:
            astream_msg.event = EV_CEN_TX_UAUDIO_STOPPED;
            astream_msg.data1 =
                (uintptr_t)EEP_GetCourtesyBeepParms(pChan->chanID);
            ASTREAM_MessageDestDevicesByActivePath(
                                pChan->pathID, &astream_msg);

            // Transition to Squelched: Resetting the paths
            // may cause the transmitter(s) to go unassigned.
            ASTREAM_RemovePathByID(pChan->pathID);

            // This marks squelch closed:
            RxTransition_Squelched(pChan);
            break;

        // Here, a DTMF character is received.
        case EV_CEN_DTMF_PRESS:
            // Only transition if the character is the "start
            // of command character".
            if((char)data == EEP_GetDTMFCommandChar())
            {
                dusiChan = pChan->chanID;
                DUSI_DTMFCharReceived(pChan->chanID, (char)data);
                RxTransition_DTMF(pChan);
            }
            break;

        // There is a request to record a voice file on the
        // selected channel:
        case EV_CEN_RECORD:
            SUSI_NotifyRecordComplete(SUSI_REC_STS_BUSY);
            break;

        // Operator gave disable command while receiver was active.
        // Remove the paths but keep listening for DTMF.
        case EV_CEN_RX_DISABLE:
            astream_msg.event = EV_CEN_TX_UAUDIO_STOPPED;
            astream_msg.data1 =
                (uintptr_t)EEP_GetCourtesyBeepParms(pChan->chanID);
            ASTREAM_MessageDestDevicesByActivePath(
                                pChan->pathID, &astream_msg);

            // Transition to Muted: Resetting the paths
            // may cause the transmitter(s) to go unassigned.
            ASTREAM_RemovePathByID(pChan->pathID);
            RxTransition_Disabled(pChan);
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */ /**
    (C) Prepare the state machine to listen for more DTMF characters,
    having already received the leading DTMF command character.
    */
static void RxTransition_DTMF(struct TAG_CEN_RX_CHAN *pChan)
{
    astream_msg_t astream_msg;

    // Shut off all unused DTMF paths:
    if(pChan->chanID == RX_CHAN_1)
    {
        DTMF_Off(RX_CHAN_2_DTMF_CHAN(RX_CHAN_2));
    }
    else
    {
        DTMF_Off(RX_CHAN_2_DTMF_CHAN(RX_CHAN_1));
    }

    // Signal all transmitters assigned to this receiver
    // that a DTMF command code was received.
    astream_msg.event = EV_CEN_TX_DTMF;
    ASTREAM_MessageDestDevicesByPathID(
                                pChan->pathID, &astream_msg);

    SET_RX_STATE(RxState_DTMF, 'C');
}

    /** In this state, the receiver is processing DTMF
     * characters.
     */
static void RxState_DTMF(enum TAG_CEN_EVIDS event,
    struct TAG_CEN_RX_CHAN *pChan, uint32_t data)
{
    switch(event)
    {
        //If a DTMF character is received, send it to DUSI.
        case EV_CEN_DTMF_PRESS:
            DUSI_DTMFCharReceived(pChan->chanID, (char)data);
            break;

         // Here, we have been informed that the squelch
         // has closed.
         case EV_CEN_RX_SQUELCHED:
            // Go to a state where we will wait for DUSI to
            // signal that the command was either OK or an ERR.
            RxTransition_DUSI_Complete(pChan);
            break;

        // There is a request to record a voice file on the
        // selected channel:
        case EV_CEN_RECORD:
            SUSI_NotifyRecordComplete(SUSI_REC_STS_BUSY);
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */ /**
     * (D) Prepare to enter a state in which we will Wait for DUSI to
     * evaluate the DTMF command and respond with an ERR or OK.
     */
static void RxTransition_DUSI_Complete(
                                    struct TAG_CEN_RX_CHAN *pChan)
{
    // Send an "end of command" character to DUSI.
    DUSI_DTMFCharReceived(pChan->chanID, '\0');
    DTMF_Off(RX_CHAN_2_DTMF_CHAN(pChan->chanID));
    SET_RX_STATE(RxState_DUSI_Complete, 'D');
}

    /** In this state, we are waiting for a response from DUSI.  If
     * the command was properly processed, an OK response will be
     * sent.  Otherwise an ERR response will be sent.
     */
static void RxState_DUSI_Complete(enum TAG_CEN_EVIDS event,
    struct TAG_CEN_RX_CHAN *pChan, uint32_t unused)
{
    astream_msg_t astream_msg;

    switch(event)
    {
        // Process a DUSI ERR response.
        case EV_CEN_DUSI_ERR:
            // Cause the transmitter(s) to output the DTMF result
            // and hang time.
            astream_msg.event = EV_CEN_TX_DTMF_ERR;
            ASTREAM_MessageDestDevicesByPathID(
                                    pChan->pathID, &astream_msg);

            // Reset the crosspoint.
            ASTREAM_RemovePathByID(pChan->pathID);

            RxTransition_Squelched(pChan);
            break;

        // Process a DUSI OK response.
        case EV_CEN_DUSI_OK:
            // Cause the transmitter(s) to output the DTMF result
            // and hang time.
            astream_msg.event = EV_CEN_TX_DTMF_OK;
            ASTREAM_MessageDestDevicesByPathID(
                                    pChan->pathID, &astream_msg);

            // Reset the crosspoint.
            ASTREAM_RemovePathByID(pChan->pathID);

            RxTransition_Squelched(pChan);
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** (E) Transition into the Arm-Recorder state.
    */
static void
    RxTransition_RecorderArmed(struct TAG_CEN_RX_CHAN *pChan)
{
    astream_destsig_t dest;

    // Set arm-record timer.  If timer times out before the
    // user breaks squelch, the record is disarmed.
    Apex_StartTmr(&recordTimer, CEN_TIME_ARM_RECORD);

    // Lay in a path to record the file.
    pChan->pathID = ASTREAM_GetPathID();

    // Pick the recording channel:
    if(pChan->chanID == RX_CHAN_1)
        dest = ASTREAM_DSIG_VOICE1;
    else
        dest = ASTREAM_DSIG_VOICE2;

    ASTREAM_AddPath(
            pChan->pathID, pChan->srcDev, dest, 'H');

    // Set Record as the new state.
    SET_RX_STATE(RxState_RecorderArmed, 'E');
}

    /** In this state, the receiver is waiting for an unsquelched
     * event.  When it occurs, the receiver will transition to
     * the 'recording' state.
     */
static void RxState_RecorderArmed(enum TAG_CEN_EVIDS event,
    struct TAG_CEN_RX_CHAN *pChan, uint32_t unused)
{
    switch(event)
    {
        // The carrier squelch for a channel has opened.
        case EV_CEN_RX_UNSQUELCHED:
            Apex_StopTmr(&recordTimer);
            RxTransition_Recording(pChan);
            break;

        // The user canceled the recording before it started.
        case EV_CEN_CANCEL_RECORD:
            Apex_StopTmr(&recordTimer);
            pRecordingChan = NULL;
            // Reset the crosspoint.
            ASTREAM_RemovePathByID(pChan->pathID);
            RxTransition_Squelched(pChan);
            break;

        // Here, the arm-record timer has timed out.  Disarm
        // the recorder.
        case EV_CEN_RECORD_TIMER:
            SUSI_NotifyRecordComplete(SUSI_REC_STS_TIMEOUT);
            pRecordingChan = NULL;
            // Reset the crosspoint.
            ASTREAM_RemovePathByID(pChan->pathID);
            RxTransition_Squelched(pChan);
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** (F) Transition into the Recording state.
    */
static void
    RxTransition_Recording(struct TAG_CEN_RX_CHAN *pChan)
{
    // Start the recording:
    VOI_RecordFile(
            (enum TAG_VOI_CHAN)pChan->chanID, recordFileNum);

    SUSI_NotifyRecordStarted();

    // Set Record as the new state.
    SET_RX_STATE(RxState_Recording, 'F');
}

    /** In this state, the receiver is recording a voice file.
     */
static void RxState_Recording(enum TAG_CEN_EVIDS event,
    struct TAG_CEN_RX_CHAN *pChan, uint32_t unused)
{
    switch(event)
    {
        // The channel squelch closed, indicating end of the
        // recording.
        case EV_CEN_RX_SQUELCHED:
            VOI_StopRecording((enum TAG_VOI_CHAN)pChan->chanID);
            SUSI_NotifyRecordComplete(SUSI_REC_STS_OK);
            pRecordingChan = NULL;
            // Reset the crosspoint.
            ASTREAM_RemovePathByID(pChan->pathID);
            RxTransition_Squelched(pChan);
            break;

        // Here, the voice recorder stopped because it ran
        // out of memory in the file.
        case EV_CEN_RECORD_FINI:
            SUSI_NotifyRecordComplete(SUSI_REC_STS_OVF);
            pRecordingChan = NULL;
            // Reset the crosspoint.
            ASTREAM_RemovePathByID(pChan->pathID);
            RxTransition_Squelched(pChan);
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** (G) Transition into Muted State.
    */
static void
    RxTransition_Disabled(struct TAG_CEN_RX_CHAN *pChan)
{

    // Get a path ID so that the DTMF states can pretend there
    // is a path. That way we don't have to test for muted
    // receivers in DTMF.
    pChan->pathID = ASTREAM_GetPathID();
    DTMF_On(RX_CHAN_2_DTMF_CHAN(pChan->chanID));

    // Set Muted as the new state.
    SET_RX_STATE(RxState_Disabled, 'G');
}

    /** In this state, the receiver is muted but still listening
     * for DTMF.
     */
static void RxState_Disabled(enum TAG_CEN_EVIDS event,
    struct TAG_CEN_RX_CHAN *pChan, uint32_t data)
{
    switch(event)
    {
        // The channel squelch closed
        case EV_CEN_RX_SQUELCHED:
            RxTransition_Squelched(pChan);
            break;

        // Here, a DTMF character is received.
        case EV_CEN_DTMF_PRESS:
            // Only transition if the character is the "start
            // of command character".
            if((char)data == EEP_GetDTMFCommandChar())
            {
                dusiChan = pChan->chanID;
                DUSI_DTMFCharReceived(pChan->chanID, (char)data);
                RxTransition_DTMF(pChan);
            }
            break;

        // No processing for other events in this state.
        default:
            break;
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                     UTILITY FUNCTIONS
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

#if 0
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Using the crosspoint switch, turn off the audio
        stream to the transmitter(s) that are currently
        transmitting.
    */
static void ResetCrosspoint(enum TAG_XPOINT_CODE xpoint)
{
    if(xpoint == XPOINT_TX1 || xpoint == XPOINT_ALL)
    {
        FILT_CrosspointSelect(CP_SOURCE_OFF, TX_CHAN_1);
        TXSM_NotifyUserAudioOff(TX_CHAN_1);
        IDSM_NotifyAudioOff(TX_CHAN_1);

        // Turn off subaudible tone encoder if enabled.
        if(EEP_GetTxFlags(TX_CHAN_1, EEP_TX_FLAG_SATE) &&
           (EEP_GetCTCSSEncodeMode(TX_CHAN_1) == SATE_MODE_RUS))
        {
            SATE_Off(TX_CHAN_1);
        }
    }

    if(xpoint == XPOINT_TX2 || xpoint == XPOINT_ALL)
    {
        FILT_CrosspointSelect(CP_SOURCE_OFF, TX_CHAN_2);
        TXSM_NotifyUserAudioOff(TX_CHAN_2);
        IDSM_NotifyAudioOff(TX_CHAN_2);

        // Turn off subaudible tone encoder if enabled.
        if(EEP_GetTxFlags(TX_CHAN_2, EEP_TX_FLAG_SATE) &&
           (EEP_GetCTCSSEncodeMode(TX_CHAN_2) == SATE_MODE_RUS))
        {
            SATE_Off(TX_CHAN_2);
        }
    }
}
#endif

///@}
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                        END
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
