/***************************************************************/
/**
    @file console.c
    Console Task and Support Functions

    @defgroup console Console Task and Support Functions

    @author ALP

    @par History
        @oo 081018 Modified console.c to include XModem protocol.

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

*/
//***************************************************************
//      Copyright 2008 by Pion and Simon Electronics.  All
//      rights reserved.
//***************************************************************

//      *** INCLUDE FILES ***

#include <stdint.h>
#include <stdbool.h>        //lint -esym(766, stdbool.h)
#include <string.h>

#include "std.h"
//#include "tp_system.h"      //lint -esym(766, tp_system.h)
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"
#include "apex.h"

#include "app.h"
#include "apparse.h"
#include "apprint.h"
#include "io_main.h"
#include "console.h"

#include "lpc213x.h"    //lint -esym(766, lpc213x.h)
#include "io_debug.h"   //lint -esym(766, io_debug.h)
#include "io_map.h"     //lint -esym(766, io_map.h)


//***************************************************************
//      *** MACROS ****


//***************************************************************

//      *** GENERAL DEFINITIONS ****

    // Stack Size:
    // It overflowed on 7/9/2012 with a size of 540
    //      bytes.  I increased it to 768.
    // Measured use of 432 bytes on 3/13/12.

#define Z_STK_CON       768         // Stack size in bytes


//***************************************************************

//      *** DATA STRUCTURE Definitions ***

//***************************************************************

//      *** VARIABLES ***

static TCB conTCB;

    // Con task stack;
static uint32_t conStack[Z_STK_CON/sizeof(uint32_t)];

static MBOX mailbox;

    // State machine to call when a character arrives.
static void (*pStateMachine)(void);

static char lineBuf[Z_LINE_BUF];
static char *pBuf;      // Pointer into buffer being parsed
static char charCount;  // Number of characters parsed.
static bool echo;

static char wordBuf[Z_WORD];
static const
    console_cmd_tbl_t *pContext;    // Points to current cmd table.
static char prompt;                 // Current prompt char
static void (*pStream)(char c);     // Current output device


//***************************************************************
//      *** CONSTANTS ***

static char * const errorMessage[] =
{
    " OK",                                  // 0: CON_STS_OK
                                            // 1: CON_STS_SYNTAX
    " - Syntax error or extra spaces at end of line.",
    " - Parameter out of range.",           // 2: CON_STS_RANGE
    "",                                     // 3: CON_STS_NULL
    "",                                     // 4: CON_STS_XMODEM
    " - Command not found.",                // 5: CON_STS_NOTFOUND
};

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Variable table for debugger access

const struct TAG_VARIABLE_TABLE consoleVarTab[] =
{
    {"stack", conStack, VAR_TYPE_NONE, sizeof(conStack)},
    {"tcb", &conTCB, VAR_TYPE_NONE, sizeof(conTCB)},
    {"mbox", &mailbox, VAR_TYPE_NONE, 1},
    {NULL, NULL, VAR_TYPE_NONE, 0}
};


//***************************************************************
//                  Function Prototypes
//***************************************************************

static void Con_Task(void);
static char *LineEdit(char c);
static void LineInit(void);
//static void NullStateMachine(void);

//***************************************************************
//           APPLICATION INTERFACE FUNCTIONS (API)
//***************************************************************

void Con_Reset(void)
{
    Apex_InitMailbox(&mailbox);
    pStateMachine = NULL;

    echo = true;            // Initially set up for echo to console.

    pContext = NULL;
    prompt = '>';
    pStream = STDOUT;
    LineInit();
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Con_Start() cannot make a call that could block because
    // we want to be able to call it from Apex_Target_Start().

void Con_Start(void)
{
    Apex_CreateTask(&conTCB, TID_CON, Con_Task,
        conStack, (sizeof(conStack)), TPRI_CON);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void Con_SetContext(const console_cmd_tbl_t *pNewContext,
        char newPrompt, void (*pNewStream)(char c),
        void (*pNewStateMachine)(void))
{
    prompt = newPrompt;
    pStream = pNewStream;
    pContext = pNewContext;
    pStateMachine = pNewStateMachine;
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Turn echo feature on or off.

void Con_Echo(bool state) {echo = state;}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Accept a character from the user and use it to build and
     * execute a command.
     */

void Con_CharacterSink(char c)
{
    const console_cmd_tbl_t *pCmdTbl;
    char *pLine;
    bool found;
    unsigned wordLen;
    uint8_t errorCode;

    if((pLine = LineEdit(c)) != NULL)
    {
        // Parse the first token from the line:

        if(strlen(pLine) != 0)
        {
            wordLen = 0;

            if(APParse(pLine, &wordLen,
                           GEN_FORMAT(w, Z_WORD), wordBuf) == 1)

            {
                pCmdTbl = pContext;
                ASSERT(pCmdTbl != NULL);

                found = false;

                while((pCmdTbl->pName != NULL) && !found)
                {
                    if(strcmp(wordBuf, pCmdTbl->pName) == 0)
                        found = true;
                    else
                        pCmdTbl++;
                }

                if(found && (pCmdTbl->pCommand != NULL))
                {
                    errorCode =
                        (*pCmdTbl->pCommand)(pLine+wordLen);
                }
                else
                    errorCode = CON_STS_NOTFOUND;

                // Process the error code.

                switch(errorCode)
                {
                    case CON_STS_NULL:
                        APPrint_CRLF(pStream);
                        Con_Prompt();
                        break;

                    case CON_STS_STATECHG:
                        break;

                    default:
                        APPrint_String(
                                pStream, errorMessage[errorCode]);
                        APPrint_CRLF(pStream);
                        Con_Prompt();
                        break;
                }
            }
        }

        else
        {
            APPrint_CRLF(pStream);
            Con_Prompt();
        }
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Output the current prompt.
     */
void Con_Prompt(void)
{
    LineInit(); // Init command line.
//    APPrint_CRLF(pStream);
    APPrint_Char(pStream, prompt);
    APPrint_Char(pStream, ' ');
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void LineInit(void)
{
    charCount = 0;          // Clear char count
    pBuf = lineBuf;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//  Put characters into a line.  Handle backspace and newline.

//  returns: Pointer to line buffer if 'enter' received.
//              NULL otherwise.

static char *LineEdit(char c)
{
    if(c != '\r' && c != '\n')
    {
        if(c == '\b')
        {
            // Here, we received a backspace.
            if(charCount > 0)
            {
                charCount--;
                pBuf--;
                if(echo)
                    APPrint_String(pStream, "\b \b");
            }
        }

        // Here, a space or printable character,

        else if(c >= ' ' && c <= '~')
        {
            if(charCount < Z_LINE_BUF-1)
            {
                *pBuf++ = c;
                charCount++;
                if(echo)
                    APPrint_Char(pStream, c);  // Echo the char.
            }
        }

        else
        {
            if(echo)
                APPrint_Char(pStream, '\a');   // Ring the bell
        }
        return(NULL);
    }

    // Here, we received a new line character.  Terminate the
    // buffer.

    else
    {
        *pBuf = '\0';
        return(lineBuf);
    }
}

//***************************************************************

static void Con_Task(void)
{
    FOREVER
    {
        ASSERT(pStateMachine != NULL);
        (*pStateMachine)();
    }
}


//***************************************************************
//                          *** END ***
//***************************************************************
