/*******************************************************************/
/**
    @file dtmfdet.c
    DTMF Tone Detector Using Goertzel's Algorithm.

    @defgroup dtmfdet DTMF Decoder

    $Id: dtmfdet.c 68 2010-01-24 22:59:11Z albert $

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

    @author ALP

    @verbatim
               DTMF Tone Table

             1209  1336  1477  1633

      697      1     2     3     A
      770      4     5     6     B
      852      7     8     9     C
      941      *     0     #     D


       General Goertzel Transfer Function:

              K * (1 - K_C * Z^(-1))
   H(Z) = ---------------------------------
             1 - K_A * Z^(-1) + Z^(-2)
    @endverbatim


    <P><A href="dtmfdet1.svg"> DTMF Detector ISR State Diagram</A></P>
    <P><A href="dtmfdet2.svg"> DTMF Detector Task State Diagram</A></P>

********************************************************************/

//********************************************************************

//      *** INCLUDE FILES ***

#include <stdint.h>
#include <limits.h>
#include <stdbool.h>

#include "lpc213x.h"        //lint -esym(766, lpc213x.h)
#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"
#include "app.h"

#include "apex.h"
#include "aplog2.h"
#include "dtmfdet.h"
#include "io_codec.h"
#include "filter.h"
#include "astream.h"        // eep.h needs this
#include "eep.h"            //lint -esym(766, eep.h)
                            // Need this to get command char.

    // Headers used during debug:
#include "io_debug.h"       //lint -esym(766, io_debug.h)
#include "apprint.h"        //lint -esym(766, apprint.h)
#include "io_map.h"         //lint -esym(766, io_map.h)
#include "io_main.h"        //lint -esym(766, io_main.h)

//***************************************************************
//      *** MACROS ****

//lint -esym(750, PRINT_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, SET_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, STATE_CHAN) (Macro not always referenced)

#define PRINT_STATE_CHAR(x)  if(STATE_CHAN == pChan->chanID) \
        IO_DBPrintChar(x)

#if ((TP_DTMF_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
  || (TP_DTMF_STATE_FLAGS == STATE_FLAGS_LOWER_CH1))
#   define STATE_CHAN  RX_CHAN_1
#else
#   define STATE_CHAN  RX_CHAN_2
#endif

#if (TP_DTMF_STATE_FLAGS == STATE_FLAGS_OFF)
#   define SET_STATE(X,Y)   pChan->pTaskState = (X)

#elif ((TP_DTMF_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
    || (TP_DTMF_STATE_FLAGS == STATE_FLAGS_UPPER_CH2))
#   define SET_STATE(X,Y)   pChan->pTaskState = (X); \
                                            PRINT_STATE_CHAR(Y)

#elif ((TP_DTMF_STATE_FLAGS == STATE_FLAGS_LOWER_CH1) \
    || (TP_DTMF_STATE_FLAGS == STATE_FLAGS_LOWER_CH2))
#   define SET_STATE(X,Y)   pChan->pTaskState = (X); \
                                            PRINT_STATE_CHAR((Y)|0x20)
#else
#error "TP_DTMF_STATE_FLAGS improperly defined."
#endif

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // DTMF Decoder Debugging Helps

#define DBHLP_PRINT_SIG_PWR         0
#define DBHLP_PRINT_RESULT_CHAR     0

#define DBHLP_DISABLE_CMD_DECODE    ( \
                                    DBHLP_PRINT_SIG_PWR | \
                                    DBHLP_PRINT_RESULT_CHAR \
                                    )

#define DBHLP_CRLF  ( \
                    DBHLP_PRINT_SIG_PWR | \
                    DBHLP_PRINT_RESULT_CHAR \
                    )

//***************************************************************

//      *** APEX RELATED DEFINITIONS ****

    // Stack Size: Measured use of 224 bytes on 3/13/12.

#define Z_STK_DTMF      300     // Stack size in bytes


    // State Machine Event IDs:

enum TAG_DTMF_EVIDS
{
    EV_DTMF_PP = DTMF_ID(0),    // Got a buffer to post process.
    EV_DTMF_TIMEOUT,
    EV_DTMF_ON,
    EV_DTMF_OFF,
    EV_DTMF_TIMEOUT1,
    EV_DTMF_TIMEOUT2,
    EV_DTMF_NULL               // Signal that the event was processed.
};

//#define TIME_DTMF_DIGIT   (1*TU_10SEC)    // Inter-digit timer.

//***************************************************************

//      *** GENERAL DEFINITIONS ***

#define LO_FREQS        0   // Low frequencies are 0 through 3
#define HI_FREQS        4   // High frequencies are 4 through 7

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // The K_A table contains the filter constants for the
    // Goertzel filters.  The constants are scaled by the
    // following binary exponent (E2_):

#define E2_K_A          10      // 2^10 = 1024

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // The following definitions contain the K_A constants
    // shown in the goertzel transfer function.  These constants
    // are scaled by E2_K_A (above).

    //  Constants calculated for sample rate of 9375 Hz:

#if CDX_PARAMETER_FREQ_SEL == 9375

#define N           232     // This many samples per post-process

#define K_A_697     14629       // 697 Hz
#define K_A_770     14250       // 770 Hz
#define K_A_852     13785       // 852 Hz
#define K_A_941     13232       // 941 Hz

#define K_A_1209    11293       // 1209 Hz
#define K_A_1336    10243       // 1336 Hz
#define K_A_1477    8991        // 1477 Hz
#define K_A_1633    7513        // 1633 Hz


#elif CDX_PARAMETER_FREQ_SEL == 11719

    // The following N is calculated for a sample rate of
    // 11718.75 Hz:

#define N           290     // This many samples per post-process

#   if E2_K_A == 13

//      The following constants are calculated for a sample
//      rate of 11718.75 Hz and a shift constant of 13.

#define K_A_697     15253       // 697 Hz
#define K_A_770     15007       // 770 Hz
#define K_A_852     14704       // 852 Hz
#define K_A_941     14343       // 941 Hz

#define K_A_1209    13061       // 1209 Hz
#define K_A_1336    12357       // 1336 Hz
#define K_A_1477    11509       // 1477 Hz
#define K_A_1633    10495       // 1633 Hz

#   elif E2_K_A == 10

//      The following constants are calculated for a sample
//      rate of 11718.75 Hz and a shift constant of 10.

#define K_A_697     1907        // 697 Hz
#define K_A_770     1876        // 770 Hz
#define K_A_852     1838        // 852 Hz
#define K_A_941     1793        // 941 Hz

#define K_A_1209    1633        // 1209 Hz
#define K_A_1336    1545        // 1336 Hz
#define K_A_1477    1439        // 1477 Hz
#define K_A_1633    1312        // 1633 Hz

#   endif // E2_K_A

#else // CDX_PARAMETER_FREQ_SEL
#   error "CDX_PARAMETER_FREQ_SEL is not a valid value."
#endif // CDX_PARAMETER_FREQ_SEL

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // At the end of each series of N samples we must multiply
    // each Goertzel output by 2/(N*N) (or 1/21,012.5 in the
    // case of 205 samples).  In practice this will be done by
    // multiplying by 25/524288 which is the same as multiplying
    // by 25 and downshifting 19.

    // Note: If N is 290, we need to multiply by 25 and
    // downshift 20.

    // "NUMR_" stands for "numerator", and "Q_DNOM" stands for
    // "demoninator-power-of-two".  NSQ stands for N squared.

#if N == 232
#define NUMR_NSQ        39      // Multiply by this to divide by N^2
#define Q_DNOM_NSQ      20      // ...then downshift this many.

#elif N == 290
    // The constants below are calculated for N=290.

#define NUMR_NSQ        25      // Multiply by this to divide by N^2
#define Q_DNOM_NSQ      20      // ...then downshift this many.

#endif // N

    // Similarly, we must divide the accumulated squared magnitude
    // of the entire signal by N.  For N=205 this is accomplished
    // by multiplying by 10 and downshifting 11 (10/2048). However
    // to prevent overflow of the squared magnitude accumulator,
    // we divide each accumulated square by 256.  That means in the
    // end, when we divide by N, we multiply by NUMR_N and then
    // downshift 8 bits less than we otherwise would.

    // Constants below are for N=232:
#if N == 232

#define NUMR_N        141   // Multiply by this to divide by N
#define Q_DNOM_N      7     // ...then downshift this many.

#elif N == 290

    // Constants below are for N=290:

#define NUMR_N        113   // Multiply by this to divide by N
#define Q_DNOM_N        7   // ...then downshift this many.
#endif // N

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//              *** Tone Specification Parameters ***
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Parameters for digit-pressed detection: These parameters
    // qualify tones for the presence of a digit.  The parameters
    // all begin with DET_.

    // DET_TOT_SIG_PWR is the absolute level that the square of
    // the RMS signal amplitude (i.e. the RMS signal power) must
    // rise above in order to enable detection of DTMF codes.
    // Nominally, DTMF tones will give a modulation level of
    // approximately 3000 Hz.

    // This constant should reflect a low deviation. It's purpose
    // is to insure that we have some valid signal before we
    // enable detection.

#define DET_TOT_SIG_PWR         14913   // Deviation of approx.
                                        // 1000 Hz peak.

    // DET_TONE_PWR_LOG2 is the amount (expressed in log-base-2) by
    // which the ratio of total signal power to tone power must be
    // less than in order to qualify a valid digit.

#define DET_TONE_PWR_LOG2       DBP_2_LOG2(3.0)

    // DET_RESIDUAL_PWR_LOG2 is the ratio (expressed in log-base-2)
    // of a particular tone's power and the sum of the powers of
    // all other tones in its group (that is, the high and low tone
    // groups).  If a valid digit is present, the digit's tone's
    // power should be much higher than the sum of all other tone's
    // powers in the same group.

#define DET_RESIDUAL_PWR_LOG2   DBP_2_LOG2(12.0)

    // DET_TWIST_PWR_LOG2 is the amount by which the absolute value
    // of the ratio (expressed in log-base-2) between the tones in
    // each group must be less than, in order to qualify a valid
    // digit.

    // 1/26/2013 ALP: Changed twist spec from 4.0 to 10.0 as a
    // result of testing with Steve Fritts.

#define DET_TWIST_PWR_LOG2      DBP_2_LOG2(10.0)

    // DET_TOT_PWR_CONSISTANCY_LOG2 is the amount by which the
    // absolute value of the ratio (expressed in log-base-2)
    // between the total signal power on the current and previous
    // samples must be less than in order to qualify a valid digit.

#define DET_TOT_PWR_STABILITY_LOG2    DBP_2_LOG2(1.0)

    // DET_TONE_PWR_CONSISTANCY_LOG2 is the amount by which the
    // absolute value of the ratio (expressed in log-base-2)
    // between the total tone power on the current and previous
    // samples must be less than in order to qualify a valid digit.

#define DET_TONE_PWR_STABILITY_LOG2   DBP_2_LOG2(1.0)

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Parameters for digit-released detection.  These parameters
    // qualify the signal for the absence of digit tones.  All such
    // parameters begin with NDET_.

    // NDET_TOT_SIG_PWR is the power level that the total signal
    // power must fall below in order to qualify that the digit
    // has been released.  It's defined in terms of the positive
    // detection power and provides some hysteresis.

#define NDET_TOT_SIG_PWR    (DET_TOT_SIG_PWR/2)

//***************************************************************

//      *** DATA STRUCTURE DEFINITIONS ***

typedef struct TAG_DETECTOR_CHAN
{
    bool enable;        ///< DTMF channel enabled if true.

    TMR digitTimer;     // Times out waiting for a digit.

    // Channel's current task and ISR state:

    enum TAG_DTMF_EVIDS (*pTaskState)(enum TAG_DTMF_EVIDS event,
        struct TAG_DETECTOR_CHAN *pChan);

    enum TAG_DTMF_CHAN chanID;

    void (*pISRState)(
        struct TAG_DETECTOR_CHAN *pChan, int16_t sample);

    // The following arrays are used to accumulate samples
    // in the Goertzel filters.  There are two of them because
    // one will always be assigned to the interrupt service
    // routine and one to the background task.  When the
    // interrupt finishes with one bank, it switches banks
    // and sends a message to the background task to tell it
    // to begin post-processing the bank just finished.

    volatile struct TAG_DTMF_BANK gf_bank_a;
    volatile struct TAG_DTMF_BANK gf_bank_b;

    volatile struct TAG_DTMF_BANK * pISRBank;  // Bank ISR is working on.
    volatile struct TAG_DTMF_BANK * pTaskBank; // Bank task is working on.

    int toneDuration;       // Num of times in a row tone met spec

    int maxLoTone;          // Low group filter # with greatest power
    int maxHiTone;          // High group filter # with greatest power

    // The power in the tones in the low and high groups, and
    // the total tone power:

    uint32_t totSigPower, totTonePower;
    uint32_t loTonePower, hiTonePower;

    // The following contain log-base-2 of the above.

    int32_t logTotSigPower, logTotTonePower;
    int32_t logLoTonePower, logHiTonePower;
    int32_t logLoGroupResidualPower, logHiGroupResidualPower;

    // These values are saved from the previous post-processing
    // samples:

    int32_t prevLogTotSigPower, prevLogTotTonePower;

} DTMF_CHAN;

//***************************************************************
//      *** Signals ***

DEFINE_SIGNAL(DTMF_Signal_Digit)        // Say digit was pressed
DEFINE_SIGNAL(DTMF_Signal_NotDigit)     // Say digit was released
DEFINE_SIGNAL(DTMF_Signal_RMS2)         // Report RMS^2 signal level

//***************************************************************

//      *** STATIC FUNCTION PROTOTYPES ****

static void Chan_Reset(DTMF_CHAN *pChan);
static void GFilter(
    int32_t sample, volatile GOERTZEL *pFilter, uint16_t k_a);
static void ResetDTMFBank(volatile struct TAG_DTMF_BANK *pBank);

static void DTMFDetectorTask(void);
static void ChanPostProcess(volatile DTMF_CHAN *pChan);
static void GFilterPostProcess(
    volatile GOERTZEL *pFilter, uint16_t k_a);
static void ChanFindFMax(volatile DTMF_CHAN *pChan);
static dtmf_result_t TonesMeetDetSpec(const DTMF_CHAN *pChan);
//static dtmf_result_t TonesMeetDetSpec(DTMF_CHAN *pChan);
static bool TonesMeetNotDetSpec(const DTMF_CHAN *pChan);

    // Task States:

static void Transition_Idle(DTMF_CHAN *pChan);  // A
static enum TAG_DTMF_EVIDS State_Idle(enum TAG_DTMF_EVIDS event,
        struct TAG_DETECTOR_CHAN *pChan);

static void Transition_CommandDetected(DTMF_CHAN *pChan);   // B
static enum TAG_DTMF_EVIDS State_CommandDetected(
    enum TAG_DTMF_EVIDS event, struct TAG_DETECTOR_CHAN *pChan);

static void Transition_AnyDigit(DTMF_CHAN *pChan);  // C
static enum TAG_DTMF_EVIDS State_AnyDigit(
    enum TAG_DTMF_EVIDS event, struct TAG_DETECTOR_CHAN *pChan);

static void Transition_DigitGone(DTMF_CHAN *pChan); // D
static enum TAG_DTMF_EVIDS State_DigitGone(
    enum TAG_DTMF_EVIDS event, struct TAG_DETECTOR_CHAN *pChan);

    // ISR States:

static void DTMF_ISR_Idle(const DTMF_CHAN *pChan, int16_t sample);

#if DBHLP_DISABLE_CMD_DECODE == 0
static void DTMF_ISR_Cmd(DTMF_CHAN *pChan, int16_t sample);
#endif

static void DTMF_ISR_All(DTMF_CHAN *pChan, int16_t sample);

//***************************************************************

//      *** VARIABLES ***

    // Variables for both channels.
static DTMF_CHAN channels[MAX_DTMF_CHAN];

static TCB dtmfTCB;

    // DD task stack;
static uint32_t ddStack[Z_STK_DTMF/sizeof(uint32_t)];
static MBOX mailbox;

//***************************************************************

//                  *** CONSTANTS IN FLASH ***

    // The table below maps the detected tone numbers into
    // actual digits and symbols:

static const uint8_t digitTable[] = "147*2580369#ABCD";

    // Debugger access to variables:

const struct TAG_VARIABLE_TABLE dtmfVarTab[] =
{
    {"stack", ddStack, VAR_TYPE_NONE, sizeof(ddStack)},
    {"tcb", &dtmfTCB, VAR_TYPE_NONE, sizeof(dtmfTCB)},
    {"mb", &mailbox, VAR_TYPE_NONE, sizeof(mailbox)},
    {"sp1", (void *)&channels[DTMF_CHAN_1].pTaskState,
                                            VAR_TYPE_HEXINT, 4},
    {"sp2", (void *)&channels[DTMF_CHAN_2].pTaskState,
                                            VAR_TYPE_HEXINT, 4},
    {NULL, NULL, VAR_TYPE_NONE, 0}
};


//***************************************************************
//              *** INITIALIZATION ***
//***************************************************************

    // Power-Up Initialization of DTMF Detector

void DTMF_Reset(void)
{
    DTMF_CHAN *pChan;

    Apex_InitMailbox(&mailbox);

    // Some of the per-channel initialization is done in-line
    // below instead of inside Chan_Reset().  This enables
    // both DTMF_Reset() and Transition_A() to call Chan_Reset()
    // and still work interrupt disable/enable properly.

    pChan = &channels[DTMF_CHAN_1];
    pChan->chanID = DTMF_CHAN_1;
    Apex_InitMsgTimer(
        &pChan->digitTimer, &mailbox, EV_DTMF_TIMEOUT, 0, 0);
    Chan_Reset(pChan);
    pChan->pISRState = DTMF_ISR_Idle;
    SET_STATE(State_Idle,'A');     // Sets state for RX_CHAN_1

    pChan = &channels[DTMF_CHAN_2];
    pChan->chanID = DTMF_CHAN_2;
    Apex_InitMsgTimer(
        &pChan->digitTimer, &mailbox, EV_DTMF_TIMEOUT, 0, 0);
    Chan_Reset(pChan);
    pChan->pISRState = DTMF_ISR_Idle;
    SET_STATE(State_Idle,'A');     // Sets state for RX_CHAN_2

    DTMF_UnsubscribeAllSignals();
}

void DTMF_Start(void)
{
    Apex_CreateTask(&dtmfTCB, TID_DTMF, DTMFDetectorTask,
        ddStack, (sizeof(ddStack)), TPRI_DTMF);
}

void DTMF_UnsubscribeAllSignals(void)
{
    Unsubscribe_DTMF_Signal_Digit();
    Unsubscribe_DTMF_Signal_NotDigit();
    Unsubscribe_DTMF_Signal_RMS2();
}
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Init channel variables:
    // This function is also called from Transition_A() and
    // Transition_C(), but there it is bracketed with interrupt
    // disable/enable.  That way Chan_Reset() may be called by
    // DTMF_Reset() also, saving a little code space.

static void Chan_Reset(DTMF_CHAN *pChan)
{
    ResetDTMFBank(&pChan->gf_bank_a);
    ResetDTMFBank(&pChan->gf_bank_b);
    pChan->pISRBank = &pChan->gf_bank_a;
    pChan->pTaskBank = &pChan->gf_bank_b;
    pChan->toneDuration = 3;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // Turn on DTMF processing for the indicated channel.

void DTMF_On(enum TAG_DTMF_CHAN chan)
{
    ASSERT(VALID_DTMF_CHAN(chan));
    Apex_SendMsgExt(&mailbox, EV_DTMF_ON, chan, 0);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// Turn off DTMF processing for the indicated channel.

void DTMF_Off(enum TAG_DTMF_CHAN chan)
{
    ASSERT(VALID_DTMF_CHAN(chan));
    Apex_SendMsgExt(&mailbox, EV_DTMF_OFF, chan, 0);
}

//***************************************************************
//              *** INTERRUPT SERVICE ROUTINES ***
//***************************************************************

    // DTMF Decoder Interrupt Service Routine

    // Fourth order composite high-pass pre-filter.  This filter
    // consists of a 2nd order notch followed by a 2nd order
    // high-pass.  The pre-filter removes audio below about 500 Hz

    // Fs=11718.75,
    // First biquad: Highpass, F0=310 Hz, Q=4.5, double zeros at
    //      0 Hz (DC).
    // Second biquad: Notch, F0=300 Hz, Q=0.67, double zeros at
    //      223 Hz.
    // Gain adjusted to 0 dB at Nyquist frequency.

void DTMF_SampleSink_ISR(enum TAG_DTMF_CHAN chan, int16_t sample)
{
    static const biquad_coeff_t
        hipassCoefficients = { 11, 1947, -3843, 1898, 0, 0 };

    static const biquad_coeff_t
            notchCoefficients = { 11, 1603, -3108, 1230, -4062, 0 };

    static biquad_delay_t highpassDelays[MAX_DTMF_CHAN];

    static biquad_delay_t notchDelays[MAX_DTMF_CHAN];

    // Prefilter the sample to remove freqs below about 500 Hz:
    sample = FILT_2ndOrderNotch(
                    &notchCoefficients, &notchDelays[chan], sample);

    sample = FILT_2ndOrderHighPass(
                &hipassCoefficients, &highpassDelays[chan], sample);

    // Execute the state machine for the channel using the
    // filtered sample.
    (*channels[chan].pISRState)(&channels[chan], sample);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // DTMF_ISR_Idle: Detector is off. (ISR State A)

static void DTMF_ISR_Idle(const DTMF_CHAN *unused1, int16_t unused2)
{}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // DTMF_ISR_Cmd: Process Start-of-Command Frequencies.
    // (ISR State B)

    // Detection starts by looking for the Start-of-Command
    // character.

    // The Start-of-Command character is 'C' or '#'.  The 'C'
    // character is encoeded using 852 Hz and 1633 Hz.  The #
    // character is encoded using 941 Hz and 1477 Hz.

#if DBHLP_DISABLE_CMD_DECODE == 0

static void DTMF_ISR_Cmd(DTMF_CHAN *pChan, int16_t sample)
{
    volatile struct TAG_DTMF_BANK *pBank;

    sample = sample >> 4;
    pBank = pChan->pISRBank;

    // Accumulate squared magnitude for later RMS calculation.

    pBank->totSigPower += (
        (uint32_t)(sample * sample) >> 8);

    // Run sample through detector (filter) banks.

    switch(EEP_GetDTMFCommandChar())
    {
        case '*':
            GFilter(sample, &pBank->filter[3], K_A_941);
            GFilter(sample, &pBank->filter[4], K_A_1209);
            break;

        case '#':
            GFilter(sample, &pBank->filter[3], K_A_941);
            GFilter(sample, &pBank->filter[6], K_A_1477);
            break;

        case 'A':
            GFilter(sample, &pBank->filter[0], K_A_697);
            GFilter(sample, &pBank->filter[7], K_A_1633);
            break;

        case 'B':
            GFilter(sample, &pBank->filter[1], K_A_770);
            GFilter(sample, &pBank->filter[7], K_A_1633);
            break;

        case 'C':
            GFilter(sample, &pBank->filter[2], K_A_852);
            GFilter(sample, &pBank->filter[7], K_A_1633);
            break;

        case 'D':
            GFilter(sample, &pBank->filter[3], K_A_941);
            GFilter(sample, &pBank->filter[7], K_A_1633);
            break;

        default:
            FATAL();
    }

    // Now check to see if all the filters are done.  If so, send
    // a message to the task.

    if(--pBank->count == 0)
    {
        Apex_SendMsgExt(&mailbox, EV_DTMF_PP, pChan->chanID, 0);
        pChan->pISRBank = pChan->pTaskBank; // Set new ISR bank.
        pChan->pTaskBank = pBank;           // Set new task bank
        pChan->pISRBank->count = N;         // Init new count.
    }
}

#endif

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // DTMF_ISR_All: Process All 8 DTMF Frequencies. (ISR State C)

static void DTMF_ISR_All(DTMF_CHAN *pChan, int16_t sample)
{
    volatile struct TAG_DTMF_BANK *pBank;

    sample = sample >> 4;
    pBank = pChan->pISRBank;

    // Accumulate squared magnitude for later RMS calculation.

    pBank->totSigPower += (
        (uint32_t)(sample * sample) >> 8);

    // Run sample through detector (filter) banks.

    GFilter(sample, &pBank->filter[0], K_A_697);
    GFilter(sample, &pBank->filter[1], K_A_770);
    GFilter(sample, &pBank->filter[2], K_A_852);
    GFilter(sample, &pBank->filter[3], K_A_941);
    GFilter(sample, &pBank->filter[4], K_A_1209);
    GFilter(sample, &pBank->filter[5], K_A_1336);
    GFilter(sample, &pBank->filter[6], K_A_1477);
    GFilter(sample, &pBank->filter[7], K_A_1633);

    // Now check to see if all the filters are done.  If so, send
    // a message to the task.

    if(--pBank->count == 0)
    {
        Apex_SendMsgExt(&mailbox, EV_DTMF_PP, pChan->chanID, 0);
        pChan->pISRBank = pChan->pTaskBank; // Set new ISR bank.
        pChan->pTaskBank = pBank;           // Set new task bank
        pChan->pISRBank->count = N;         // Init new count.
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // This function runs a sample through a Goertzel filter.
    // The frequency of the filter is determined by the constant
    // k_a.

static void GFilter(
    int32_t sample, volatile GOERTZEL *pFilter, uint16_t k_a)
{
    int32_t e_of_z;

    e_of_z =
        sample - pFilter->z_2 + ((k_a * pFilter->z_1) >> E2_K_A);

    pFilter->z_2 = pFilter->z_1;
    pFilter->z_1 = e_of_z;
}

//***************************************************************
//                          *** TASK ***
//***************************************************************

static void DTMFDetectorTask(void)
{
    unsigned id;
    uintptr_t data1;
    DTMF_CHAN *pChan;

    FOREVER
    {
        // Wait for a msg from ISR.
        Apex_PendMsg(&mailbox, &id, &data1, NULL, 0);

        // data1 contains the channel ID.

        ASSERT(data1 < MAX_RXCHAN);
        pChan = &channels[data1];

        // Here, send the event to the state machine. If the state
        // machine does not process the event, process it statelessly.

        ASSERT(pChan->pTaskState != NULL);

        switch((*pChan->pTaskState)((enum TAG_DTMF_EVIDS)id, pChan))
        {
            // Enable DTMF processing of the audio signal.
            case EV_DTMF_ON:
                if(
                    ((pChan->chanID == DTMF_CHAN_1) &&
                    (channels[DTMF_CHAN_2].pTaskState == State_Idle)) ||
                    ((pChan->chanID == DTMF_CHAN_2) &&
                    (channels[DTMF_CHAN_1].pTaskState == State_Idle))
                    )
                {
                    DI();
#if DBHLP_PRINT_RESULT_CHAR != 0
                    pChan->pISRState = DTMF_ISR_All;
#else
                    pChan->pISRState = DTMF_ISR_Cmd;
#endif
                    pChan->enable = true;
                    EI();
#if DBHLP_CRLF != 0
                    IO_DB_CRLF();
#endif
                }
                break;

            // Disable further DTMF processing.
            case EV_DTMF_OFF:
                DI();
                pChan->pISRState = DTMF_ISR_Idle;
                pChan->enable = false;
                EI();
                Transition_Idle(pChan);
#if DBHLP_CRLF != 0
                IO_DB_CRLF();
#endif
                break;

            default:
                break;
        }
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // State A: The detector channel is looking for a valid
    // start of command digit.  Could be 'C' or "#'.

static void Transition_Idle(DTMF_CHAN *pChan)
{
    // Stop audio interrupts while changing interrupt
    // variables:

    DI();
    Chan_Reset(pChan);
    EI();

    SET_STATE(State_Idle,'A');
}

static enum TAG_DTMF_EVIDS State_Idle(
        enum TAG_DTMF_EVIDS event, struct TAG_DETECTOR_CHAN *pChan)
{
    int digitIndex;
    dtmf_result_t result;
    uint8_t digit;

    if(event == EV_DTMF_PP)
    {
        // Do the tones meet the specification for detection?
        ChanPostProcess(pChan);

        result = TonesMeetDetSpec(pChan);

#if DBHLP_PRINT_RESULT_CHAR != 0
        switch(result)
        {
            // Signal passes all checks.
            case DTMF_RESULT_OK:
                IO_DBPrintChar('K');
                break;

            // Failed total power
            case DTMF_RESULT_TOT_SIG:
                IO_DBPrintChar('L');
                break;

            // Failed tone power
            case DTMF_RESULT_TONE_PWR:
                IO_DBPrintChar('M');
                break;

            // Failed low group residual
            case DTMF_RESULT_LRESIDUAL:
                IO_DBPrintChar('N');
                break;

            // Failed high group residual
            case DTMF_RESULT_HRESIDUAL:
                IO_DBPrintChar('P');
                break;

            // Failed twist
            case DTMF_RESULT_TWIST:
                IO_DBPrintChar('R');
                break;

            // Failed total pwr consistency
            case DTMF_RESULT_TOT_STABILITY:
                IO_DBPrintChar('S');
                break;

            // Failed tone pwr consistency
            case DTMF_RESULT_TONE_STABILITY:
                IO_DBPrintChar('T');
                break;

            default:
                IO_DBPrintChar('U');
                break;
        }

        IO_DBPrintChar(' ');
#endif

        if(result == DTMF_RESULT_OK)
        {
            // Tones have to meet detection spec this many times
            // in a row to encode a digit:

            if(--pChan->toneDuration == 0)
            {
                digitIndex = pChan->maxLoTone
                    + ((pChan->maxHiTone-4) << 2);
                digit = digitTable[digitIndex];

#if DBHLP_PRINT_RESULT_CHAR != 0
                IO_DBPrintChar((char)digit);
#endif
                ASSERT(DTMF_Signal_Digit != NULL);
                EMIT_SIGNAL(DTMF_Signal_Digit,
                        pChan->chanID, (uintptr_t)digit);

                // If we discover a '#' on one channel, turn off
                // the other channel.
                if(pChan->chanID == DTMF_CHAN_1)
                    DTMF_Off(DTMF_CHAN_2);
                else
                    DTMF_Off(DTMF_CHAN_1);

                // A '#' was detected.
                Transition_CommandDetected(pChan);
            }
        }
        else
            pChan->toneDuration = 3;

#if DBHLP_CRLF != 0
        IO_DB_CRLF();
#endif

        return(EV_DTMF_NULL);   // Say we processed the event.
    }

    else
        return(event);  // Say we didn't process the event.
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // In state B, a valid command character was detected.
    // Now we are waiting for it to go away.

static void Transition_CommandDetected(DTMF_CHAN *pChan)
{
    pChan->toneDuration = 3;
    SET_STATE(State_CommandDetected,'B');
}

static enum TAG_DTMF_EVIDS State_CommandDetected(
    enum TAG_DTMF_EVIDS event, struct TAG_DETECTOR_CHAN *pChan)
{
    if(event == EV_DTMF_PP)
    {
        ChanPostProcess(pChan);
        if(TonesMeetNotDetSpec(pChan))
        {
            // Tones have to meet not-detected spec this many times
            // in a row to unlock:

            if(--pChan->toneDuration == 0)
            {
                ASSERT(DTMF_Signal_NotDigit != NULL);
                EMIT_SIGNAL(DTMF_Signal_NotDigit,
                                            pChan->chanID, 0);
                Transition_AnyDigit(pChan);
            }
        }
        else
            pChan->toneDuration = 3;

        return(EV_DTMF_NULL);
    }

    else
        return(event);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // State C: The detector channel is looking for any valid
    // DTMF digit.

static void Transition_AnyDigit(DTMF_CHAN *pChan)
{
    DI();
    Chan_Reset(pChan);
    pChan->pISRState = DTMF_ISR_All;
    EI();

    SET_STATE(State_AnyDigit,'C');
}

static enum TAG_DTMF_EVIDS State_AnyDigit(
        enum TAG_DTMF_EVIDS event, struct TAG_DETECTOR_CHAN *pChan)
{
    int digitIndex;
    dtmf_result_t result;
    uint8_t digit;

    if(event == EV_DTMF_PP)
    {
        ChanPostProcess(pChan);

        // Do the tones meet the specification for detection?

        result = TonesMeetDetSpec(pChan);

#if DBHLP_PRINT_RESULT_CHAR != 0
        switch(result)
        {
            // Signal passes all checks.
            case DTMF_RESULT_OK:
                IO_DBPrintChar('K');
                break;

            // Failed total power
            case DTMF_RESULT_TOT_SIG:
                IO_DBPrintChar('L');
                break;

            // Failed tone power
            case DTMF_RESULT_TONE_PWR:
                IO_DBPrintChar('M');
                break;

            // Failed low group residual
            case DTMF_RESULT_LRESIDUAL:
                IO_DBPrintChar('N');
                break;

            // Failed high group residual
            case DTMF_RESULT_HRESIDUAL:
                IO_DBPrintChar('P');
                break;

            // Failed twist
            case DTMF_RESULT_TWIST:
                IO_DBPrintChar('R');
                break;

            // Failed total pwr consistency
            case DTMF_RESULT_TOT_STABILITY:
                IO_DBPrintChar('S');
                break;

            // Failed tone pwr consistency
            case DTMF_RESULT_TONE_STABILITY:
                IO_DBPrintChar('T');
                break;

            default:
                IO_DBPrintChar('U');
                break;
        }
        IO_DBPrintChar(' ');
#endif

        if(result == DTMF_RESULT_OK)
        {
            // Tones have to meet detection spec this many times
            // in a row to encode a digit:

            if(--pChan->toneDuration == 0)
            {
                digitIndex = pChan->maxLoTone
                    + ((pChan->maxHiTone-4) << 2);
                digit = digitTable[digitIndex];

#if DBHLP_PRINT_RESULT_CHAR != 0
                IO_DBPrintChar((char)digit);
#endif
                ASSERT(DTMF_Signal_Digit != NULL);
                EMIT_SIGNAL(DTMF_Signal_Digit,
                        pChan->chanID, (uintptr_t)digit);

                Transition_DigitGone(pChan);  // A digit was detected.
            }
        }

        else
            pChan->toneDuration = 3;

#if DBHLP_CRLF != 0
        IO_DB_CRLF();
#endif
        return(EV_DTMF_NULL);
    }

    else
        return(event);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // In state D, a valid digit was already detected.  Now we are
    // waiting for it to go away.

static void Transition_DigitGone(DTMF_CHAN *pChan)
{
    pChan->toneDuration = 3;
    SET_STATE(State_DigitGone,'D');
}

static enum TAG_DTMF_EVIDS State_DigitGone(
    enum TAG_DTMF_EVIDS event, struct TAG_DETECTOR_CHAN *pChan)
{
    if(event == EV_DTMF_PP)
    {
        ChanPostProcess(pChan);

        if(TonesMeetNotDetSpec(pChan))
        {
            // Tones have to meet not-detected spec this many times
            // in a row to unlock:

            if(--pChan->toneDuration == 0)
            {
                ASSERT(DTMF_Signal_NotDigit != NULL);
                EMIT_SIGNAL(DTMF_Signal_NotDigit,
                                            pChan->chanID, 0);
                Transition_AnyDigit(pChan);  // A digit was released.
            }
        }
        else
            pChan->toneDuration = 3;

        return(EV_DTMF_NULL);
    }

    else
        return(event);

}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Clear the Goertzel Filters

static void ResetDTMFBank(volatile struct TAG_DTMF_BANK *pBank)
{
    int i;
    volatile GOERTZEL *pGoert;

    pBank->totSigPower = 0;
    pBank->count = N;

    pGoert = pBank->filter;

    for(i=0; i<NUM_FREQS; i++, pGoert++)
        pGoert->z_2 = pGoert->z_1 = 0;
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Post processing for 8 filters:

static void ChanPostProcess(volatile DTMF_CHAN *pChan)
{
    GFilterPostProcess(&pChan->pTaskBank->filter[0], K_A_697);
    GFilterPostProcess(&pChan->pTaskBank->filter[1], K_A_770);
    GFilterPostProcess(&pChan->pTaskBank->filter[2], K_A_852);
    GFilterPostProcess(&pChan->pTaskBank->filter[3], K_A_941);
    GFilterPostProcess(&pChan->pTaskBank->filter[4], K_A_1209);
    GFilterPostProcess(&pChan->pTaskBank->filter[5], K_A_1336);
    GFilterPostProcess(&pChan->pTaskBank->filter[6], K_A_1477);
    GFilterPostProcess(&pChan->pTaskBank->filter[7], K_A_1633);

    // Calculate the overall-signal squared-magnitude:

    pChan->pTaskBank->totSigPower =
        (pChan->pTaskBank->totSigPower * NUMR_N) >> Q_DNOM_N;

    EMIT_SIGNAL(DTMF_Signal_RMS2, pChan->chanID,
                        pChan->pTaskBank->totSigPower);

    ChanFindFMax(pChan);
    ResetDTMFBank(pChan->pTaskBank);    // Done with bank
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Post-process a single Goertzel filter.  This operation
    // uses the filter output at sample N to compute the
    // squared-magnitude of the signal at the frequency
    // of the filter.  The frequency is determined by the
    // constant k_a.

static void GFilterPostProcess(
                    volatile GOERTZEL *pFilter, uint16_t k_a)
{
    int64_t term0, term1, term2;  // Post-processing variables

    // Calculate squared magnitudes of the signal at the
    // frequency of this filter:

    term0 =
        (int64_t)pFilter->z_2 * (int64_t)pFilter->z_2;

    term1 =
        (int64_t)pFilter->z_1 * (int64_t)pFilter->z_1;

    term2 =
        (((int64_t)pFilter->z_1 * (int64_t)pFilter->z_2)    * k_a)
        >> E2_K_A;

    pFilter->power = (uint32_t)
        (((term0 + term1 - term2) * NUMR_NSQ) >> Q_DNOM_NSQ);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Find the frequencies with the maximum power in each
    // of the low frequency and high frequency groups:

static void ChanFindFMax(volatile DTMF_CHAN *pChan)
{
    uint32_t maxLoTonePower, maxHiTonePower;
    uint32_t loGroupResidualPower, hiGroupResidualPower;
    volatile GOERTZEL *pFilter;
    int i;

    loGroupResidualPower = 0;
    hiGroupResidualPower = 0;
    maxLoTonePower = maxHiTonePower = 0;

    pFilter = &pChan->pTaskBank->filter[0];

    for(i=LO_FREQS; i<HI_FREQS; i++)
    {
        if(pFilter->power > maxLoTonePower)
        {
            maxLoTonePower = pFilter->power;
            pChan->maxLoTone = i;
        }

        loGroupResidualPower += pFilter->power;

        pFilter++;
    }

    for(i=HI_FREQS; i<NUM_FREQS; i++)
    {
        if(pFilter->power > maxHiTonePower)
        {
            maxHiTonePower = pFilter->power;
            pChan->maxHiTone = i;
        }

        hiGroupResidualPower += pFilter->power;

        pFilter++;
    }

    // Save values from the previous sample:

    pChan->prevLogTotSigPower = pChan->logTotSigPower;
    pChan->prevLogTotTonePower = pChan->logTotTonePower;

    // Now extract the max tone power levels from each group
    // and add them together to give the total tone power.

    pChan->loTonePower = maxLoTonePower;
    pChan->hiTonePower = maxHiTonePower;
    pChan->totTonePower = maxLoTonePower + maxHiTonePower;
    pChan->totSigPower = pChan->pTaskBank->totSigPower;

    // Subtract off the high and low tone powers from the
    // accumulated group powers so we get the residual power.

    loGroupResidualPower -= maxLoTonePower;
    hiGroupResidualPower -= maxHiTonePower;

    // Compute log-base-2 of power levels:

    pChan->logTotSigPower =
        pChan->totSigPower > 0? APLog2(pChan->totSigPower) : 0;
    pChan->logTotTonePower =
        pChan->totTonePower > 0? APLog2(pChan->totTonePower) : 0;
    pChan->logLoTonePower =
        maxLoTonePower > 0? APLog2(maxLoTonePower) : 0;
    pChan->logHiTonePower =
        maxHiTonePower > 0? APLog2(maxHiTonePower) : 0;
    pChan->logLoGroupResidualPower =
        loGroupResidualPower > 0? APLog2(loGroupResidualPower) : 0;
    pChan->logHiGroupResidualPower =
        hiGroupResidualPower > 0? APLog2(hiGroupResidualPower) : 0;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Return OK if the selected tones meet spec for
    // detection, otherwise return reason for failure.


static dtmf_result_t TonesMeetDetSpec(const DTMF_CHAN *pChan)
{
    int32_t logTotSigPwrToToneRatio;
    int32_t logTwistRatio;
    int32_t logTotPwrStability;
    int32_t logTonePwrStability;
    dtmf_result_t result;

    // Calculate the statistics we will use to judge the tones:

    logTotSigPwrToToneRatio =
                pChan->logTotSigPower - pChan->logTotTonePower;

    logTwistRatio =
            pChan->logLoTonePower - pChan->logHiTonePower;

    logTotPwrStability =
        pChan->prevLogTotSigPower - pChan->logTotSigPower;

    logTonePwrStability =
        pChan->prevLogTotTonePower - pChan->logTotTonePower;

    // Here, the total signal power must be greater than a
    // minimum level necessary for detection:

#if DBHLP_PRINT_SIG_PWR != 0
    IO_DBPrintDWord(pChan->totSigPower);
    IO_DBPrintChar(' ');
#endif

    if(pChan->totSigPower < DET_TOT_SIG_PWR)
        result = DTMF_RESULT_TOT_SIG;

    // Here, the total tone power must be a minimum fraction
    // of the total signal power.

    else if(ABS(logTotSigPwrToToneRatio) > DET_TONE_PWR_LOG2)
    {
        result = DTMF_RESULT_TONE_PWR;
    }

    // Check the tones to insure that they are on frequency.  An
    // off-frequency tone will bleed into the other filters.
    // While this test is not conclusive (if the tone is less
    // than 697 for example, it won't bleed into any other filter
    // because there are no other filters lower than 697), it
    // should help differentiate between voice/noise and valid tone.

    else if(
        (pChan->logLoTonePower - pChan->logLoGroupResidualPower)
        < DET_RESIDUAL_PWR_LOG2)
    {
        result = DTMF_RESULT_LRESIDUAL;
    }

    else if((pChan->logHiTonePower - pChan->logHiGroupResidualPower)
            < DET_RESIDUAL_PWR_LOG2)
    {
        result = DTMF_RESULT_HRESIDUAL;
    }

    // Twist must be within spec:
    else if(ABS(logTwistRatio) > DET_TWIST_PWR_LOG2)
    {
        result = DTMF_RESULT_TWIST;
    }

    // The total signal power must be stable:
    else if(ABS(logTotPwrStability) > DET_TOT_PWR_STABILITY_LOG2)
    {
        result = DTMF_RESULT_TOT_STABILITY;
    }

    // The total tone power must be stable:
    else if(ABS(logTonePwrStability) > DET_TONE_PWR_STABILITY_LOG2)
    {
        result = DTMF_RESULT_TONE_STABILITY;
    }

    // Here, passed all tests.
    else
    {
        result = DTMF_RESULT_OK;
    }

    return(result);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Return true if the selected tones meet spec for
    // not-detection, otherwise return false.

static bool TonesMeetNotDetSpec(const DTMF_CHAN *pChan)
{
    // Here, if the total signal power is less than or
    // equal to a maximum, there is no digit.

    if(pChan->totSigPower > NDET_TOT_SIG_PWR)
        return(false);

    else
        return(true);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                           END
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
