/*********************************************************************/
/**
    @file dtmfdet.h
    DTMF Tone Detector Using Goertzel's Algorithm.

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

    @verbatim
                            DTMF Tone Table

                         1209  1336  1477  1633

                697     1       2       3       A
                770     4       5       6       B
                852     7       8       9       C
                941     *       0       #       D

    @endverbatim


****************************************************************/

//  *** DTMF Detector Module Number and Related Definitions ***

#define DTMF_ID(x)  (MODULE(MID_DTMF) + (x))

#define TID_DTMF            DTMF_ID(1)  // DTMF Detector task ID

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

#define NUM_FREQS       8       // This many freqs are detected.

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// DTMF Channel Designation

enum TAG_DTMF_CHAN
{
    DTMF_CHAN_1 = 0,      ///< Channel 1
    DTMF_CHAN_2,          ///< Channel 2
    MAX_DTMF_CHAN         ///< Number of channels
};

// The definition below is used with the ASSERT macro to
// qualify channel numbers:

#define VALID_DTMF_CHAN(x)  ((enum TAG_DTMF_CHAN)(x) < MAX_DTMF_CHAN)

    /// This macro provides a mapping from DTMF channels to
    /// receiver channels:
#define DTMF_CHAN_2_RX_CHAN(x)      ((enum TAG_RX_CHAN)(x))

#define RX_CHAN_2_DTMF_CHAN(x)      ((enum TAG_DTMF_CHAN)(x))

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//      *** DATA STRUCTURE DEFINITIONS ***

    // The structure definitions below are contained in this h
    // file so the unit test shell can print a report from the
    // structures.

    // The following structure is used to contain the delay
    // blocks of each Goertzel filter.

typedef struct TAG_GORT_FILTER
{
    int32_t z_1;        // Delay block Z^(-1)
    int32_t z_2;        // Delay block Z^(-2)
    uint32_t power;     // Squared magnitude at this frequency.

} GOERTZEL;


    // The following structure represents a bank of Goertzel
    // filters plus a buffer for the overall magnitude-squared of
    // the signal.

struct TAG_DTMF_BANK
{
    GOERTZEL filter[NUM_FREQS];
    uint32_t totSigPower;       // Total power in signal
    int count;                 // Number of samples accumulated.
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                  SIGNALS
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Tell subscriber a digit was pressed
DECLARE_SIGNAL(DTMF_Signal_Digit)

    /// Tell Subscriber a digit was released.
DECLARE_SIGNAL(DTMF_Signal_NotDigit)

    /// Report RMS^2 signal level.
DECLARE_SIGNAL(DTMF_Signal_RMS2)

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//      *** PUBLIC FUNCTION PROTOTYPES ***

    // The following functions are in dtmfdet.c:

void DTMF_Reset(void);
void DTMF_Start(void);

void DTMF_On(enum TAG_DTMF_CHAN chan);
void DTMF_Off(enum TAG_DTMF_CHAN chan);

void DTMF_UnsubscribeAllSignals(void);

void DTMF_SampleSink_ISR(enum TAG_DTMF_CHAN chan, int16_t sample);

extern const struct TAG_VARIABLE_TABLE dtmfVarTab[];

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // The following codes are returned by the digit qualification
    // function to indicate which test failed, or if all tests
    // passed.  This information is contained in the h file so
    // that the external unit test report function can see them.

typedef enum TAG_DTMF_RESULT
{
    DTMF_RESULT_OK,               ///< Signal passes all checks.
    DTMF_RESULT_TOT_SIG,          ///< Failed total power
    DTMF_RESULT_TONE_PWR,         ///< Failed tone power
    DTMF_RESULT_LRESIDUAL,        ///< Failed low group residual
    DTMF_RESULT_HRESIDUAL,        ///< Failed high group residual
    DTMF_RESULT_TWIST,            ///< Failed twist
    DTMF_RESULT_TOT_STABILITY,    ///< Failed total pwr stability
    DTMF_RESULT_TONE_STABILITY,   ///< Failed tone pwr stability
    DTMF_RESULT_MAX

} dtmf_result_t;

//*********************************************************************
//                     End of File
//*********************************************************************
