/****************************************************************//**

    @file eep.c

    EEPROM Functions and Data for PSE-702

    @defgroup eep EEPROM Parameter Storage and Retrieval

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

********************************************************************/


//***************************************************************

//#define DEBUG_SHORT_TIMEOUTS

//***************************************************************

//      *** INCLUDE FILES ***

#include <stdint.h>         //lint -esym(766, stdint.h)
#include <stdbool.h>        //lint -esym(766, stdbool.h)

#include "lpc213x.h"        //lint -esym(766, lpc213x.h)
#include "std.h"
#include "target.h"

#include "apex.h"
#include "app.h"
#include "aplib.h"

#include "io_flash.h"
#include "io_main.h"   //lint -esym(766, io_main.h)

#include "ssm_sate.h"
#include "astream.h"
#include "ssm_idsm.h"
#include "eep.h"
#include "user_parms.h"

#include "io_debug.h"   //lint -esym(766, io_debug.h)
#include "apprint.h"    //lint -esym(766, apprint.h)

//***************************************************************
//                 *** MACROS ****

//***************************************************************
//             *** GENERAL DEFINITIONS ****

//***************************************************************
//             *** Default User Parameters ***
//***************************************************************

static const uint16_t defaultSystemFlags = 0x0000;  ///< System flags
static const char defaultPassWord[] = "8958";   ///< Password
static const char defaultDtmfCmdChar = '#';

    // Receiver Structure (TAG_EEP_RX_CHAN)

static const struct TAG_EEP_RX_CHAN defaultRxChan =
{
    EEP_RX_FLAG_RUSLEVEL | EEP_RX_FLAG_ENABLE,     ///< Rx flags
#if 0
        EEP_RX_FLAG_DEEMPH | EEP_RX_FLAG_SATR // Deprecated.
#endif

    45,                 ///< Receiver channel pot default, percent.
    10,                 ///< CTCSS frequency index
    1,                  ///< Deemphasis filter index
    1,                  ///< CTCSS reject filter index
    {
        700,                ///< 1st beep tone freq in Hz
        100,                ///< 1st beep tone duration in ms.
        800,                ///< 2nd beep tone freq in Hz
        100,                ///< 2nd beep tone duration in ms
        500,                ///< Pre-beep delay in ms.
        CB_MODE_SINGLE,     ///< Mode
        "K"                 ///< Morse character
    }
};

    // Transmitter Structure (TAG_EEP_TX_CHAN)

static const struct TAG_EEP_TX_CHAN defaultTxChan =
{
    // Word aligned items:

    // Transmitter Flags: set the ones that should be ON by default
    (EEP_TX_FLAG_TXENA | EEP_TX_FLAG_DMUTE | EEP_TX_CB_ENABLE),

    1500,               ///< Hang time, mS
    180,                ///< Tx timeout in Sec (3 Min)
    420,                ///< ID timeout in Sec (7 Min).
    800,                ///< Default morse freq, Hz

    // Byte aligned items:

    {'N','O',' ','C','W','I','D','\0'},  //lint !e785 ///<ID string
    SATE_MODE_RUS,      ///< SATE Mode
    10,                 ///< SATE frequency index
    58,                  ///< Tx level index, equiv to 0 dB.
    0,                  ///< Voice ID file number
    20,                 ///< Courtesy beep level in percent (100 max)
    20,                 ///< Morse level in percent (100 max)
    20                  ///< Morse rate, wpm
};


    /// This definition makes it easy to define unused paths.

#define UNUSED_PATH  \
                ASTREAM_NULL_SOURCE, ASTREAM_NULL_DEST, '\0'

    /// Default Paths (Paths are numbered 1-16 in SUSI):
    /// Receiver 1 deemphasized to transmitter 1 high priority.
    /// Receiver 2 deemphasized to transmitter 2 high priority.

static const struct TAG_EEP_PATH defaultPaths[] =
{
    {ASTREAM_SSIG_RX1, ASTREAM_DSIG_TX1, 'H'},     // Path 1
    {ASTREAM_SSIG_RX2, ASTREAM_DSIG_TX2, 'H'},     // Path 2
    {UNUSED_PATH},                                       // Path 3
    {UNUSED_PATH},                                       // Path 4
    {UNUSED_PATH},                                       // Path 5
    {UNUSED_PATH},                                       // Path 6
    {UNUSED_PATH},                                       // Path 7
    {UNUSED_PATH},                                       // Path 8
    {UNUSED_PATH},                                       // Path 9
    {UNUSED_PATH},                                       // Path 10
    {UNUSED_PATH},                                       // Path 11
    {UNUSED_PATH},                                       // Path 12
    {UNUSED_PATH},                                       // Path 13
    {UNUSED_PATH},                                       // Path 14
    {UNUSED_PATH},                                       // Path 15
    {UNUSED_PATH}                                        // Path 16
};

/// This definition makes it easy to define unused announcements.

#define UNUSED_ANNO     0, TXCHAN_INVALID, 'L', 0, false

    /// Default Announcements:

static const struct TAG_EEP_ANNO defaultAnno[] =
{
    {UNUSED_ANNO},                                       // Anno 1
    {UNUSED_ANNO},                                       // Anno 2
    {UNUSED_ANNO},                                       // Anno 3
    {UNUSED_ANNO},                                       // Anno 4
    {UNUSED_ANNO},                                       // Anno 5
    {UNUSED_ANNO},                                       // Anno 6
    {UNUSED_ANNO},                                       // Anno 7
    {UNUSED_ANNO},                                       // Anno 8
    {UNUSED_ANNO},                                       // Anno 9
    {UNUSED_ANNO},                                       // Anno 10
    {UNUSED_ANNO},                                       // Anno 11
    {UNUSED_ANNO},                                       // Anno 12
    {UNUSED_ANNO},                                       // Anno 13
    {UNUSED_ANNO},                                       // Anno 14
    {UNUSED_ANNO},                                       // Anno 15
    {UNUSED_ANNO}                                        // Anno 16
};

//***************************************************************
//                *** Static Variables ****
//***************************************************************

    // Make this a union so that there is always at least one
    // page of memory here.  That's because we write to the
    // flash in units of pages.

    /** Image of user parameters copied from flash.
     *
     */
static union TAG_EEP_USER_FLASH_IMAGE
{
    struct TAG_EEP_USER_PARMS image;
    uint8_t pageArray[Z_FLASH_PAGE*2];  // Give enough room for more

} userFlashImage;

#define UFI             userFlashImage.image
#define UFI_RX(x)       userFlashImage.image.rxChanStruct[x]
#define UFI_TX(x)       userFlashImage.image.txChanStruct[x]

#define Z_USER_PARMS    sizeof(union TAG_EEP_USER_FLASH_IMAGE)

static uint8_t oldUserFlashImage[Z_FLASH_PAGE*2];

//lint -esym(750, POUFI, POUFI_TX)

#define POUFI       pOldUserParms
#define POUFI_RX(x) pOldUserParms->rxChanStruct[(x)]
#define POUFI_TX(x) pOldUserParms->txChanStruct[(x)]

static unsigned pathSearchIndex;  ///< Index into path array

static bool eepPwrUpChecksumWasOK; ///< true if EEPROM checksum was
                                    ///< OK on power-up check.

//***************************************************************
//              *** STATIC FUNCTION PROTOTYPES ****
//***************************************************************

static uint16_t Checksum(void);
static void ReviseUserParms_0_1(void);

//***************************************************************
//                    *** API ****
//***************************************************************

//  Read the EEPROM data and compute the checksum

void EEP_Start(void)
{
    uint16_t crc1, crc2;
    uint8_t *pByte;
    unsigned i;
    uint16_t revision;

#if DEBUG_PORT != 0
    if(sizeof(struct TAG_EEP_USER_PARMS) > Z_FLASH_PAGE) //lint !e506 !e774
    {
        APPrint(STDERR,
            "EEP user parms: %d\n\r",
            sizeof(struct TAG_EEP_USER_PARMS));
    }
#endif

    Flash_ReadPage(
        BLOCK_2_PAGE_ADDR(FLASH_EEP_BLOCK),
        &userFlashImage.pageArray[0]);

    Flash_ReadPage(
        BLOCK_2_PAGE_ADDR(FLASH_EEP_BLOCK)+1,
        &userFlashImage.pageArray[Z_FLASH_PAGE]);

    crc1 = UFI.crc;   // Read the CRC from the flash image.
    UFI.crc = 0;      // Clear it for the calculation.

    crc2 = Checksum();

    if(crc1 != crc2)
    {
        eepPwrUpChecksumWasOK = false;
        EEP_SetDefaults();
        EEP_WriteOperatorParameters();
    }

    else
    {
        eepPwrUpChecksumWasOK = true;

        // Check to see if the user parameter structure is of the
        // same revision level as the firmware, and  update  it
        // if not:

        while(UFI.revision != EEP_STRUCT_REVISION)
        {
            revision = UFI.revision;
            memcpy(oldUserFlashImage, &UFI, sizeof(UFI));

            /// Clear user flash image:
            pByte = (uint8_t *)&UFI;
            for(i=0; i<sizeof(UFI); i++)
                *pByte++ = 0xFF;

            switch(revision)
            {
                case 0:
                    ReviseUserParms_0_1();
                    break;

                default:
                    break;
            }
        }
    }

    pathSearchIndex = 0;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
 * Revise user parameter structure 0 to structure 1:
 *
 * From version 0 to 1 we added optional CTCSS reject filters,
 * and changed the DTMF filter option from a flag in the 'rxf'
 * command, into index 0 in the 'filt' command.
 */

static void ReviseUserParms_0_1(void)
{
    struct TAG_EEP_USER_PARMS_0 *pOldUserParms;

    pOldUserParms = (struct TAG_EEP_USER_PARMS_0 *)
                                &oldUserFlashImage; //lint !e545 !e826

    userFlashImage.image.revision = 1;

    UFI.systemFlags = pOldUserParms->systemFlags;
    strcpy(UFI.passWord, pOldUserParms->passWord);
    UFI.dtmfCmdChar = pOldUserParms->dtmfCmdChar;

    /// Copy TX parameters
    memcpy(&UFI.txChanStruct[TX_CHAN_1],
                &pOldUserParms->txChanStruct[TX_CHAN_1],
                                    sizeof(struct TAG_EEP_TX_CHAN));
    memcpy(&UFI.txChanStruct[TX_CHAN_2],
                &pOldUserParms->txChanStruct[TX_CHAN_2],
                                    sizeof(struct TAG_EEP_TX_CHAN));

    /// Copy paths
    memcpy(&UFI.pathStruct[0],
                    &pOldUserParms->pathStruct[0],
                                    sizeof(struct TAG_EEP_PATH));

    /// Copy announcements
    memcpy(&UFI.annoStruct[0],
                    &pOldUserParms->annoStruct[0],
                                        sizeof(struct TAG_EEP_ANNO));

    /// Reformat RX parameters
    UFI_RX(RX_CHAN_1).beep = POUFI_RX(RX_CHAN_1).beep;
    UFI_RX(RX_CHAN_2).beep = POUFI_RX(RX_CHAN_2).beep;

    UFI_RX(RX_CHAN_1).rxLevel = POUFI_RX(RX_CHAN_1).rxLevel;
    UFI_RX(RX_CHAN_2).rxLevel = POUFI_RX(RX_CHAN_2).rxLevel;

    UFI_RX(RX_CHAN_1).satdIndex = POUFI_RX(RX_CHAN_1).satdIndex;
    UFI_RX(RX_CHAN_2).satdIndex = POUFI_RX(RX_CHAN_2).satdIndex;

        /// Preset the flags:
    UFI_RX(RX_CHAN_1).flags = POUFI_RX(RX_CHAN_1).flags
                        & ~(EEP_RX_FLAG_DEEMPH_0 | EEP_RX_FLAG_SATR_0);
    UFI_RX(RX_CHAN_2).flags = POUFI_RX(RX_CHAN_2).flags
                        & ~(EEP_RX_FLAG_DEEMPH_0 | EEP_RX_FLAG_SATR_0);

    /// Deemphasis enable

    UFI_RX(RX_CHAN_1).deemphIndex =
        ((POUFI_RX(RX_CHAN_1).flags & EEP_RX_FLAG_DEEMPH_0) != 0)?
        1 : 0;

    UFI_RX(RX_CHAN_2).deemphIndex =
        ((POUFI_RX(RX_CHAN_2).flags & EEP_RX_FLAG_DEEMPH_0) != 0)?
        1 : 0;

    /// CTCSS reject filtering

    UFI_RX(RX_CHAN_1).rejectIndex =
        ((POUFI_RX(RX_CHAN_1).flags & EEP_RX_FLAG_SATR_0) != 0)?
        1 : 0;

    UFI_RX(RX_CHAN_2).rejectIndex =
        ((POUFI_RX(RX_CHAN_2).flags & EEP_RX_FLAG_SATR_0) != 0)?
        1 : 0;

    EEP_WriteOperatorParameters();
}

//****************************************************************
//                          Path Commands
//****************************************************************
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
 *
    Initialize path search by source
*/
void EEP_InitFindPathBySrc(void)
{
    pathSearchIndex = 0;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    * Search the RxTx path table for the next path that matches
    * the given signal source.
    *
    * This function may be used to find all occurances in the
    * path table for a given signal source.  Call
    * EEP_InitPathSearch() first to initialize the search,
    * then this function may be called as many times
    * as necessary to obtain pointers to all paths containing
    * the given source.
    *
    *  srcSig The ASTREAM source device to search for in the path
    *  table.
    *
    *  Returns: NULL if no additional source paths exist in the
    *  table.
    */
struct TAG_EEP_PATH *const EEP_FindPathBySrc(
                                            astream_srcsig_t srcSig)
{
    struct TAG_EEP_PATH *pResult = NULL;

    while(pathSearchIndex < MAX_PATHS && pResult == NULL)
    {
        // If the path matches, return it.
        if(srcSig == UFI.pathStruct[pathSearchIndex].src)
        {
            pResult = &UFI.pathStruct[pathSearchIndex];
        }

        pathSearchIndex++;
    }

    // Return the path or NULL if not found.

    return(pResult);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
 * Search the RxTx path table for a path that matches the given
 * source and destination.
 *
 * Paths with a given source and destination are unique so there
 * is no need to search more than once to get all of them.  The
 * only exception is the empty path, signified by values of
 * ASTREAM_NULL_SOURCE and ASTREAM_NULL_DEST.  Thus this function
 * may be used to find unused table locations by searching for
 * the null source and destination signatures.
 *
 *  srcSig: The ASTREAM source device to search for.
 *
 *  destSig: The ASTREAM destination device to search for.
 *
 *  Returns a pointer to the path that matches the source and
 *      destination, or NULL if none are found.
 */
eep_path_t *const EEP_FindPath(
            astream_srcsig_t srcSig, astream_destsig_t destSig)
{
    eep_path_t *pResult = NULL;
    unsigned i = 0;

    while(i < MAX_PATHS && pResult == NULL)
    {
        // If the path matches, return it.
        if(srcSig == UFI.pathStruct[i].src
            && destSig == UFI.pathStruct[i].dest)
        {
            pResult = &UFI.pathStruct[i];
            break;
        }

        i++;
    }

    // Return the path or NULL if not found.

    return(pResult);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /* Set a path to the parameters specified.
     *
     * pPath: A pointer to an empty path slot returned by
     *  EEP_FindEmptyPath() to contain the new path data.
     * pNewPath: A pointer to the new path to copy into the empty slot.
     */
void EEP_SetPath(eep_path_t *pPath, const eep_path_t *pNewPath)
{
    DI();
    *pPath = *pNewPath;
    EI();
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /* Delete a path.
     *
     * pPath: A pointer to the path to be deleted.
     */
void EEP_DeletePath(eep_path_t *pPath)
{
    eep_path_t path = {UNUSED_PATH};
    EEP_SetPath(pPath, &path);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /* Get a pointer to the path specified by the index.
     *
     * param index An index to the path to be retrieved.  Must be
     * less than MAX_PATHS
     */
struct TAG_EEP_PATH *const EEP_GetPath(unsigned index)
{
    ASSERT(index < MAX_PATHS);
    return(&UFI.pathStruct[index]);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   /*   Set an announcement to the parameters specified.
    *
    *   pAnno A pointer to an empty announcement slot
    *   returned by EEP_GetAnno(), to contain the new
    *   announcement data.
    *   pNewAnno A pointer to the new announcement to
    *   copy into the empty slot.
    */

void EEP_SetAnno(eep_anno_t *pAnno, const eep_anno_t *pNewAnno)
{
    DI();
    *pAnno = *pNewAnno;
    EI();
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   /*  Find an announcement by channel number and file number.
    *
    * If the announcement is found, an index to it is returned,
    * otherwise -1 is returned.
    */

short EEP_FindAnno(enum TAG_TX_CHAN chan, unsigned fileNum)
{
    short i;

    for(i=0; i<MAX_ANNO; i++)
    {
        if((UFI.annoStruct[i].chan == chan) &&
           (UFI.annoStruct[i].fileNum == fileNum))
        {
            return(i);
        }
    }
    return(-1);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   /*   Get a pointer to the announcement specified by the index.
    *
    * index An index to the announcement to be retrieved.
    * Index must be less than MAX_ANNO.
    */

eep_anno_t *const EEP_GetAnno(unsigned index)
{
    ASSERT(index < MAX_ANNO);
    return(&UFI.annoStruct[index]);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   /* Delete the announcement pointed to by the parameter.
    *
    * pAnno A pointer to the announcement to be deleted.
    */

void EEP_DeleteAnno(eep_anno_t *pAnno)
{
    eep_anno_t anno = {UNUSED_ANNO};
    EEP_SetAnno(pAnno, &anno);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   /* Enable or disable an announcement.
    * pAnno: A pointer to the announcement to enable/disable.
    * enable: A boolean: true to enable the announcement, false
    * to disable it.
    */

void EEP_EnableAnno(eep_anno_t *pAnno, bool enable)
{
    DI();
    pAnno->enabled = enable;
    EI();
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // Get EEPROM Checksum Power-Up Status

bool EEP_GetPwrUpChecksumWasOK(void)
{
    return(eepPwrUpChecksumWasOK);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // Get, Inq, and Set Voice ID Parameters

uint8_t EEP_GetVidFileNum(enum TAG_TX_CHAN chan)
{
    return(UFI_TX(chan).vidFileNum);
}

#if 0
bool EEP_GetVidDeemphasis(enum TAG_TX_CHAN chan)
{
    return((UFI_TX(chan).flags & EEP_TX_VID_DEEMPHASIS) != 0);
}
#endif

void EEP_SetVidFileNum(
    enum TAG_TX_CHAN chan, uint8_t fileNum)
{
    UFI_TX(chan).vidFileNum = fileNum;
}


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Get, Inq, and Set Timers.

long EEP_GetTxTimeout(enum TAG_TX_CHAN chan)
{
    return((long)UFI_TX(chan).txTimeout * TU_1SEC);
}

long EEP_GetHangTime(enum TAG_TX_CHAN chan)
{
    return((long)UFI_TX(chan).hangTime);
}

long EEP_GetIDTime(enum TAG_TX_CHAN chan)
{
    return((long)UFI_TX(chan).idTime * TU_1SEC);
}

void EEP_InqTimers(enum TAG_TX_CHAN txChan,
    uint32_t *pIdTime, uint32_t *pTxTimeout, uint32_t *pHangTime)
{
    *pIdTime = (uint32_t)UFI_TX(txChan).idTime;
    *pTxTimeout = (uint32_t)UFI_TX(txChan).txTimeout;
    *pHangTime = (uint32_t)UFI_TX(txChan).hangTime;
}

void EEP_SetTimers(enum TAG_TX_CHAN txChan,
    uint16_t idTime, uint16_t txTimeout, uint16_t hangTime)
{
    UFI_TX(txChan).idTime = idTime;
    UFI_TX(txChan).txTimeout = txTimeout;
    UFI_TX(txChan).hangTime = hangTime;
    IDSM_NotifyChangedID(txChan);
}


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // Get, Inq, and Set courtesy beep parameters

uint8_t EEP_GetCourtesyBeepLevel(enum TAG_TX_CHAN txChan)
{
    return(UFI_TX(txChan).cbLevel);
}


void EEP_SetCourtesyBeepLevel(
                    enum TAG_TX_CHAN txChan, uint8_t level)
{
    UFI_TX(txChan).cbLevel = level;
}

eep_courtsey_beep_t
    *EEP_GetCourtesyBeepParms(enum TAG_RX_CHAN chan)
{
    return(&UFI_RX(chan).beep);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // Get, Inq, and Set CW ID parameters

uint8_t EEP_GetMorseRate(enum TAG_TX_CHAN txChan)
{
    return(UFI_TX(txChan).morseRate);
}

uint16_t EEP_GetMorseFreq(enum TAG_TX_CHAN txChan)
{
    return(UFI_TX(txChan).morseFreq);
}

uint8_t EEP_GetMorseLevel(enum TAG_TX_CHAN txChan)
{
    return(UFI_TX(txChan).morseLevel);
}

void EEP_SetMorseLevel(enum TAG_TX_CHAN txChan, uint8_t level)
{
    UFI_TX(txChan).morseLevel = level;
}

void EEP_SetMorseFreq(enum TAG_TX_CHAN txChan, uint16_t freq)
{
    UFI_TX(txChan).morseFreq = freq;
}

void EEP_SetMorseRate(enum TAG_TX_CHAN txChan, uint8_t rate)
{
    UFI_TX(txChan).morseRate = rate;
}


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // Get, Inq, and Set transmitter flags

bool EEP_GetTxFlags(enum TAG_TX_CHAN chan, uint16_t mask)
{
    return((UFI_TX(chan).flags & mask) != 0);
}

void EEP_InqTxFlags(enum TAG_TX_CHAN txChan, uint16_t *pFlags)
{
    *pFlags = UFI_TX(txChan).flags;
}

void EEP_SetTxFlags(enum TAG_TX_CHAN txChan, uint16_t flags)
{
    UFI_TX(txChan).flags = flags;
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // Get, Inq, and Set receiver flags

bool EEP_GetRxFlags(enum TAG_RX_CHAN chan, uint16_t mask)
{
    return((UFI_RX(chan).flags & mask) != 0);
}

void EEP_InqRxFlags(enum TAG_RX_CHAN rxChan, uint16_t *pFlags)
{
    *pFlags = UFI_RX(rxChan).flags;
}

void EEP_SetRxFlags(enum TAG_RX_CHAN rxChan, uint16_t flags)
{
    UFI_RX(rxChan).flags = flags;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Get Deemphasis Filter Index
unsigned EEP_GetDeemphasisFilterIndex(enum TAG_RX_CHAN rxChan)
{
    return((unsigned)UFI_RX(rxChan).deemphIndex);
}

    /// Set Deemphasis Filter Index
void EEP_SetDeemphasisFilterIndex(
                enum TAG_RX_CHAN rxChan, unsigned index)
{
    UFI_RX(rxChan).deemphIndex = (uint8_t)index;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Get CTCSS Rejection Filter Index
unsigned EEP_GetCtcssRejectionFilterIndex(enum TAG_RX_CHAN rxChan)
{
    return((unsigned)UFI_RX(rxChan).rejectIndex);
}

    /// Set CTCSS Rejection Filter Index
void EEP_SetCtcssRejectionFilterIndex(
                enum TAG_RX_CHAN rxChan, unsigned index)
{
    UFI_RX(rxChan).rejectIndex = (uint8_t)index;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Functions for setting and testing system flags

bool EEP_GetSysFlags(uint16_t mask)
{
    return((UFI.systemFlags & mask) != 0);
}

void EEP_InqSysFlags(uint16_t *pFlags)
{
    *pFlags = UFI.systemFlags;
}

void EEP_SetSysFlags(uint16_t flags)
{
    UFI.systemFlags = flags;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // Functions to get and set DUSI start of command character

char EEP_GetDTMFCommandChar(void)
{
    return(UFI.dtmfCmdChar);
}

void EEP_SetDTMFCommandChar(char c)
{
    UFI.dtmfCmdChar = c;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // CTCSS Encode Parameters

void EEP_SetCTCSSEncodeMode(
    enum TAG_TX_CHAN chan, enum TAG_SATE_MODE mode)
{
    UFI_TX(chan).sateMode = mode;
}

enum TAG_SATE_MODE EEP_GetCTCSSEncodeMode(enum TAG_TX_CHAN chan)
{
    return(UFI_TX(chan).sateMode);
}

uint8_t EEP_GetCTCSSEncodeIndex(enum TAG_TX_CHAN chan)
{
    return(UFI_TX(chan).sateIndex);
}

void EEP_SetCTCSSEncodeIndex(enum TAG_TX_CHAN chan, uint8_t index)
{
    UFI_TX(chan).sateIndex = index;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

uint8_t EEP_GetCTCSSDecodeIndex(enum TAG_RX_CHAN chan)
{
    return(UFI_RX(chan).satdIndex);
}

void EEP_SetCTCSSDecodeIndex(enum TAG_RX_CHAN chan, uint8_t index)
{
    UFI_RX(chan).satdIndex = index;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//  Get, Inq, and Set ID string

char *EEP_GetIDString(enum TAG_TX_CHAN txChan)
{
    return(UFI_TX(txChan).idString);
}

void EEP_SetIDString(enum TAG_TX_CHAN chan, const char *string)
{
    strcpy(UFI_TX(chan).idString, string);
}

void EEP_InqIDString(enum TAG_TX_CHAN chan, char **ppString)
{
     *ppString = UFI_TX(chan).idString;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//  Get, Inq, and Set DTMF password

char *EEP_GetDTMFPwd(void)
{
    return(UFI.passWord);
}

void EEP_SetDTMFPwd(const char *string)
{
    strcpy(UFI.passWord, string);
}

void EEP_InqDTMFPwd(char **ppString)
{
     *ppString = UFI.passWord;
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // Get, Inq, and Set receiver level

uint8_t EEP_GetRxLevel(enum TAG_RX_CHAN rxChan)
{
    return((uint8_t)UFI_RX(rxChan).rxLevel);
}

void EEP_InqRxLevel(enum TAG_RX_CHAN rxChan, uint32_t *pRxLevel)
{
    *pRxLevel = (uint32_t)UFI_RX(rxChan).rxLevel;
}

void EEP_SetRxLevel(enum TAG_RX_CHAN rxChan, uint8_t rxLevel)
{
    UFI_RX(rxChan).rxLevel = rxLevel;
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // Get, Inq, and Set transmitter level

uint8_t EEP_GetTxLevel(enum TAG_TX_CHAN txChan)
{
    return((uint8_t)UFI_TX(txChan).txLevelIndex);
}

void EEP_InqTxLevel(enum TAG_TX_CHAN txChan, uint32_t *pTxLevel)
{
    *pTxLevel = (uint32_t)UFI_TX(txChan).txLevelIndex;
}

void EEP_SetTxLevel(enum TAG_TX_CHAN txChan, uint8_t index)
{
    UFI_TX(txChan).txLevelIndex = index;
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Audio Delay
#if 0
uint16_t EEP_GetAudioDelay(enum TAG_TX_CHAN txChan)
{
    return((uint16_t)UFI_TX(txChan).audioDelay);
}

void EEP_InqAudioDelay(enum TAG_TX_CHAN txChan, uint16_t *pAudioDelay)
{
    *pAudioDelay = (uint32_t)UFI_TX(txChan).audioDelay;
}

void EEP_SetAudioDelay(enum TAG_TX_CHAN txChan, uint16_t audioDelay)
{
    UFI_TX(txChan).audioDelay = audioDelay;
}
#endif

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void EEP_WriteOperatorParameters(void)
{
    UFI.crc = 0;            // Clear it for the calculation.

    UFI.crc = Checksum();   // Load crc into the image.

    Flash_EraseBlock(FLASH_EEP_BLOCK);

    Flash_WritePage(
            BLOCK_2_PAGE_ADDR(FLASH_EEP_BLOCK),
            (uint8_t *)&userFlashImage.pageArray[0]);

    Flash_WritePage(
            BLOCK_2_PAGE_ADDR(FLASH_EEP_BLOCK)+1,
            (uint8_t *)&userFlashImage.pageArray[Z_FLASH_PAGE]);
}

//***************************************************************
//              *** STATIC FUNCTIONS ****
//***************************************************************

void EEP_SetDefaults(void)
{
    uint32_t i;
    uint8_t *pByte;

    /// Clear user flash image:
    pByte = (uint8_t *)&userFlashImage;
    for(i=0; i<sizeof(userFlashImage); i++)
        *pByte++ = 0xFF;

    /// Set system wide parameters:

    userFlashImage.image.revision = EEP_STRUCT_REVISION;
    userFlashImage.image.systemFlags = defaultSystemFlags;
    strcpy(userFlashImage.image.passWord, defaultPassWord);
    userFlashImage.image.dtmfCmdChar = defaultDtmfCmdChar;

    /// Set default RX parameters
    memcpy(&userFlashImage.image.rxChanStruct[0],
        &defaultRxChan, sizeof(struct TAG_EEP_RX_CHAN));
    memcpy(&userFlashImage.image.rxChanStruct[1],
        &defaultRxChan, sizeof(struct TAG_EEP_RX_CHAN));

    /// Set default  TX parameters
    memcpy(&userFlashImage.image.txChanStruct[0],
        &defaultTxChan, sizeof(struct TAG_EEP_TX_CHAN));
    memcpy(&userFlashImage.image.txChanStruct[1],
        &defaultTxChan, sizeof(struct TAG_EEP_TX_CHAN));

    /// Set default paths
    memcpy(&userFlashImage.image.pathStruct[0],
                    defaultPaths, sizeof(defaultPaths));

    /// Set default announcements
    memcpy(&userFlashImage.image.annoStruct[0],
                        defaultAnno, sizeof(defaultAnno));
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /*
     * CRC Algorithm courtesy of Ashley Roll
     * Digital Nemesis Pty Ltd
     * www.digitalnemesis.com
     */

    // The 16 bit CRC occupies the last 2 bytes of the flash page.

    /*
     * CRC16 Lookup tables (High and Low Byte) for 4 bits per
     * iteration.
     */
static const unsigned short CRC16_LookupHigh[16] =
{
    0x00, 0x10, 0x20, 0x30, 0x40, 0x50, 0x60, 0x70,
    0x81, 0x91, 0xA1, 0xB1, 0xC1, 0xD1, 0xE1, 0xF1
};

static const unsigned short CRC16_LookupLow[16] =
{
    0x00, 0x21, 0x42, 0x63, 0x84, 0xA5, 0xC6, 0xE7,
    0x08, 0x29, 0x4A, 0x6B, 0x8C, 0xAD, 0xCE, 0xEF
};

static unsigned char crc16_High, crc16_Low; // Results accumulator

static void CRC16_Update4Bits( unsigned char val );
static void CRC16_Update( unsigned char val );


    // Calculate checksum of EEPROM image.

static uint16_t Checksum(void)
{
    uint32_t i;
    uint8_t *pMem = userFlashImage.pageArray;

    // Initialize the CRC to 0xFFFF for the CCITT specification

    crc16_High = 0xFF;
    crc16_Low = 0xFF;

    for(i=0; i<Z_USER_PARMS; i++)
        CRC16_Update(*pMem++);

    return((crc16_High << 8) | crc16_Low);
}

/*
 * Process one Message Byte to update the current CRC Value
 */
static void CRC16_Update( unsigned char val )
{
    CRC16_Update4Bits( val >> 4 );      // High nibble first
    CRC16_Update4Bits( val & 0x0F );    // Low nibble
}

/*
 * Process 4 bits of the message to update the CRC Value.
 *
 * Note that the data must be in the low nibble of val.
 */
static void CRC16_Update4Bits( unsigned char val )
{
    unsigned char   t;

    // Step one, extract the Most significant 4 bits of the CRC register
    t = crc16_High >> 4;

    // XOR in the Message Data into the extracted bits
    t = t ^ val;

    // Shift the CRC Register left 4 bits
    crc16_High = (unsigned char)(crc16_High << 4) | (crc16_Low >> 4);
    crc16_Low = (unsigned char)(crc16_Low << 4);

    // Do the table lookups and XOR the result into the CRC Tables
    crc16_High = (unsigned char)(crc16_High ^ CRC16_LookupHigh[t]);
    crc16_Low = (unsigned char)(crc16_Low ^ CRC16_LookupLow[t]);
}



//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                      END
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
