/****************************************************************//**

    @file eep.h

    EEPROM Functions and Data for PSE-702

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

********************************************************************/

/// @todo Add DTMF enable flag on a per-channel  basis.


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//      Operator Interface Parameter Limits
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
///@todo Make naming convention consistent

#define MIN_IDTIME          60      ///< Minimum ID period in seconds
#define MAX_IDTIME          1920    ///< Maximum ID period in seconds

#define MIN_TXTIME          10      ///< Minimum Tx timeout in seconds
#define MAX_TXTIME          600     ///< Maximum Tx timeout in seconds

#define MIN_HANGTIME        1       ///< Minimum hang period in ms
#define MAX_HANGTIME        10000   ///< Maximum hang period in ms

#define CB_MIN_DELAY        100     ///< Min courtesy beep delay in ms
#define CB_MAX_DELAY        10000   ///< Max courtesy beep delay in ms

#define CB_MIN_DURATION     50      ///< Min courtesy beep duration, ms
#define CB_MAX_DURATION     1000    ///< Max courtesy beep duration, ms

#define MIN_PL_TIMEOUT      20      ///< Min PL Detector Timeout, ms
#define MAX_PL_TIMEOUT      2000    ///< Max PL Detector Timeout, ms

#define MIN_ANN_PERIOD      1       ///< Min Announcement Period, min.
#define MAX_ANN_PERIOD      4320    ///< Max Announcement Period, min.

#define FG_MIN_FREQ         400     ///< Min possible frequency, Hz
#define FG_MAX_FREQ         2000    ///< Max possible frequency, Hz

#define MAX_ID_LENGTH       20      ///< Max ID string length.
                                     ///< (Incl trailing null)
#define MIN_ID_LENGTH       3       ///< ID length must be >= this.

#define MAX_TX_LEVEL        100     ///< Max transmitter audio level
#define MIN_TX_LEVEL        0       ///< Min transmitter audio level

#define MAX_RX_LEVEL        100     ///< Max receiver audio level
#define MIN_RX_LEVEL        0       ///< Min receiver audio level

#define MAX_ANNO_PERIOD     525960  ///< Max announcement period: 1 yr.

#define Z_DTMF_PASSWORD     9       ///< 8 character password + '\0'
#define MIN_DTMF_PASSWORD   4       ///< Must be at least 4 chars.

#define MAX_PATHS           16      ///< Maximum number of paths

#define MAX_ANNO            16      ///< Max number of announcements


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//  Transmitter flag definitions

#define EEP_TX_FLAG_TXENA       BV(0)   ///< Transmitter enable
#define EEP_TX_FLAG_REVBURST    BV(1)   ///< Reverse burst enable
//#define                        BV(2)   ///< Unused
#define EEP_TX_FLAG_DMUTE       BV(3)   ///< DTMF muting enable
#define EEP_TX_FLAG_VID         BV(4)   ///< Voice ID enable
#define EEP_TX_FLAG_PTTLEVEL    BV(5)   ///< PTT active high if set
#define EEP_TX_FLAG_SATE        BV(6)   ///< CTCSS encode enable
#define EEP_TX_CB_ENABLE        BV(7)   ///< Enable courtesy beeps.
//#define                        BV(8)   ///< Unused
//#define                        BV(9)   ///< Unused

// Receiver flag definitions

#define EEP_RX_FLAG_RUSLEVEL    BV(0)   ///< RUS active high enable
#define EEP_RX_FLAG_SATD        BV(1)   ///< CTCSS tone decode enable
#define EEP_RX_FLAG_ENABLE      BV(3)   ///< Mute a receiver.

#if 0   // DEEMPH and SATR flags deprecated.
#define EEP_RX_FLAG_DEEMPH      BV(2)   ///< Deemphasis enable
#define EEP_RX_FLAG_SATR        BV(4)   ///< Enable tone reject filter
#endif

// System level flag definitions

#define EEP_SYS_FLAG_CH1_HALF   BV(0)   ///< Rx1 and Tx1 half duplex
#define EEP_SYS_FLAG_CH2_HALF   BV(1)   ///< Rx2 and Tx2 half duplex
#define EEP_SYS_FLAG_DTMF_PASS  BV(2)   ///< Require DTMF password.

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//              *** PUBLIC FUNCTION PROTOTYPES ***
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void EEP_Start(void);
void EEP_SetDefaults(void);
bool EEP_GetPwrUpChecksumWasOK(void);

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
 * Write operator parameters to flash.
 *
 * SUSI calls this to write the operator parameters to flash
 * after a delay.
 */
void EEP_WriteOperatorParameters(void);

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    @name Path Related Definitions
    @{
    */

    /// Path structure

typedef struct TAG_EEP_PATH
{
    astream_srcsig_t src;       ///< Source of audio stream
    astream_destsig_t dest;     ///< Destination of audio stream
    char priority;              ///< H M or L for destination priority
} eep_path_t;

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
     * Initialize path search by source.
     *
     * This function initializes the search index so that successive
     * calls to EEP_FindPathBySrc() will start searching from the
     * beginning of the path table.
     */
void EEP_InitFindPathBySrc(void);

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    * Search the RxTx path table for the next path that matches
    * the given signal source.
    *
    * This function may be used to find all occurances in the
    * path table for a given signal source.  Call
    * EEP_InitPathSearch() first to initialize the search,
    * then this function may be called as many times
    * as necessary to obtain pointers to all paths containing
    * the given source.
    *
    *  srcSig The ASTREAM source device to search for in the path
    *  table.
    *
    *  Returns: NULL if no additional source paths exist in the
    *  table.
    */
eep_path_t *const EEP_FindPathBySrc(astream_srcsig_t srcSig);

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    * Search the RxTx path table for a path that matches the given
    * source and destination.
    *
    * Paths with a specific source and destination are unique so there
    * is no need to search more than once to get all of them.  The
    * only exception is the empty path, signified by values of
    * ASTREAM_NULL_SOURCE and ASTREAM_NULL_DEST.  Thus this function
    * may be used to find unused table locations by searching for
    * the null source and destination signatures.
    *
    *  @param srcSig The ASTREAM source device to search for.
    *
    *  @param destSig The ASTREAM destination device to search for.
    *
    *  @returns A pointer to the path that matches the source and
    *      destination is returned, or NULL is returned if none
    *      are found.
    */
eep_path_t *const EEP_FindPath(
            astream_srcsig_t srcSig, astream_destsig_t destSig);

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    * Set a path to the parameters specified.
    *
    *   @param pPath A pointer to an empty path slot returned by
    *       EEP_FindEmptyPath() to contain the new path data.
    *   @param pNewPath A pointer to the new path to copy into
    *   the empty slot.
    */
void EEP_SetPath(eep_path_t *pPath, const eep_path_t *pNewPath);

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
     * Get a pointer to the path specified by the index.
     *
     * @param index An index to the path to be retrieved.  Index
     * must be less than MAX_PATHS.
     */
eep_path_t *const EEP_GetPath(unsigned index);

    /** Delete the path pointed to by the parameter.
     *
     * @param pPath A pointer to the path to be deleted.
     */
void EEP_DeletePath(eep_path_t *pPath);

/// @}  Paths

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    @name Courtesy Beep Related Definitions
    @{
    */

    /// Courtesy beep mode definitions

enum TAG_CB_MODE
{
    CB_MODE_DISABLED,       ///< Courtesy beep is disabled
    CB_MODE_SINGLE,         ///< Single tone beep
    CB_MODE_DOUBLE,         ///< Double tone beep
    CB_MODE_MORSE,          ///< Morse character
    CB_MODE_MAX             ///< End of list
};

    /// Structure to define courtesy beep parameters

typedef struct TAG_EEP_COURTESY_BEEPS
{
    uint16_t freq1;           ///< First beep frequency, Hz
    uint16_t duration1;       ///< Duration of first beep, ms
    uint16_t freq2;           ///< Second beep frequency, Hz
    uint16_t duration2;       ///< Duration of second beep, ms
    uint16_t delay;           ///< Pre-courtesy beep delay, ms
    enum TAG_CB_MODE mode;    ///< Mode: single, double, morse
    char morse[2];            ///< Morse character followed by '\0'

} eep_courtsey_beep_t;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Functions for setting courtesy beep parameters

eep_courtsey_beep_t
    *EEP_GetCourtesyBeepParms(enum TAG_RX_CHAN chan);

uint8_t EEP_GetCourtesyBeepLevel(enum TAG_TX_CHAN chan);

void EEP_SetCourtesyBeepLevel(
                    enum TAG_TX_CHAN chan, uint8_t level);

/// @}  Courtesy Beeps

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    @name Announcement Related Definitions
    @{
    */

/// A structure to represent announcements.

typedef struct TAG_EEP_ANNO
{
    uint32_t period;     ///< Number of minutes between announcements.
    enum TAG_TX_CHAN chan; ///< Destination of audio stream
//    bool deemphasis;     ///< Should the stream be deemphasized?
    char priority;       ///< H M or L for destination priority
    uint8_t fileNum;     ///< Announcement file number.
    bool enabled;        ///< True if announcement is enabled.
} eep_anno_t;

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    * Set an announcement to the parameters specified.
    *
    *   @param pAnno A pointer to an empty announcement slot
    *   returned by EEP_GetAnno(), to contain the new
    *   announcement data.
    *   @param pNewAnno A pointer to the new announcement to
    *   copy into the empty slot.
    */
void EEP_SetAnno(eep_anno_t *pAnno, const eep_anno_t *pNewAnno);

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
     * Get a pointer to the announcement specified by the index.
     *
     * @param index An index to the announcement to be retrieved.
     * Index must be less than MAX_ANNO.
     */
eep_anno_t *const EEP_GetAnno(unsigned index);

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
     * Find an announcement by channel and file number.
     *
     * @param chan The channel number to which the announcement
     * is assigned.
     * @param fileNum The file number to which the announcement
     * is assigned.
     * @returns Returns an index to the announcement if found,
     * or -1 if not found.
     */
short EEP_FindAnno(enum TAG_TX_CHAN chan, unsigned fileNum);

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
     * Delete the announcement pointed to by the parameter.
     *
     * @param pAnno A pointer to the announcement to be deleted.
     */
void EEP_DeleteAnno(eep_anno_t *pAnno);

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
     * Enable or disable an announcement.
     * @param pAnno A pointer to the announcement to enable/disable.
     * @param enable A boolean: true to enable the announcement, false
     * to disable it.
     */
void EEP_EnableAnno(eep_anno_t *pAnno, bool enable);

/// @} Announcements

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
 * Get Voice ID File Number
 *
 * This function returns a voice ID file number given a
 * transmitter channel.
 *
 * @param txChan The channel for which the voice ID file
 * number shall be returned.
 *
 * @returns The voice ID file number for the given channel
 * is returned.
 */
uint8_t EEP_GetVidFileNum(enum TAG_TX_CHAN txChan);

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
 * Set Voice ID File Number
 *
 * This function sets a voice ID file number
 * for a selected transmitter channel.
 *
 * @param chan The transmitter channel to which the command
 * applies.
 *
 * @param fileNum A file number to set for that channel.
 */
void EEP_SetVidFileNum(
                    enum TAG_TX_CHAN chan, uint8_t fileNum);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // EEP Functions for Timers

long EEP_GetTxTimeout(enum TAG_TX_CHAN txChan);
long EEP_GetHangTime(enum TAG_TX_CHAN txChan);
long EEP_GetIDTime(enum TAG_TX_CHAN txChan);
void EEP_SetTimers(enum TAG_TX_CHAN txChan,
    uint16_t idTmr, uint16_t txTimeout, uint16_t hangTime);
void EEP_InqTimers(enum TAG_TX_CHAN txChan,
    uint32_t *pIdTmr, uint32_t *pTxTimeout, uint32_t *pHangTime);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Functions for setting CW ID paramaters

uint8_t EEP_GetMorseLevel(enum TAG_TX_CHAN txChan);
uint16_t EEP_GetMorseFreq(enum TAG_TX_CHAN txChan);
uint8_t EEP_GetMorseRate(enum TAG_TX_CHAN txChan);

void EEP_SetMorseLevel(enum TAG_TX_CHAN txChan, uint8_t level);
void EEP_SetMorseFreq(enum TAG_TX_CHAN txChan, uint16_t freq);
void EEP_SetMorseRate(enum TAG_TX_CHAN txChan, uint8_t rate);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Functions for setting and testing transmitter flags
/// @todo change name from EEP_GetTxFlags to EEP_TestTxFlag
bool EEP_GetTxFlags(enum TAG_TX_CHAN txChan, uint16_t mask);

/// @todo change name from EEP_InqTxFlags to EEP_GetTxFlags
void EEP_InqTxFlags(enum TAG_TX_CHAN txChan, uint16_t *pFlags);
void EEP_SetTxFlags(enum TAG_TX_CHAN txChan, uint16_t flags);

#define EEP_TX_CB_ENABLED(chan)     EEP_GetTxFlags( \
                                        (chan), EEP_TX_CB_ENABLE)

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Functions for setting and testing receiver flags

/// @todo change name from EEP_GetRxFlags to EEP_TestRxFlag
bool EEP_GetRxFlags(enum TAG_RX_CHAN rxChan, uint16_t mask);

/// @todo change name from EEP_InqRxFlags to EEP_GetRxFlags
void EEP_InqRxFlags(enum TAG_RX_CHAN rxChan, uint16_t *pFlags);
void EEP_SetRxFlags(enum TAG_RX_CHAN rxChan, uint16_t flags);

#define EEP_RX_IS_ENABLED(chan)     EEP_GetRxFlags( \
                                       (chan), EEP_RX_FLAG_ENABLE)

#define EEP_RX_IS_SATD_REQUIRED(chan)  EEP_GetRxFlags( \
                                            (chan), EEP_RX_FLAG_SATD)
#if 0
#define EEP_RX_TONE_REJECT_ENABLED(chan) EEP_GetRxFlags( \
                                            (chan), EEP_RX_FLAG_SATR)

#define EEP_RX_IS_DEEMPHASIS_ENABLED(chan)  EEP_GetRxFlags( \
                                            (chan), EEP_RX_FLAG_DEEMPH)

#endif

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Get Deemphasis Filter Index
unsigned EEP_GetDeemphasisFilterIndex(enum TAG_RX_CHAN rxChan);

    /// Set Deemphasis Filter Index
void EEP_SetDeemphasisFilterIndex(
                enum TAG_RX_CHAN rxChan, unsigned index);

/// Get CTCSS Rejection Filter Index
unsigned EEP_GetCtcssRejectionFilterIndex(enum TAG_RX_CHAN rxChan);

/// Set CTCSS Rejection Filter Index
void EEP_SetCtcssRejectionFilterIndex(
            enum TAG_RX_CHAN rxChan, unsigned index);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Functions for setting and testing system flags

/// @todo change name from EEP_GetSysFlags to EEP_TestSysFlag
bool EEP_GetSysFlags(uint16_t mask);
/// @todo change name from EEP_InqSysFlags to EEP_GetSysFlags
void EEP_InqSysFlags(uint16_t *pFlags);
void EEP_SetSysFlags(uint16_t flags);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // Functions to get and set DUSI start of command character

char EEP_GetDTMFCommandChar(void);
void EEP_SetDTMFCommandChar(char c);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Functions for setting receiver audio levels (in %)

uint8_t EEP_GetRxLevel(enum TAG_RX_CHAN rxChan);
void EEP_InqRxLevel(enum TAG_RX_CHAN rxChan, uint32_t *pRxLevel);
void EEP_SetRxLevel(enum TAG_RX_CHAN rxChan, uint8_t rxLevel);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Functions for setting transmitter audio levels (in %)

uint8_t EEP_GetTxLevel(enum TAG_TX_CHAN txChan);
void EEP_InqTxLevel(enum TAG_TX_CHAN txChan, uint32_t *pTxLevel);
void EEP_SetTxLevel(enum TAG_TX_CHAN txChan, uint8_t txLevel);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Functions for CTCSS encoder.

void EEP_SetCTCSSEncodeIndex(enum TAG_TX_CHAN chan, uint8_t index);
uint8_t EEP_GetCTCSSEncodeIndex(enum TAG_TX_CHAN chan);

void EEP_SetCTCSSEncodeMode(
    enum TAG_TX_CHAN chan, enum TAG_SATE_MODE mode);
enum TAG_SATE_MODE EEP_GetCTCSSEncodeMode(enum TAG_TX_CHAN chan);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Functions for controlling CTCSS decoder

//bool EEP_GetCTCSSDecodeEnable(enum TAG_RX_CHAN chan);

void EEP_SetCTCSSDecodeIndex(enum TAG_RX_CHAN chan, uint8_t index);
uint8_t EEP_GetCTCSSDecodeIndex(enum TAG_RX_CHAN chan);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Functions for Setting ID strings

char *EEP_GetIDString(enum TAG_TX_CHAN chan);
void EEP_SetIDString(enum TAG_TX_CHAN chan, const char *string);
void EEP_InqIDString(enum TAG_TX_CHAN chan, char **ppString);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Functions for setting DTMF password

char *EEP_GetDTMFPwd(void);
void EEP_SetDTMFPwd(const char *string);
void EEP_InqDTMFPwd(char **ppString);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                              END
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


