/*********************************************************************/
/**
    @file fgen.c
    Frequency Generator for ARM Applications

    This frequency generator uses 32 bit phase accumulators
    and phase increments.  It produces a 16 bit output sample.
    The sample is encoded in two's compliment format.

    Physical Channel Assignements:

    @o PWM channel 0 is the Tx1 CTCSS encoder.
    @o PWM channel 1 is the Tx2 CTCSS encoder.
    @o Codec channel 1 is the Tx1 audio tone channel.
    @o Codec channel 2 is the Tx2 audio tone channel.

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

****************************************************************/

//      *** INCLUDE FILES ***

#include <stdint.h>
#include <stdbool.h>

#include "std.h"
#include "lpc213x.h"        //lint -esym(766,lpc213x.h)
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"
#include "fgen.h"

#include "io_map.h"         //lint -esym(766,io_map.h)
#include "io_debug.h"       //lint -esym(766,io_debug.h)

/***************************************************************/

//      *** MACROS ****

#define INTERPOLATION   1   // Set to 1 to enable, 0 to disable

/***************************************************************/

//      *** GENERAL DEFINITIONS ****

    /**
        The following constant provides wave shaping for
        CW generation.  RAMP_CONSTANT_A is calculated as follows:

        RAMP_CONSTANT_A = (exp(-T/tau) * 256)

        where T is the sample period in seconds and tau is
            the desired time constant of the amplitude
            envelope.

        A value for RAMP_CONSTANT_A of 252 gives a tau of
        approximately 5 ms at a sample rate of 11,718.75
        samples per second.  249 gives a tau of approx. 3 ms.
    */
#define RAMP_CONSTANT_A         249
#define RAMP_CONSTANT_B         (256-RAMP_CONSTANT_A)

/***************************************************************/

//      *** DATA STRUCTURE Definitions ***

struct TAG_FGEN_CHAN
{
    bool enable;          // true if channel is enabled.
    uint16_t freq;        // The output frequency in 1/10 Hz.
    uint16_t sampleFreq;  // Sample frequency in Hz.
    uint16_t maxFreq;     // Maximum output freq in 1/10 Hz.
    uint8_t userGain;     // The gain (atten) requested for the channel.
    uint8_t gain;         // Effective gain applied to the channel
    int32_t z_1;          // Ramp-up/ramp-down filter delay element.
    uint32_t phaseK;      // Phase inc per sample per 1/10 Hz
    uint32_t phaseInc;    // Amount phase is incremented each sample.
    uint32_t phase;       // The accumulated phase.
};

/***************************************************************/

//      *** STATIC FUNCTION PROTOTYPES ****

static int16_t CalculateNextSample_ISR(struct TAG_FGEN_CHAN *pChan);
static void ChannelGainRamp(
    struct TAG_FGEN_CHAN *pChan, uint8_t gain);
static int16_t ProcessSample_ISR(unsigned chan);

/***************************************************************/

//      *** VARIABLES ***

static struct TAG_FGEN_CHAN channels[FGEN_NUM_CHAN];

/***************************************************************/

//      *** STATIC DATA ****

    // 16 Bit Sine table.  Stores 2 quadrants only.

static const int16_t SineTable[] =
{
    0, 804, 1608, 2410, 3212, 4011, 4808, 5602,
    6393, 7179, 7962, 8739, 9512, 10278, 11039, 11793,
    12539, 13279, 14010, 14732, 15446, 16151, 16846, 17530,
    18204, 18868, 19519, 20159, 20787, 21403, 22005, 22594,
    23170, 23731, 24279, 24811, 25329, 25832, 26319, 26790,
    27245, 27683, 28105, 28510, 28898, 29268, 29621, 29956,
    30273, 30571, 30852, 31113, 31356, 31580, 31785, 31971,
    32137, 32285, 32412, 32521, 32609, 32678, 32728, 32757,
    32767, 32757, 32728, 32678, 32609, 32521, 32412, 32285,
    32137, 31971, 31785, 31580, 31356, 31113, 30852, 30571,
    30273, 29956, 29621, 29268, 28898, 28510, 28105, 27683,
    27245, 26790, 26319, 25832, 25329, 24811, 24279, 23731,
    23170, 22594, 22005, 21403, 20787, 20159, 19519, 18868,
    18204, 17530, 16846, 16151, 15446, 14732, 14010, 13279,
    12539, 11793, 11039, 10278, 9512, 8739, 7962, 7179,
    6393, 5602, 4808, 4011, 3212, 2410, 1608, 804
};

    // Slope table: Used for linear interpolation between
    // samples.

static const int16_t SlopeTable[] =
{
    804, 804, 803, 801, 799, 797, 794, 791,
    787, 782, 778, 772, 767, 760, 754, 747,
    739, 731, 723, 714, 705, 695, 685, 674,
    663, 652, 640, 628, 615, 602, 589, 576,
    562, 547, 533, 518, 503, 487, 471, 455,
    439, 422, 405, 388, 370, 353, 335, 317,
    299, 280, 262, 243, 224, 205, 186, 167,
    147, 128, 108, 89, 69, 49, 30, 10,
    -10, -30, -49, -69, -89, -108, -128, -147,
    -167, -186, -205, -224, -243, -262, -280, -299,
    -317, -335, -353, -370, -388, -405, -422, -439,
    -455, -471, -487, -503, -518, -533, -547, -562,
    -576, -589, -602, -615, -628, -640, -652, -663,
    -674, -685, -695, -705, -714, -723, -731, -739,
    -747, -754, -760, -767, -772, -778, -782, -787,
    -791, -794, -797, -799, -801, -803, -804, -804
};

/*****************************************************************
                                Program Code
*****************************************************************/


void FGen_Reset(void)
{

}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    //  Interrupt service routines that calculate and deliver
    // samples for channels.  Channels 0 and 1 are for the
    // PWM DACs.  Channels 2 and 3 are for the audio filter
    // stream.



// FGen Channel 0 Interrupt Service Routine

int16_t FGen_SourceChan0_ISR(void)
{
    return(ProcessSample_ISR(0));
}

// FGen Channel 1 Interrupt Service Routine

int16_t FGen_SourceChan1_ISR(void)
{
    return(ProcessSample_ISR(1));
}

// FGen Channel 2 Interrupt Service Routine

int16_t FGen_SourceChan2_ISR(void)
{
    return(ProcessSample_ISR(2));
}

    // FGen Channel 3 Interrupt Service Routine

int16_t FGen_SourceChan3_ISR(void)
{
    return(ProcessSample_ISR(3));
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static int16_t ProcessSample_ISR(unsigned chan)
{
    if(channels[chan].enable)
    {
        ChannelGainRamp(&channels[chan], channels[chan].userGain);
        return((int16_t)CalculateNextSample_ISR(&channels[chan]));
    }
    else if(channels[chan].gain > 0)
    {
        ChannelGainRamp(&channels[chan], 0);
        return((int16_t)CalculateNextSample_ISR(&channels[chan]));
    }
    else
        return(0);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void ChannelGainRamp(
                    struct TAG_FGEN_CHAN *pChan, uint8_t gain)
{
    pChan->z_1 = gain + ((pChan->z_1 * RAMP_CONSTANT_A) >> 8);

    pChan->gain =
        (uint8_t)((pChan->z_1 * RAMP_CONSTANT_B) >> 8);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Calculate the next sample for the given channel.

static int16_t CalculateNextSample_ISR(
                                    struct TAG_FGEN_CHAN *pChan)
{
    uint8_t index, offset, sign;
    int32_t sample;
    int16_t slope;

    pChan->phase += pChan->phaseInc;  // Increment the phase.
    index = pChan->phase >> 24;       // extract table index.
    sign = index & 0x80;              // Remember the sign
    index &= 0x7F;

    // Sine table contains 2 quadrants of a full wave.

    sample = SineTable[index];

    // Linear interpolation between points:

#if INTERPOLATION == 1
    slope = SlopeTable[index];
    offset = (pChan->phase >> 16) & 0xFF;
    sample += ((slope * offset) + 128) >> 8;
#endif

    sample = ((sample * pChan->gain) + 128) >> 8;

    // Now map the sample to 4 quadrants

    return((int16_t)(sign? (-sample) : sample));
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Initialize a channel:

    // Define Phase Constant.  This is the number that the
    // desired frequency (in Hz) must be multiplied by, to obtain
    // the phase increment.  It is equal to 2^n/Rs where
    // n is the number of bits in the phase increment, and
    // Rs is the sample rate in Hz.

    // In the case of n=16 and Rs=8.000 KHz.

    // (2^n/Rs is multiplied by 65536 to give better resolution).

    //      K.phi = 536870.9

    // Since we want the increment in 10ths of a Hz, we divide
    // the phase constant by 10 and round up.

void FGen_Channel_Init(uint8_t chan, uint16_t sampleFreq)
{
    ASSERT(chan < FGEN_NUM_CHAN);
    FGen_ChannelOff(chan);  // Stop it if it's running.

    // Sample frequency in Hz:
    channels[chan].sampleFreq = sampleFreq;

    // Calculate the maximum output frequency in 10ths of
    // a Hz.
    ASSERT(sampleFreq < 65535/5);
    channels[chan].maxFreq = (uint16_t)(sampleFreq * 5);

    // Frequency generator constants:
    channels[chan].phaseK = ((uint32_t)1 << 31)/(sampleFreq * 5);
    channels[chan].phase = 0;
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//  Turn on waveform generation.

void FGen_ChannelOn(uint8_t chan)
{
    ASSERT(chan < FGEN_NUM_CHAN);

    DI();
    channels[chan].z_1 = 0;
    channels[chan].enable = true;
    EI();
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Stop waveform generation.

void FGen_ChannelOff(uint8_t chan)
{
    ASSERT(chan < FGEN_NUM_CHAN);

    DI();
    channels[chan].enable = false;
    EI();
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // FGen_IncrementPhase is used to implement "reverse-burst".
    // The phase is incremented by the fraction of a cycle
    // contained in phaseInc.  phaseInc is a fraction with the
    // high order bit representing 1/2.

void FGen_IncrementPhase(uint8_t chan, uint16_t phaseInc)
{
    ASSERT(chan < FGEN_NUM_CHAN);

    DI();
    channels[chan].phase += (phaseInc << 16);
    EI();
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Set the frequency of a waveform.

    // chan 0 to 3
    // freq is in 10ths of a Hz.

void FGen_SetChannelFreq(uint8_t chan, uint16_t outputFreq)
{
    ASSERT(chan < FGEN_NUM_CHAN);

    // Clip the frequency to the limits
    outputFreq =
        CLIP(outputFreq, channels[chan].maxFreq, FGEN_MIN_FREQ);

    DI();
    channels[chan].freq = outputFreq;
    channels[chan].phaseInc =
        channels[chan].phaseK * outputFreq;
    EI();
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Set the gain of the waveform.  Either gain or attenuation
    // may be obtained.

    // Each sample is multiplied by the gain and then divided
    // by 256, so gain values between 0 and 255 actually
    // result in attenuation.  A gain of 255 gives a 16 bit full-
    // scale signal.

    // Because we need to take into account the codec's reference
    // variation over devices and temperatures, we must reduce
    // the gain by the amount of CDX_0DB_REFERENCE/32768 to insure
    // that we don't clip.

void FGen_SetChannelGain(uint8_t chan, uint8_t gain)
{
    ASSERT(chan < FGEN_NUM_CHAN);

    DI();
    channels[chan].userGain = gain;
    EI();
}

/****************************************************************
                End of Waveform Generator Module
****************************************************************/
