/****************************************************************//**

    @file fgen.h
    Frequency Generator for ARM Applications

    This frequency generator uses 32 bit phase accumulators
    and phase increments.  It produces a 16 bit output sample.
    The sample is encoded in two's compliment format.

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

********************************************************************/

//      *** GENERAL DEFINITIONS ***

    /// Channel minimum frequency in 1/10 Hz.  Channel maximum
    /// frequency is determined by sample rate.  See
    /// FGen_Channel_Init().

#define FGEN_MIN_FREQ       1

#define FGEN_MIN_GAIN       1       ///< Minimum gain
#define FGEN_MAX_GAIN   UINT8_MAX   ///< Maximum gain

    // Logical Channel Assignments:

#define FGEN_NUM_CHAN           4   ///< 4 channels are supported

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Post-Reset Initialization

    This function initializes the function generator after a
    hardware reset.

    @return Nothing.
    @context This function may be called anytime after reset and
        before Apex_Start(). It does not make any system calls nor
        does it enable interrupts.
*/
void FGen_Reset(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Function Generator Channel Startup Initialization

    This function initializes a function generator channel by
    setting its sampling frequency and initializing its phase
    to zero.

    @param chan A byte specifying which channel to initialize.
    @param sampleFreq An unsigned integer specifying the channel's
        sample frequency in hertz.

    @return Nothing.

    @context Must be called after multitasking startup since it disables
        and re-enables interrupts.
*/

void FGen_Channel_Init(uint8_t chan, uint16_t sampleFreq);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void FGen_ChannelOff(uint8_t chan);
void FGen_ChannelOn(uint8_t chan);

    // Freq in 1/10 Hz
void FGen_SetChannelFreq(uint8_t chan, uint16_t freq);

void FGen_SetChannelGain(uint8_t chan, uint8_t gain);

void FGen_IncrementPhase(uint8_t chan, uint16_t cycleFraction);

    // FGen ISR Routines:

int16_t FGen_SourceChan0_ISR(void);
int16_t FGen_SourceChan1_ISR(void);
int16_t FGen_SourceChan2_ISR(void);
int16_t FGen_SourceChan3_ISR(void);
