//**************************************************************
/**
    @file filesys.c
    API for Flash File System for PSE-702

    @author ALP

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

    @section fsys_req Requirements:

    Since this file system is being hosted by a flash memory,
    writes may only occur into locations that have been erased.
    This module currently assumes that the flash may be erased
    in units of 4KB blocks.

    Because of the nature of the erase-write paradigm, all
    information necessary to operate the file system must be
    contained within the files themselves (that is, no common
    directory).  This avoids the need to erase and re-write
    some common area such as a directory or free-sector list
    every time a data write operation occurs.

    Therefore, erasure only occur when a file is deleted, and
    only the blocks associated with that file are erased.

    The first several blocks are base blocks for the files.
    For example, if the file system supports a maximum of 4
    files, blocks 0 through 3 are the base blocks for the
    four files.

    The first two pages of each base block contain block tables
    and free page maps for the file.  The remaining pages
    and blocks contain data.

    A byte level read/write capability must be provided by the
    flash driver.

    @section fsys_header File Header Map:

    Each file begins with the base block.  This is block 0
    for file 0, block 1 for file 1, etc. up to block 3 in a
    4 file system.  The base block contains the directory
    information for its file.

    The directory is an array of entries, one for each logical
    block allocated to the file.  Each entry consists of two
    16 bit words: the physical block address, and the free
    page bitmap.

    Since there are 4096 bytes in the base block, there are
    1024 possible directory entries for each file.  Thus a
    file could contain 1024 physical blocks of 4096 bytes
    each, or 4 megabytes of memory.

    When the file system is started, a free-block bitmap is
    constructed in SRAM to aid in the allocation of new
    blocks to files.  This is done by scanning each directory
    block for each file, and recording all allocated blocks.

    When a new page is to be written to a file, the system first
    determines if there is a free page in the last physical block
    allocated to the file, and if so, that page is written.  Then
    the free-page bitmap in the file's directory for that
    logical block is changed to reflect the allocation of the new
    page.

    If all pages in the current physical block are already
    allocated, a new physical block is allocated from the free-
    block bitmap, and a new directory entry is recorded in the
    file's directory.

    @defgroup filesys File System
    @ingroup module_design

****************************************************************/

//      *** INCLUDE FILES ***

#include <stdint.h>
#include <stdbool.h>

#include "std.h"            //lint -esym(766, std.h)
//#include "tp_system.h"      //lint -esym(766, tp_system.h)
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"
#include "apex.h"
#include "filesys.h"
#include "io_flash.h"

#include "io_debug.h"       //lint -esym(766, io_debug.h)
#include "apprint.h"        //lint -esym(766, apprint.h)
#include "io_main.h"        //lint -esym(766, io_main.h)

/***************************************************************/
//      *** MACROS ****

/***************************************************************/

//      *** GENERAL DEFINITIONS ****

    /// Maximum number of blocks allowed per file.  This takes
    /// into account the fact that one block is allocated
    /// for the directory.

#define MAX_BLOCKS_PER_FILE ( \
        NUM_BLOCKS/sizeof(struct TAG_DIRECTORY_ENTRY))

    /// Convert a file number and logical block number
    /// to the physical flash byte address of the corresponding
    /// 'struct TAG_DIRECTORY_ENTRY' in the directory.

#define DIRECTORY_ENTRY_ADDR(f,b) ((struct TAG_DIRECTORY_ENTRY *) \
   (((f)*Z_FLASH_BLOCK) + ((b)*sizeof(struct TAG_DIRECTORY_ENTRY))))

/***************************************************************/
//      *** DATA STRUCTURE Definitions ***

    /** Assigned Block Structure

        An array of these structures is used in the first block
        of each file to keep track of which blocks have been
        allocated to the file, and which pages within those
        blocks have been written.
    */
struct TAG_DIRECTORY_ENTRY
{
    uint16_t phyBlockNum;
    uint16_t freePageMap;
};

/***************************************************************/
//      *** CONSTANTS ***

/***************************************************************/

//      *** VARIABLES ***

    //  Free Block Map

    /// Number of words in freeBlockMap

#define Z_FREE_BLOCK_MAP    (NUM_BLOCKS >> 4)

    /// Free block map.  16 bit word is used because searching
    /// functions can be used to search this map as well as the
    /// free page map used elsewhere.

static uint16_t freeBlockMap[Z_FREE_BLOCK_MAP];

    /// Allocate a free block by clearing the corresponding bit
    /// in the freeBlockMap
    /// @a x is the block number to allocate.

#define ALLOCATE_FREE_BLOCK(x)   freeBlockMap[(x) >> 4] &= \
                            ~((uint16_t)(1 << ((x) & 0x000F)))

    /// Compute a free block number given the bitmap index and
    /// bit number within the indexed word.
    /// @a i is the bitmap index
    /// @a b is the bit number within the indexed word.

#define FREE_BLOCK_BIT_2_BLOCK_NUM(i,b)     ((i << 4) + (b))


    /** Opened File Information

        An array of these structures is used to keep track of
        certain pieces of information about each opened file.
    */
static struct TAG_OPENED_FILE
{
    uint16_t readPtrLogBlock;   ///< Logical block address corresponding
                                ///< to physical block below.
    uint16_t readPtrPhyBlock;   ///< Current read pointer physical block

    /// Last physical block and page bitmap allocated to file.
    struct TAG_DIRECTORY_ENTRY lastDirEntry;

    uint16_t numBlocks;         ///< Number of blocks in file.
    uint16_t numPages;          ///< Number of pages in file.

} openedFiles[NUM_FILES];

    /// Allocate a free page in a file by clearing the corresponding bit
    /// in the freePageMap
    /// @a f is the file number
    /// @a p is the page number to allocate.

#define ALLOCATE_FREE_PAGE(f,p) openedFiles[fileNum]. \
    lastDirEntry.freePageMap &= ~((uint16_t)(1 << ((p) & 0x000F)))

static MUTEX fileSystemAccess;  // Only one file system user at a time.

const struct TAG_VARIABLE_TABLE filesysVarTab[] =
{
//  AVARIABLE(freeBlockMap),
//  VARIABLE(openFileNum),
//  VARIABLE(openLogBlockNum),
//  VARIABLE(openPhyBlockNum),
//  VARIABLE(openPageBitMapAddr),
//  VARIABLE(openPageBitMap),
//  VARIABLE(openNextLogPageNum),
//  VARIABLE(fileBusySema),
    {"blocks", freeBlockMap, VAR_TYPE_NONE, sizeof(freeBlockMap)},

    {NULL, NULL, VAR_TYPE_NONE, 0}
};


/***************************************************************/

//      *** STATIC FUNCTION PROTOTYPES ****

static void BuildFreeBlockMap(void);
static bool ReadFileDirectory(unsigned fileNum,
    unsigned logBlockNum, struct TAG_DIRECTORY_ENTRY *pBlock);
static uint16_t Bitmap2Num(uint16_t bitmap);
static bool AllocatePhysicalBlockToFile(unsigned fileNum,
    uint16_t logBlockNum, struct TAG_DIRECTORY_ENTRY *pBlock);


//***************************************************************
//                          API Functions
//***************************************************************

    // Initialize the file system.

void FSys_Reset(void)
{
    // Protect the file system with a mutex that only allows
    // access to one task at a time.  Initialize it as locked
    // so that no access may occur until FSys_Start is called.
    APEX_INIT_MUTEX_LOCKED(&fileSystemAccess);
}


    // Start running the file system: Must be called by a
    // task after APEX_Start() because Apex_Pend() is used.
    // (Note: cannot be called from Apex_Target_Start()
    // either).

    // This function unlocks the mutex, so that no task
    // can actually use the file system until this function
    // is executed.

void FSys_Start(void)
{
    BuildFreeBlockMap();

    // Unprotect the flash so we can write to it.

    Flash_UnprotectChip();
    APEX_UNLOCK_MUTEX(&fileSystemAccess);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Erase the flash and create a file system.
    */
void FSys_Format(void)
{
    unsigned i;

    APEX_LOCK_MUTEX(&fileSystemAccess, 0);

    // Erase the entire flash:
    // (except don't erase last block where EEPROM data is stored).

    for(i=0; i<(NUM_BLOCKS-1); i++)
    {
        Flash_EraseBlock(i);
    }

    BuildFreeBlockMap();

    APEX_UNLOCK_MUTEX(&fileSystemAccess);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell the caller how many data pages are allocated to a
    // file.

unsigned FSys_FileSize(unsigned fileNum)
{
    return(openedFiles[fileNum].numPages);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell the caller how many pages are unallocated.

unsigned FSys_FreePages(void)
{
    unsigned i;
    unsigned usedPages = 0;

    for(i=0; i<NUM_FILES; i++)
    {
        // The number of pages used in a file is equal to the
        // directory pages plus the number of user data pages.
        usedPages +=
            (NUM_PAGES_PER_BLOCK + openedFiles[i].numPages);
    }

    // Calculate free pages.  This is equal to the total number
    // of pages in the flash, minus pages that store user
    // parameters, minus the pages used in files.

    return( NUM_PAGES
        - (NUM_USER_BLOCKS * NUM_PAGES_PER_BLOCK)
        - usedPages);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Read a Page from a File

        This function reads the logical page addressed by the
        pageNum parameter.
    */

bool FSys_ReadPage(unsigned fileNum, unsigned pageNum, uint8_t *pPage)
{
    uint16_t logBlockNum;
    uint16_t phyPageNum;
    struct TAG_DIRECTORY_ENTRY dirEntry;
    bool result = false;     // Default result.

    ASSERT(fileNum < NUM_FILES);

    APEX_LOCK_MUTEX(&fileSystemAccess, 0);

    // Check to see if end of file.
    // If not end of file, read addressed physical page.

    if(pageNum < openedFiles[fileNum].numPages)
    {
        // Develop the physical address given the logical address:

        logBlockNum = PAGE_2_BLOCK_ADDR(pageNum);

        // See if we can read the next physical page in the current
        // block, or if we need to get a new block address.
        if(logBlockNum != openedFiles[fileNum].readPtrLogBlock)
        {
            // Here, the new page goes beyond the current
            // logical block, so read a new physical block
            // address.
            ReadFileDirectory(fileNum, logBlockNum, &dirEntry); //lint !e534

            openedFiles[fileNum].readPtrPhyBlock = dirEntry.phyBlockNum;
            openedFiles[fileNum].readPtrLogBlock = logBlockNum;
        }

        phyPageNum = LOG_PAGE_2_PHY_PAGE(
                    openedFiles[fileNum].readPtrPhyBlock, pageNum);
        Flash_ReadPage(phyPageNum, pPage);
        result = true;
    }

    APEX_UNLOCK_MUTEX(&fileSystemAccess);
    return(result);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Append the page pointed to by pPage to the end of the
    // currently open file.

    // Returns true if the page was written, or false if memory
    // is full.

bool FSys_AppendPage(unsigned fileNum, uint8_t *pPage)
{
    uint16_t pageNum;
    uint16_t pageAddr;
    uint32_t directoryAddr;
    uint8_t *pPageBitmapByte;

//    static unsigned breakpointCount = 0;

    ASSERT(fileNum < NUM_FILES);

    APEX_LOCK_MUTEX(&fileSystemAccess, 0);

    // Here, if freePageBitmap was zero, we need to allocate
    // a new block.

    if(openedFiles[fileNum].lastDirEntry.freePageMap == 0)
    {
        // Here, check to make sure we have not already allocated
        // the maximum file length.
        if(openedFiles[fileNum].numBlocks == MAX_BLOCKS_PER_FILE)
        {
            APEX_UNLOCK_MUTEX(&fileSystemAccess);
            return(false);
        }

        // Allocate a new block to the file.

        if(!AllocatePhysicalBlockToFile(
            fileNum, openedFiles[fileNum].numBlocks,
            &openedFiles[fileNum].lastDirEntry))
        {
            APEX_UNLOCK_MUTEX(&fileSystemAccess);
            return(false);
        }

        openedFiles[fileNum].numBlocks++;
    }

    // We have a non-zero page bitmap now, so allocate
    // a new page.

    pageNum = Bitmap2Num(
        openedFiles[fileNum].lastDirEntry.freePageMap);

    ALLOCATE_FREE_PAGE(fileNum, pageNum);
    openedFiles[fileNum].numPages++;

    // Write the page bitmap back to the directory for the
    // file.  Only write the byte that changed.

    directoryAddr = (uint32_t)&((DIRECTORY_ENTRY_ADDR(fileNum,
        (openedFiles[fileNum].numBlocks-1)))->freePageMap);

    pPageBitmapByte =
        (uint8_t *)&openedFiles[fileNum].lastDirEntry.freePageMap;

    if((pageNum & 0x000F) > 7)
    {
        directoryAddr++;
        pPageBitmapByte++;
    }

    Flash_WriteBytes(directoryAddr, pPageBitmapByte, 1);

    pageAddr = (uint16_t)BLOCK_2_PAGE_ADDR(
        openedFiles[fileNum].lastDirEntry.phyBlockNum) + pageNum;

    // Finally, write the page data

    Flash_WritePage(pageAddr, pPage);

    APEX_UNLOCK_MUTEX(&fileSystemAccess);

    return(true);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Delete File: Erases all blocks used by a file
    // and initializes the block table and page bit
    // map.

    // File cannot be deleted if currently opened.

void FSys_Delete(unsigned fileNum)
{
    uint8_t logBlockAddr;
    struct TAG_DIRECTORY_ENTRY dirEntry;

    ASSERT(fileNum < NUM_FILES);

    APEX_LOCK_MUTEX(&fileSystemAccess, 0);

    // Erase all blocks in the file (except for the base block).

    logBlockAddr=0;
    while(ReadFileDirectory(fileNum, logBlockAddr++, &dirEntry))
    {
        Flash_EraseBlock(dirEntry.phyBlockNum);

        // Deallocate the block in the free block map.
        freeBlockMap[dirEntry.phyBlockNum >> 4] |=
                ((uint16_t)(1 << (dirEntry.phyBlockNum & 0x000F)));
    }

    // Now erase the base block.

    Flash_EraseBlock(fileNum);

    // Reset the open file parameters to those of an empty file.

    openedFiles[fileNum].numBlocks=0;
    openedFiles[fileNum].numPages=0;
    openedFiles[fileNum].readPtrLogBlock = 0xFFFF;
    openedFiles[fileNum].readPtrPhyBlock = 0xFFFF;
    openedFiles[fileNum].lastDirEntry.phyBlockNum = 0;
    openedFiles[fileNum].lastDirEntry.freePageMap = 0;

    APEX_UNLOCK_MUTEX(&fileSystemAccess);
}

//***************************************************************
//                      Static Functions
//***************************************************************

/** Build Free Block Map

        This function creates a bit map for used and unused
        flash blocks.  Unused blocks are represented by a set
        bit, and unused blocks are represented by a clear bit.

        The map is built by examining all existing files and
        recording the blocks used by those files as 'used' in
        the bit map.
     */
static void BuildFreeBlockMap(void)
{
    uint16_t i;
    uint16_t logBlockAddr;      ///< Logical block address
    struct TAG_DIRECTORY_ENTRY dirEntry;  // Put block allocation data
                                          // here

    // Initialize free block map to 'all blocks free':

    for(i=0; i<(Z_FREE_BLOCK_MAP); i++)
        freeBlockMap[i] = 0xFFFF;

    // Build free block map.  Go through all base blocks of
    // each file:

    for(i=0; i<NUM_FILES; i++)
    {
        // Allocate base block for each file.
        ALLOCATE_FREE_BLOCK(i);

        // Now for each file, go through the Block-Map
        // and mark each allocated block in the Free-Block
        // map.

        openedFiles[i].numBlocks = 0;
        openedFiles[i].numPages = 0;
        openedFiles[i].readPtrLogBlock = 0xFFFF;
        openedFiles[i].readPtrPhyBlock = 0xFFFF;
        openedFiles[i].lastDirEntry.phyBlockNum = 0;
        openedFiles[i].lastDirEntry.freePageMap = 0;

        for(logBlockAddr=0;
            logBlockAddr<MAX_BLOCKS_PER_FILE; logBlockAddr++)
        {
                // Read the directory entry, and test to make sure
                // it was actually allocated.
            if(ReadFileDirectory(i, logBlockAddr, &dirEntry))
            {
                // Here, a block is allocated to the file, so
                // clear the corresponding bits in the bitmap.

                ALLOCATE_FREE_BLOCK(dirEntry.phyBlockNum);

                // Here, record the address parameters so that when
                // we have looked at the last allocated block, we have
                // it's address parameters recorded for the open file.
                // This information will be used when the user tries
                // to append more data.

                openedFiles[i].lastDirEntry = dirEntry;
                openedFiles[i].numBlocks++;
            }
            else
            {
                // Here, we are at the end of the allocations for this
                // file.  The Append parameters should be up to date.

                if(openedFiles[i].numBlocks != 0)
                {
                    openedFiles[i].numPages = (uint16_t)(
                        openedFiles[i].numBlocks *
                        NUM_PAGES_PER_BLOCK);

                    if(openedFiles[i].lastDirEntry.freePageMap != 0)
                    {
                        openedFiles[i].numPages -=
                                NUM_PAGES_PER_BLOCK - Bitmap2Num(
                                openedFiles[i].lastDirEntry.freePageMap);
                    }
                }

                // Here, the file is of zero size.  No writes have
                // occured yet.
                else
                {
                    // A zero freePageMap will cause the Append
                    // function to allocate the first block.
                    openedFiles[i].numPages = 0;
                    openedFiles[i].lastDirEntry.phyBlockNum = 0;
                    openedFiles[i].lastDirEntry.freePageMap = 0;
                }
                break; // for logBlockAddr

            } // if
        } // for logBlockAddr
    } // for filenum

    // Allocate user parameter blocks.
    for(i=0; i<NUM_USER_BLOCKS; i++)
    {
        ALLOCATE_FREE_BLOCK((NUM_BLOCKS-1)-i);
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Convert A 16 Bit Map to Bit Number

        This function examines a 16 bit map and returns the ordinal
        bit number of the least significant set bit.  It assumes that
        at least one bit is set, and asserts if no bits are set.

        @param bitmap A non-zero value containing bits
        @returns A number indicating the lowest order bit that
            was set in the map.
    */
static uint16_t Bitmap2Num(uint16_t bitmap)
{
    uint16_t i = 0;

    // Not using the assert because if the bitmap is zero, we still
    // want to initialize so we can format the flash.

    ASSERT(bitmap != 0);

//    if(bitmap == 0)
//        return(16);
//    else
//    {
        while((bitmap & 0x0001) == 0)
        {
            bitmap = bitmap >> 1;
            i++;
        }
        return(i);
//    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Read a Record in the File's Directory

        This function reads the block allocation directory for a
        file.  It returns the physical block address and page
        bitmap corresponding to a file's logical block address.

        @param fileNum The file number for the addressed file.
        @param logBlockNum The logical block number for the block
            who's directory record is to be read.
        @param pDirEntry A pointer to the buffer into which the
            directory entry should be placed.

        @returns Returns true if the logical block is
            allocated in the directory.  In that case the
            data in the record pointed to by pDirEntry is valid.
            Returns false if the logical block is unallocated,
            in which case the data in the record pointed to
            by pDirEntry is undefined.
     */
static bool ReadFileDirectory(unsigned fileNum,
    unsigned logBlockNum, struct TAG_DIRECTORY_ENTRY *pDirEntry)
{
    struct TAG_DIRECTORY_ENTRY *pDirectoryAddr;

    /// Check to make sure that the logical block number is
    /// not too big.

    if(logBlockNum <= MAX_BLOCKS_PER_FILE)
    {
        /// The physical address of the allocation record is
        /// the displacement into the first file's block to the
        /// first byte that comprises the record.  Data is stored
        /// little-endian.

        pDirectoryAddr =
            DIRECTORY_ENTRY_ADDR(fileNum, logBlockNum);

        Flash_ReadBytes((uint32_t)pDirectoryAddr,
            (uint8_t *)pDirEntry, sizeof(struct TAG_DIRECTORY_ENTRY));

        if(pDirEntry->phyBlockNum != 0xFFFF)
            return(true);   // Block is allocated.
        else
            return(false);  // Block not allocated.
    }
    else
        return(false);  // Too many blocks allocated.
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Allocate a New Physical Block to a File

        This function writes a new block allocation table entry
        for a file.

        @param fileNum The file number for the addressed file.
        @param logBlockNum The logical block number for the block
            who's allocation is to be written.
        @param pDirEntry A pointer to the buffer into which the
            allocation data should be placed.
        @return The physical block number of the block that was
            allocated is returned.
     */
static bool AllocatePhysicalBlockToFile(unsigned fileNum,
    uint16_t logBlockNum, struct TAG_DIRECTORY_ENTRY *pDirEntry)
{
    uint16_t freeBlockBits;
    struct TAG_DIRECTORY_ENTRY *pDirectoryAddr;
    unsigned freeBlockBitNum;

    // Search the free block map for a free block.

    unsigned bitMapIndex = 0;
    while((freeBlockBits = freeBlockMap[bitMapIndex]) == 0)
    {
        // Bump the free block number.  If it goes over,
        // tell the caller we are out of memory.

        if(++bitMapIndex == Z_FREE_BLOCK_MAP)
            return(false);
    }

    // Here, found room.  Get the block address and clear
    // the bit in the free-block bitmap.

    freeBlockBitNum = Bitmap2Num(freeBlockBits);

    pDirEntry->phyBlockNum =
            (uint16_t)FREE_BLOCK_BIT_2_BLOCK_NUM(
            bitMapIndex, freeBlockBitNum);

    ALLOCATE_FREE_BLOCK(pDirEntry->phyBlockNum);

    // Now write the new block in the file's directory.

    pDirEntry->freePageMap = 0xFFFF;

    /// The physical address of the allocation record is
    /// the displacement into the file's first block to the
    /// first byte that comprises the record.

    pDirectoryAddr = DIRECTORY_ENTRY_ADDR(fileNum, logBlockNum);

    Flash_WriteBytes(
        (uint32_t)&pDirectoryAddr->phyBlockNum,
        (uint8_t *)&pDirEntry->phyBlockNum,
        sizeof(pDirEntry->phyBlockNum));

    /// It isn't necessary to write the page bitmap because
    /// it should already be set to 0xFFFF since it is
    /// erased.

    return(true);
}
//***************************************************************
//                              End of File
//***************************************************************
