//**************************************************************
/**

    @file filesys.h

    API for Flash File System for PSE-702

    @author ALP

    @version 110916 Modifications for large flashes.
    @version 080923 File started.

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

    @defgroup filesys_api File System API
    @ingroup module_api

    @{
****************************************************************/

//              *** GENERAL DEFINITIONS ****

#define NUM_FILES   16 // Number of files supported.  File numbers
                        // range from 0 to NUM_FILES-1.

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/**
    Description: Initializes the file system after reset.

    Calling Context: Reset Initialization.
 */
void FSys_Reset(void);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Initialize File System

    Initializes SRAM image of file system parameters,
    such as free sector map.  Requires Flash driver operation and
    task-based system calls to read Flash and build tables.

    Calling Context: Task
*/
void FSys_Start(void);


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Read a page from a file.

    Reads the page addressed by pageNum and returns the data
    it in the buffer pointed to by pPage.

    @param fileNum: The file number of the file to read from.

    @param pageNum: The page number of the page to read.

    @param pPage: A pointer to a page-sized buffer in SRAM
        into which the file page shall be read.

    @return true if the page was read, false if the addressed
        page is beyond the end of file.
*/
bool FSys_ReadPage(
        unsigned fileNum, unsigned pageNum, uint8_t *pPage);


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Append a page to a file.
 *
 * Appends the data pointed to by pPage to the end of the file
 * addressed by fileNum.

    @param fileNum: The file number of the file to append to.

    @param pPage: A pointer to a page-sized buffer
    in SRAM containing the data to append.

    @return true if the page was appended, false if the page
    was not appended (flash full).
*/
bool FSys_AppendPage(unsigned fileNum, uint8_t *pPage);


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Inquire of the size of a file.

    @param fileNum: The file number of the file to inquire of.

    @return The number of pages in the file is returned.
*/
unsigned FSys_FileSize(unsigned fileNum);


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Reutn the number of free pages left in the file system.

    @return The number of free pages is returned.
*/
unsigned FSys_FreePages(void);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Format the file system.

    Clears all data from all files and prepares the file system
    for operation.  Does not clear the last block in the flash
    as this is used for "eeprom" storage.
*/

void FSys_Format(void);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Delete a file

    Clears all data from the addressed file.  The file is thus
    empty and ready to be re-written.

    @param fileNum: The file number of the file to delete.
*/

void FSys_Delete(unsigned fileNum);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

extern const struct TAG_VARIABLE_TABLE filesysVarTab[];

//*********************************************************************
///@}                          END OF FILE
//*********************************************************************
