/*********************************************************************/
/**
    @file filter.h
    Audio Processing Stream Interrupt for PSE-702

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************/


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** First Order Filter Coefficients Structure

    The 1st order diagrams in the Mathcad spreadsheets show a
    term b0, which, in the implementations below, is factored
    out of the numerator terms to give an overall gain factor.
    The following additional simplifications are also performed:

    High-pass: b1 = -1. The value of b1 in the parameter
        structure is ignored.

    Low-Pass: b1 = +1. The value of b1 in the parameter
        structure is ignored.

    General: All parameters are used with the signs provided
        in the constant structure.
*/

typedef struct TAG_FIRST_ORDER_COEFF
{
    uint16_t scale_factor;  ///< Pwr of 2 by which parms are upscaled
    uint16_t k;             ///< Overall gain
    int16_t a;              ///< Feedback parm
    int16_t b;              ///< Feed forward parm

} first_order_coeff_t;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** First Order Delay Term Structure
     *
     * This structure declares the delay terms to be used with a
     * 1st order filter.  Each instance of such a filter requires
     * its own delay term structure, even if they share the same
     * parameter structure.
     */
typedef struct TAG_FIRST_ORDER_DELAYS
{
    int32_t zz;       ///< 'Zero' delay element
    int32_t zp;       ///< 'Pole' delay element

} first_order_delay_t;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** 2nd Order Biquad Coefficients Structure

        The biquad diagrams in the Mathcad spreadsheets show a
        term b0, which, in the implementations below, is factored
        out of the numerator to give an overall gain factor k.
        The following additional simplifications are also performed:

        High-pass: b1 = -2 and b2 = +1. The value of b1 and b2
        in the coefficient structure are ignored.

        Notch: b2 = 1. The value of b2 in the coefficient structure
            is ignored.

        Low-Pass: b1 = +2 and b2 = +1, and the values of b1 and b2
            in the coefficient structure are ignored.

        Band-pass: b1 = 0 and b2 = +1, and the values of b1 and b2
            in the coefficient structure are ignored.

        General: All parameters are used with the signs provided
            in the coefficient structure.
*/

typedef struct TAG_BIQUAD_COEFF
{
    uint16_t scale_factor;  ///< Amount by which parms are upscaled
    uint16_t k;             ///< Overall gain
    int16_t a1, a2;         ///< Feedback parms
    int16_t b1, b2;         ///< Feed forward parms

} biquad_coeff_t;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Biquad Delay Term Structure
     *
     * This structure declares the delay terms to be used with a
     * 2nd order filter.  Each instance of such a filter requires
     * its own delay term structure, even if they share the same
     * parameter structure.
     */
typedef struct TAG_BIQUAD_DELAYS
{
    int32_t zz1, zz2;       ///< 'Zeros' delay elements
    int32_t zp1, zp2;       ///< 'Pole' delay elements

} biquad_delay_t;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void FILT_Reset(void);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void FILT_Start(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Define a first order filter type so we can use pointers
    /// to the filter in structures.

typedef int16_t first_order_filter_t(first_order_coeff_t *pParms,
        first_order_delay_t *pDelay, int16_t x);

int16_t FILT_1stOrderLowPass(const first_order_coeff_t *pParms,
                            first_order_delay_t *pDelay, int16_t x);

int16_t FILT_1stOrderHighPass(const first_order_coeff_t *pParms,
                            first_order_delay_t *pDelay, int16_t x);

int16_t FILT_1stOrderGeneralized(const first_order_coeff_t *pParms,
                            first_order_delay_t *pDelay, int16_t x);

int16_t FILT_2ndOrderLowPass(
    const biquad_coeff_t *pParms, biquad_delay_t *pDelay, int16_t x);

int16_t FILT_2ndOrderHighPass(
    const biquad_coeff_t *pParms, biquad_delay_t *pDelay, int16_t x);

int16_t FILT_2ndOrderBandPass(
    const biquad_coeff_t *pParms, biquad_delay_t *pDelay, int16_t x);

int16_t FILT_2ndOrderNotch(
    const biquad_coeff_t *pParms, biquad_delay_t *pDelay, int16_t x);

int16_t FILT_2ndOrderGeneralized(
    const biquad_coeff_t *pParms, biquad_delay_t *pDelay, int16_t x);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

extern const struct TAG_VARIABLE_TABLE filtVarTab[];

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                          END
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
