/*********************************************************************/
/**
    @file io_codec.c

    Driver for ARM SSI device as used in PSE-702 to
    communicate with the AD73322L codec.

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

    The codec driver controls a two channel codec.  The channels
    are numbered 1 and 2 in the hardware.

****************************************************************/

//      *** INCLUDE FILES ***

#include <stdint.h>
#include <stdbool.h>

#include "lpc213x.h"
#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"
#include "io_map.h"
#include "io_codec.h"

    // Headers used during debug:

#include "io_debug.h"       //lint -esym(766, io_debug.h)
#include "io_main.h"        //lint -esym(766, io_main.h)

//***************************************************************
//      *** COMPILE SWITCHES ***


//***************************************************************
//      *** MACROS ****

//***************************************************************
//      *** GENERAL DEFINITIONS ****

    //
    //                  3322 2222 2222 1111 1111 1100 0000 0000
    // Bit Num:         1098 7654 3210 9876 5432 1098 7654 3210

    // Control Reg 0 (SSPCR0):
    //      Select 16 bit word, SSI Frame.
    //                                      0000 0000 0001 1111
    //                                         0    0    1    F

#define SSP_CONTROL_0   0x001F  // Control Reg 0 initial value.

    // Control Register 1 (SSPCR1):
    //      Select SSP enabled, Slave mode.
    //                                                0000 0110
    //                                                   0    6
#define SSP_CONTROL_1   0x06

    // Enable and disable codec interrupts:

#define CDX_DI()    VICIntEnClr = BV(VIC_SSP)   // Disable SSP Ints
#define CDX_EI()    VICIntEnable = BV(VIC_SSP)  // Enable SSP Ints

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                      Codec Control Registers
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    //lint -esym(750, CRE, CRG, CRH)

    // Codec Control Register Addresses:

    //                                          1111 1100 0000 0000
    // Bit Num:                             5432 1098 7654 3210

#define CDX_CHAN1       0x0000      // 0000 0000 0000 0000
#define CDX_CHAN2       0x0800      // 0000 1000 0000 0000

#define CRA             0x8000      // 1000 0000 0000 0000
#define CRB             0x8100      // 1000 0001 0000 0000
#define CRC             0x8200      // 1000 0010 0000 0000
#define CRD             0x8300      // 1000 0011 0000 0000
#define CRE             0x8400      // 1000 0100 0000 0000
#define CRF             0x8500      // 1000 0101 0000 0000
#define CRG             0x8600      // 1000 0110 0000 0000
#define CRH             0x8700      // 1000 0111 0000 0000


    //lint -esym(750, CDX_INIT_CRE, CDX_INIT_CRG, CDX_INIT_CRH)

    // CRA: Set DATA mode, 1 device in cascade.
    //      (Last register to be programmed).
    //                                 Bit Num:    7654 3210
#define CDX_INIT_CRA        (CRA | 0x11)        // 0000 0001

    // CRB: FS=DMCLK/256, SCLK=DMCLK/2, DMCLK=MCLK/5,
    //                                 Bit Num:    7654 3210
#if CDX_PARAMETER_FREQ_SEL == 9375
#define CDX_INIT_CRB      (CRB | 0x4B)          // 0100 1011
#elif CDX_PARAMETER_FREQ_SEL == 11719
// MCLK divider = 1, DMCLK divider = 1024, Fx = 11718.75
//#define CDX_INIT_CRB        (CRB | 0x01)        // 0000 0001
// MCLK divider = 2, DMCLK divider = 512, Fs = 11718.75
#define CDX_INIT_CRB        (CRB | 0x12)        // 0001 0010
// MCLK divider = 4, DMCLK divider = 256, Fs = 11718.75
//#define CDX_INIT_CRB        (CRB | 0x33)        // 0011 0011
#else
#   error "CDX_PARAMETER_FREQ_SEL is not a valid value."
#endif // CDX_PARAMETER_SELECTION

    // CRC: PU, PUIA, PUADC, PUDAC, PUREF, RU
    //                                 Bit Num:    7654 3210
#define CDX_INIT_CRC        (CRC | 0x7D)        // 0111 1101

    // CRD: IGS=0dB; OGS=0dB; Unmute output.
    //                                 Bit Num:    7654 3210
#define CDX_INIT_CRD        (CRD | 0x20)        // 0010 0000

    // CRE: DA=0; IBYP=0; DGTE=0;
    //                                 Bit Num:    7654 3210
#define CDX_INIT_CRE        (CRE | 0x00)        // 0000 0000

    // CRF: SEEN (Single ended enable)
    //
#define CDX_INIT_CRF        (CRF | 0x00)        // 0010 0000

    // CRG: No digital gain tap.
    //                                 Bit Num:    7654 3210
#define CDX_INIT_CRG        (CRG | 0x00)        // 0000 0000

    // CRH: No digital gain tap.
    //                                 Bit Num:    7654 3210
#define CDX_INIT_CRH        (CRH | 0x00)        //    0    0

//***************************************************************
//      *** DATA TYPE DEFINITIONS ***

//***************************************************************

    // Table of initialization words.

static const uint16_t cdxInitWordTable[] =
{
    (CDX_INIT_CRB | CDX_CHAN2), (CDX_INIT_CRB | CDX_CHAN1),
    (CDX_INIT_CRC | CDX_CHAN2), (CDX_INIT_CRC | CDX_CHAN1),
    (CDX_INIT_CRD | CDX_CHAN2), (CDX_INIT_CRD | CDX_CHAN1),
    (CDX_INIT_CRF | CDX_CHAN2), (CDX_INIT_CRF | CDX_CHAN1),
    (CDX_INIT_CRA | CDX_CHAN2), (CDX_INIT_CRA | CDX_CHAN1),

    0, 0,   // Dummys.  Must send to clear out initial samples.
    0, 0,
    0, 0,
    0, 0
};

//***************************************************************

//      *** VARIABLES ***

static void (*volatile pState)(void);  // State variable

static volatile uint16_t initCount;         // Count of init words.
static volatile const uint16_t *pInitWords; // Codec setup words

static void (*pAudioSampleTickHandler_ISR)(void);

static int16_t sample1, sample2;

//***************************************************************

//      *** STATIC FUNCTION PROTOTYPES ****

static void SSP_ISR(void) __attribute__ ((interrupt));
static void CodecProgramMode_ISR(void);
static void CodecDataMode_ISR(void);
static void CodecFatal_ISR(void);
static void NullHandler_ISR(void);

//***************************************************************
//                      *** API ***
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Reset initialization of the SSP State Machine.

void CDX_Reset(void)
{
    O_CODEC_RST();      // Hold the codec in reset until we are
                        // ready for it.

    // Set Up Vectored Interrupt Controller:

    VIC_PRI_SSP = (uint32_t)SSP_ISR;            // set vector
    VIC_CTRL_SSP = VIC_IRQ_EN | VIC_SSP;    // Select SSP

    CDX_Unsubscribe();
    pState = CodecFatal_ISR;    // Fatal error unless CDX_Start()
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Start or restart the Codec

void CDX_Start(void)
{
    CDX_DI();

    pState = CodecProgramMode_ISR; // Initial interrupt mode = Program.

    pInitWords = cdxInitWordTable;  // Point at init words.
    initCount = sizeof(cdxInitWordTable)/4; // word pairs.

     // Set Up Control Registers:
    SSPIMSC = TXIM;
    SSPCR0 = SSP_CONTROL_0;
    SSPCR1 = SSP_CONTROL_1;

    // Preload the SSP and start it up.
    SSPDR = *pInitWords++; SSPDR = *pInitWords++;
    initCount--;
    O_CODEC_RUN();
    CDX_EI();
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

int16_t CDX_GetSample1(void) {return(sample1);}
int16_t CDX_GetSample2(void) {return(sample2);}

void CDX_PutSample1(int16_t sample) {sample1 = sample;}
void CDX_PutSample2(int16_t sample) {sample2 = sample;}


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//      Enable/Disable Individual Audio Stream Components:
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // Attach a filter to a slot.  The filter will get a
    // call every time a sample arrives.
void CDX_Subscribe(void (*pAudioSampleTickHandlerFunction)(void))
{
    ieint_t ieStatus;
    ieStatus = NDI();   // Nested disable audio stream interrupts
    pAudioSampleTickHandler_ISR = pAudioSampleTickHandlerFunction;
    NEI(ieStatus);      // Enable audio stream interrupts
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void CDX_Unsubscribe(void)
{
    ieint_t ieStatus;
    ieStatus = NDI();   // Nested disable audio stream interrupts
    pAudioSampleTickHandler_ISR = NullHandler_ISR;
    NEI(ieStatus);      // Enable audio stream interrupts
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                      STATIC FUNCTIONS
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void NullHandler_ISR(void){}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Codec Interrupt Service Routine

static void SSP_ISR(void)
{
    TP_INTERRUPT_LATENCY_DI_SIGNAL
    TP_CODEC_INTERRUPT_START

    VICVectAddr = 0;            // Acknowledge Interrupt in VIC

    (*pState)();

    TP_CODEC_INTERRUPT_END
    TP_INTERRUPT_LATENCY_EI_SIGNAL
}

    // Interrupt comes here if an attempt is made to run the
    // codec without resetting it first.

static void CodecFatal_ISR(void) {FATAL();}

    // Interrupt comes here when in the codec's program mode.

static void CodecProgramMode_ISR(void)
{
    volatile uint16_t bitBucket;

    // See if Tx fifo has room.

    ASSERT(SSPRIS & TXRIS); // Is Tx fifo <= half empty?

    if(initCount-- > 0)
    {
        // Send two channels, read and throw away any received
        // words.

        SSPDR = *pInitWords++; SSPDR = *pInitWords++;

        // Now if the Rx fifo has samples, read them out in
        // pairs (they are garbage while the codec is being
        // initialized).

        if(SSPRIS & RXRIS)      // Is Rx fifo >= half full?
        {
            bitBucket = SSPDR;
            bitBucket = SSPDR;  // Read 2 samples and thorw away
        }
    }

    // Here, we put the last of the initialization words into
    // the transmit buffer.  Assume we are now in data mode.

    else
    {
        pState = CodecDataMode_ISR; // Codec is in data mode now.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Interrupt comes here when in the codec's data mode.

#if TP_FILTER_UNIT_TEST != 0
int16_t ASTREAM_TestFilter(int16_t sample);
#endif

static void CodecDataMode_ISR(void)
{
    // See if Tx fifo has room and there are Rx samples to read.
#if EMULATOR == 0
    ASSERT(SSPRIS & TXRIS);  // Is Tx fifo <= half empty?
    ASSERT(SSPRIS & RXRIS);  // Is Rx fifo >= half full?
    ASSERT(!(SSPRIS & RORRIS)); // Was there an overrun?
#endif
    // Read the samples and send them to the filters.

    sample2 = (int16_t)SSPDR;
    sample1 = (int16_t)SSPDR;

#if TP_FILTER_UNIT_TEST != 0
    sample1 = ASTREAM_TestFilter(sample1);
#else
    (*pAudioSampleTickHandler_ISR)();
#endif
    // Send the samples to the DAC.  Scale them to the 0 dB reference
    // to prevent clipping because the codec I/O levels have unit
    // and temperature variation.

//    SSPDR = (uint16_t)((sample2 * CDX_0DB_REFERENCE) >> 15);
//    SSPDR = (uint16_t)((sample1 * CDX_0DB_REFERENCE) >> 15);

    // Forget that stuff above.  The samples are scaled in Astream.
    SSPDR = (uint16_t)sample2;
    SSPDR = (uint16_t)sample1;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                          END
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
