/****************************************************************//**

    @file io_debug.h

    Contains I/O functions designed to aid debugging of the
    PSE-702 application.

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

********************************************************************/

//      *** INCLUDE FILES ***

#include <stdint.h>
#include <stdbool.h>

#include "lpc213x.h"        //lint -esym(766, lpc213x.h)
#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"
#include "apprint.h"
#include "apparse.h"
#include "apex.h"           //lint -esym(766, apex.h)
#include "io_map.h"         //lint -esym(766, io_map.h)
#include "io_main.h"
#include "io_debug.h"       //lint -esym(766, io_debug.h)
#include "console.h"

/***************************************************************/

//      *** MACROS ****

/***************************************************************/

//      *** GENERAL DEFINITIONS ****

/***************************************************************/

//      *** STATIC FUNCTION PROTOTYPES ****

static uint8_t CmdReadMemory(const char *pLine);

//***************************************************************

//      *** DATA STRUCTURE Definitions ***

//***************************************************************

//      *** EEPORM VARIABLES ***

//***************************************************************

//      *** CONSTANTS ***

static const console_cmd_tbl_t crashCmdTable[] =
{
    {   // Set courtesy beep parameters
        "r", CmdReadMemory,
        "Read memory",
        "r {addr}",
        ""
    },

    // Marks end of table.
    {"", NULL, NULL, NULL, NULL}
};

//***************************************************************

//      *** VARIABLES ***

//***************************************************************
//          DEBUG SERIAL I/O FORMATTING FUNCTIONS
//***************************************************************

void IO_DB_Reset(void)
{
    IO_DB_BreakpointDisarm();
}

void IO_DBPause(char idChar)
{
    ieint_t ieStatus;

    ieStatus = NDI();
    APPrint(STDERR, "\r\nPause %c:\n\r", idChar);

    IO_DBRxChar();  //lint !e534 Wait for user to type a char.
    NEI(ieStatus);
}

//void IO_DBPrintMessage(char taskID, MSG *pMsg)
//{
//  APPrint(STDERR, "\n\rM-%d-%d ", taskID, pMsg->id);
//}

// Print one character to the debug console.

void IO_DBPrintChar(char c)
{
    void (*pStream)(char c) = STDERR;

    (*pStream)(c);
}

void IO_DB_CRLF(void)
{
    IO_DBPrintChar('\n');
    IO_DBPrintChar('\r');
}


void IO_DBPrintString(const char *pC)
{
    while(*pC != '\0') {IO_DBPrintChar(*pC++);}
}


void IO_DBPrintDWord(uint32_t data)
{
    IO_DBPrintWord((uint16_t)(data >> 16));
    IO_DBPrintWord((uint16_t)(data & 0xFFFF));
}

void IO_DBPrintWord(uint16_t data)
{
    IO_DBPrintByte((uint8_t)(data >> 8));
    IO_DBPrintByte((uint8_t)(data & 0xFF));
}

void IO_DBPrintByte(uint8_t data)
{
    IO_DBPrintNibble(data >> 4);
    IO_DBPrintNibble(data & 0x0F);
}

void IO_DBPrintNibble(uint8_t nibble)
{
    switch(nibble)
    {
        case 0x00: case 0x01: case 0x02: case 0x03: case 0x04:
        case 0x05: case 0x06: case 0x07: case 0x08: case 0x09:
            IO_DBPrintChar((char)(nibble+'0'));
            break;

        case 0x0A: case 0x0B: case 0x0C:
        case 0x0D: case 0x0E: case 0x0F:
            IO_DBPrintChar((char)('A'-0x0A+nibble));
            break;

        default:
            IO_DBPrintChar('?');
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Display processor memory in byte format on the stream
        pointed to by 'pStream'.

        The memory is pointed to by 'pData' and is 'length'
        bytes long.

        The listing will consist of several lines each with
        the value contained in 'addr' followed by 16 bytes
        displayed in hex.  'addr' is icremented by 16 for
        each successive line.
    */

void IO_DisplayMemoryBytes(
    void (*pStream)(char c), const uint8_t *pMem, uintptr_t addr, uint32_t length)
{
    uint32_t j;

    while(length > 0)
    {
        APPrint(pStream, "\n\r\t%xd:  ", addr);

        j = 0;
        while((j<16) && (length>0))
        {
            APPrint(pStream, "%xb ", *pMem++);
            j++;
            length--;
        }

        addr += 16;
    }

    APPrint(pStream, "\n\r");
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Display processor memory in dword format on the stream
        pointed to by 'pStream'.

        The memory is pointed to by 'pData' and is 'length'
        bytes long.

        The listing will consist of several lines each with
        the value contained in 'addr' followed by four dwords
        displayed in hex.  'addr' is icremented by 16 for
        each successive line.
    */

void IO_DisplayMemoryDwords(void (*pStream)(char c),
        const unsigned *pMem, uintptr_t addr, uint32_t length)
{
    uint32_t j;

    while(length > 0)
    {
        APPrint(pStream, "\n\r\t%xw:  ", addr);

        j = 0;
        while((j<16) && (length>0))
        {
            APPrint(pStream, "%xb ", *pMem++);
            j++;
            length--;
        }

        addr += 16;
    }

    APPrint(pStream, "\n\r");
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Crash monitor
     *
     * Provides some basic memory examining tools for use when
     * the code crashes.
     */
void IO_DBCrashMonitor(uintptr_t stackPointer)
{
    Con_Reset();
    Con_SetContext(
            crashCmdTable, '$', STDERR, Slot_Null);   //lint !e64
    APPrint_HexDWord(STDERR, stackPointer);
    APPrint_CRLF(STDERR);
    Con_Prompt();

    FOREVER
    {
        Con_CharacterSink((char)IO_PolledRxUART1());
    }
}   //lint !e715    Unused parameter

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Read an arbitrary memory location

        @clsyntax $ r {addr}

        @par Command Line Parameters:
            @oo {addr}: The hexadecimal memory address to read.

        @description This command reads the supplied memory
        address and prints the contents on the console.
*/

static uint8_t CmdReadMemory(const char *pLine)
{
    uint32_t addr, contents;
    unsigned charCount;

    if(APParse(pLine, &charCount, " %8x|", &addr) != 1)
        return(CON_STS_RANGE);

    else
    {
        APPrint_Char(STDERR, ' ');
        contents = *(uint32_t *)addr;
        APPrint_HexDWord(STDERR, contents);
        return(CON_STS_OK);
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// Print the character on the debug screen and halt.  Interrupts
// are not explicitly disabled during the halt.

void IO_DBHalt(char c)
{
    void (*pStream)(char c) = STDERR;

    (*pStream)(c);
    FOREVER ;           // Loop forever.
}

static uint8_t markCount = 0;  // Number of passes before IO_DBMark fires.

bool IO_DBMark(uint8_t count)
{
    if(++markCount == count)
        return(true);
    else
        return(false);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Here, assertion failed.

void Assert(const char *pFile, int line, const char *pString)
{
    DI();   // non-nested disable interrupts

    APPrint(STDERR, "\r\nAssertion Failed: %s line %d, %s\n\r",
        pFile, (uint16_t)line, (char *)pString);

    APPrint(STDERR, "\tCurrent task ID: %xw\n\r",
        Apex_GetCurrentTask()->id);

    IO_DBPrintString("Return addresses:\n\r");
    IO_DBPrintDWord((uint32_t)__builtin_return_address(0));
    IO_DB_CRLF();
    IO_DBPrintDWord((uint32_t)__builtin_return_address(1));
    IO_DB_CRLF();
    IO_DBPrintDWord((uint32_t)__builtin_return_address(2));
    IO_DB_CRLF();

#   if TP_ASSERT_TRAP_ENABLE != 0
    while(1)            //lint !e716
        ;               // Hang here with interrupts disabled.
#   endif
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                  Breakpoint  Functionality
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static bool bpArmed;

void IO_DB_BreakpointDisarm(void) {bpArmed = false;}
void IO_DB_BreakpointArm(void) {bpArmed = true;}
bool IO_DB_BreakpointIsArmed(void) {return(bpArmed);}

void IO_DB_Breakpoint(void)
{
    if(bpArmed)
        PutBreakpointHere();
}
void PutBreakpointHere(void){}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                      Trace  Functionality
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#if TRACE != 0

    // Variables

#define Z_TRACEBUF  128 // Trace buffer size

static struct TAG_TRACE     // Trace structure and buffer.
{
    uintptr_t data;
    uint32_t tag;
} traceBuf[Z_TRACEBUF];

static uint32_t inTrace, outTrace;  // Buffer pointers.
static uint32_t countTrace;             // Number of traces in buffer

static bool enaTrace;               // Trace enable flag

    // Reset trace system, and Enable/Disable tracing:

void IO_DBTraceReset(void)
{
    enaTrace = false;
    inTrace = outTrace = countTrace = 0;
}

void IO_DBTraceEnable(void)
{
    enaTrace = true;
}

void IO_DBTraceDisable(void)
{
    enaTrace = false;
}

    // Get count of entries in trace buffer:

uint32_t IO_DBTraceGetCount(void){return(countTrace);}


    // Put an entry into the trace buffer if tracing
    // is enabled, else skip.

void IO_DBTraceIn(uintptr_t d, uint32_t t)
{
    ieint_t ieStatus;

    ieStatus = NDI();

    if(enaTrace)
    {
        traceBuf[inTrace].data = d;
        traceBuf[inTrace++].tag = t;

        if(inTrace == Z_TRACEBUF)
            inTrace = 0;

        if(countTrace < Z_TRACEBUF)
            countTrace++;
        else
            outTrace = inTrace;
    }
    NEI(ieStatus);
}

    // Remove the next entry from the trace buffer
    // (for display)

void IO_DBTraceOut(uintptr_t *pD, uint32_t *pT)
{
    ieint_t ieStatus;

    ieStatus = NDI();

    *pD = traceBuf[outTrace].data;
    *pT = traceBuf[outTrace++].tag;

    if(outTrace == Z_TRACEBUF)
        outTrace = 0;

    NEI(ieStatus);
}

#endif // TRACE

//***************************************************************
//                     END OF FILE
//***************************************************************

