/*********************************************************************/
/**
    @file io_f321.c

    Driver for Atmel AT25DF321A flash memory device as used in
    the PSE-702.

    This driver assumes the flash device is interfaced to an
    LPC213X microcontroller SPI bus using io_spi.c.  This driver
    is used by filesys.c to implement a file system on the flash.
    This file provides the low level read/write and memory
    allocation functions.

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

    @author ALP

*****************************************************************

    @section AT25DF321A Flash Structure:

    @par Introduction
    The AT25DF321A (hereinafter refered to as 'Flash') provides
    32 megabits (4 megabytes) of flash storage.  This storage is
    divided into several subdivisions for the purposes of write
    protection, erasure, reading, and writing.

    @par Memory Protection Sectors
    The Flash is divided into 64 64KB units called Sectors
    for the purpose of write protection.  The sectors are numberd
    sequentially from 0 to 63.


    @par Erase Blocks
    The Flash device itself may be erased in its entirety by one
    command, or may be erased in units of 64KB, 32KB, or 4KB.
    This driver implements commands to erase the entire flash,
    or to erase any single 4KB block.  The 4KB blocks are
    numbered sequentially from 0 to 1023.

    @par Reading and Writing
    The Flash may be read or written a single byte at a time, or
    in units of 256 bytes, called Pages.  To write a byte or a page,
    that byte or page must first have been erased by using one of
    the erasure commands mentioned previously.  Once erased, a page
    may be written one byte at a time.  Once a byte has been written
    it may not be re-written until the block in which it resides
    has been erased.

    @defgroup io_flash Flash Driver
    @ingroup filesys
    @{
**********************************************************************/


//      *** INCLUDE FILES ***

#include <stdint.h>
#include <stdbool.h>

#include "lpc213x.h"        // LPC213x definitions
#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"
#include "apex.h"
#include "io_map.h"
#include "io_spi.h"
#include "io_flash.h"

#include "io_main.h"  //lint -esym(766, io_main.h)
#include "aplib.h"      //lint -esym(766, aplib.h)
#include "io_debug.h"   //lint -esym(766, io_debug.h)
#include "apprint.h"    //lint -esym(766, apprint.h)


//***************************************************************

//      *** MACROS ****

//***************************************************************

//      *** GENERAL DEFINITIONS ****

    // Flash Instructions:

#define FLASH_READ          0x03    ///< Page read command
#define FLASH_WRITE         0x02    ///< Page write command

#define FLASH_READ_BYTE     0x03    ///< Byte read command
#define FLASH_WRITE_BYTE    0x02    ///< Byte write command

#define FLASH_READ_SR       0x05    ///< Read status register cmd

#define FLASH_WRITE_ENA     0x06    ///< Write-enable command

#define FLASH_BLOCK_ERASE   0x20    ///< Erase a 4KB block

#define FLASH_UNPROT_SECTOR 0x39    ///< Unprotect a sector

#define FLASH_WRITE_SR1     0x01    ///< Wr9te status byte 1

#define FLASH_WRITE_SR2     0x31    ///< Write status byte 2

#define FLASH_READ_SN       0x77    ///< Read serial number string


    // Flash Status Register Flags

#define FLASH_BUSY_FLAG     0x01    // Busy flag in status register.


//***************************************************************

//      *** STATIC FUNCTION PROTOTYPES ****

static void SelectFlash(bool select);
static void Flash_UnprotectSector(uint32_t addr);
static void Flash_EnableWrites(void);
static void SetSPICommandParameters(uint8_t command, uint32_t address);

//***************************************************************

//      *** VARIABLES ***

static uint8_t mCommand[6]; // Flash instruction byte and address.

static MUTEX flashMutex;    // Only one task uses flash at a time.

const struct TAG_VARIABLE_TABLE flashVarTab[] =
{
    VARIABLE(flashMutex),
    {NULL, NULL, VAR_TYPE_NONE, 0}
};



//***************************************************************
//                          FLASH DRIVER API
//***************************************************************


void Flash_Reset(void)
{
    APEX_INIT_MUTEX_UNLOCKED(&flashMutex);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** Read a Page from Flash

    This function reads a 256 byte page from flash and puts it
    in the buffer pointed to by the input parameter.

    @param addr The page address of the page to read. 0x00
    will be appended to the right hand side to create a 24
    bit byte address.

    @param pPage A pointer to a 256 byte page buffer into which
    the data read from the flash will be placed.
*/
void Flash_ReadPage(uint32_t addr, uint8_t *pPage)
{
    ASSERT(pPage != NULL);
    APEX_LOCK_MUTEX(&flashMutex, 0);
    SetSPICommandParameters(FLASH_READ, PAGE_2_BYTE_ADDR(addr));
    SPI_Read(SelectFlash, mCommand, 4, pPage, Z_FLASH_PAGE);
    APEX_UNLOCK_MUTEX(&flashMutex);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Write a 256 byte page

    // "addr" is the ordinal number of the page to write. 0x00
    // will be appended to the right hand side to create a 24
    // bit flash address.

void Flash_WritePage(uint32_t addr, uint8_t *pPage)
{
    uint8_t statusReg;

    ASSERT(pPage != NULL);

    APEX_LOCK_MUTEX(&flashMutex, 0);

    Flash_EnableWrites();
    SetSPICommandParameters(FLASH_WRITE, PAGE_2_BYTE_ADDR(addr));
    SPI_Write(SelectFlash, mCommand, 4, pPage, Z_FLASH_PAGE);

    // Now read the flash status register and determine the
    // busy/done status.  We can do this in a loop because the
    // SPI driver will suspend us while waiting for the SPI.

    mCommand[0] = FLASH_READ_SR;
    statusReg = FLASH_BUSY_FLAG;

    while((statusReg & FLASH_BUSY_FLAG) != 0)
    {
        SPI_Read(SelectFlash, mCommand, 1, &statusReg, 1);
    }
    APEX_UNLOCK_MUTEX(&flashMutex);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Read the status register from external modules.
    // (I say external because this function is bracketed with
    // mutex lock/unlock, so you have to be careful from where
    // you call it.  If you call it with the mutex already
    // locked, a deadlock will occur.)

void Flash_ReadStatusRegister(uint8_t *pBuf, unsigned length)
{
    APEX_LOCK_MUTEX(&flashMutex, 0);
    mCommand[0] = FLASH_READ_SR;
    SPI_Read(SelectFlash, mCommand, 1, pBuf, length);
    APEX_UNLOCK_MUTEX(&flashMutex);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Write the status register from external modules.
    // (I say external because this function is bracketed with
    // mutex lock/unlock, so you have to be careful from where
    // you call it.  If you call it with the mutex already
    // locked, a deadlock will occur.)

void Flash_WriteStatusRegister(const flash_status_t *pStatus)
{
    uint8_t loStatus = LO_BYTE(*pStatus);
    uint8_t hiStatus = HI_BYTE(*pStatus);

    APEX_LOCK_MUTEX(&flashMutex, 0);

    // Write low order status byte:
    Flash_EnableWrites();
    mCommand[0] = FLASH_WRITE_SR1;
    SPI_Write(SelectFlash, mCommand, 1, &loStatus, 1);

    // Write high order status byte:
    Flash_EnableWrites();
    mCommand[0] = FLASH_WRITE_SR2;
    SPI_Write(SelectFlash, mCommand, 1, &hiStatus, 1);

    APEX_UNLOCK_MUTEX(&flashMutex);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Read one or more bytes.

    // Reads continuously up to one page.

void Flash_ReadBytes(
    uint32_t addr, uint8_t *pBuf, unsigned length)
{
    ASSERT(pBuf != NULL);
    ASSERT(length >  0 && length <  Z_FLASH_PAGE);

    APEX_LOCK_MUTEX(&flashMutex, 0);
    SetSPICommandParameters(FLASH_READ_BYTE, addr);
    SPI_Read(SelectFlash, mCommand, 4, pBuf, length);
    APEX_UNLOCK_MUTEX(&flashMutex);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Write one or more bytes within a single page.

    // Wraps around if low order 8 bits of address plus length
    // is greater than 0xFF.

void Flash_WriteBytes(
    uint32_t addr, uint8_t *pBuf, unsigned length)
{
    uint8_t statusReg;

    ASSERT(pBuf != NULL);
    ASSERT(length >  0 && length < Z_FLASH_PAGE);

    APEX_LOCK_MUTEX(&flashMutex, 0);

    // Now write the page or fraction thereof.

    Flash_EnableWrites();

    SetSPICommandParameters(FLASH_WRITE_BYTE, addr);
    SPI_Write(SelectFlash, mCommand, 4, pBuf, length);

    // Now read the flash status register and determine the
    // busy/done status.  We can do this in a loop because the
    // SPI driver will suspend us while waiting for the SPI.

    mCommand[0] = FLASH_READ_SR;
    statusReg = FLASH_BUSY_FLAG;

    while((statusReg & FLASH_BUSY_FLAG) != 0)
    {
        SPI_Read(SelectFlash, mCommand, 1, &statusReg, 1);
    }

    APEX_UNLOCK_MUTEX(&flashMutex);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Unprotect Entire Flash Chip

        This function unprotects the entire chip one sector at a
        time by looping through all sectors.
     */

void Flash_UnprotectChip(void)
{
    unsigned i, addr;

    APEX_LOCK_MUTEX(&flashMutex, 0);

    for(i=0, addr=0; i<NUM_SECTORS; i++, addr+=Z_FLASH_SECTOR)
        Flash_UnprotectSector(addr);
    APEX_UNLOCK_MUTEX(&flashMutex);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Erase a 4K block:

void Flash_EraseBlock(uint32_t addr)
{
    uint8_t statusReg;

    APEX_LOCK_MUTEX(&flashMutex, 0);

    Flash_EnableWrites();

    SetSPICommandParameters(FLASH_BLOCK_ERASE,
        BLOCK_2_BYTE_ADDR(addr));
    SPI_Command(SelectFlash, mCommand, 4);

    // Now read the flash status register and determine the
    // busy/done status.  We can do this in a loop because the
    // SPI driver will suspend us while waiting for the SPI.

    mCommand[0] = FLASH_READ_SR;
    statusReg = FLASH_BUSY_FLAG;

    while((statusReg & FLASH_BUSY_FLAG) != 0)
    {
        SPI_Read(SelectFlash, mCommand, 1, &statusReg, 1);
    }
    APEX_UNLOCK_MUTEX(&flashMutex);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/** Read the Serial Number String from Flash

    This function reads a unique serial number string
    from the flash and puts it in the buffer pointed to by the
    input parameter.

    The serial number begins at location 0x40 in the OTP security
    register.

*/
void Flash_ReadSerialString(uint8_t *pBuf)
{
    mCommand[0] = FLASH_READ_SN;
    mCommand[1] = 0x00;
    mCommand[2] = 0x00;
    mCommand[3] = 0x40;
    mCommand[4] = 0x00;
    mCommand[5] = 0x00;

    ASSERT(pBuf != NULL);
    APEX_LOCK_MUTEX(&flashMutex, 0);
    SPI_Read(SelectFlash, mCommand, 6, pBuf, FLASH_SERIAL_Z);
    APEX_UNLOCK_MUTEX(&flashMutex);
}



// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                      STATIC FUNCTIONS
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Enable the flash for writing:

static void Flash_EnableWrites(void)
{
    mCommand[0] = FLASH_WRITE_ENA;
    SPI_Command(SelectFlash, mCommand, 1);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Unprotect a Flash Sector:

    // "addr" is the starting address of a 64K block.

static void Flash_UnprotectSector(uint32_t addr)
{
    Flash_EnableWrites();
    SetSPICommandParameters(FLASH_UNPROT_SECTOR, addr);
    SPI_Command(SelectFlash, mCommand, 4);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void SetSPICommandParameters(uint8_t command, uint32_t address)
{
    mCommand[0] = command;
    mCommand[1] = LO_BYTE(HI_WORD(address));
    mCommand[2] = HI_BYTE(LO_WORD(address));
    mCommand[3] = LO_BYTE(LO_WORD(address));
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void SelectFlash(bool select)
{
    if(select)
        O_FLASH_ENA();          // Select Flash
    else
        O_FLASH_DIS();          // Deelect Flash
}

//*********************************************************************
///@}                          END OF FILE
//*********************************************************************
