/*********************************************************************/
/**
    @file io_flash.h

    API for External Flash Driver as used in PSE-702

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

    @author ALP

    @version 110914 ALP Generalized the parameters passed in
        function calls to accomodate larger flash chips,
        specifically the AT25DF321A.

    @version 071205 ALP Modified for PSE-702 PCB V1.0
        prototype hardware.

    @defgroup io_flash_api Flash Driver API
    @ingroup module_api
    @{

**********************************************************************/

/* Change the definitions below for the particular flash in use.
*/
    /** This is the total size in bytes of the flash:
     */
#define Z_FLASH ((32*1024*1024)/8)

    /** A Page is the convinient size to read and write large
        `quantities of data.
    */
#define Z_FLASH_PAGE        256     ///< Bytes per page

    /** A Block is the smallest unit that may be erased.
    */
#define Z_FLASH_BLOCK       4096    ///< Bytes per block

    /** A Sector is the largest unit (other than the entire flash
        that may be protected.
    */
#define Z_FLASH_SECTOR      65536   ///< Bytes per sector

    /** Number of pages per block.
    */
#define NUM_PAGES_PER_BLOCK     16

    /** Length of flash serial number string.
    */
#define FLASH_SERIAL_Z      64

    /** Convert a page address to a block address.
    */
#define PAGE_2_BLOCK_ADDR(x)     ((uint16_t)((x) >> 4))

    /** Convert a block address to a page address.
    */
#define BLOCK_2_PAGE_ADDR(x)     ((uint16_t)((x) << 4))

    /** Convert a page address to a byte address.
    */
#define PAGE_2_BYTE_ADDR(x)     (((uint32_t)(x)) << 8)

    /** Convert a byte address to a page address.
    */
#define BYTE_2_PAGE_ADDR(x)     ((uint16_t)((x) >> 8))

    /** Given a byte address, return the displacement
     * into the page.
     */
#define BYTE_DISP_IN_PAGE(x)     (((x) & (uint32_t)0xFF))

    /** Convert a block address to a byte address.
    */
#define BLOCK_2_BYTE_ADDR(x)    (((uint32_t)(x)) << 12)

    /** Convert a logical page (p) and a physical block (b)
        into a physical page.
    */
#define LOG_PAGE_2_PHY_PAGE(b,p) ((uint16_t) \
                              (BLOCK_2_PAGE_ADDR(b) + ((p) & 0x000F)))

#define NUM_BLOCKS          1024    ///< Number of blocks in flash

#define NUM_PAGES       (NUM_BLOCKS*16) ///< Number of pages in flash

    /// Number of write protect sectors
#define NUM_SECTORS (Z_FLASH/Z_FLASH_SECTOR)

    /// Blocks for User Parameters
#define NUM_USER_BLOCKS     2   ///< Number blocks for user parms
#define FLASH_EEP_BLOCK     (NUM_BLOCKS-1)  ///< User parameters
#define FLASH_EEP_COPY      (NUM_BLOCKS-2)  ///< User parm copy block

typedef uint16_t        flash_status_t;     ///< Flash status type

    /** The size in bytes of the flash status register.
    */
#define Z_FLASH_STATUS      sizeof(flash_status_t)


    // Flash Driver API

/** Initialize File System

    This function initializes the file system upon power-up reset.
*/
void Flash_Reset(void);


/** Read a Page from Flash

    This function reads a page of data from the flash to SRAM.

    @param addr The page address in the flash of the page to
        read.  Page addresses start at zero and increment by
        1 for each page through all pages in the flash.

    @param pPage A pointer to a buffer in SRAM into which
        the page will be read.
*/
void Flash_ReadPage(uint32_t addr, uint8_t *pPage);


/** Write a Page to Flash

    This function writes a page of data from SRAM to the flash.

    @param addr The page address in the flash of the page to
        write.  Page addresses start at zero and increment by
        1 for each page through all pages in the flash.

    @param pBuf A pointer to a buffer in SRAM from which
        the page will be written.
*/
void Flash_WritePage(uint32_t addr, uint8_t *pBuf);


/** Read Bytes from Flash

    This function reads one or more bytes of data from
    the flash.  The read may cross page or block boundries.

    @param addr The byte address in the flash of the first byte
        to read.  Byte addresses start at zero and increment
        by 1 for each byte through all bytes in the flash.

    @param pBuf A pointer to a buffer containing sufficient
        memory space to contain the bytes to be read.

    @param length An unsigned integer containing the number
        of bytes to read.  The buffer pointed to by pBuf
        must be at least this long.
*/
void Flash_ReadBytes(
    uint32_t addr, uint8_t *pBuf, unsigned length);


/** Write Bytes to Flash

    This function writes one or more bytes of data to
    the flash.  The write should not cross page boundries
    or it will wrap around the page.  This function
    returns only after the flash has finished writing
    all of the bytes.

    Note that depending on the type of flash, this
    operation could consume a lot of time, since the flash
    may need to write each byte individually.

    @param addr The byte address in the flash of the first byte
        to write.  Byte addresses start at zero and increment
        by 1 for each byte through all bytes in the flash.

    @param pBuf A pointer to a buffer containing the bytes
        to be written.

    @param length An unsigned integer containing the number
        of bytes to write.  Must be greater than zero and
        less than or equal to the length of a page.
*/

void Flash_WriteBytes(
    uint32_t addr, uint8_t *pBuf, unsigned length);

/** Erase a Block

    This function erases the flash block pointed to by the
    parameter.

    @param addr: The physical block number of the block to
        be erased. Block numbers begin at 0 and increment
        by 1 for each block from the beginning to the end
        of the flash.
*/
void Flash_EraseBlock(uint32_t addr);

/** Unprotect the Entire Flash Memory

    This function unprotects all blocks of the flash memory
    rendering it suitable for writing.
*/
void Flash_UnprotectChip(void);

/** Read the Flash Status Register

    This function reads the contents of the flash status
    register.  If the particular flash in use has a
    multi-byte status register, you must supply the
    number of bytes to read.

    @param pBuf: A pointer to a buffer into which the
        status register data shall be placed.

    @param length: The number of status register bytes to read.
*/
void Flash_ReadStatusRegister(uint8_t *pBuf, unsigned length);

/** Write the Flash Status Register

    This function writes the contents of the flash status
    register.  If the particular flash in use has a
    multi-byte status register, you must supply the
    number of bytes to write.

    @param pStatus: A pointer to a buffer containing the
        status register.  The status register should be
        declared with the type 'flash_status_t'.
*/
void Flash_WriteStatusRegister(const flash_status_t *pStatus);

/** Read Serial Number String

    This function reads a serial number string from the flash.
    The string is FLASH_SERIAL_Z bytes long.

    @param pBuf A pointer to a byte buffer at least FLASH_SERIAL_Z
    bytes long, into which the serial number string shall be placed.

 */
void Flash_ReadSerialString(uint8_t *pBuf);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

extern const struct TAG_VARIABLE_TABLE flashVarTab[];

//*********************************************************************
///@}                      END OF FILE
//*********************************************************************

