/******************************************************************//**
    @file io_main.c
        Contains I/O definitions for LPC-213X processor for
        the PSE-702 application.

    This file hides the details of the LPC-213X from the
    application modules in the PSE-702 application.

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

    @defgroup io_devices I/O Devices

**********************************************************************/

//      *** INCLUDE FILES ***

#include <stdint.h>
#include <stdbool.h>        //lint -esym(766, stdbool.h)

#include "lpc213x.h"        // LPC213x definitions
#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"
#include "apex.h"
#include "io_map.h"
#include "io_main.h"

//#include "apprint.h"  //lint -esym(766, apprint.h)
#include "io_debug.h"   //lint -esym(766, io_debug.h)


/***************************************************************/

//      *** MACROS ****

/***************************************************************/

//      *** GENERAL DEFINITIONS ****

    // Timer 0 count for system timer (1 ms).
#define CNT_SYS_TIMER   (unsigned)(PCLK + 0.5)

/***************************************************************/

//      *** STATIC FUNCTION PROTOTYPES ****

static void Timer0_ISR(void) __attribute__ ((interrupt));
static void UART0_ISR(void) __attribute__ ((interrupt));
static void PWM_ISR(void) __attribute__ ((interrupt));

static void IntDrivenTxUART0(char c);
static void RxCharSink(char notUsed);

//***************************************************************

//      *** DATA STRUCTURE Definitions ***

//***************************************************************

//      *** EEPORM VARIABLES ***

//***************************************************************

//      *** CONSTANTS ***

//***************************************************************

//      *** VARIABLES ***

//***************************************************************
//              SERIAL I/O ROUTINES
//***************************************************************

        // Baud Rates:

        //lint -esym(750, U0DLL_9600) (Not referenced)
#define U0DLL_9600  ((uint8_t)((PCLK/(16*9.6))+0.5))   // 9600 Baud
#define U0DLL_19200 ((uint8_t)((PCLK/(16*19.2))+0.5))  // 19200 Baud
#define U0DLL_38400 ((uint8_t)((PCLK/(16*38.4))+0.5))  // 38400 Baud

//***************************************************************

    // Variables:

    // Uart 0:
static SEMA uart0TxSema;            // Free if room in Uart 0 Tx buffer.
static void (*pTxUART0)(char c);    // Pointer to service routine.
static bool enaIntDrivenRx0;        // Enable interrupt driven Rx
static bool enaIntDrivenTx0;        // Enable interrupt driven Rx

// UART Chan 0 Received Char Subscription
static void (*pRxUART0)(char c);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Initialize Debug UART (UART1)

void IO_UART1_Reset(void)
{
    U1LCR = 0x83;               // 8 bits, no Parity, 1 Stop bit
    U1DLL = U0DLL_38400;        // Baud rate constant
    U1LCR = 0x03;               // DLAB = 0
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Functions to Support UART 1 (Debug Port).

    // IO_PolledTxUART1 is called by the application to write a
    // byte to the debugging output device under polled I/O.

void IO_PolledTxUART1(char c)
{
    LOOP_TIL_true(U1LSR & 0x20) ;   // Wait until buffer empty.
    U1THR = (uint8_t)c;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // IO_PolledRxUART1 is called by the application to read a
    // byte from the debugging input device under polled I/O.

uint8_t IO_PolledRxUART1(void)
{
    LOOP_TIL_true(U1LSR & 0x01);        // Wait until buffer full.
    return (U1RBR);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Functions to Support UART 0 (User Console).

    // Initialize User Interface UART (UART0)

void IO_UART0_Reset(void)
{
    U0LCR = 0x83;               // 8 bits, no Parity, 1 Stop bit
    U0DLL = U0DLL_19200;        // Baud rate constant
    U0LCR = 0x03;               // DLAB = 0

    pTxUART0 = IO_PolledTxUART0;

    enaIntDrivenRx0 = false;
    enaIntDrivenTx0 = false;

    Unsubscribe_IO_RxUart0();
}


void IO_TxUART0(char c)
{
    (*pTxUART0)(c);
}

void Subscribe_IO_RxUart0(void (*pRxCharSink)(char c))
{
    pRxUART0 = pRxCharSink;
}

void Unsubscribe_IO_RxUart0(void) {pRxUART0 = RxCharSink;}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // IO_PolledTxUART0 and IO_PolledRxUART0 are called by the
    // application to write and read a byte to/from the user
    // console port using polled I/O.

void IO_PolledTxUART0(char c)
{
    LOOP_TIL_true(U0LSR & 0x20);        // Wait until buffer empty.
    U0THR = (uint8_t)c;
}

uint8_t IO_PolledRxUART0(void)
{
    LOOP_TIL_true(U0LSR & 0x01);        // Wait until char available.
    return (U0RBR);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // The following functions control Interrupt Driven Serial I/O
    // using UART0 (user console).

void IO_EnableIntDrivenTxUART0(void)
{
    enaIntDrivenTx0 = true;

    pTxUART0 = IntDrivenTxUART0;

    // The sema is initialized with a limit of +1 because an
    // interrupt could occur before the call to Pend() in
    // IntDrivenTxUART0() below.
    Apex_InitSema(&uart0TxSema, 0, 1);
    VIC_PRI_UART0 = (uint32_t)UART0_ISR;        // set vector
    VIC_CTRL_UART0 = VIC_IRQ_EN | VIC_UART0;    // Select device
    VICIntEnable = BV(VIC_UART0); // Enable UART Interrupts
}

void IO_EnableIntDrivenRxUART0(void)
{
    enaIntDrivenRx0 = true;

    VIC_PRI_UART0 = (uint32_t)UART0_ISR;        // set vector
    VIC_CTRL_UART0 = VIC_IRQ_EN | VIC_UART0;    // Select device
    VICIntEnable = BV(VIC_UART0); // Enable UART Interrupts

    U0IER = 0x00000001;         // Enable Rx interrupts.
}


void IO_DisableIntDrivenUART0(void)
{
    enaIntDrivenRx0 = false;
    enaIntDrivenTx0 = false;
    pTxUART0 = IO_PolledTxUART0;
    U0IER = 0x00000000;     // Disable Rx and Tx interrupts.
    VICIntEnClr = BV(VIC_UART0); // Disable Interrupts at VIC
}

    // Transmit a character via interrupt driven serial I/O
    // on UART0 (user console).

static void IntDrivenTxUART0(char c)
{
    if(!(U0LSR & 0x20))     // If there is no room, wait.
    {
        U0IER |= 0x0002;     // Enable Tx interrupt.
        Apex_PendSema(&uart0TxSema, 0);
    }

    U0THR = (uint8_t)c;
}

static void RxCharSink(char unused) {}

    // UART0 Interrupt Handler:

static void UART0_ISR(void)
{
    volatile uint8_t bitBucket;
    uint32_t iir;

    TP_INTERRUPT_LATENCY_DI_SIGNAL
    TP_UART0_INTERRUPT_START

    // First determine the cause of the interrupt:

    VICVectAddr = 0;        // Acknowledge Interrupt in VIC

    iir = (U0IIR & 0x0E);   // Read Interrupt ID Register

    if(enaIntDrivenRx0)
    {
        // Interrupt Service Routine for UART 0 Receiver.
        if(iir == 0x04)         // Receiver interrupt
        {
            uint32_t lsr = U0LSR;

            if(lsr & 0x0000000E)        // Test for errors
                bitBucket = (uint8_t)U0RBR; // Read it and throw it away.
            else
                (*pRxUART0)((char)U0RBR);  // Send it to the subscriber.
        }
    }

    if(enaIntDrivenTx0)
    {
        // Interrupt Service Routines for UART0 Transmitter.
        if(iir == 0x02)    // Transmitter HR empty interrupt
        {
            U0IER &= ~0x0002;             // Disable Tx interrupts.
            Apex_PostSema(&uart0TxSema);  // Post the semaphone
        }
    }

    TP_UART0_INTERRUPT_END
    TP_INTERRUPT_LATENCY_EI_SIGNAL
}

//***************************************************************
//              GENERAL PURPOSE I/O ROUTINES
//***************************************************************

    // Initialize General Purpose I/O

void IO_GPIO_Reset(void)
{
    // Initialize Pin Selections (See LPC213X manual Chap. 7)

    PINSEL0 = PORT0_FUNC_A;
    PINSEL1 = PORT0_FUNC_B;
    //  PINSEL2 =

    // Initialize Port Directions and Defaults

    IODIR0 = PORT0_DIR;         // Set port 0 direction
    IODIR1 = PORT1_DIR;         // Set port 1 direction

    IOPIN0 = PORT0_DEFAULTS;
    IOPIN1 = PORT1_DEFAULTS;
}


//***************************************************************
//              DEFAULT INTERRUPT HANDLERS
//***************************************************************

//#define IRQ_HANDLER_TRAP       0
//#define FIQ_HANDLER_TRAP       0
//#define SWI_HANDLER_TRAP       0
//#define ABT_HANDLER_TRAP       0
//#define UND_HANDLER_TRAP       0

void IO_IRQ_Handler(void)
{
    IO_DBPrintString("IRQ handler trap. Return addresses:\n\r");
    IO_DBPrintDWord((uint32_t)__builtin_return_address(0));
    IO_DB_CRLF();
    FOREVER
        ;               // Hang here with interrupts disabled.
}

void IO_FIQ_Handler(void)
{
    IO_DBPrintString("FIQ handler trap. Return addresses:\n\r");
    IO_DBPrintDWord((uint32_t)__builtin_return_address(0));
    IO_DB_CRLF();
    FOREVER
        ;               // Hang here with interrupts disabled.
}

void IO_SWI_Handler(void)
{
    IO_DBPrintString("SWI handler trap. Return addresses:\n\r");
    IO_DBPrintDWord((uint32_t)__builtin_return_address(0));
    IO_DB_CRLF();
    FOREVER
        ;               // Hang here with interrupts disabled.
}

void IO_ABORT_Handler(void)
{
    IO_DBPrintString("ABORT handler trap. Return addresses:\n\r");
    IO_DBPrintDWord((uint32_t)__builtin_return_address(0));
    IO_DB_CRLF();
    FOREVER
        ;               // Hang here with interrupts disabled.
}

void IO_UNDEF_Handler(void)
{
    IO_DBPrintString("UNDEF handler trap. Return addresses:\n\r");
    IO_DBPrintDWord((uint32_t)__builtin_return_address(0));
    IO_DB_CRLF();
    FOREVER
        ;               // Hang here with interrupts disabled.
}

//***************************************************************
//           TIMER 0 SUPPORT & SYSTEM TIMER INTERRUPT
//***************************************************************

    // Start System Timer.  It is initialized for a 1 ms timebase.

void IO_SystemTimer_Start(void)
{
    T0MR0 = T0TC + CNT_SYS_TIMER;   // 1.0 mSec
    T0MCR = 1;                      // Interrupt on MR0
    T0TCR = 1;                      // Start Timer0 counting.

    // Init the interrupt controller

    VIC_PRI_TMR0 = (uint32_t)Timer0_ISR;          // set vector
    VIC_CTRL_TMR0 = VIC_IRQ_EN | VIC_TIMER0;    // Select Timer 0
    VICIntEnable = BV(VIC_TIMER0); // Enable Timer0 Interrupts
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // System Timer Interrupt via Timer 0 MR0:

static void Timer0_ISR(void)
{
    TP_INTERRUPT_LATENCY_DI_SIGNAL
    TP_TMR0_INTERRUPT_START

    T0IR = 1;               // Reset MR0 Interrupt in CPU
    VICVectAddr = 0;        // Acknowledge Interrupt in VIC
    T0MR0 += CNT_SYS_TIMER; // Set for next interrupt
    Apex_Tick();           // Call service routine.

    TP_TMR0_INTERRUPT_END
    TP_INTERRUPT_LATENCY_EI_SIGNAL
}

//***************************************************************
//                      PWM DRIVER
//***************************************************************

    //  The PWM driver supplies 2 channels of PWM to the user.
    // To use, call PWM_EnableSampleSource() with a pointer to
    // your interrupt handler.  Your interrupt handler must
    // supply a sample each time it is called.  Samples are 16
    // bits twos-complement.

//      LPC2138 PWM Channel 2 (P0.7) is Logical PWM Channel 0.
//      LPC2138 PWM Channel 5 (P0.21) is Logical PWM Channel 1.

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//              PWM Static Function Prototypes
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static int16_t MuteFilter_ISR(void);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                 PWM Definitions and Macros
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // PWM Driver accepts a 16 bit sample and normalizes it to
    // the value of PWM_BASE_PERIOD.

    // The number we need to load into the PWM timer is
    // equal to the PCLK value divided by the desired sample
    // rate:

#define PWM_BASE_PERIOD     (int32_t)( \
                            (PCLK*1000.0)/PWM_SAMPLE_RATE_HZ)

#define PWM_TCR_START       0x09    // Enable PWM and start counter
//#define PWM_TCR_STOP      0x00    // Stop PWM.

    // Midscale PWM count:

#define PWM_MIDSCALE    (int32_t)( \
                        ((PCLK*1000.0)/PWM_SAMPLE_RATE_HZ) * 0.5)

    //                  3322 2222 2222 1111 1111 1100 0000 0000
    // Bit Num:         1098 7654 3210 9876 5432 1098 7654 3210

    // Interrupt and restart TC on match 0.
    // PWMMCR:          0000 0000 0000 0000 0000 0000 0000 0011
    //                     0    0    0    0    0    0    0    3
#define PWM_MCR_INIT        0x00000003

    //                  1111 1100 0000 0000
    // Bit Num:         5432 1098 7654 3210

    // PWMPCR:          0010 0100 0000 0000
    //                     2    4    0    0
#define PWM_PCR_INIT         0x2400

    // PWMPR:   Shouldn't need to init prescaler.

    // Bit Num:         7654 3210

    // Enable writing to PWM registers 2 and 5:
    // PWMLER:          0010 0100
    //                         2    4
#define PWM_LER_SET 0x24

    // PWM Signal Sources
static pwm_source_t *pPWMChan1Source_ISR;
static pwm_source_t *pPWMChan2Source_ISR;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                      PWM Driver Text
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Initialize the PWM system:

void IO_PWM_Reset(void)
{
    // Set up VIC here.

    VIC_PRI_PWM = (uint32_t)PWM_ISR;          // set vector
    VIC_CTRL_PWM = VIC_IRQ_EN | VIC_PWM0;   // Select VIC_PWM

    PWMMCR = PWM_MCR_INIT;      // Init PWM behaviour and interrupts
    PWMPCR = PWM_PCR_INIT;      // Init pin enables

    PWMMR0 = PWM_BASE_PERIOD;   // Init base period.
    PWMMR2 = PWM_MIDSCALE;      // Init midscale zeros
    PWMMR5 = PWM_MIDSCALE;

    IO_PWM_Unsubscribe(PWM_CHAN_1);
    IO_PWM_Unsubscribe(PWM_CHAN_2);

}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void IO_PWM_Unsubscribe(enum TAG_PWM_CHAN chan)
{
    if(chan == PWM_CHAN_1)
    {
        pPWMChan1Source_ISR = MuteFilter_ISR;
    }
    else if(chan == PWM_CHAN_2)
    {
        pPWMChan2Source_ISR = MuteFilter_ISR;
    }
    else
        FATAL();
}

static int16_t MuteFilter_ISR(void)
    {return((int16_t)0);}   //lint !e715


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Attach a signal source to a PWM output channel.
     *
     */
void IO_PWM_Subscribe(enum TAG_PWM_CHAN chan, pwm_source_t *pSource)
{
    ASSERT(pSource != NULL);

    if(chan == PWM_CHAN_1)
    {
        pPWMChan1Source_ISR = pSource;
    }
    else if(chan == PWM_CHAN_2)
    {
        pPWMChan2Source_ISR = pSource;
    }
    else
        FATAL();
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void IO_PWM_Start(void)
{
    PWMTCR = PWM_TCR_START;     // Start PWM.

    PWM_EI();                   // Enable interrupts and go!
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // When an interrupt occurs, get new samples for
    // channel 1 and 2 and write them.  Samples are
    // buffered and the hardware will load them into
    // the PWM just prior to the next interrupt.

static void PWM_ISR(void)
{
    TP_INTERRUPT_LATENCY_DI_SIGNAL
    TP_PWM_INTERRUPT_START

    int16_t sample0, sample1;

    PWMIR = 0x0001;         // Reset PWM interrupt.
    VICVectAddr = 0;        // Acknowledge Interrupt in VIC

    sample0 = (*pPWMChan1Source_ISR)();
    sample1 = (*pPWMChan2Source_ISR)();

    // Normalize according to base period

    PWMMR2 = (uint32_t)(
        ((sample0 * PWM_BASE_PERIOD) >> 16) + PWM_MIDSCALE);
    PWMMR5 = (uint32_t)(
        ((sample1 * PWM_BASE_PERIOD) >> 16) + PWM_MIDSCALE);

    PWMLER = PWM_LER_SET;   // Enable latches for 2 and 5.

    TP_PWM_INTERRUPT_END
    TP_INTERRUPT_LATENCY_EI_SIGNAL
}

//***************************************************************
//                      REAL TIME CLOCK DRIVER
//***************************************************************

    // Initialize and set the Real-Time Clock

void IO_RTC_SetTime(
    uint8_t min, uint8_t hr, uint8_t dom, uint8_t month, uint16_t year)
{
    CCR = 0x02;
    CCR = 0x00;     // Reset the RTC.

    MINUTE = min;
    HOUR = hr;
    DOM = dom;
    MONTH = month;
    YEAR = year;

    CCR = 0x11;     // Start counting.
    PCONP = PCONP & ~BV(PCRTC); // Turn off power to RTC.
}

    // Read the current time.

void IO_RTC_ReadTime(
    uint8_t *pMin, uint8_t *pHr, uint8_t *pDom, uint8_t *pMonth, uint16_t *pYear)
{
    ieint_t ieStatus;
    ieStatus = NDI();
    CCR = 0x10;     // Stop the counters

    *pMin = (uint8_t)MINUTE;
    *pHr = (uint8_t)HOUR;
    *pDom = (uint8_t)DOM;
    *pMonth = (uint8_t)MONTH;
    *pYear = YEAR;

    CCR = 0x11;     // Start counting again.

    NEI(ieStatus);
}


//***************************************************************
//                  WATCHDOG TIMER DRIVER
//***************************************************************



//***************************************************************
//                     END OF FILE
//***************************************************************





