/*********************************************************************/
/**
    @file io_spi.c

    Driver for ARM SPI Device

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

    @defgroup io_spi SPI Driver
    @ingroup io_devices
    @{
**********************************************************************/


//      *** INCLUDE FILES ***

#include <stdint.h>
#include <stdbool.h>

#include "lpc213x.h"        // LPC213x definitions
#include "std.h"
//#include "tp_system.h"      //lint -esym(766, tp_system.h)
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"
#include "apex.h"
#include "io_map.h"
#include "io_spi.h"

#include "io_debug.h"       //lint -esym(766, io_debug.h)

#include "aplib.h"          //lint -esym(766, aplib.h)
#include "apprint.h"        //lint -esym(766, apprint.h)
#include "io_main.h"        //lint -esym(766, io_main.h)


//***************************************************************

//      *** MACROS ****

//***************************************************************

//      *** GENERAL DEFINITIONS ****

#define SPI_CLOCK_DIV   8       // Clock divider: PCLK/8.
                                        // ...Must be even and >= 8.
                                        // Freq = 3.75 MHz
    // SPI Control Register

    // This register controls various aspects of the SPI
    // such as clock polarity, interrupt enable, etc.

        //                 1111 11
        //                 5432 1098 7654 3210

        //                 0000 1000 X010 0000
        //  Reserved_______|||| |||| |||| ||||
        //  8 Bit Xfer__________|||| |||| ||||
        //                           |||| ||||
        //  Enable SPI Interrupts____|||| ||||
        //  MSB First_________________||| ||||
        //  MSTR Mode__________________|| ||||
        //  SCK Act High________________| ||||
        //                                ||||
        //  SCK Phase Std_________________||||
        //  8 Bit Xfers____________________|||
        //  Reserved________________________||

//      *       *       *       *       *       *       *       *

    // Say SPI semaphore uses the common semaphore event:

#define SPI_CTRL        0x08A0      // SPI Control, Enable Ints

//#define SPI_CTRL        0x0820    // SPI Control, Disable Ints
//#define SPI_IE          0x0080    // SPI Int Enable Bit

#define SPI_DI() VICIntEnClr = BV(VIC_SPI0)   // Disable SPI Ints
#define SPI_EI() VICIntEnable = BV(VIC_SPI0)  // Enable SPI Ints

//#define SPI_DI() S0SPCR = SPI_CTRL             // Disable SPI Ints
//#define SPI_EI() S0SPCR = (SPI_IE | SPI_CTRL)  // Enable SPI Ints

//***************************************************************
//      *** STATIC FUNCTION PROTOTYPES ****

static void SPI_ISR(void) __attribute__ ((interrupt));

static void SPI_WaitBusy(void);
static void SPI_RW_Common(void);

static void State_Idle(void);
static void State_Command(void);
static void State_SendReadCommand(void);
static void State_ReadData(void);
static void State_SendWriteCommand(void);
static void State_WriteData(void);

//***************************************************************
//      *** STATIC VARIABLES ***

    ///< Pointer to a function that selects or deselects the device.
static spi_sel_func_t *mpDevice;

static uint8_t volatile *mpCommand;  ///< Buffer for pointer to command
static unsigned volatile mCmdLength; ///< Buffer for command length
static uint8_t volatile *mpData;     ///< Buffer for pointer to data
static unsigned volatile mDataLength; ///< Buffer for data length

static void (*volatile pState)(void); ///< State vector
static MUTEX spiMutex;          ///< One task at a time in SPI.
static SEMA spiBusy;            ///< Wait for SPI to finish.

//***************************************************************

const struct TAG_VARIABLE_TABLE spiVarTab[] =
{
    VARIABLE(mCmdLength),
    {NULL, NULL, VAR_TYPE_NONE, 0}
};

//***************************************************************
//                 SPI DRIVER API FUNCTIONS
//***************************************************************

/** Initialize the SPI Driver from Reset

    This function is called after reset to initialize the SPI
    driver.
 */
void SPI_Reset(void)
{
    APEX_INIT_BUSY_SEMA(&spiBusy);
    APEX_INIT_MUTEX_UNLOCKED(&spiMutex);

    // Set Up Vectored Interrupt Controller:

    VIC_PRI_SPI0 = (uint32_t)SPI_ISR;         // set vector
    VIC_CTRL_SPI0 = VIC_IRQ_EN | VIC_SPI0;  // Select SPI0
    SPI_DI();                   // Disable SPI Interrupts

    // Set Up Control Registers:

    S0SPCCR = SPI_CLOCK_DIV;    // Set clock divider.
    S0SPCR = SPI_CTRL;          // Set up ctrl reg, enable ints.
                                // Note: SPI ints are disabled
                                // at the VIC.
    pState = State_Idle;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Send a Command to an SPI Device

    A command is sent to the addressed SPI device.  No additional
    data is read or written.

    @see SPI_Command.
*/

void SPI_Command(spi_sel_func_t *pDevice,
            uint8_t *pCommand, unsigned cmdLength)
{
    ASSERT(pCommand != NULL);
    ASSERT(cmdLength != 0);

    // Only one task at a time can use SPI
    APEX_LOCK_MUTEX(&spiMutex, 0);

    SPI_DI();   // Disable SPI Interrupts

    mpDevice = pDevice;
    mpCommand = pCommand;
    mCmdLength = cmdLength;

    (*mpDevice)(true);              // Select the device.
    S0SPDR = (uint8_t)*mpCommand++;    // Write the first command uint8_t to
                                    // the spi.
    pState = State_Command;         // Set up to wait for SPI
                                    // complete irq.
    SPI_EI();           // Enable SPI Interrupts

    SPI_WaitBusy();     // Wait for SPI to finish.
    APEX_UNLOCK_MUTEX(&spiMutex);   // Let another task use the SPI.
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Read Data from SPI Device

        Send a command to an SPI device and read the results
        of the command into a buffer.

        @see SPI_Read.
    */
void SPI_Read(spi_sel_func_t *pDevice,
    uint8_t *pCommand, unsigned cmdLength,
    uint8_t *pData, unsigned dataLength)
{
    ASSERT(pCommand != NULL);
    ASSERT(cmdLength != 0);
    ASSERT(pData != NULL);
    ASSERT(dataLength != 0);

    APEX_LOCK_MUTEX(&spiMutex, 0); // Only one task at a time

    SPI_DI();   // Disable SPI Interrupts

    mpDevice = pDevice;
    mpCommand = pCommand;
    mCmdLength = cmdLength;
    mpData = pData;
    mDataLength = dataLength;

    (*mpDevice)(true);              // Select the device.

    S0SPDR = (uint8_t)*mpCommand++; // Write the 1st command uint8_t to
                                    // the spi.
    // Select next state
    pState = State_SendReadCommand;     // Set up to wait for SPI
                                                    // complete irq.
    SPI_EI();   // Enable SPI Interrupts

    SPI_WaitBusy();     // Wait for SPI to finish.

    APEX_UNLOCK_MUTEX(&spiMutex);   // Let another task use the SPI.
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Write Data to SPI Device

    Send a command to an SPI device and write a buffer of data
    following the command.

    @see SPI_Write.
*/
void SPI_Write(spi_sel_func_t *pDevice,
    uint8_t *pCommand, unsigned cmdLength,
    uint8_t *pData, unsigned dataLength)
{
    ASSERT(pCommand != NULL);

    ASSERT(cmdLength != 0);
    ASSERT(pData != NULL);
    ASSERT(dataLength != 0);

    APEX_LOCK_MUTEX(&spiMutex, 0); // Only one task at a time

    SPI_DI();   // Disable SPI Interrupts

    mpDevice = pDevice;
    mpCommand = pCommand;
    mCmdLength = cmdLength;
    mpData = pData;
    mDataLength = dataLength;

    (*mpDevice)(true);          // Select the device.
    S0SPDR = (uint8_t)*mpCommand++;    // Write the first command uint8_t to
                                    // the spi.
    // Select next state
    pState = State_SendWriteCommand;        // Set up to wait for SPI
                                                    // complete irq.
    SPI_EI();   // Enable SPI Interrupts

    SPI_WaitBusy();     // Wait for SPI to finish.

    APEX_UNLOCK_MUTEX(&spiMutex);   // Let another task use the SPI.
}

//***************************************************************
//                      SPI HELPER FUNCTIONS
//***************************************************************

    // Wait on the SPI busy semaphore.

static void SPI_WaitBusy(void)
{
    APEX_WAIT_BUSY(&spiBusy, 0);  // Wait for semaphore
}

//***************************************************************
//                      SPI STATE MACHINES
//***************************************************************

    // The SPI interrupt comes here.  This function simply calls
    // the routine that handles the current state.

static void SPI_ISR(void)
{
    TP_INTERRUPT_LATENCY_DI_SIGNAL
    TP_SPI_INTERRUPT_START
    VICVectAddr = 0;            // Acknowledge Interrupt in VIC
    (*pState)();
    TP_SPI_INTERRUPT_END
    TP_INTERRUPT_LATENCY_EI_SIGNAL
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void SPI_RW_Common(void)
{
    volatile uint32_t bitBucket;

    bitBucket = S0SPSR;     // Read port status and toss.
    bitBucket = S0SPDR;     // Read data and toss.
    S0SPINT = 1;            // Clear interrupt flag.
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    //  SPI Idle State.

static void State_Idle(void)
{
    FATAL();                // Spurious Interrupt.
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Wait for SPI interrupt, then send remaining command
    // bytes, then end command.

static void State_Command(void)
{
    SPI_RW_Common();
    if(--mCmdLength > 0)
    {
        S0SPDR = (uint8_t)*mpCommand++;    // Write the next command byte
    }
    else
    {
        (*mpDevice)(false);         // Deselect the device.
        pState = State_Idle;        // Make the interrupt go idle.
        APEX_MARK_DONE(&spiBusy);   // Say SPI no longer busy.
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Wait for SPI interrupt, then send remaining command
    // bytes, then go to the read data state.

static void State_SendReadCommand(void)
{
    SPI_RW_Common();
    if(--mCmdLength > 0)
        S0SPDR = (uint8_t)*mpCommand++;    // Write the next command
                                                // uint8_t to SPI
    else
    {
        S0SPDR = 0; // Dummy write so we can read data.
        pState = State_ReadData;    // Make the interrupt go idle.
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void State_ReadData(void)
{
    volatile uint32_t bitBucket;

    bitBucket = S0SPSR;     // Read port status and toss.
    *mpData++ = (uint8_t) S0SPDR;       // Read data and put in buffer.
    S0SPINT = 1;                // Clear interrupt flag.

    if(--mDataLength > 0)
        S0SPDR = 0; // Dummy write so we can read data.
    else
    {
        (*mpDevice)(false);         // Deselect the device.
        pState = State_Idle;        // Make the interrupt go idle.
        APEX_MARK_DONE(&spiBusy);   // Say SPI no longer busy.
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Wait for SPI interrupt, then send remaining command
    // bytes, then go to the write data state.

static void State_SendWriteCommand(void)
{
    SPI_RW_Common();

    if(--mCmdLength > 0)
        S0SPDR = (uint8_t)*mpCommand++;    // Write the next command
                                            // uint8_t to SPI
    else
    {
        S0SPDR = (uint8_t)*mpData++;   // Write first data byte
        pState = State_WriteData;   // Make the interrupt go idle.
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void State_WriteData(void)
{
    volatile uint32_t bitBucket;

    bitBucket = S0SPSR;     // Read port status and toss.
    bitBucket = S0SPDR;     // Read data and throw away.
    S0SPINT = 1;            // Clear interrupt flag.

    if(--mDataLength > 0)
        S0SPDR = (uint8_t)*mpData++;   // Write another data byte
    else
    {
        (*mpDevice)(false);       // Deselect the device.
        pState = State_Idle;      // Make the interrupt go idle.
        APEX_MARK_DONE(&spiBusy); // Say SPI no longer busy.
    }
}

//*********************************************************************
///@}                          END OF FILE
//*********************************************************************
