/*********************************************************************/
/**
    @file io_spi.h

    API for SPI Driver for ARM as used in PSE-702.

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

    @author ALP

    @version 110916 ALP Cloned from flash driver and io_spi.h


    @defgroup io_spi_api SPI Driver API
    @ingroup module_api
    @{
**********************************************************************/

    /** Device Select/Deselect Function

        The SPI driver calls a user supplied function to either
        select or de-select the device being addressed by an SPI
        operation.  The function should select the device if the
        'select' argument is true, and de-select the device if
        the argument is false.  This is how specific devices
        are addressed by the SPI driver.
    */
typedef void spi_sel_func_t(bool select);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//      *** PUBLIC FUNCTION PROTOTYPES ***

/** Initialize the SPI Driver from Reset

    This function is called after reset to initialize the SPI
    driver.
 */
void SPI_Reset(void);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Send a Command to an SPI Device

    A command is sent to the addressed SPI device.  No additional
    data is read or written.

    @param pDevice A pointer to the function that selects or
        deselects the SPI device to be sent the command.  @see
        SPI_SEL_FUNC for a description of this function.
    @param pCommand A pointer to a buffer containing the command to
        write to the device.
    @param cmdLength The number of bytes pointed to by pCommand
*/
void SPI_Command(spi_sel_func_t *pDevice,
            uint8_t *pCommand, unsigned cmdLength);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Read Data from SPI Device

    Send a command to an SPI device and read the results
    of the command into a buffer.

    @param pDevice A pointer to the function that selects or
        deselects the SPI device to be sent the command.  @see
        SPI_SEL_FUNC for a description of this function.
    @param pCommand A pointer to a buffer containing the command to
        write to the device.
    @param cmdLength The number of bytes pointed to by pCommand
    @param pData A pointer to a buffer into which data read from the
        device shall be placed.
    @param dataLength The number of bytes to read from the device
        into pData.
*/
void SPI_Read(spi_sel_func_t *pDevice,
    uint8_t *pCommand, unsigned cmdLength,
    uint8_t *pData, unsigned dataLength);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** Write Data to SPI Device
 *
    Send a command to an SPI device and then write a buffer of data
    to the device.

    @param pDevice A pointer to the function that selects or
        deselects the SPI device to be sent the command.  @see
        SPI_SEL_FUNC for a description of this function.
    @param pCommand A pointer to a buffer containing the command to
        write to the device.
    @param cmdLength The number of bytes pointed to by pCommand
    @param pData A pointer to a buffer from which data should be
        written to the device.
    @param dataLength The number of data bytes to write to the device.
*/
void SPI_Write(spi_sel_func_t *pDevice,
    uint8_t *pCommand, unsigned cmdLength,
    uint8_t *pData, unsigned dataLength);

extern const struct TAG_VARIABLE_TABLE spiVarTab[];

//*********************************************************************
///@}                          END OF FILE
//*********************************************************************
