/****************************************************************
*                                                               *
*   target.h: Contains definitions that are target hardware     *
*   dependant, as well as definitions that interface Apex       *
*   to the target hardware.                                     *                                                               *
    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

*                                                               *
****************************************************************/

//****************************************************************
//                      Lint Statements
//****************************************************************

    //lint

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // The following semaphore event is used throughout the
    // application.

//#define EV_SEMA_COMMON  0x8000  // Common semaphore event

//****************************************************************

typedef unsigned ieint_t;

//****************************************************************

        // *** LENGTH-CERTAIN MACROS ****

#define LO_NIBBLE(x)    ( (uint8_t) ( (x) & 0x0F) )
#define HI_NIBBLE(x)    ( (uint8_t) ( ((x) >> 4) & 0x0F ) )
#define MAKE_BYTE(x,y)  ( (uint8_t) ((((x))<<4) | ((y)&0x0F)))

#define LO_BYTE(x)      ((uint8_t)(x))
#define HI_BYTE(x)      ((uint8_t)((x) >> 8))
#define MAKE_WORD(x,y)  ((uint16_t)( \
                            (((uint16_t)(x))<<8) | ((uint8_t)(y))))

#define LO_WORD(x)      ((uint16_t)(x))
#define HI_WORD(x)      ((uint16_t)(((uint32_t)(x)) >> 16))
#define MAKE_DWORD(x,y) ((uint32_t)( \
                            (((uint32_t)(x))<<16) | ((uint16_t)(y))))

//****************************************************************

#define CLIP2INT16(x)   (int16_t)(CLIP((x), INT16_MAX, INT16_MIN))

#define CLIP2BYTE(x)        (uint8_t)((x) > BYTE_MAX ? BYTE_MAX, (x))
#define CLIP2WORD(x)        (uint16_t)((x) > WORD_MAX ? WORD_MAX, (x))

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//      *** MEMORY MAP DEFINITIONS ***

#define SRAM_START      0x40000000  // Start of SRAM
#define SRAM_LENGTH     0x00003C00  // Length of SRAM

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    //  *** System Clock Related Definitions ***

    // XCLK is the crystal frequency in KHz.  It is not the same
    // frequency as the CPU or peripheral clock because there
    // is a PLL multiplier for the CPU clock and a divider for
    // the peripheral clock.

#define XCLK            12000.0       // Crystal freq in KHz

    // The instruction cycle clock CCLK is 60 MHz as initialized in
    // startup.s.

#define CCLK            60000.0       // CPU frequency in KHz

    // PCLK, defined below, is the peripheral clock rate
    // in the LPC-213X.  It is a fraction of the instruction
    // cycle clock as determined by the value written to the
    // APBDIV register during initialization.  This is the
    // frequency used to clock timers, pwm, etc.

    // For the PSE-702 application, the codec outputs a
    // minimum SCLK frequency of 1.5 MHz.  In order to be
    // able to communicate with the codec, PCLK must operate
    // at a higher frequency than the default 15 MHz.  The
    // next fastest possible frequency is 30 MHz which
    // corresponds to an APB divider of 2 (instead of the
    // default value of 4).

    // PCLK_DIV can be either 1, 2, or 4.

#define PCLK_DIV        2           // PCLK is CCLK divided by this.
#define PCLK        (CCLK/PCLK_DIV) // Peripheral clock freq in KHz

    // Resolve APB_DIV:
#if PCLK_DIV == 1
#   define APB_DIV      0x01
#elif PCLK_DIV == 2
#   define APB_DIV      0x02
#elif PCLK_DIV == 4
#   define APB_DIV      0x00
#else
#   error "PCLK_DIV is not a legal value."
#endif


    // The following time units can be used to correct for the
    // fact that the processor's timer tick period may not be
    // evenly divisible into 1.  To use these units, multiply
    // your timer's duration by the appropriate unit below.
    // (The timer tick period is defined as CNT_SYS_TIMER in
    // io.c):

#define TU_1MS          1           // 1 millisecond
#define TU_100MS        100         // 100 milliseconds
#define TU_1SEC         1000        // One second.
#define TU_10SEC        10000       // Ten seconds.

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Variable tables are used to facilitate display of variables
    // on the Test and Troubleshooting Console.

    // AVARIABLE defines array variables.  VARIABLE defines
    // all other variables.

#define AVARIABLE(x) {#x, (uint8_t *)(x), \
                                    VAR_TYPE_NONE, sizeof((x))}
#define VARIABLE(x)  {#x, (uint8_t *)(&(x)), \
                                    VAR_TYPE_NONE, sizeof((x))}

enum TAG_VARIABLE_TYPES
{
    VAR_TYPE_NONE,
    VAR_TYPE_CHAR,
    VAR_TYPE_STRING,
    VAR_TYPE_HEXBYTE,
    VAR_TYPE_HEXWORD,
    VAR_TYPE_HEXINT
};

struct TAG_VARIABLE_TABLE
{
    char *pName;                        // Name of variable
    void *pAddr;                        // Address of variable
    enum TAG_VARIABLE_TYPES type;       // Type of variable
    uint32_t length;                    // Length of variable in bytes.
};

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//      *** ASSERT AND DEBUG DEFINITIONS ***

#ifdef ASSERT_ENABLE
//lint -sem(Assert,r_no)
//lint -emacro(506, ASSERT) // Constant value boolean is OK.
//lint -emacro(774, ASSERT) // Constant value boolean is OK.
void Assert(const char *, int, const char *);
#define ASSERT(x)       if(!(x)) Assert(__FILE__, __LINE__, #x)
#else
#define ASSERT(x);
#endif

void PutBreakpointHere(void);

void Target_Exception(const char *pFile, const int line);
//lint -sem(Target_Exception, r_no)

#define FATAL() Target_Exception(__FILE__, __LINE__)

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /// Returns the start of heap location

uint32_t Target_GetHeapStart(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /// Returns the current value of the stack pointer.

uint32_t Target_GetStackPointer(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//   Functions to support EI/DI with interrupt latency measurement:
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Globally Disable Interrupts (Un-Nested)

        This function may be called to globally disable interrupts.

        @return Nothing.
    */
void _DI(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Globally Enable Interrupts (Un-nested)

        This function may be called to globally enable interrupts.

        @return Nothing.
    */
void _EI(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Globally Disable Interrupts (Nested)

        This function may be called to disable all interrupts, and return
        the previously set interrupt enable condition.

        @return A target system defined data item that contains the
            interrupt enable state.  It should be used in a subsequent
            call to Target_NEI to restore the interrupt state
            to the same value it had prior to the call to Target_NDI.
    */
ieint_t _NDI(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Globally Enable Interrupts (Nested)

        This function may be called to globally enable interrupts after
        they had been disabled in a nested manner.

        @param ie: A target system defined data item containing the
            interrupt enable state.  It is the returned value of a
            previous call to Target_NDI.

        @return Nothing.
    */
void _NEI(ieint_t ie);

#define DI()                _DI()
#define EI()                _EI()

#define NDI()               _NDI()
#define NEI(x)              _NEI(x)

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Function to assert port pin when interrupts are disabled.
     *
     * Define this function in the application to assert a port
     * pin when called in order to monitor interrupt latency.
     *
     * Note that this function must preserve the contents of R0.
     */
void Target_Interrupts_Disabled_Hook(void);

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Function to de-assert port pin when interrupts are enabled.
     *
     * Define this function in the application to de-assert a port
     * pin when called in order to monitor interrupt latency.
     *
     * Note that this function must preserve the contents of R0.
     */
void Target_Interrupts_Enabled_Hook(void);

//*********************************************************************
//                          END OF FILE
//*********************************************************************

