/****************************************************************//**

    @file main.c

    @defgroup main Reset and Application Startup Module

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

    @section init_philosophy Initialization Philosophy

    Initialization in the PSE-702 is performed in two steps: Reset
    initialization and Startup initialization.  Reset initialization
    performs those tasks necessary to cold-start a module, such as
    setting default variable values.  Startup initialization performs
    initializations that require system calls available only
    after multitasking startup has occurred, (such as semaphore
    pending).

    Each module (except some library modules that do not require
    any initialization) has a function assigned to perform reset
    initialization,
    designated by the module ID numonic followed by "_Reset()".  For
    example, the reset initialization function for the DTMF decoder
    module is called DTMF_Reset().  Startup initialization functions
    are designated by the module ID numonic followed by "_Start()".
    The DTMF module's startup initialization function is called
    DTMF_Start().  Not every module requires a startup initialization
    function.

    @subsection reset_init Reset Initialization
    Module reset initialization
    functions are called after the call to Apex_Reset(), and before
    the call to Apex_Start() and the enabling of global interrupts.
    Reset initialization functions must execute with interrupts
    disabled and must not enable interrupts before returning.

    Reset initialization functions may use a subset of Apex system
    calls. For example, reset initialization functions are
    usually the place where a module's task is created, and therefore
    call Apex_CreateTask().  They also are often the place where
    mailboxes and semaphores are initialized, so they call
    Apex_InitMailbox() and Apex_InitSema().  See the discussion on
    "context" in the Apex documentation for more details about
    which system calls may be used before multitasking startup.

    @subsection start_init Startup Initialization
    As stated earlier, not every module requires a specific startup
    initialization
    function.  Modules may perform initialization sequences that
    take advantage of the full range of system calls by placing
    such initialization in the first few lines of a task's body,
    before it enters it's "FOREVER" loop.  Such code is guarrenteed
    to execute after the call to Apex_Start() and can therefore
    use all of the system calls.  However, such code will be
    executed in the priority order of the tasks created during
    reset initialization.

    For example, if the DTMF task has a higher priority than the
    central task, the DTMF task will execute for the first time
    before central task, and DTMF's task initialization will
    occur first.  It is sometimes desirable or necessary to
    initialize the tasks in an order different than their task's
    priority order.  That is one reason to put initialization
    sequences in a startup initialization function in main.c.

    Startup initialization functions are called after Apex_Start()
    and with global interrupts enabled.  Startup initialization
    functions may take advantage of all system calls, (including
    calls that pend on events) and may enable and disable interrupts
    as necessary.

    @section init_bugs Improper Initialization Sequencing Problems
    A bug could easily occur by not carefully following the rules
    outlined above.  Consider the following example, which has
    actually occurred:

    @par
    In the process of performing its initialization, the DTMF_Reset()
    function calls another function inside the DTMF module that
    disables, then re-enables the codec interrupts.  This leaves
    the codec interrupt enabled upon return from DTMF_Reset(), so
    that when global interrupts are enabled after calling
    Apex_Start(), a codec interrupt immediately occurs.  However,
    the interrupt vector is not set (because it doesn't get set
    until a call to CDX_Start()).  This causes an assert.

    The fix for this is to not enable interrupts within DTMF_Reset().

********************************************************************/

//      *** INCLUDE FILES ***

//lint -e766

#include <stdint.h>
#include <stdbool.h>

#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)

#include "lpc213x.h"            // LPC21xx definitions
#include "target.h"

#include "io_debug.h"   //lint -esym(766, io_debug.h)

#include "apex.h"
#include "aplib.h"
#include "app.h"

#include "io_map.h"
#include "io_main.h"
#include "io_spi.h"
#include "io_codec.h"
#include "io_flash.h"
#include "io_misc.h"
#include "apprint.h"
#include "dtmfdet.h"
#include "console.h"
#include "apalloc.h"
#include "fgen.h"
//#include "ttcon.h"
#include "filesys.h"
#include "central.h"
#include "voice.h"
#include "filter.h"
#include "astream.h"
#include "eep.h"
#include "audiosm.h"
//#include "morse.h"
//#include "squelch.h"
//#include "idsm.h"
//#include "txsm.h"
#include "ssm_morse.h"
#include "ssm_squelch.h"
#include "ssm_idsm.h"
#include "ssm_txsm.h"
#include "ssm_sate.h"
#include "ssm_satd.h"
#include "susi.h"
#include "new_dusi.h"
#include "ansi_term.h"
#include "signon.h"

#undef APEX_H               // This prevents the error msg
#include "simplesm.h"

/***************************************************************/

#if WATCHDOG_ENABLE != 0
#define WDOG_TIME   100.0       // Watchdog timer in ms
#define WDOG_COUNT  ((uint32_t)((PCLK*WDOG_TIME)/4))
#endif

/***************************************************************/
//      *** STATIC FUNCTION PROTOTYPES ****

#if TP_TASK_MONITOR_ID != 0
static void ContextSwitchHook(const TCB *pTCB);
#endif

#if WATCHDOG_ENABLE != 0 && EMULATOR != 0
static void Watchdog_ISR(void) __attribute__ ((interrupt));
#endif

/***************************************************************/
//      *** VARIABLES ***

   //lint -esym(551, resetCode) Sometimes not accessed.

static uint8_t resetCode;

/***************************************************************/

        //      *** MAIN ***

    //  This function is passed control by the C start-up
    //  file.  It performs power-up initialization, then
    //  goes into a loop in which it retreives an event
    //  and distributes the event to the proper state
    //  machine.

int main(void)
{
    // Reset source ID bits (note that these are bit numbers, not
    // binary encoded integers):
    //  00 POR: Power-on reset
    //  01 EXTR: External reset
    //  02 WDTR: Watchdog timer reset
    //  03 BODR: Brownout detector reset

    // Note that if the reset code is 00, no reset actually occured.
    // In this case, the code vectored to location 0x00000000.

    resetCode = RSID;
    RSID = 0x0F;

    IO_DB_Reset();      // Init debugger if it exists.
    Apex_Reset();       // Init OS.

    // Init APB divider (VPB) here.

    VPBDIV = APB_DIV;      // Divide CCLK to get PCLK.

    // Reset Initialization Section

    IO_GPIO_Reset();

    IO_UART0_Reset();
    IO_UART1_Reset();

    FILT_Reset();
    ASTREAM_Reset();    // Reset the ASTREAM module.
    SPI_Reset();        // Init the SPI bus.
    IO_PWM_Reset();     // Init PWM Outputs
    CDX_Reset();        // Init communications w/ codec
    SSM_Reset();        // Init simple state machine support

    // Sign On to Debug Console:

#if EMULATOR != 0
    IO_DBPause('M');

#   if WATCHDOG_ENABLE != 0
    // Setup Watchdog Timer
    WDTC = WDOG_COUNT; // Initialize watchdog timer register

    WDMOD = 0x01;   // Watchdog gives interrupt.
    VIC_PRI_WDOG = (uint32_t)Watchdog_ISR;   // set vector
    VIC_CTRL_WDOG = VIC_IRQ_EN | VIC_WDT;    // Select device
    VICIntEnable = BV(VIC_WDT); // Enable watchdog interrupts
#   endif
#else
#   if WATCHDOG_ENABLE != 0
    // Setup Watchdog Timer
    WDTC = WDOG_COUNT; // Initialize watchdog timer register
    WDMOD = 0x03;   // Watchdog gives reset.
#   endif
#endif

#if DEBUG_PORT != 0
//    ANSI_ClearScreen(STDERR);
//    ANSI_MoveCursor(STDERR, 0, 0);
    APPrint(STDERR, "%xb DEBUG %s\n\r", resetCode, signon);

#   if TP_STRUCTURE_SIZES != 0
    APPrint(STDERR, "LINK: %d bytes\n\r", sizeof(LINK));
    APPrint(STDERR, "MSG: %d bytes\n\r", sizeof(MSG));
    APPrint(STDERR, "TMR: %d bytes\n\r", sizeof(TMR));
    APPrint(STDERR, "TCB: %d bytes\n\r", sizeof(TCB));
    APPrint(STDERR, "SEMA: %d bytes\n\r", sizeof(SEMA));
#   endif

#endif // DEBUG_PORT

    MEM_Reset();          // Init memory allocator.

    // Init Application Tasks:

    FGen_Reset();
    Flash_Reset();
    FSys_Reset();       // Init flash file system.
    CEN_Reset();        // Init central control module.

    DTMF_Reset();       // Init the DTMF detector.
    SATD_Reset();       // Init sub-audible tone module.
    VOI_Reset();        // Init voice record/playback module
    Con_Reset();        // Init console first to clear context
    SUSI_Reset();       // Init serial user interface.  Should come
                        // after Con_Reset() since it sets the initial
                        // context table pointer.
    DUSI_Reset();
//    TTCon_Reset();      // Init Test & Troubleshooting Console
    AUD_Reset();        // Init the audio state machine.
    IDSM_Reset();
    Morse_Reset();
    SQU_Reset();        // Init squelch detector.
    TXSM_Reset();       // Init the transmitter state machine.
    SATE_Reset();       // Init subaudible tone encoder.

    // Strobe a firmware testpoint at the end of the reset sequence.

    TP_RESET_COMPLETE

    // If we are monitoring a task, set the context switch
    // hook:
#if TP_TASK_MONITOR_ID != 0
    Apex_Hook_SwitchContexts(ContextSwitchHook);    //lint !e534
#endif

    // Start Multitasking:

    Apex_Start(TPRI_APEX);       // Start Apex and don't return.
    FATAL();            // (Register fatal error if it does)

    return(0);  //lint !e527 // Unreachable, yes.  But this
                    // statement prevents gnu C from issuing the
                    // error about "main() must return a value".
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// Nothing to do for Apex_Target_Start.

void Apex_Target_Start(void){}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // When Apex_Target_StartSystemTimer() is called by the
    // timer task, interrupts are already enabled.

void Apex_Target_StartSystemTimer(void)
{
    DI();
    IO_SystemTimer_Start();
    EI();
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Create a new context

void Apex_Target_CreateTask(
    TCB *pTask, void (*pStartAddr)(void), uint16_t stackSize)
{
    uint32_t *pStack;
    uint32_t *pFill;
    int i, length;

    // Fill stack area

    pFill = pTask->pStack;
    length = stackSize/sizeof(uint32_t);

    for(i=0; i<length; i++)
        *pFill++ = 0xA5A5A5A5;

    // Point at start of stack.
    pStack = pTask->pStack + (stackSize - sizeof(uint32_t));    //lint !e124

    *pStack-- = (uint32_t)pStartAddr; // R14 (LR)
    *pStack-- = 0xCCCCCCCC;             // R12
    *pStack-- = 0xBBBBBBBB;             // R11
    *pStack-- = 0xAAAAAAAA;             // R10
    *pStack-- = 0x99999999;             // R09
    *pStack-- = 0x88888888;             // R08
    *pStack-- = 0x77777777;             // R07
    *pStack-- = 0x66666666;             // R06
    *pStack-- = 0x55555555;             // R05
    *pStack-- = 0x44444444;             // R04
    *pStack-- = 0x33333333;             // R03
    *pStack   = 0x22222222;             // R02

    pTask->pStack = pStack;   // Set task's stack pointer
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Do Fatal Error Stuff

void Target_Exception(const char *pFile, const int line)
{
    DI();   // non-nested disable interrupts

    IO_GPIO_Reset();    // Put I/O ports in a quiescent state.

    // Notify the User via the Console Here.

    APPrint(STDERR,
        "\r\nFatal Error: %s line %d, current task ID is %xw\n\r",
        pFile, (uint16_t)line, Apex_GetCurrentTask()->id);

    FOREVER
        ;               // Hang here with interrupts disabled.
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#if WATCHDOG_ENABLE != 0 && EMULATOR != 0
static void Watchdog_ISR(void)
{
    IO_DBPrintString("Watchdog trap. Return addresses:\n\r");
    IO_DBPrintDWord((uint32_t)__builtin_return_address(0));
    IO_DB_CRLF();
    FOREVER
        ;               // Hang here with interrupts disabled.
}
#endif
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Enable and disable Apex System Timer interrupts:

void Apex_Target_SystemTimer_EI(void)
    {VICIntEnable = BV(VIC_TIMER0);}

void Apex_Target_SystemTimer_DI(void)
    {VICIntEnClr = BV(VIC_TIMER0);}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // Functions to assert a firmware testpoint during
    // periods of interrupt latency:

void Target_Interrupts_Disabled_Hook(void)
{
    TP_INTERRUPT_LATENCY_DI_SIGNAL
}

void Target_Interrupts_Enabled_Hook(void)
{
    TP_INTERRUPT_LATENCY_EI_SIGNAL
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /// A slot that does nothing when called.
void Slot_Null(
        uintptr_t slotParm1, uintptr_t slotParm2){} //lint !e715

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

#if TP_TASK_MONITOR_ID != 0

static void ContextSwitchHook(const TCB *pTCB)
{
    if(pTCB->id == TP_TASK_MONITOR_ID)
    {
        TP_TASK_MONITOR_START
    }

    else
    {
        TP_TASK_MONITOR_END
    }
}
#endif

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                      END
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

