/**************************************************************/
/**
    @file new_dusi.c

    dusi.c: DTMF User Interface for PSE-702

    @author EAS/NLH/alp

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************/
//                     INCLUDE FILES
//***************************************************************

#include <stdint.h>         //lint -esym(766, stdint.h)
#include <stdbool.h>        //lint -esym(766, stdbool.h)

#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"

#include "aplib.h"
#include "apex.h"
#include "apparse.h"

#include "app.h"
#include "central.h"
#include "ssm_satd.h"
#include "ssm_sate.h"
#include "astream.h"
#include "eep.h"
#include "ssm_morse.h"
#include "new_dusi.h"
#include "filesys.h"
#include "audiosm.h"
//#include "io_misc.h"
//#include "filter.h"

#include "io_debug.h"       //lint -esym(766, io_debug.h)
#include "apprint.h"        //lint -esym(766, apprint.h)
#include "io_main.h"        //lint -esym(766, io_main.h)

//***************************************************************
//                       MACROS
//***************************************************************

//lint -esym(750, PRINT_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, SET_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, STATE_CHAN) (Macro not always referenced)

#define PRINT_STATE_CHAR(x)   IO_DBPrintChar(x)

#if (TP_DUSI_STATE_FLAGS == STATE_FLAGS_OFF)
#   define SET_STATE(X,Y)   pState = (X)

#elif ((TP_DUSI_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
    || (TP_DUSI_STATE_FLAGS == STATE_FLAGS_UPPER_CH2))
#   define SET_STATE(X,Y)   pState = (X); PRINT_STATE_CHAR(Y)

#elif ((TP_DUSI_STATE_FLAGS == STATE_FLAGS_LOWER_CH1) \
    || (TP_DUSI_STATE_FLAGS == STATE_FLAGS_LOWER_CH2))
#   define SET_STATE(X,Y)   pState = (X); PRINT_STATE_CHAR((Y)|0x20)
#else
#error "TP_DUSI_STATE_FLAGS improperly defined."
#endif


//***************************************************************
    // Stack Size: Measured stack overflow on 3/13/12.
    // Increasing size from 384 to 512.

#define Z_STK_DUSI      512     // Stack size in bytes


    // State Machine Event IDs:

enum TAG_DUSI_EVIDS
{
    EV_DUSI_CHAR = DUSI_ID(0),      // A character arrived.
    EV_DUSI_TIMEOUT                 // A password timeout occured.
//  EV_DUSI_PWDOK                       // The password was accepted
};


    // Correspondance between Message IDs and State Machine
    // Events (Gives correspondance to state diagram):

#define MSG_DUSI_CHAR       ((int16_t)EV_DUSI_CHAR)


#define TIME_DUSI_PASSWORD   (5*60*TU_1SEC)  // Password timeout

#define LOGIN_COMMAND       "00"    // Special command to log in.
#define LOGOFF_COMMAND      "99"    // Special command to log out.

    // DUSI Entry Error Messages

#define DUSI_STS_OK         0       // "OK",
#define DUSI_STS_SYNTAX     1       // "Syntax error.",
#define DUSI_STS_RANGE      2       // "Parameter out of range."
#define DUSI_STS_PASSWD     3       // "Bad password."
#define DUSI_STS_LOGOFF     4       // "Logoff command issued"
#define DUSI_STS_NULL       5       // Take no action

//***************************************************************
//              DATA STRUCTURE DEFINITIONS
//***************************************************************


typedef struct TAG_CMD_TABLE
{
    char *pName;                       // Name of command
    uint8_t (*pCommand)(char *pLine);  // Function to perform command

} CMD_TABLE;


//***************************************************************
//                STATIC VARIABLES
//***************************************************************

static TCB dusiTCB;
static uint32_t dusiStack[Z_STK_DUSI/sizeof(uint32_t)];
static MBOX mailbox;

    // Task state:
static void (*pState)(enum TAG_DUSI_EVIDS event, char c);

static TMR passwordTmr;     // Password timer
static bool passwordValid;  // Password is valid
//static enum TAG_RX_CHAN activeChan; // current active channel

    // variables used by state machine to accumulate commands

static int count;           // String length counter

    // Max cmd length for cmd + parameters + null
static char myCmdString[Z_CMD_LINE];

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//              Signals
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */ /**

DEFINE_SIGNAL(DUSI_CommandOK)
DEFINE_SIGNAL(DUSI_CommandERR)
DEFINE_SIGNAL(DUSI_ArmRecorder)

//***************************************************************
//                       CONSTANTS
//***************************************************************

    // Debugging Table for Variable Display

const struct TAG_VARIABLE_TABLE dusiVarTab[] =
{
    {"stack", dusiStack, VAR_TYPE_NONE, sizeof(dusiStack)},
    {"tcb", &dusiTCB, VAR_TYPE_NONE, sizeof(dusiTCB)},
    {NULL, NULL, VAR_TYPE_NONE, 0}
};


//***************************************************************
//             *** STATIC FUNCTION PROTOTYPES ****
//***************************************************************

static void DusiTask(void);

static void Transition_Idle(void);              // State A
static void State_Idle(enum TAG_DUSI_EVIDS event, char c);
static void Transition_AccumulateLine(void);    // State B
static void State_AccumulateLine(
                        enum TAG_DUSI_EVIDS event, char c);
static void Transition_Error(void);             // State C
static void State_Error(enum TAG_DUSI_EVIDS event, char c);

static void RunDusi (char *pLine);
static uint8_t CmdSetTxFlagsCommon(
                           const char *pLine, uint16_t mask);

static uint8_t Cmdlogin(const char *pLine);

static uint8_t CmdSetBeepFreq(const char *pLine);
static uint8_t CmdSetBeepDuration(const char *pLine);
static uint8_t CmdSetBeepLevel(const char *pLine);
static uint8_t CmdSetBeepMode(const char *pLine);

static uint8_t CmdSetMorseRate(const char *pLine);
static uint8_t CmdSetMorseFreq(const char *pLine);
static uint8_t CmdSetMorseLevel(const char *pLine);

static uint8_t CmdSetFilterIndex(const char *pLine);

//static uint8_t CmdSetCWID(const char *pLine);
static uint8_t CmdSetIDString(const char *pLine);
static uint8_t CmdPlayID(const char *pLine);
static uint8_t CmdSetVoiceIDFileNum(const char *pLine);

static uint8_t CmdSetRxFlagP(const char *pLine);
static uint8_t CmdSetRxFlagR(const char *pLine);

static uint8_t CmdSetTxFlagB(const char *pLine);
static uint8_t CmdSetTxFlagD(const char *pLine);
static uint8_t CmdSetTxFlagP(const char *pLine);
static uint8_t CmdSetTxFlagR(const char *pLine);
static uint8_t CmdSetTxFlagT(const char *pLine);
static uint8_t CmdSetTxFlagV(const char *pLine);

static uint8_t CmdSetTxAudioLevel(const char *pLine);

static uint8_t CmdSetIdTimer(const char *pLine);
static uint8_t CmdSetTxTimeout(const char *pLine);
static uint8_t CmdSetHangTime(const char *pLine);

static uint8_t CmdLogOff(const char *pLine);

static uint8_t CmdSetRxTxPath(const char *pLine);
static uint8_t CmdSchedAnno(const char *pLine);
static uint8_t CmdDeletePath(const char *pLine);
static uint8_t CmdDeleteAnno(const char *pLine);
static uint8_t CmdEnableAnno(const char *pLine);
//static uint8_t CmdPlayAnno(const char *pLine);
static uint8_t CmdPlayFile(const char *pLine);

static uint8_t CmdSetCTCSSDecodeIndex(const char *pLine);
static uint8_t CmdSetCTCSSEncodeIndex(const char *pLine);
static uint8_t CmdSetCTCSSEncodeMode(const char *pLine);

static uint8_t CmdFileRecord(const char *pLine);
static uint8_t CmdFileDelete(const char *pLine);

#if EMULATOR == 0
static uint8_t CmdReset(const char *pLine);
#endif

//***************************************************************
///    Command Table

///    The following table identifies the command numeric codes and
///    functions available via DTMF access. (* is used for 'space').

static const CMD_TABLE dusiCmdTable[] =
{
    {LOGIN_COMMAND, Cmdlogin},  ///< Validate DTMF Password
    // 01-09 unused
    {"10", CmdSetTxFlagT},      ///< Transmitter enable
    {"11", CmdSetTxFlagP},      ///< CTCSS encoder master enable
    {"12", CmdSetTxFlagB},      ///< Courtesy beep enable
    {"13", CmdSetTxFlagD},      ///< DTMF mute
//    {"14", CmdSetTxFlagA},      ///< Announcement mode enable
//    {"15", CmdSetTxFlagL},      ///< PTT active level
    {"16", CmdSetTxFlagR},      ///< CTCSS reverse burst enable
    {"17", CmdSetTxFlagV},      ///< Set/Clear voice ID flag
    // 18-19 unused
    {"20", CmdSetRxTxPath},     ///< Set Rx-Tx paths
    {"21", CmdSchedAnno},       ///< Schedule announcements.
    {"22", CmdDeletePath},      ///< Delete a path
    {"24", CmdDeleteAnno},      ///< Delete an announcement schedule
    {"25", CmdSetIDString},     ///< Set ID string
    {"26", CmdPlayID},          ///< Play current ID setting
    {"27", CmdEnableAnno},      ///< Enable/disable announcement
//    {"28", CmdPlayAnno},        ///< .
    // 29 unused
    {"30", CmdSetTxAudioLevel}, ///< Set Tx audio level
    {"31", CmdSetBeepLevel},    ///< Set courtesy beep audio level
    {"32", CmdSetMorseLevel},   ///< Set Morse audio level
    // 33-39 unused
    {"40", CmdSetRxFlagP},      ///< CTCSS decoder master enable
    {"41", CmdSetRxFlagR},      ///< Mute a receiver.
    // 42-44 unused
    {"45", CmdSetFilterIndex},  ///< Set deemphasis filter index.
    // 46-49 unused
    {"50", CmdSetBeepFreq},     ///< Set courtesy beep frequency
    {"51", CmdSetBeepDuration}, ///< Set courtesy beep duration
    {"52", CmdSetBeepMode},     ///< Set courtesy beep mode
    {"53", CmdSetMorseRate},    ///< Set CW rate
    {"54", CmdSetMorseFreq},    ///< Set CW tone frequency
    // 55 unused
    {"56", CmdSetIdTimer},      ///< Set ID timer
    {"57", CmdSetTxTimeout},    ///< Set Tx timeout timer
    {"58", CmdSetHangTime},     ///< Set hang timer
    {"59", CmdSetVoiceIDFileNum}, ///< Set voice ID file number

    {"60", CmdSetCTCSSDecodeIndex}, ///< Set Rx CTCSS decoder frequency
    {"61", CmdSetCTCSSEncodeIndex}, ///< Set Tx CTCSS encoder frequency
    {"62", CmdSetCTCSSEncodeMode},  ///< Set Tx CTCSS encoder mode
    // 63-69 unused
    {"70", CmdFileRecord},      ///< Record a voice file
    {"71", CmdFileDelete},      ///< Delete a voice file
    {"72", CmdPlayFile},        ///< Play an audio file

    // 73-89 unused
#if EMULATOR == 0
    {"90", CmdReset},           ///< Reset the microcontroller
#endif

    // 91-98 unused
    {LOGOFF_COMMAND, CmdLogOff},    ///< Log off

    {NULL, NULL}                ///< Marks end of table.
};


//***************************************************************
//                      *** API ***
//***************************************************************

void DUSI_Reset(void)
{
    Apex_InitMailbox(&mailbox);

    Apex_InitMsgTimer(
        &passwordTmr, &mailbox, EV_DUSI_TIMEOUT, 0, 0);

//    activeChan = MAX_RXCHAN;    // This arms an assert in DusiTask().

    DUSI_UnsubscribeAll();
}

void DUSI_Start(void)
{
    Apex_CreateTask(&dusiTCB, TID_DUSI, DusiTask,
        dusiStack, (sizeof(dusiStack)), TPRI_DUSI);
}

    // Tell DUSI there is a character to process.

void DUSI_DTMFCharReceived(enum TAG_RX_CHAN chan, char c)
{
    Apex_SendMsgExt(&mailbox,
            MSG_DUSI_CHAR, (uintptr_t)chan, (uintptr_t)(uint8_t)c);
}

void DUSI_UnsubscribeAll(void)
{
    Unsubscribe_DUSI_CommandOK();
    Unsubscribe_DUSI_CommandERR();
    Unsubscribe_DUSI_ArmRecorder();
}

//***************************************************************
//               *** STATIC FUNCTIONS ***
//***************************************************************

static void StartPasswordTimer(void)
{
    if(passwordTmr.state == TMS_COUNTING)
        Apex_StopTmr(&passwordTmr);

    Apex_StartTmr(&passwordTmr, TIME_DUSI_PASSWORD);
}

static void StopPasswordTimeout(void)
{
    Apex_StopTmr(&passwordTmr);
}

//***************************************************************
//               DUSI State Machine Graph
//***************************************************************

/**
    @dot
    digraph G {
        rankdir=LR;
        ratio=0.7;
        size="8,10"

        node [shape = rect];
        node [style=rounded];
        State_A State_B State_E;

        State_Idle [label="State_Idle:\nWait for #"];
        State_AccumulateLine [label="State_AccumulateLine"];
        State_Error [label="State_Error"];

        edge [fontsize=12];

        State_Idle -> State_AccumulateLine [ label = "#" ];
        State_Idle -> State_Error [ label = "Other Char" ];

        State_AccumulateLine -> State_Idle
                                [label = "Successful Cmd" ];
        State_AccumulateLine -> State_Error
                                [label = "Cmd Entry Error" ];
        State_AccumulateLine -> State_Error [ label = "#" ];

        State_Error -> State_Idle [label = "Unkeyed & ERR" ];
    }
    @enddot
*/
//***************************************************************

static void DusiTask(void)
{
    unsigned event;
    uintptr_t data1, data2;

    Transition_Idle();
    passwordValid = false;         // clear password flag

    FOREVER
    {
        // Call the current state for the transmitter
        // addressed by the message with the message ID as
        // the state's event.

        Apex_PendMsg(&mailbox, &event, &data1, &data2, 0);

        ASSERT(pState != NULL);

        // If we've had a timeout, invalidate the password.

        if(event == EV_DUSI_TIMEOUT)
        {
            passwordValid = false;
        }

        else
        {
            (*pState)((enum TAG_DUSI_EVIDS)event, (char)data2);
        }
    }
}

//***************************************************************
// State A: Wait for start of command (#)
//***************************************************************


static void Transition_Idle(void)
{
    count = 0;                  // Clear char count
    SET_STATE(State_Idle,'A');
}

static void State_Idle(enum TAG_DUSI_EVIDS event, char c)
{
    if(event == EV_DUSI_CHAR)
    {
        if(c == EEP_GetDTMFCommandChar())
            Transition_AccumulateLine();

        else
        {
            if(c != '\0')
                Transition_Error();
        }
    }
}


//***************************************************************
// State B: Accumulate Command line
//***************************************************************

static void Transition_AccumulateLine(void)
{
    count = 0;
    SET_STATE(State_AccumulateLine,'B');
}

static void State_AccumulateLine(enum TAG_DUSI_EVIDS event, char c)
{
    switch(event)
    {
        case EV_DUSI_CHAR: // if DUSI receives a character
            switch(c)
            {
                // Here, the user unkeyed.  Take the command line
                // and run with it.
                case '\0':
                    // add the null to password string
                    myCmdString[count] = '\0';
                    RunDusi(myCmdString); // Then run the command
                    break;

                // Accumulate a new character
                default:
                    myCmdString[count] = c;
                    // check length of string
                    if(count++ >= Z_CMD_LINE-1)
                    {
                        Transition_Error();  // error
                    }
                    break;
            }
            break;

        default:
            break;  // if any other event, then break
    }
}


//***************************************************************
// State E: Command Error: Wait for null
//***************************************************************

static void Transition_Error(void)
{
    SET_STATE(State_Error,'C');
}

static void State_Error(enum TAG_DUSI_EVIDS event, char c)
{
    switch(event)
    {
        case EV_DUSI_CHAR:
            if(c == '\0')
            {
                EMIT_SIGNAL(DUSI_CommandERR, 0, 0);
                Transition_Idle();
            }
            break;

        default:
            break;  // if any other event, then break

    }
}

//***************************************************************
// Run a DUSI command
//***************************************************************

static void RunDusi(char *pLine)
{
    const CMD_TABLE *pCmdTbl;
    char cmdWord[3];
    bool found;
    uint8_t result;

    count = 0;

    // Here, if the line is empty, it's an error.

    if(strlen(pLine) < 2)
    {
        EMIT_SIGNAL(DUSI_CommandERR, 0, 0);
        Transition_Idle(); //  wait for another command
        return;
    }

    cmdWord[0] = *pLine++;
    cmdWord[1] = *pLine++;
    cmdWord[2] = '\0';

    // Look up the specified command
    pCmdTbl = dusiCmdTable;
    found = false;

    while((pCmdTbl->pName != NULL) && !found)
    {
        if(strcmp(cmdWord, pCmdTbl->pName) == 0)
        {
            // Here, found the function in the table

            // If it's the login command
            // or the logoff command, we don't need
            // a valid password.

            if((strcmp(pCmdTbl->pName, LOGIN_COMMAND) == 0) ||
               (strcmp(pCmdTbl->pName, LOGOFF_COMMAND) == 0))
            {
                found = true;
            }

            // Otherwise, is a password required?
            else if(EEP_GetSysFlags(EEP_SYS_FLAG_DTMF_PASS))
            {
                // If a password is required, has one been
                // entered?
                if(passwordValid)
                    found = true;

                // If not, flag an error and exit.
                else
                {
                    EMIT_SIGNAL(DUSI_CommandERR, 0, 0);
                    Transition_Idle(); //  wait for another command
                    return;
                }
            }

            // Here, a password is not required.
            else
                found = true;
        }
        else
            pCmdTbl++;
    }

    if(found)
    {
        // Run the command function specified and return
        result = (*pCmdTbl->pCommand)(pLine);

        switch(result)
        {
            case DUSI_STS_OK:
                EMIT_SIGNAL(DUSI_CommandOK, 0, 0);
                StartPasswordTimer();   // Refresh the timer
                break;

            // exit and wait for another pwd
            case DUSI_STS_LOGOFF:
                EMIT_SIGNAL(DUSI_CommandOK, 0, 0);
                StopPasswordTimeout();
                break;

            // Here, the command itself gave the status signal
            // so no need to do it here.  This is usually
            // because the command must sequence the OK signal
            // before doing something else it has to do.  See
            // the CmdFileRecord() function, for example.
            case DUSI_STS_NULL:
                StartPasswordTimer();   // Refresh the timer
                break;

            default:
                EMIT_SIGNAL(DUSI_CommandERR, 0, 0);
                break;
        }
    }

    else
    {
        EMIT_SIGNAL(DUSI_CommandERR, 0, 0);
    }

    Transition_Idle(); //  wait for another command
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/// @defgroup dusi DTMF Operator Commands
/// @{
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Log In

        @clsyntax #00{password}

        @par Parameters:
            @oo {password}: A digit string corresponding to the
            DTMF password that was set on the operator console
            using the 'pwd' command.

        @description This command checks the entered password for
        validity, and if valid, allows further command entry.
    */

static uint8_t Cmdlogin(const char *pLine)
{
    if(strcmp(pLine, EEP_GetDTMFPwd()) == 0)
    {
        passwordValid = true;         // Validate password.
        return(DUSI_STS_OK);
    }

    else
        return(DUSI_STS_PASSWD);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set courtesy beep frequency

        @clsyntax #50{chan}{tone}{freq}

        @par Command Line Parameters:
            @oo {chan}: Receiver channel number (1 or 2) to which
                the setting applies.

            @oo {tone}: A digit 1 or 2 indicating which
                courtesy beep tone is being changed.

            @oo {freq}: Courtesy beep tone frequency in Hz.
                FG_MIN_FREQ is min, FG_MAX_FREQ is max

        @description This command sets courtesy beep frequency.
    */

static uint8_t CmdSetBeepFreq(const char *pLine)
{
    uint32_t freq;
    unsigned charCount;
    eep_courtsey_beep_t *pBeep;
    char toneNum;
    enum TAG_RX_CHAN chan;

    chan = ORDINAL_2_RX_CHAN(*pLine - '0');
    pLine++;

    toneNum = *pLine++;

    if(APParse(pLine, &charCount, "%d|", &freq) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_RX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if((freq < FG_MIN_FREQ) || (freq > FG_MAX_FREQ))
        return(DUSI_STS_RANGE);

    else
    {
        pBeep = EEP_GetCourtesyBeepParms(chan);

        if(toneNum == '1')
        {
            pBeep->freq1 = (uint16_t)freq;
            EEP_WriteOperatorParameters();
            return(DUSI_STS_OK);
        }

        else if(toneNum == '2')
        {
            pBeep->freq2 = (uint16_t)freq;
            EEP_WriteOperatorParameters();
            return(DUSI_STS_OK);
        }

        else
            return(DUSI_STS_RANGE);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set courtesy beep duration.

        @clsyntax #51{chan}{duration}

        @par Command Line Parameters:
            @oo {chan}: Receiver channel number (1 or 2) to which
                the courtesy beep settings apply.
            @oo {duration}: Length of beep in milliseconds.
                Min = 50; max = 1000.

        @description This command sets the Length of time
        the selected courtesy beep tone will be on after squelch
        closes on the receiver.
    */

static uint8_t CmdSetBeepDuration(const char *pLine)
{
    uint32_t duration;
    unsigned charCount;
    eep_courtsey_beep_t *pBeep;
    char toneNum;
    enum TAG_RX_CHAN chan;

    chan = ORDINAL_2_RX_CHAN(*pLine - '0');
    pLine++;

    toneNum = *pLine++;

    if(APParse(pLine, &charCount, "%d|", &duration) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_RX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if((duration < CB_MIN_DURATION)
                                || (duration > CB_MAX_DURATION))
        return(DUSI_STS_RANGE);

    else
    {
        pBeep = EEP_GetCourtesyBeepParms(chan);

        if(toneNum == '1')
        {
            pBeep->duration1 = (uint16_t)duration;
            EEP_WriteOperatorParameters();
            return(DUSI_STS_OK);
        }

        else if(toneNum == '2')
        {
            pBeep->duration2 = (uint16_t)duration;
            EEP_WriteOperatorParameters();
            return(DUSI_STS_OK);
        }

        else
            return(DUSI_STS_RANGE);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set courtesy beep mode.

        @clsyntax #52{chan}{mode}

        @par Command Line Parameters:
            @oo {chan}: Receiver channel number (1 or 2) to which
                the courtesy beep settings apply.
            @oo {mode}: A code from the following table identifying
                the desired courtesy beep mode.

        @verbatim
            Code
            Number  Mode Description
            ======  ================
              0     Courtesy beep disabled
              1     Single tone beep
              2     Double tone beep
              3     Morse code character
        @endverbatim

        @description This command sets the type of courtesy beep
        desired for the given receiver channel.
    */

static uint8_t CmdSetBeepMode(const char *pLine)
{
    eep_courtsey_beep_t *pBeep;
    enum TAG_CB_MODE mode;
    enum TAG_RX_CHAN chan;

    chan = ORDINAL_2_RX_CHAN(*pLine - '0');
    pLine++;

    mode = (enum TAG_CB_MODE)(*pLine - '0');
    pLine++;

    if(*pLine != '\0')
        return(DUSI_STS_SYNTAX);

    else if(!VALID_RX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if(mode > CB_MODE_MAX)
        return(DUSI_STS_RANGE);

    else
    {
        pBeep = EEP_GetCourtesyBeepParms(chan);
        pBeep->mode = mode;
        EEP_WriteOperatorParameters();
        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set courtesy beep level.

        @clsyntax #31{chan}{level}

        @par Command Line Parameters:
            @oo {chan}: Transmitter channel number (1 or 2) to which
                the courtesy beep settings apply.
            @oo {level}: Volume level in percent (0 to 100).

        @description This command sets courtesy beep audio level.
    */

static uint8_t CmdSetBeepLevel(const char *pLine)
{
    uint32_t level;
    unsigned charCount;
    enum TAG_TX_CHAN chan;

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;

    if(APParse(pLine, &charCount, "%d|", &level) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if(level > 100)
        return(DUSI_STS_RANGE);

    else
    {
        EEP_SetCourtesyBeepLevel(chan, (uint8_t)level);
        EEP_WriteOperatorParameters();
        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set Morse rate

        @clsyntax #53{chan}{rate}

        @par Command Line Parameters:
            @oo {chan}: Transmitter channel number to which the
                setting applies.
            @oo {rate}: Morse rate in words per minute.
                MIN_MORSE_CODE_RATE is min, MAX_MORSE_CODE_RATE
                is max

        @description This command sets the Morse code
        rate (in wpm).
    */

static uint8_t CmdSetMorseRate(const char *pLine)
{
    uint32_t rate;
    unsigned charCount;
    enum TAG_TX_CHAN chan;

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;

    if(APParse(pLine, &charCount, "%d|", &rate) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if((rate < MIN_MORSE_CODE_RATE) ||
            (rate > MAX_MORSE_CODE_RATE))
        return(DUSI_STS_RANGE);

    else
    {
        EEP_SetMorseRate(chan, (uint8_t)rate);
        EEP_WriteOperatorParameters();
        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set Morse tone frequency

        @clsyntax #54{chan}{freq}

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) to which the
                setting applies.
            @oo {freq}: Morse tone frequency in Hz. FG_MIN_FREQ
                is min and FG_MAX_FREQ is max

        @description This command sets the Morse code tone
            frequency in Hz.
    */

static uint8_t CmdSetMorseFreq(const char *pLine)
{
    uint32_t freq;
    unsigned charCount;
    enum TAG_TX_CHAN chan;

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;

    if(APParse(pLine, &charCount, "%d|", &freq) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if((freq < FG_MIN_FREQ) || (freq > FG_MAX_FREQ))
        return(DUSI_STS_RANGE);

    else
    {
        EEP_SetMorseFreq(chan, (uint16_t)freq);
        EEP_WriteOperatorParameters();
        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set Morse audio level

        @clsyntax #32{chan}{level}

        @par Command Line Parameters:
            @oo {chan}: Transmitter channel number to which the
                setting applies.
            @oo {level}: Morse volume level in
                percent (0 to 100).

        @description This command sets CW (Morse code) rate (in wpm),
            frequency. and
            volume level for transmitter channels 1 and 2 respectively.
    */

static uint8_t CmdSetMorseLevel(const char *pLine)
{
    uint32_t level;
    unsigned charCount;
    enum TAG_TX_CHAN chan;

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;

    if(APParse(pLine, &charCount, "%d|", &level) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if(level > 100)
        return(DUSI_STS_RANGE);

    else
    {
        EEP_SetMorseLevel(chan, (uint8_t)level);
        EEP_WriteOperatorParameters();
        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set transmitter flag D

        @clsyntax #13{chan}{x}

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) to which the
                transmitter flag settings apply.
            @oo {x}: A 0 or 1 to disable or enable (respectively)
                transmitted DTMF audio on the selected channel.
    */

static uint8_t CmdSetTxFlagD(const char *pLine)
{
    return(CmdSetTxFlagsCommon(pLine, EEP_TX_FLAG_DMUTE));
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set transmitter flag P

        @clsyntax #11{chan}{x}

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) to which the
                transmitter flag settings apply.
            @oo {x}: A 0 or 1 to disable or enable (respectively)
                CTCSS output (encode) tone on the selected
                channel.
    */

static uint8_t CmdSetTxFlagP(const char *pLine)
{
    unsigned charCount;
    uint16_t flags;                    // Flag word
    bool flag;                         // Boolean value
    enum TAG_TX_CHAN chan;

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;

    if(APParse(pLine, &charCount, "%b|", &flag) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else
    {
        EEP_InqTxFlags(chan, &flags);

        // Tell Central.
        if(flag)
        {
            flags |= EEP_TX_FLAG_SATE;
        }

        else
        {
            flags &= ~EEP_TX_FLAG_SATE;
            SATE_Off(chan);
        }

        EEP_SetTxFlags(chan, flags);
        EEP_WriteOperatorParameters();
        return(DUSI_STS_OK);
    }
}



// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set transmitter flag B

        @clsyntax #12{chan}{x}

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) to which the
                transmitter flag settings apply.
            @oo {x}: A 0 or 1 to disable or enable (respectively)
                a courtesy beep on the selected transmitter.
    */

static uint8_t CmdSetTxFlagB(const char *pLine)
{
    return(CmdSetTxFlagsCommon(pLine, EEP_TX_CB_ENABLE));
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set transmitter flag R

        @clsyntax #16{chan}{x}

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) to which the
                transmitter flag settings apply.
            @oo {x}: A 0 or 1 to disable or enable (respectively)
                a CTCSS reverse burst on the selected channel.
    */

static uint8_t CmdSetTxFlagR(const char *pLine)
{
    return(CmdSetTxFlagsCommon(pLine, EEP_TX_FLAG_REVBURST));
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set transmitter flag T

        @clsyntax #10{chan}{x}

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) to which the
                transmitter flag settings apply.
            @oo {x}: A 0 or 1 to disable or enable (respectively)
                the selected transmitter itself.
    */

static uint8_t CmdSetTxFlagT(const char *pLine)
{
    unsigned charCount;
    uint16_t flags;                    // Flag word
    bool flag;                         // Boolean value
    enum TAG_TX_CHAN chan;

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;

    if(APParse(pLine, &charCount, "%b|", &flag) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else
    {
        EEP_InqTxFlags(chan, &flags);

        // Tell Central.
        if(flag)
        {
            flags |= EEP_TX_FLAG_TXENA;
            CEN_TxEnable(chan);
        }

        else
        {
            flags &= ~EEP_TX_FLAG_TXENA;
            CEN_TxDisable(chan);
        }

        EEP_SetTxFlags(chan, flags);
        EEP_WriteOperatorParameters();
        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set transmitter flag V

        @clsyntax #17{chan}{x}

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) to which the
                transmitter flag settings apply.
            @oo {x}: A 0 or 1 to disable or enable (respectively)
                voice ID on the selected transmitter.
    */

static uint8_t CmdSetTxFlagV(const char *pLine)
{
    return(CmdSetTxFlagsCommon(pLine, EEP_TX_FLAG_VID));
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Common functionality for setting/clearing flags.

static uint8_t CmdSetTxFlagsCommon(
                           const char *pLine, uint16_t mask)
{
    unsigned charCount;
    uint16_t flags;                    // Flag word
    bool flag;                         // Boolean value
    enum TAG_TX_CHAN chan;

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;

    if(APParse(pLine, &charCount, "%b|", &flag) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else
    {
        EEP_InqTxFlags(chan, &flags);

        if(flag)
            flags |= mask;
        else
            flags &= ~mask;

        EEP_SetTxFlags(chan, flags);
        EEP_WriteOperatorParameters();
        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set receiver flag P

        @clsyntax #40{chan}{x}

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) to which the transmitter
                flag settings apply.
            @oo {x}: A 0 or 1 to disable or enable (respectively) CTCSS
                input (decode) tone on the selected channel.
    */

static uint8_t CmdSetRxFlagP(const char *pLine)
{
    unsigned charCount;
    uint16_t flags;                    // Flag word
    bool flag;                         // Boolean value
    enum TAG_RX_CHAN chan;

    chan = ORDINAL_2_RX_CHAN(*pLine - '0');
    pLine++;

    if(APParse(pLine, &charCount, "%b|", &flag) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_RX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else
    {
        EEP_InqRxFlags(chan, &flags);

        if(flag)
            flags |= EEP_RX_FLAG_SATD;
        else
            flags &= ~EEP_RX_FLAG_SATD;

        EEP_SetRxFlags(chan, flags);
        EEP_WriteOperatorParameters();
        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set receiver flag R

        @clsyntax #41{chan}{x}

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) to which the
                receiver flag settings apply.
            @oo {x}: A 0 or 1 to disable or enable (respectively)
                audio on the selected channel.
    */

static uint8_t CmdSetRxFlagR(const char *pLine)
{
    unsigned charCount;
    uint16_t flags;                    // Flag word
    bool flag;                         // Boolean value
    enum TAG_RX_CHAN chan;

    chan = ORDINAL_2_RX_CHAN(*pLine - '0');
    pLine++;

    if(APParse(pLine, &charCount, "%b|", &flag) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_RX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else
    {
        EEP_InqRxFlags(chan, &flags);

        if(flag)
            flags |= EEP_RX_FLAG_ENABLE;
        else
            flags &= ~EEP_RX_FLAG_ENABLE;

        CEN_RxEnable(chan, flag);

        EEP_SetRxFlags(chan, flags);
        EEP_WriteOperatorParameters();
        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set a Receiver's Filter Index

        @clsyntax #45{chan}{type}{index}

        @par Command Line Parameters:
            @oo {chan}: 1 or 2, identifying the
                receiver channel to which the command applies.
            @oo {type}: 0 = Deemphasis, 1 = CTCSS Reject
            @oo {index}: Deemphasis: 0 or 1, for disable or enable,
                CTCSS Reject: 0 = disable, 1 = 3rd order,
                2 = 4th order

        @description This command sets the filter profile
            for the addressed receiver's filter type.
*/

static uint8_t CmdSetFilterIndex(const char *pLine)
{
    uint32_t index;
    enum TAG_RX_CHAN chan;
    char type;

    if(strlen(pLine) != 3)
        return(DUSI_STS_SYNTAX);

    chan = ORDINAL_2_RX_CHAN(*pLine - '0');
    pLine++;
    type = *pLine++;
    index = (uint32_t)*pLine - '0';

    if(!VALID_RX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else
    {
        switch(type)
        {
            // Deemphasis filter:
            case '0':
                if(index >= MAX_DEEMPH_FILTERS)
                    return(DUSI_STS_RANGE);
                else
                {
                    EEP_SetDeemphasisFilterIndex(chan, index);
                    ASTREAM_SetDeemphasisFilterIndex(chan, index);
                }
                break;

            case '1':
                if(index >= MAX_CTCSS_REJECT_FILTERS)
                    return(DUSI_STS_RANGE);
                else
                {
                    EEP_SetCtcssRejectionFilterIndex(chan, index);
                    ASTREAM_SetCtcssRejectionFilterIndex(chan, index);
                }
                break;

            default:
                return(DUSI_STS_RANGE);
        }

        EEP_WriteOperatorParameters();

        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set ID timer.

        @clsyntax #56{chan}{idTmr}

        @par Command Line Parameters:
            @oo {chan}: Transmitter channel number (1 or 2) to
                which the timer settings apply.
            @oo {idTmr}: ID timer setting in seconds.  A setting
                of zero means no ID timer. MIN_IDTIME is min;
                MAX_IDTIME is max.

        @description This command sets the ID timer on the addressed
            channel to the value specified by the parameters.

        @note The limits on the above parameters are given in
            uilimits.h.
    */

static uint8_t CmdSetIdTimer(const char *pLine)
{
    uint32_t idTmr, unused, txTimeout, hangTime;
    unsigned charCount;
    enum TAG_TX_CHAN chan;

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;

    if(APParse(pLine, &charCount, "%d|", &idTmr) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if((idTmr != 0) &&
            (idTmr < MIN_IDTIME || idTmr > MAX_IDTIME))
        return(DUSI_STS_RANGE);

    else
    {
        EEP_InqTimers(chan, &unused, &txTimeout, &hangTime);

        EEP_SetTimers(chan, (uint16_t)idTmr,
                (uint16_t)txTimeout, (uint16_t)hangTime);

        EEP_WriteOperatorParameters();
        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set Tx Timeout Timer

        @clsyntax #57{chan}{txTimeout}

        @par Command Line Parameters:
            @oo {chan}: Transmitter channel number (1 or 2) to
                which the timer settings apply.
            @oo {txTimeout}: Transmitter timeout setting in seconds.
                A setting of zero means no timeout. MIN_TXTIME
                is min; MAX_TXTIME is max.

        @description This command sets the transmitter timeout
            timers on the addressed channel.

        @note The limits on the above parameters are given in
            uilimits.h.
    */

static uint8_t CmdSetTxTimeout(const char *pLine)
{
    uint32_t txTimeout, idTmr, hangTime, unused;
    unsigned charCount;
    enum TAG_TX_CHAN chan;

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;

    if(APParse(pLine, &charCount, "%d|", &txTimeout) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if((txTimeout != 0)
        && (txTimeout < MIN_TXTIME || txTimeout > MAX_TXTIME))
        return(DUSI_STS_RANGE);

    else
    {
        EEP_InqTimers(chan, &idTmr, &unused, &hangTime);

        EEP_SetTimers(chan, (uint16_t)idTmr,
                    (uint16_t)txTimeout, (uint16_t)hangTime);

        EEP_WriteOperatorParameters();
        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set Hang Timer.

        @clsyntax #58{chan}{hangTime}

        @par Command Line Parameters:
            @oo {chan}: Transmitter channel number (1 or 2) to
                which the timer settings apply.
            @oo {hangTime}: Duration of hang-timer in milliseconds.
                MIN_HANGTIME
                is min; MAX_HANGTIME is max.

        @description This command sets the hang time on the addressed
            transmitter channel.

        @note The limits on the above parameters are given in
            uilimits.h.
    */

static uint8_t CmdSetHangTime(const char *pLine)
{
    uint32_t hangTime, unused, idTmr, txTimeout;
    unsigned charCount;
    enum TAG_TX_CHAN chan;

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;

    if(APParse(pLine, &charCount, "%d|", &hangTime) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if((hangTime != 0)
        && (hangTime < MIN_HANGTIME || hangTime > MAX_HANGTIME))
        return(DUSI_STS_RANGE);

    else
    {
        EEP_InqTimers(chan, &idTmr, &txTimeout, &unused);

        EEP_SetTimers(chan, (uint16_t)idTmr,
                    (uint16_t)txTimeout, (uint16_t)hangTime);

        EEP_WriteOperatorParameters();
        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set transmiter channel audio level

        @clsyntax #30{chan}{level}

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) to which the
                transmitter level setting applies.
            @oo {level}: Volume level (0 to 100).

        @description This command sets the audio volume level
        associated with the specified transmitter channel.  The
        level is a logrithmic scale.  100 has a gain of +20 dB
        and 0 has a gain of -28 dB.
    */

static uint8_t CmdSetTxAudioLevel(const char *pLine)
{
    uint32_t txLevel;
    unsigned charCount;
    enum TAG_TX_CHAN chan;

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;

    if(APParse(pLine, &charCount, "%d|", &txLevel) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if(txLevel > MAX_TX_LEVEL)
        return(DUSI_STS_RANGE);

    else
    {
        EEP_SetTxLevel(chan, (uint8_t)txLevel);

        // Set the new level in the hardware.
        if(chan == TX_CHAN_1)
        {
            ASTREAM_SetTx1Level((uint8_t)txLevel);
        }
        else
        {
            ASTREAM_SetTx2Level((uint8_t)txLevel);
        }

        EEP_WriteOperatorParameters();
        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Log off

        @clsyntax #99

        @par Command Line Parameters:
            @oo None.

        @description This command ends the user DTMF setting.

    */

static uint8_t CmdLogOff(const char *pLine)
{
    if(*pLine != '\0')
        return(DUSI_STS_SYNTAX);
    else
    {
        passwordValid = false;         // Invalidate password.
        return(DUSI_STS_LOGOFF);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#if EMULATOR == 0
    /** Reset the microcontroller

        @clsyntax #90

        @par Command Line Parameters:
            @oo None.

        @description This command hangs the microcontroller and
            forces the watchdog timer to reset.
    */

static uint8_t CmdReset(const char *pLine)
{
    if(*pLine != '\0')
        return(DUSI_STS_SYNTAX);
    else
    {
        APPrint(STDOUT, " - Resetting. Press 'Enter'.");
        DI();
        FOREVER
            ;
    }
}
#endif

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /// ID String Table

    /// This table is used in the Set ID String command, (25) to
    /// decode the {string} parameter into ASCII codes.

static const char idTable[] =
    "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789/-.? ";

    /** Set ID string

        @clsyntax #25{chan}{string}

        @par Command Line Parameters:
            @oo {chan}: Transmitter channel number (1 or 2) to
                which the ID string applies.
            @oo {string}: A string of digit pairs containing
                codes that translate into the ID.

        @description This command sets the ID string for either
            transmitter channel 1 or 2.

        @note Max length of the string is 20 characters (including
            terminating null).
    */
static uint8_t CmdSetIDString(const char *pLine)
{
    unsigned charCount;
    int i;
    char *pString;
    char oddChar, evenChar;
    char string[40];       // ID string entry variable
    char wordBuf[40];      // ID string parse variable
    uint8_t c;
    enum TAG_TX_CHAN chan;

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;

    pString = string;

    // check for correct number of parameters
    if(APParse(pLine, &charCount, "%40w|", string) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    // Check to make sure there are
    // an even number of characters (pairs)
    else if((strlen(string) & 0x01) != 0)
        return(DUSI_STS_SYNTAX);

    // if string length is an even number of characters
    else
    {
        i = 0;
        while(*pString != '\0')
        {
            // parse the string one pair at a time
            oddChar = *pString++;
            evenChar = *pString++;

            // now convert the pair from decimal to BCD
            c = ((oddChar & 0x0F) * 0x0A) + (evenChar & 0x0F);

            // make sure character is in the table
            if(c > (sizeof(idTable)-1))
                return(DUSI_STS_SYNTAX);

            // add the character to the buffer
            wordBuf[i] = idTable[c];

            // increment the loop counter
            i++;

        }
        // terminate the buffer string with a null
        wordBuf[i] = '\0';
        EEP_SetIDString(chan, wordBuf);
        EEP_WriteOperatorParameters();
        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Play a transmitter's ID

        @clsyntax #26{chan}

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) indicating on
                which channel to play the ID,

        @description This command plays the ID for the selected
        transmitter.
    */
static uint8_t CmdPlayID(const char *pLine)
{
    enum TAG_TX_CHAN chan;

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;


    if(*pLine != '\0')
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else
    {
        CEN_PlayID(chan);
        return(DUSI_STS_OK);
    }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    Set Voice ID File Number

    @clsyntax #59{chan}{filenum}

    @par Command Line Parameters:
        @oo {chan}: Transmitter number (1 or 2) indicating which
            transmitter to which the ID pertains.
        @oo {filenum}: File number (1 through 8) indicating
            which file to play for the ID.

    @description This command sets the file number for the
    selected transmitter channel's voice ID.
    */

static uint8_t CmdSetVoiceIDFileNum(const char *pLine)
{
    uint8_t fileNum;
    enum TAG_TX_CHAN chan;

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;
    fileNum = ORDINAL_2_FILENUM(*pLine - '0');
    pLine++;

    if(*pLine != '\0')
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if(!VALID_USER_FILENUM(fileNum))
        return(DUSI_STS_RANGE);

    else
    {
        EEP_SetVidFileNum(chan, fileNum);
        EEP_WriteOperatorParameters();
        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set Receiver-Transmitter Path

    @clsyntax #20{source}{dest}{priority}

        @par Command Line Parameters:
        @oo {source}: 1 for receiver 1, 2 for receiver 2.
        @oo {dest}: 1 for transmitter 1, 2 for transmitter 2.
        @oo {priority} 1 for low, 2 for medium (mix) or 3 for high.

        @description This command selects which transmitters are
        to be activated when a particular receiver is unsquelched.
    */

static uint8_t CmdSetRxTxPath(const char *pLine)
{
    eep_path_t *pPath;
    eep_path_t path;
    char priority;
    enum TAG_RX_CHAN rxChan;
    enum TAG_TX_CHAN txChan;

    rxChan = ORDINAL_2_RX_CHAN(*pLine - '0');
    pLine++;
    txChan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;

    priority = *pLine++;

    if(*pLine != '\0')
        return(DUSI_STS_SYNTAX);

    else if(priority != '1' && priority != '2' && priority != '3')
        return(DUSI_STS_RANGE);

    else if(!VALID_TX_CHAN(txChan))
        return(DUSI_STS_RANGE);

    else if(!VALID_RX_CHAN(rxChan))
        return(DUSI_STS_RANGE);

    else
    {
        switch(rxChan)
        {
            case RX_CHAN_1: path.src = ASTREAM_SSIG_RX1; break;
            case RX_CHAN_2: path.src = ASTREAM_SSIG_RX2; break;
            default: return(DUSI_STS_RANGE);
        }

        switch(txChan)
        {
            case TX_CHAN_1: path.dest = ASTREAM_DSIG_TX1; break;
            case TX_CHAN_2: path.dest = ASTREAM_DSIG_TX2; break;
            default: return(DUSI_STS_RANGE);
        }

        switch(priority)
        {
            case '1': path.priority = 'L'; break;
            case '2': path.priority = 'M'; break;
            case '3': default: path.priority = 'H'; break;
        }

        // Make sure the path is unique:

        // Use an existing path with the same source and destination
        // if one exists:
        if((pPath = EEP_FindPath(path.src, path.dest)) != NULL)
        {
            EEP_SetPath(pPath, &path);
            EEP_WriteOperatorParameters();
            return(DUSI_STS_OK);
        }

        // If no existing path with same src and dest exists, find
        // an empty path slot to use.
        else if((pPath = EEP_FindPath(
                     ASTREAM_NULL_SOURCE, ASTREAM_NULL_DEST)) != NULL)
        {
            EEP_SetPath(pPath, &path);
            EEP_WriteOperatorParameters();
            return(DUSI_STS_OK);
        }

        // If we can't find any empty path slots, give a
        // message and return.
        else
        {
            return(DUSI_STS_RANGE);
        }
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Schedule an Announcement

    @clsyntax #21{chan}{fileNum}{priority}{period}

        @par Command Line Parameters:
        @oo {chan}: A 1 or 2 for transmitter 1 or 2 respectively,
            on which the announcement is to be heard.
        @oo {fileNum}: File number identifying the announcement
            file.
        @oo {priority} The number 1, 2, or 3, corresponding to
            low, medium, or high priority, respectively.
        @oo {period}: The amount of time, in minutes, between
            successive playbacks of the announcement.

        @description This command schedules an announcement to
        play on a selected transmitter.
    */

static uint8_t CmdSchedAnno(const char *pLine)
{
    eep_anno_t *pAnno;
    eep_anno_t anno;
    unsigned i, charCount, period;
    enum TAG_TX_CHAN chan;
    char priority;
    uint8_t fileNum;

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;

    fileNum = ORDINAL_2_FILENUM(*pLine - '0');
    pLine++;

    priority = *pLine++;

    if(APParse(pLine, &charCount, "%d|", &period) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if(period == 0 || period > MAX_ANNO_PERIOD)
        return(DUSI_STS_RANGE);

    // Here, add a new path or edit an existing one.
    else
    {
        anno.fileNum = fileNum;
        anno.period = period;
        anno.enabled = false;
        anno.chan = chan;

        // Decode the priority parameter:
        switch(priority)
        {
            case '3': anno.priority = 'H'; break;
            case '2': anno.priority = 'M'; break;
            case '1': anno.priority = 'L'; break;
            default: return(DUSI_STS_RANGE);
        }

        // Make sure the announcement is unique:

        // If this announcement is the same as an existing one,
        // replace the existing one.
        for(i=0; i<MAX_ANNO; i++)
        {
            pAnno = EEP_GetAnno(i);
            if(pAnno->fileNum == anno.fileNum)
            {
                // If another announcement exists with the same file
                // number, set the enable parameter for the new
                // announcement to be the same as the other one.
                anno.enabled = pAnno->enabled;

                // If this one is a duplicate, use the same slot.
                if(pAnno->chan == anno.chan)
                    break;
            }
        }

        // If we didn't find a duplicate, go through the list
        // and find an empty slot.
        if(i == MAX_ANNO)
        {
            for(i=0; i<MAX_ANNO; i++)
            {
                pAnno = EEP_GetAnno(i);
                if(pAnno->chan == TXCHAN_INVALID)
                    break;
            }
        }

        // Give an error if we didn't find an open slot.
        if(i == MAX_ANNO)
        {
            return(DUSI_STS_SYNTAX);
        }
        else
        {
            // Add the announcement in EEP before adding it in
            // audio task because audio task looks in EEP to
            // get the announcement parameters.
            EEP_SetAnno(pAnno, &anno);
            EEP_WriteOperatorParameters();
            AUD_AddAnno(i);
            return(DUSI_STS_OK);
        }
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Delete Path

    @clsyntax #22{source}{dest}

    @par Command Line Parameters:
    @oo {source}: 1 for receiver 1, 2 for receiver 2.
    @oo {dest}: 1 for transmitter 1, 2 for transmitter 2.

    @description This command deletes any path matching
    the given source and destination parameters.
    */

static uint8_t CmdDeletePath(const char *pLine)
{
    eep_path_t *pPath;
    astream_srcsig_t srcSig;
    astream_destsig_t destSig;
    char rxChan, txChan;

    // Parse the parameters.

    rxChan = *pLine++;
    txChan = *pLine++;

    if(*pLine != '\0')
        return(DUSI_STS_SYNTAX);

    // Here, add a new path or edit an existing one.
    else
    {
        switch(rxChan)
        {
            case '1': srcSig = ASTREAM_SSIG_RX1; break;
            case '2': srcSig = ASTREAM_SSIG_RX2; break;
            default: return(DUSI_STS_RANGE);
        }

        switch(txChan)
        {
            case '1': destSig = ASTREAM_DSIG_TX1; break;
            case '2': destSig = ASTREAM_DSIG_TX2; break;
            default: return(DUSI_STS_RANGE);
        }

        if((pPath = EEP_FindPath(srcSig, destSig)) != NULL)
        {
            EEP_DeletePath(pPath);
            EEP_WriteOperatorParameters();
        }
        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Delete an announcement schedule

        @clsyntax #24{chan}{fileNum}

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) corresponding to the
            transmitter to which the schedule applies.
            @oo {fileNum}: A number 1 through 8 corresponding to the
            file containing the announcement.

        @description This command deletes the schedule for the
        indicated announcement.  It does not delete the announcement
        audio file itself.  To do that, see command 71.
    */

static uint8_t CmdDeleteAnno(const char *pLine)
{
    short i;
    enum TAG_TX_CHAN chan;
    uint8_t fileNum;

    // Parse the parameters.

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;
    fileNum = ORDINAL_2_FILENUM(*pLine - '0');
    pLine++;

    if(*pLine != '\0')
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if(!VALID_USER_FILENUM(fileNum))
        return(DUSI_STS_RANGE);

    // Here, find the announcement and delete it if found.
    else
    {
        if((i = EEP_FindAnno(chan, fileNum)) != -1)
        {
            // Delete in audio task before deleting in
            // EEP because audio task reads the record in
            // EEP to determine which announcement to delete.
            AUD_DeleteAnno((unsigned short)i);
            EEP_DeleteAnno(EEP_GetAnno((unsigned short)i));
            EEP_WriteOperatorParameters();
            return(DUSI_STS_OK);
        }
        else
            return(DUSI_STS_RANGE);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Enable an announcement schedule

        @clsyntax #27{fileNum}{enable}

        @oo {fileNum}: A number from 1 to 8 indicating the file
            that contains the announcement audio.
        @oo {enable}: 1 to enable playback, or 0 to disable
            playback of the announcement.

        @description This command enables or disables the playback
        of an announcement.  The announcement will be enabled
        or disabled for all channels on which the file number is
        assigned.  Disabling an announcement does not delete the
        schedule or the audio file.  It only prevents playback
        until the announcement is re-enabled.
    */

static uint8_t CmdEnableAnno(const char *pLine)
{
    eep_anno_t *pAnno;
    unsigned i;
    bool found;
    uint8_t fileNum;
    char enable;

    // Parse the parameters.

    fileNum = ORDINAL_2_FILENUM(*pLine - '0');
    pLine++;
    enable = *pLine++;

    if(*pLine != '\0')
        return(DUSI_STS_SYNTAX);

    else if(!VALID_USER_FILENUM(fileNum))
        return(DUSI_STS_RANGE);

    else if(enable != '0' && enable != '1')
        return(DUSI_STS_RANGE);

    // Here, find the announcement and enable or disable it if found.
    else
    {
        found = false;

        for(i=0; i<MAX_ANNO; i++)
        {
            pAnno = EEP_GetAnno(i);
            if(pAnno->chan != TXCHAN_INVALID &&
                pAnno->fileNum == fileNum)
            {
                found = true;

                // The expression (enable != '0') evaluates to
                // true if enable is not zero (i.e. enable == '1'),
                // and false if enable equals zero.
                EEP_EnableAnno(pAnno, (enable != '0'));

                // Tell audio task to add the timer if it's being
                // enabled, or delete it if it's being disabled.
                if(enable)
                    AUD_AddAnno(i);

                else
                    AUD_DeleteAnno(i);

                EEP_WriteOperatorParameters();
            }
        }

        if(found)
            return(DUSI_STS_OK);
        else
            return(DUSI_STS_RANGE);
    }
}

#if 0   // No need for this command.  See fplay.
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    Play an Announcement File

    @clsyntax #28{chan}{fileNum}

        @par Command Line Parameters:
        @oo {chan}: The transmitter channel (1 or 2) to
        which the announcement should be played.
        @oo {fileNum}: A number from 1 to 8 indicating the file
            that contains the announcement audio.
 */
static uint8_t CmdPlayAnno(const char *pLine)
{
    short i;
    enum TAG_TX_CHAN chan;
    uint8_t fileNum;

    // Parse the parameters.

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;
    fileNum = ORDINAL_2_FILENUM(*pLine - '0');
    pLine++;

    if(*pLine != '\0')
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if(!VALID_USER_FILENUM(fileNum))
        return(DUSI_STS_RANGE);

    // Here, find the announcement and play it if found.
    else
    {
        if((i = EEP_FindAnno(chan, fileNum)) != -1)
        {
            CEN_PlayAnno((unsigned short)i);
            return(DUSI_STS_OK);
        }

        else
            return(DUSI_STS_RANGE);
    }
}
#endif

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    Play a File

    @clsyntax #72{chan}{fileNum}

        @par Command Line Parameters:
        @oo {chan}: The transmitter channel (1 or 2) to
        which the file should be played.
        @oo {fileNum}: A number from 1 to 8 indicating the file
            to be played.
 */
static uint8_t CmdPlayFile(const char *pLine)
{
    enum TAG_TX_CHAN chan;
    uint8_t fileNum;

    // Parse the parameters.

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;
    fileNum = ORDINAL_2_FILENUM(*pLine - '0');
    pLine++;

    if(*pLine != '\0')
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if(!VALID_USER_FILENUM(fileNum))
        return(DUSI_STS_RANGE);

    // Here, find the announcement and play it if found.
    else
    {
        CEN_PlayFile(chan, fileNum);
        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set a Receiver's CTCSS Decoder Frequency Index

        @clsyntax > 60{rx_chan}{index}

        @par Command Line Parameters:
            @oo {rx_chan}: 1 or 2, identifying the
                receiver channel to which the command applies.
            @oo {index}: 0 to 36, to select tone decoder frequency.

        @description This command sets the CTCSS frequency index
            for the addressed receiver's decoder.
*/

static uint8_t CmdSetCTCSSDecodeIndex(const char *pLine)
{
    uint32_t index;
    unsigned charCount;
    enum TAG_RX_CHAN chan;

    chan = ORDINAL_2_RX_CHAN(*pLine - '0');
    pLine++;

    if(APParse(pLine, &charCount, "%d|", &index) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_RX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if(index > SATD_FILTER_TABLE_SIZE)
        return(DUSI_STS_RANGE);

    else
    {
        EEP_SetCTCSSDecodeIndex(chan, (uint8_t)index);
        EEP_WriteOperatorParameters();
        SATD_LoadDecoderParms(chan);

        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set a Transmitter's CTCSS Frequency Index

        @clsyntax #61{tx_chan}{index}

        @par Command Line Parameters:
            @oo {tx_chan}: 1 or 2, identifying the transmitter
                channel.
            @oo {index}: 0 to 36, to select tone encoder frequency.

        @description This command sets the CTCSS frequency index
            for the addressed transmitter's encoder.
*/

static uint8_t CmdSetCTCSSEncodeIndex(const char *pLine)
{
    uint32_t index;
    unsigned charCount;
    enum TAG_TX_CHAN chan;

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;

    if(APParse(pLine, &charCount, "%d|", &index) != 1)
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if(index > SATD_FILTER_TABLE_SIZE)
        return(DUSI_STS_RANGE);

    else
    {
        EEP_SetCTCSSEncodeIndex(chan, (uint8_t)index);
        EEP_WriteOperatorParameters();
        SATE_UpdateFrequency(chan);

        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set a Transmitter's CTCSS Mode

        @clsyntax #62{tx_chan}{mode}

        @par Command Line Parameters:
            @oo {tx_chan}: 1 or 2, identifying the
                transmitter channel to which the command applies.
            @oo {mode}: 1 or 2, depending on the desired mode:
                '1' if the transmitter should encode only while
                the receiver is unsquelched, or '2' if the
                transmitter should encode whenever PTT is asserted
        @description This command sets the CTCSS operation mode.
            The mode determines the circumstances under which the
            transmitter encodes tone.
*/

static uint8_t CmdSetCTCSSEncodeMode(const char *pLine)
{
    enum TAG_TX_CHAN chan;
    char mode;

    chan = ORDINAL_2_TX_CHAN(*pLine - '0');
    pLine++;
    mode = *pLine++;

    if(*pLine != '\0')
        return(DUSI_STS_SYNTAX);

    else if(!VALID_TX_CHAN(chan))
        return(DUSI_STS_RANGE);

    else if(mode != '1' && mode != '2')
        return(DUSI_STS_RANGE);

    else
    {

        EEP_SetCTCSSEncodeMode(chan, ORDINAL_2_SATE_MODE(mode-'0'));
        EEP_WriteOperatorParameters();

        return(DUSI_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Start recording a voice file on the current channel.

        @clsyntax #70{filenum}

        @par Command Line Parameters:
            @oo {filenum}: File number (1 through 8) indicating
                which file to record.

        @description This command tells central to arm the voice
        recorder on the channel receiving this DUSI command.
        Central sets a timer for 10 seconds and waits for the user
        to key the receiver after the OK message is played.  When
        the receiver is keyed, the audio is captured and stored in
        the selected file.
    */

static uint8_t CmdFileRecord(const char *pLine)
{
    uint8_t fileNum;

    // Parse the parameters.

    fileNum = ORDINAL_2_FILENUM(*pLine - '0');
    pLine++;

    if(*pLine != '\0')
        return(DUSI_STS_SYNTAX);

    else if(!VALID_USER_FILENUM(fileNum))
        return(DUSI_STS_RANGE);

    // Here, find the announcement and delete it if found.
    else
    {
        // Here, give the OK signal here instead of just returning
        // OK status because we must give the OK signal before
        // giving the signal to arm the recorder.  This is because
        // Central must clean up the paths from the DUSI command
        // before arming the recorder.
        EMIT_SIGNAL(DUSI_CommandOK, 0, 0);
        EMIT_SIGNAL(DUSI_ArmRecorder, 0, fileNum);
        return(DUSI_STS_NULL);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Delete a voice file

        @clsyntax #71{filenum}

        @par Command Line Parameters:
            @oo {filenum}: File number (1 through 8) indicating
            which file to delete.

        @description This command deletes all data from the file
            addressed by the filenum parameter.
    */

static uint8_t CmdFileDelete(const char *pLine)
{
    uint8_t fileNum;

    // Parse the parameters.

    fileNum = ORDINAL_2_FILENUM(*pLine - '0');
    pLine++;

    if(*pLine != '\0')
        return(DUSI_STS_SYNTAX);

    else if(!VALID_USER_FILENUM(fileNum))
        return(DUSI_STS_RANGE);

    // Here, find the announcement and delete it if found.
    else
    {
        FSys_Delete(fileNum);
        return(DUSI_STS_OK);
    }
}

/// @}
//***************************************************************
//                         END
//***************************************************************
