/******************************************************************//**
    @file simplesm.c
    Simple state machine support, implementation.

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

**********************************************************************/

//      *** COMPILE SWITCHES ***

//***************************************************************

//      *** INCLUDE FILES ***

#include <stdint.h>
#include <stdbool.h>

#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"

#include "apex.h"
#include "apalloc.h"
#include "app.h"

#undef APEX_H               // This prevents the error msg
#include "simplesm.h"

    // Headers used during debug:

#include "io_debug.h"       //lint -esym(766, io_debug.h)
#include "lpc213x.h"        //lint -esym(766, lpc213x.h)
#include "io_map.h"         //lint -esym(766, io_map.h)
#include "apprint.h"        //lint -esym(766, apprint.h)
#include "io_main.h"        //lint -esym(766, io_main.h)

//***************************************************************
//                  *** MACROS ****


//***************************************************************
//              *** MODULE ID DEFINITIONS ****
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Stack Size: Measured use of 156 bytes on 3/13/12, but
    /// module was not exercised.

#define Z_STK_SSM       512     // Stack size in bytes

//***************************************************************
//            *** DATA STRUCTURE DEFINITIONS ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

typedef struct TAG_SSM_TCB
{
    LINK link;                  ///< A link for a TCB list.
    ssm_task_t *pTaskFunc;      ///< The SSM task function itself
    uint8_t tid;                ///< The ID number of the task.
} ssm_tcb_t;

//***************************************************************
//      *** Signals ***

//***************************************************************
//                      *** VARIABLES ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static TCB ssmTCB;
static uint32_t ssmStack[Z_STK_SSM/sizeof(uint32_t)];
static MBOX mailbox;

static LINK tcbs;


//***************************************************************
//                      *** CONSTANTS ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Debugging Table for Variable Display

const struct TAG_VARIABLE_TABLE ssmVarTab[] =
{
    {"stack", ssmStack, VAR_TYPE_NONE, sizeof(ssmStack)},
    {"tcb", &ssmTCB, VAR_TYPE_NONE, sizeof(ssmTCB)},

    {NULL, NULL, VAR_TYPE_NONE, 0}
};

//***************************************************************
//            *** STATIC FUNCTION PROTOTYPES ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void SsmTask(void);

//***************************************************************
//                      API Implementation
//***************************************************************

void SSM_Reset(void)
{
    Apex_InitMailbox(&mailbox);
    Apex_InitLink(&tcbs, NULL);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void SSM_Start(void)
{
    Apex_CreateTask(&ssmTCB, TID_SSM, SsmTask,
                    ssmStack, (sizeof(ssmStack)), TPRI_SSM);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**

*/

void SSM_CreateTask(ssm_task_t *pTask, uint8_t tid)
{
    ssm_tcb_t *pNewTask;

    pNewTask = malloc(sizeof(ssm_tcb_t));
    ASSERT(pNewTask != NULL);
    Apex_InitLink(&pNewTask->link, pNewTask);
    pNewTask->pTaskFunc = pTask;
    pNewTask->tid = tid;
    DI();
    Apex_Link(tcbs.pBLink, &pNewTask->link);
    EI();

}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void SSM_SendMsg(unsigned id)
{
    Apex_SendMsgExt(&mailbox, id, 0, 0);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void SSM_SendMsgExt(
            unsigned id, uintptr_t data1, uintptr_t data2)
{
    Apex_SendMsgExt(&mailbox, id, data1, data2);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void SSM_InitTimer(TMR *pTimer,
            unsigned event, uintptr_t data1, uintptr_t data2)
{
    Apex_InitMsgTimer(pTimer, &mailbox, event, data1, data2);
}

void SSM_StopTmr(TMR *pTimer) {Apex_StopTmr(pTimer);}

void SSM_StartTmr(TMR *pTimer, int duration)
                        {Apex_StartTmr(pTimer, duration);}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                      STATIC FUNCTIONS
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void SsmTask(void)
{
    unsigned event;
    uintptr_t data1, data2;
    LINK *pNext;
    ssm_tcb_t *pTCB;
    uint8_t taskID;

    FOREVER
    {
        // Wait for a message.
        Apex_PendMsg(&mailbox, &event, &data1, &data2, 0);

        // Decode the event to determine the destination
        // task.
        taskID = ((event >> 8) & 0xFF);

        // Now, go through the list of TCBs and find the
        // one addressed by the event.
        pNext = tcbs.pFLink;

        while((pTCB = pNext->pObject) != NULL)
        {
            if(pTCB->tid == taskID)
            {
                (*pTCB->pTaskFunc)(event, data1, data2);
                break; // while
            }
            pNext = pTCB->link.pFLink;
        }
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                         END
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *



