/******************************************************************//**

    @file ssm_idsm.c

    ID State Machine using Simple State Machine host.

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

**********************************************************************/
//***************************************************************

//      *** INCLUDE FILES ***

#include <stdint.h>
#include <stdbool.h>        //lint -esym(766, stdbool.h)

#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"

#include "simplesm.h"
#include "app.h"

#include "astream.h"
#include "eep.h"
#include "ssm_idsm.h"

#include "io_main.h"        //lint -esym(766, io_main.h)
#include "io_debug.h"       //lint -esym(766, io_debug.h)
#include "apprint.h"        //lint -esym(766, apprint.h)

//***************************************************************
//      *** MACROS ****

//lint -esym(750, PRINT_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, SET_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, STATE_CHAN) (Macro not always referenced)

#define PRINT_STATE_CHAR(x)  if(STATE_CHAN == pChan->chanID) \
                                                IO_DBPrintChar(x)

#if ((TP_IDSM_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
  || (TP_IDSM_STATE_FLAGS == STATE_FLAGS_LOWER_CH1))
#   define STATE_CHAN  TX_CHAN_1
#else
#   define STATE_CHAN  TX_CHAN_2
#endif

#if (TP_IDSM_STATE_FLAGS == STATE_FLAGS_OFF)
#   define SET_STATE(X,Y)   pChan->pState = (X)

#elif ((TP_IDSM_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
    || (TP_IDSM_STATE_FLAGS == STATE_FLAGS_UPPER_CH2))
#   define SET_STATE(X,Y)   pChan->pState = (X); PRINT_STATE_CHAR(Y)

#elif ((TP_IDSM_STATE_FLAGS == STATE_FLAGS_LOWER_CH1) \
    || (TP_IDSM_STATE_FLAGS == STATE_FLAGS_LOWER_CH2))
#   define SET_STATE(X,Y)   pChan->pState = (X); PRINT_STATE_CHAR((Y)|0x20)
#else
#error "TP_IDSM_STATE_FLAGS improperly defined."
#endif

//***************************************************************

    // State Machine Event IDs:

enum TAG_IDSM_EVIDS
{
    EV_IDSM_AUDIO_ON = IDSM_ID(0),  ///< Audio on
    EV_IDSM_AUDIO_OFF,          ///< Audio off.
    EV_IDSM_PENDTO,             ///< Time to do an ID.
    EV_IDSM_TXTIMEOUT,          ///< Transmitter timed out.
    EV_IDSM_TXDISABLE,          ///< Transmitter was disabled.
    EV_IDSM_DONEID,             ///< We've finished an ID
    EV_IDSM_CHANGEID,           ///< We changed the ID timer.
    EV_IDSM_MAX
};

//***************************************************************
//          *** DATA STRUCTURE DEFINITIONS ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//***************************************************************
//                   *** VARIABLES ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static struct TAG_IDSM_CHAN
{
    enum TAG_TX_CHAN chanID;        // Must be initialized at reset.

    // Task state:
    bool (*pState)(
        enum TAG_IDSM_EVIDS event, struct TAG_IDSM_CHAN *pChan);

    TMR idPendTimer;                // ID Pending Timer

} txChan[MAX_TXCHAN];

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//              Signals
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

DEFINE_SIGNAL(IDSM_ScheduleID)
DEFINE_SIGNAL(IDSM_ScheduleOpeningID)

//***************************************************************
//                 *** CONSTANTS ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Debugging Table for Variable Display

const struct TAG_VARIABLE_TABLE idsmVarTab[] =
{
    {NULL, NULL, VAR_TYPE_NONE, 0}
};

//***************************************************************
//           *** STATIC FUNCTION PROTOTYPES ****
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void IDTask(
            unsigned event, uintptr_t data1, uintptr_t unused);

static void RestartIDPendingTimer(struct TAG_IDSM_CHAN *pChan);
static void StartIDPendingTimer(struct TAG_IDSM_CHAN *pChan);

static void Transition_A(struct TAG_IDSM_CHAN *pChan);
static bool State_A(
    enum TAG_IDSM_EVIDS event, struct TAG_IDSM_CHAN *pChan);

static void Transition_B(struct TAG_IDSM_CHAN *pChan);
static bool  State_B(
    enum TAG_IDSM_EVIDS event, struct TAG_IDSM_CHAN *pChan);

static void Transition_C(struct TAG_IDSM_CHAN *pChan);
static bool  State_C(
    enum TAG_IDSM_EVIDS event, struct TAG_IDSM_CHAN *pChan);

static void Transition_D(struct TAG_IDSM_CHAN *pChan);
static bool  State_D(
    enum TAG_IDSM_EVIDS event, struct TAG_IDSM_CHAN *pChan);

static void Transition_E(struct TAG_IDSM_CHAN *pChan);
static bool  State_E(
    enum TAG_IDSM_EVIDS event, struct TAG_IDSM_CHAN *pChan);

static void Transition_F(struct TAG_IDSM_CHAN *pChan);
static bool  State_F(
    enum TAG_IDSM_EVIDS event, struct TAG_IDSM_CHAN *pChan);

static void Transition_G(struct TAG_IDSM_CHAN *pChan);
static bool  State_G(
    enum TAG_IDSM_EVIDS event, struct TAG_IDSM_CHAN *pChan);


//***************************************************************
//                        *** API ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void IDSM_Reset(void)
{
    txChan[TX_CHAN_1].chanID = TX_CHAN_1;
    txChan[TX_CHAN_2].chanID = TX_CHAN_2;

    SSM_InitTimer(&txChan[TX_CHAN_1].idPendTimer,
                                EV_IDSM_PENDTO, TX_CHAN_1, 0);
    SSM_InitTimer(&txChan[TX_CHAN_2].idPendTimer,
                                EV_IDSM_PENDTO, TX_CHAN_2, 0);
    IDSM_UnsubscribeAll();
}

void IDSM_Start(void)
{
    SSM_CreateTask(IDTask, MID_IDSM);
    Transition_A(&txChan[TX_CHAN_1]);
    Transition_A(&txChan[TX_CHAN_2]);
}


void IDSM_UnsubscribeAll(void)
{
    Unsubscribe_IDSM_ScheduleID();
    Unsubscribe_IDSM_ScheduleOpeningID();
}

    // Signal the ID State Machine that we are transmitting
    // audio.

void IDSM_NotifyAudioOn(enum TAG_TX_CHAN chan)
{
    SSM_SendMsgExt(EV_IDSM_AUDIO_ON, (uintptr_t)chan, 0);
}

    // Signal the ID State Machine that a transmitter has
    // turned off.

void IDSM_NotifyAudioOff(enum TAG_TX_CHAN chan)
{
    SSM_SendMsgExt(EV_IDSM_AUDIO_OFF, (uintptr_t)chan, 0);
}

    // Signal the ID State Machine that the audio task has
    // finished an ID.

void IDSM_NotifyDoneID(enum TAG_TX_CHAN chan)
{
    SSM_SendMsgExt(EV_IDSM_DONEID, (uintptr_t)chan, 0);
}

void IDSM_NotifyTxTimeout(enum TAG_TX_CHAN chan)
{
    SSM_SendMsgExt(EV_IDSM_TXTIMEOUT, (uintptr_t)chan, 0);
}

void IDSM_NotifyTxDisable(enum TAG_TX_CHAN chan)
{
    SSM_SendMsgExt(EV_IDSM_TXDISABLE, (uintptr_t)chan, 0);
}

void IDSM_NotifyChangedID(enum TAG_TX_CHAN chan)
{
    SSM_SendMsgExt(EV_IDSM_CHANGEID, (uintptr_t)chan, 0);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                 *** STATIC FUNCTIONS ***
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void RestartIDPendingTimer(struct TAG_IDSM_CHAN *pChan)
{
//    if(pChan->idPendTimer.state == TMS_COUNTING)
    SSM_StopTmr(&pChan->idPendTimer);  // Restart timer.
    SSM_StartTmr(
        &pChan->idPendTimer, EEP_GetIDTime(pChan->chanID));
}

static void StartIDPendingTimer(struct TAG_IDSM_CHAN *pChan)
{
    if(pChan->idPendTimer.state != TMS_COUNTING)
    {
        SSM_StartTmr(
            &pChan->idPendTimer, EEP_GetIDTime(pChan->chanID));
    }
}

static void CancelIDPendingTimer(struct TAG_IDSM_CHAN *pChan)
{
    SSM_StopTmr(&pChan->idPendTimer);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // ID Task:

static void IDTask(
            unsigned event, uintptr_t data1, uintptr_t unused)
{
    struct TAG_IDSM_CHAN *pChan;

    ASSERT((enum TAG_TX_CHAN)data1 < MAX_TXCHAN);
    pChan = &txChan[data1];

    // Send event to the state machine.  If it's not handled
    // there, check to see if the event should be processed
    // statelessly.

    if(!(*(pChan->pState))(
            (enum TAG_IDSM_EVIDS)event, pChan))
    {
        switch(event)
        {
            case EV_IDSM_TXDISABLE:
            case EV_IDSM_TXTIMEOUT:
                CancelIDPendingTimer(pChan);
                Transition_G(pChan);
                break;

            // If the operator changes the ID timer, restart it.
            case EV_IDSM_CHANGEID:
                if(EEP_GetIDTime(pChan->chanID) != 0)
                    RestartIDPendingTimer(pChan);
                else
                {
                    CancelIDPendingTimer(pChan);
                    Transition_A(pChan);
                }
                break;

            default:
                break;
        }
    }
}


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // State A: Idle

static void Transition_A(struct TAG_IDSM_CHAN *pChan)
{
    SET_STATE(State_A,'A');
}

static bool  State_A(
    enum TAG_IDSM_EVIDS event, struct TAG_IDSM_CHAN *pChan)
{
    switch(event)
    {
        // We've been keyed.  Wait for unkey.
        case EV_IDSM_AUDIO_ON:
            if(EEP_GetIDTime(pChan->chanID) != 0)
            {
                Transition_E(pChan);
            }
            return(true);

        default:
            return(false);
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Here, done opening ID, and still repeating

static void Transition_B(struct TAG_IDSM_CHAN *pChan)
{
    SET_STATE(State_B,'B');
}

static bool  State_B(
    enum TAG_IDSM_EVIDS event, struct TAG_IDSM_CHAN *pChan)
{
    switch(event)
    {
        // Send EV_AUD_IDPEND
        case EV_IDSM_PENDTO:
            EMIT_SIGNAL(IDSM_ScheduleID, pChan->chanID, 0);
            return(true);

        // Here, the audio task is telling us that an
        // ID has just completed.
        case EV_IDSM_DONEID:
            RestartIDPendingTimer(pChan);
            return(true);

        // Here, the transmitter turned off.
        case EV_IDSM_AUDIO_OFF:
            Transition_C(pChan);
            return(true);

        default:
            return(false);
    }
}
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Wait for user audio to reappear, or the final ID
    // to occur.

static void Transition_C(struct TAG_IDSM_CHAN *pChan)
{
    SET_STATE(State_C,'C');
}

static bool  State_C(
    enum TAG_IDSM_EVIDS event, struct TAG_IDSM_CHAN *pChan)
{
    switch(event)
    {
        // Here, the ID pending timer timed out.
        case EV_IDSM_PENDTO:
            EMIT_SIGNAL(IDSM_ScheduleID, pChan->chanID, 0);
            return(true);

        // Here, audio task is telling us that it just finished
        // an ID.
        case EV_IDSM_DONEID:
            RestartIDPendingTimer(pChan); // Restart the ID timer
            Transition_D(pChan);
            return(true);

        // Here, transmitter started up again.
        case EV_IDSM_AUDIO_ON:
            StartIDPendingTimer(pChan);
            Transition_B(pChan);
            return(true);

        default:
            return(false);
    }
}
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Wait for final ID timeout to occur, or for user audio to
    // come back on.

static void Transition_D(struct TAG_IDSM_CHAN *pChan)
{
    SET_STATE(State_D,'D');
}

static bool  State_D(
    enum TAG_IDSM_EVIDS event, struct TAG_IDSM_CHAN *pChan)
{
    switch(event)
    {
        // Here, the ID pending timer timed out.
        case EV_IDSM_PENDTO:
            Transition_A(pChan);
            return(true);

        // Here, the transmitter just came back on.
        case EV_IDSM_AUDIO_ON:
            Transition_B(pChan);
            return(true);

        default:
            return(false);
    }
}
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Transmitter just turned on.
    // Schedule an ID and wait for the user to unkey.

static void Transition_E(struct TAG_IDSM_CHAN *pChan)
{
    StartIDPendingTimer(pChan);
    EMIT_SIGNAL(IDSM_ScheduleOpeningID, pChan->chanID, 0);
    SET_STATE(State_E,'E');
}

static bool  State_E(
    enum TAG_IDSM_EVIDS event, struct TAG_IDSM_CHAN *pChan)
{
    switch(event)
    {
        case EV_IDSM_PENDTO:
            EMIT_SIGNAL(IDSM_ScheduleID, pChan->chanID, 0);
            return(true);

        // Here, we have done the opening ID.
        case EV_IDSM_DONEID:
            RestartIDPendingTimer(pChan);
            Transition_B(pChan);
            return(true);

        // Here, the audio turned off.
        case EV_IDSM_AUDIO_OFF:
            Transition_F(pChan);
            return(true);

        default:
            return(false);
    }
}
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Kerchunk State

static void Transition_F(struct TAG_IDSM_CHAN *pChan)
{
    SET_STATE(State_F,'F');
}

static bool  State_F(
    enum TAG_IDSM_EVIDS event, struct TAG_IDSM_CHAN *pChan)
{
    switch(event)
    {
        // Here, the ID pending timer timed out.

        case EV_IDSM_PENDTO:
            Transition_A(pChan);
            return(true);

        // Here, the audio task is telling us that it just
        // finished an ID.
        case EV_IDSM_DONEID:
            RestartIDPendingTimer(pChan);
            return(true);

        // Here, the audio came back on.
        case EV_IDSM_AUDIO_ON:
            Transition_B(pChan);
            return(true);

        default:
            return(false);
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Timeout State
    // Wait here until audio has finished.  Go with "Done ID"
    // because that is the last thing that happens.

static void Transition_G(struct TAG_IDSM_CHAN *pChan)
{
    SET_STATE(State_G,'G');
}

static bool  State_G(
    enum TAG_IDSM_EVIDS event, struct TAG_IDSM_CHAN *pChan)
{
    switch(event)
    {
        // Here, the audio task is telling us that it just
        // finished an ID.
        case EV_IDSM_DONEID:
            Transition_A(pChan);
            return(true);

        default:
            return(false);
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                         END
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
