/******************************************************************//**

    @file ssm_morse.c

    Morse Code Encoder for PSE-702

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

**********************************************************************/

//      *** INCLUDE FILES ***

#include <stdint.h>
#include <stdbool.h>

#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"

#include "simplesm.h"

#include "app.h"
#include "astream.h"
#include "eep.h"
#include "ssm_morse.h"
#include "fgen.h"       // Waveform Generation

#include "io_debug.h"   //lint -esym(766, io_debug.h)
#include "io_main.h"    //lint -esym(766, io_main.h)
#include "apprint.h"    //lint -esym(766, apprint.h)

/***************************************************************/
//      *** MACROS ****

//lint -esym(750, PRINT_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, SET_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, STATE_CHAN) (Macro not always referenced)

#define PRINT_STATE_CHAR(x)  if(STATE_CHAN == pChan->chanID) \
        IO_DBPrintChar(x)

#if ((TP_MORSE_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
  || (TP_MORSE_STATE_FLAGS == STATE_FLAGS_LOWER_CH1))
#   define STATE_CHAN  TX_CHAN_1
#else
#   define STATE_CHAN  TX_CHAN_2
#endif

#if (TP_MORSE_STATE_FLAGS == STATE_FLAGS_OFF)
#   define SET_STATE(X,Y)   pChan->pState = (X);

#elif ((TP_MORSE_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
    || (TP_MORSE_STATE_FLAGS == STATE_FLAGS_UPPER_CH2))
#   define SET_STATE(X,Y)   pChan->pState = (X); PRINT_STATE_CHAR(Y)

#elif ((TP_MORSE_STATE_FLAGS == STATE_FLAGS_LOWER_CH1) \
    || (TP_MORSE_STATE_FLAGS == STATE_FLAGS_LOWER_CH2))
#   define SET_STATE(X,Y)   pChan->pState = (X); PRINT_STATE_CHAR((Y)|0x20)
#else
#error "TP_MORSE_STATE_FLAGS improperly defined."
#endif

/***************************************************************/

    // Morse Event IDs

enum TAG_MORSE_EVIDS
{
    EV_MORSE_ENCODE = MORSE_ID(0),  ///< Encode a string
//    EV_MORSE_WHISPER,             ///< Reduce volume
//    EV_MORSE_SHOUT,               ///< Normal volume
    EV_MORSE_DIT_TMR,               ///< Dit timer
    EV_MORSE_DLY_TMR,               ///< Delay timer
    EV_MORSE_STOP,                  ///< Stop all morse output
    EV_MORSE_MAX                    ///< End of event list
};

    // Correspondance between Message IDs and State Machine
    // Events (Gives correspondance to state diagram):

//#define MSG_MORSE_ENCODE        ((int16_t)EV_MORSE_ENCODE)
//#define MSG_MORSE_WHISPER       ((int16_t)EV_MORSE_WHISPER)
//#define MSG_MORSE_SHOUT         ((int16_t)EV_MORSE_SHOUT)

    // Morse event flags:

#if 0
#define EV_DIT_TMR1     0x0001      // DIT timer chan1 timed out.
#define EV_DIT_TMR2     0x0002      // DIT timer chan2 timed out.
#define EV_DLY_TMR1     0x0004      // Delay timer chan1 timed out.
#define EV_DLY_TMR2     0x0008      // Delay timer chan2 timed out.

#define EV_ENCODE_MAIL1 0x0010      // Mail waiting in encode box chan1.
#define EV_ENCODE_MAIL2 0x0020      // Mail waiting in encode box chan2.
#define EV_CMD_MAIL     0x0040      // Mail waiting in command box.
#endif

    //  Morse Cose Definitions

    // The prescaler stored in EEPROM is the number that
    // changes when we change the wpm rate.  The other
    // definitions determine the ratio in time between dits,
    // dahs, and spaces.

#define MCDIT_T         1   // Duration of DIT
#define MCDAH_T         3   // Duration of DAH
#define MCBIT_SPACE_T   1   // Space after DIT or DAH
#define MCCHAR_SPACE_T  3   // Space after char
#define MCWORD_SPACE_T  7   // Space after word

    // Buffer Sizes:

#define Z_MORSESTRING   20      // Max length of morse message

    // Times in mS.

#define TIME_PREMORSE   (500*TU_1MS)    // Pre-Morse timer
#define TIME_POSTMORSE  (500*TU_1MS)    // Post-Morse timer

/***************************************************************/

//      *** VARIABLES ***

    // Morse task TCB and Stack Definitions:

static struct TAG_MORSE_CHAN
{
    enum TAG_TX_CHAN chanID;    ///< Must be initialized at reset.
    uint8_t fgenChan;           ///< Function generator channel.

    /// Task state:
    void (*pState)(
        enum TAG_MORSE_EVIDS event, struct TAG_MORSE_CHAN *pChan);

    TMR ditTimer;               ///< Times durations of dits.
    TMR dlyTimer;               ///< Times durations of delays.
    uint8_t morseBuf[Z_MORSESTRING]; ///< Buf for outgoing msgs
    uint8_t *pBuf;              ///< Points to current pos in buffer
    char *pString;              ///< Points to ASCII string to encode.
    uint8_t morseByte;          ///< Byte we are currently sending
    uint8_t bitCount;           ///< Count of dits/dahs in curr char.
    uint8_t level;              ///< Volume of morse code tone.

} txChan[MAX_TXCHAN];

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                      Signals
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

DEFINE_SIGNAL(MORSE_Complete)


//***************************************************************
//                  *** CONSTANTS ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Debugging Table for Variable Display

const struct TAG_VARIABLE_TABLE morseVarTab[] =
{
    {NULL, NULL, VAR_TYPE_NONE, 0}
};


/***************************************************************/

//      *** STATIC FUNCTION PROTOTYPES ****

static void Morse_Task(
                unsigned event, uintptr_t data1, uintptr_t data2);

static void MorseXlate(struct TAG_MORSE_CHAN *pChan);
static void SendNextMorseBit(struct TAG_MORSE_CHAN *pChan);
static void SendMorseBit(
                    struct TAG_MORSE_CHAN *pChan, uint8_t time);
static void StartMDitTmr(
                struct TAG_MORSE_CHAN *pChan, uint8_t timeCount);
static void StartMDlyTmr(
                struct TAG_MORSE_CHAN *pChan, uint16_t timeCount);

static void Transition_Idle(struct TAG_MORSE_CHAN *pChan);  // A
static void State_Idle(
    enum TAG_MORSE_EVIDS event, struct TAG_MORSE_CHAN *pChan);

static void Transition_BitDelay(
                            struct TAG_MORSE_CHAN *pChan);  // B
static void State_BitDelay(
    enum TAG_MORSE_EVIDS event, struct TAG_MORSE_CHAN *pChan);

static void Transition_SendBit(
                            struct TAG_MORSE_CHAN *pChan);   // C
static void State_SendBit(
    enum TAG_MORSE_EVIDS event, struct TAG_MORSE_CHAN *pChan);

static void Transition_DelayChar(
                            struct TAG_MORSE_CHAN *pChan); // D
static void State_DelayChar(
    enum TAG_MORSE_EVIDS event, struct TAG_MORSE_CHAN *pChan);

static void Transition_EndString(
                            struct TAG_MORSE_CHAN *pChan); // E
static void State_EndString(
    enum TAG_MORSE_EVIDS event, struct TAG_MORSE_CHAN *pChan);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//  Morse Code Table

//  Morse code is encoded as follows:
//  Bits 0-2: Length of Character, 1 to 5
//  Bits 3-7: Dits and Dahs, 1=dah, 0=dit

//  Exceptions:
//      Word space: 0x80 (since 0x00 is not allowed).
//      6 dit/dah codes:
//          Bits 3-7 contain the last 5 bits of the
//              character.
//          Bits 0-2 = 6 for characters with leading
//              dit, 7 for characters with
//              leading dah.

static const uint8_t morseTbl[] =
{
    'A', 0x12,          // 0001 0010 // 00 A: .-
    'B', 0x0C,          // 0000 1100 // 01 B: -...
    'C', 0x2C,          // 0010 1100 // 02 C: -.-.
    'D', 0x0B,          // 0000 1011 // 03 D: -..
    'E', 0x01,          // 0000 0001 // 04 E: .
    'F', 0x24,          // 0010 0100 // 05 F: ..-.
    'G', 0x1B,          // 0001 1011 // 06 G: .--
    'H', 0x04,          // 0000 0100 // 07 H: ....
    'I', 0x02,          // 0000 0010 // 08 I: ..
    'J', 0x74,          // 0111 0100 // 09 J: .---
    'K', 0x2B,          // 0010 1011 // 10 K: -.-
    'L', 0x14,          // 0001 0100 // 11 L: .-..
    'M', 0x1A,          // 0001 1010 // 12 M: --
    'N', 0x0A,          // 0000 1010 // 13 N: -.
    'O', 0x3B,          // 0011 1011 // 14 O: ---
    'P', 0x34,          // 0011 0100 // 15 P: .--.
    'Q', 0x5C,          // 0101 1100 // 16 Q: --.-
    'R', 0x13,          // 0001 0011 // 17 R: .-.
    'S', 0x03,          // 0000 0011 // 18 S: ...
    'T', 0x09,          // 0000 1001 // 19 T: -
    'U', 0x23,          // 0010 0011 // 20 U: ..-
    'V', 0x44,          // 0100 0100 // 21 V: ...-
    'W', 0x33,          // 0011 0011 // 22 W: .--
    'X', 0x4C,          // 0100 1100 // 23 X: -..-
    'Y', 0x6C,          // 0110 1100 // 24 Y: -.--
    'Z', 0x1C,          // 0001 1100 // 25 Z: --..

    '0', 0xFD,          // 1111 1101 // 26 0: -----
    '1', 0xF5,          // 1111 0101 // 27 1: .----
    '2', 0xE5,          // 1110 0101 // 28 2: ..---
    '3', 0xC5,          // 1100 0101 // 29 3: ...--
    '4', 0x85,          // 1000 0101 // 30 4: ....-
    '5', 0x05,          // 0000 0101 // 31 5: .....
    '6', 0x0D,          // 0000 1101 // 32 6: -....
    '7', 0x1D,          // 0001 1101 // 33 7: --...
    '8', 0x3D,          // 0011 1101 // 34 8: ---..
    '9', 0x7D,          // 0111 1101 // 35 9: ----.

    '/', 0x4D,          // 0100 1101,   // 36 /: -..-.
    '-', 0x8D,          // 1000 1101,   // 37 BT: -...-

    // 6 dit/dah characters:

    '.', 0xAE,          // 1010 1110 // 38 .: .-.-.-
    '?', 0x36,          // 0011 0110 // 39 ?: ..--..

    ' ', 0x80,          // 1000 0000 // 40 Space

    0x00, 0x00          // marks end of table
};


//  This table encodes several rates for morse code output.

static const uint8_t morseRateTbl[] =
{
    120,    //  0: 10 wpm
    109,    //  1: 11 wpm
    100,    //  2: 12 wpm
    92,     //  3: 13 wpm
    86,     //  4: 14 wpm
    80,     //  5: 15 wpm
    75,     //  6: 16 wpm
    71,     //  7: 17 wpm
    67,     //  8: 18 wpm
    63,     //  9: 19 wpm
    60,     // 10: 20 wpm
    57,     // 11: 21 wpm
    55,     // 12: 22 wpm
    52,     // 13: 23 wpm
    50,     // 14: 24 wpm
    48      // 15: 25 wpm

    // MAX_MORSE_CODE_RATE and
    // MIN_MORSE_CODE_RATE are set in "h" file.
};


//***************************************************************
//              *** API: GLOBAL FUNCTIONS ***
//***************************************************************

    // Reset initialization of the Morse Code State Machine.
    // Called by the main.c module upon reset.

void Morse_Reset(void)
{
    txChan[TX_CHAN_1].chanID = TX_CHAN_1;
    txChan[TX_CHAN_2].chanID = TX_CHAN_2;

    txChan[TX_CHAN_1].fgenChan = FGEN_TX1_TONE_CHAN;
    txChan[TX_CHAN_2].fgenChan = FGEN_TX2_TONE_CHAN;

    // Init Dit Timers:

    SSM_InitTimer(&txChan[TX_CHAN_1].ditTimer,
                    EV_MORSE_DIT_TMR, (uintptr_t)TX_CHAN_1, 0);
    SSM_InitTimer(&txChan[TX_CHAN_2].ditTimer,
                    EV_MORSE_DIT_TMR, (uintptr_t)TX_CHAN_2, 0);

    // Init Delay Timers:

    SSM_InitTimer(&txChan[TX_CHAN_1].dlyTimer,
                    EV_MORSE_DLY_TMR, (uintptr_t)TX_CHAN_1, 0);
    SSM_InitTimer(&txChan[TX_CHAN_2].dlyTimer,
                    EV_MORSE_DLY_TMR, (uintptr_t)TX_CHAN_2, 0);

    MORSE_UnsubscribeAll();
}

void Morse_Start(void)
{
    SSM_CreateTask(Morse_Task, MID_MORSE);

    Transition_Idle(&txChan[TX_CHAN_1]);
    Transition_Idle(&txChan[TX_CHAN_2]);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void MORSE_UnsubscribeAll(void)
{
    Unsubscribe_MORSE_Complete();
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Send a String via Morse.

        Allocate a message and send it to the MORSE task.  The
        task will encode it and send it via Morse code tones.
    */
void Morse_Encode(
        enum TAG_TX_CHAN chan, uint8_t level, char *pString)
{
    ASSERT(VALID_TX_CHAN(chan));
    txChan[chan].pString = pString;
    txChan[chan].level = level;
    SSM_SendMsgExt(EV_MORSE_ENCODE, chan, 0);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Stop all morse output.
     *
     */
void Morse_Stop(enum TAG_TX_CHAN chan)
{
    SSM_StopTmr(&txChan[chan].ditTimer);
    SSM_StopTmr(&txChan[chan].dlyTimer);
    FGen_ChannelOff(txChan[chan].fgenChan);
    EMIT_SIGNAL(MORSE_Complete, chan, 0);
    Transition_Idle(&txChan[chan]);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Validate a character as being part of the Morse set.
    */

bool Morse_IsMorse(char c)
{
    unsigned i = 0;

    // Make character upper case:
    c = (c >= 'a' && c <= 'z')? c & ~0x20 : c;

    while(morseTbl[i] != 0x00)
    {
        if(morseTbl[i] == c)
            return(true);
        else
            i += 2;
    }
    return(false);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

#if 0
    // Tell morse encoder to reduce volume during a talk-over.
    // Called from a task only.

void Morse_Whisper(enum TAG_TX_CHAN chan)
{
    ASSERT(VALID_TX_CHAN(chan));
    SSM_SendMsgExt(EV_MORSE_WHISPER, (uintptr_t)chan, 0);
}

    // Tell morse encoder to use normal volume in the absence
    // of talk-over.  Called from a task only.

void Morse_Shout(enum TAG_TX_CHAN chan)
{
    ASSERT(VALID_TX_CHAN(chan));
    SSM_SendMsgExt(EV_MORSE_SHOUT, (uintptr_t)chan, 0);
}
#endif

//***************************************************************
//              *** STATIC FUNCTIONS ***
//***************************************************************

// The following functions start and stop the two timers used
// during morse code output.  These functions are called from
// within the morse code state machine.

static void StartMDitTmr(
        struct TAG_MORSE_CHAN *pChan, uint8_t timeCount)
{
    uint8_t eepRate;
    uint8_t rate;

    // Get rate from EEPROM and convert it into a table index.

    eepRate = EEP_GetMorseRate(pChan->chanID);
    rate = morseRateTbl[eepRate-MIN_MORSE_CODE_RATE];
    SSM_StartTmr(&pChan->ditTimer, (TU_1MS * rate * timeCount));
}


static void StartMDlyTmr(
        struct TAG_MORSE_CHAN *pChan, uint16_t timeCount)
{
    SSM_StartTmr(&pChan->dlyTimer, (TU_1MS * timeCount));
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//  Translate Morse Characters in Buffer

//  This function scans the morse buffer and translates
//  all characters to their morse code bit encodings.

static void MorseXlate(struct TAG_MORSE_CHAN *pChan)
{
    char c, c2;
    uint8_t *pTable;
    uint8_t *pEncodedString = pChan->morseBuf;

    while(*pChan->pString != '\0')
    {
        pTable = (uint8_t *)morseTbl;  // Point at start of table

        // Make all characters upper case:
        c2 = (char)*pChan->pString;
        c2 = (c2 >= 'a' && c2 <= 'z')? c2 & ~0x20 : c2;

        FOREVER
        {
            c = (char)*pTable++;

            if(c == c2 || c == ' ')
            {
                *pEncodedString++ = *pTable++;
                break;  // FOREVER
            }
            else
                pTable++;
        }

        pChan->pString++;
    }

    *pEncodedString = '\0';                 // Terminate the string.
    pChan->pBuf = pChan->morseBuf;     // Init channel's buf ptr
}

//***************************************************************
//                          *** TASK ***
//***************************************************************
static void Morse_Task(
                unsigned event, uintptr_t data1, uintptr_t unused)
{
    enum TAG_TX_CHAN chan;

    chan = (enum TAG_TX_CHAN)(data1);
    ASSERT(VALID_TX_CHAN(chan));

    // Here, decode the parameters depending on the
    // event.

    (*txChan[chan].pState)(
        (enum TAG_MORSE_EVIDS)event, &txChan[chan]);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//  Morse State A: Idle.

//  Get a string to send via morse.  Encode it.

static void Transition_Idle(struct TAG_MORSE_CHAN *pChan)
{
    SET_STATE(State_Idle,'A');
}


static void State_Idle(
    enum TAG_MORSE_EVIDS event, struct TAG_MORSE_CHAN *pChan)
{
    switch(event)
    {
        case EV_MORSE_ENCODE:
            MorseXlate(pChan);
            StartMDlyTmr(pChan, TIME_PREMORSE);
            Transition_DelayChar(pChan); // Start sending the string.
            break;

        default:
            break;
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//  Morse State B: Delay one inter-bit time.

//  Here, delay for one inter-bit time (same as dit).

static void Transition_BitDelay(struct TAG_MORSE_CHAN *pChan)
{
    SET_STATE(State_BitDelay,'B');
}

static void State_BitDelay(
    enum TAG_MORSE_EVIDS event, struct TAG_MORSE_CHAN *pChan)
{
    switch(event)
    {
        case EV_MORSE_DIT_TMR:
            FGen_ChannelOff(pChan->fgenChan);

            if(pChan->bitCount-- != 0)  // Are there more bits to go?
            {
                StartMDitTmr(pChan, MCBIT_SPACE_T);
                Transition_SendBit(pChan);
            }

            // Here, that character is complete so delay one inter-
            // character time, then do next character.

            else
            {
                StartMDitTmr(pChan, MCCHAR_SPACE_T);
                Transition_DelayChar(pChan);
            }
            break;

        default:
            break;
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//  Morse State C: Send a Bit

//  Here, send next dit or dah.  Goes back to state B
//  when done.


static void Transition_SendBit(struct TAG_MORSE_CHAN *pChan)
{
    // Select next state = C
    SET_STATE(State_SendBit,'C');

}

static void State_SendBit(
    enum TAG_MORSE_EVIDS event, struct TAG_MORSE_CHAN *pChan)
{
    switch(event)
    {
        case EV_MORSE_DIT_TMR:
            SendNextMorseBit(pChan);
            Transition_BitDelay(pChan);
            break;

        default:
            break;
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // State D: Delay one character time.

static void Transition_DelayChar(struct TAG_MORSE_CHAN *pChan)
{
    // Select next state = D
    SET_STATE(State_DelayChar,'D');
}

static void State_DelayChar(
    enum TAG_MORSE_EVIDS event, struct TAG_MORSE_CHAN *pChan)
{
    switch(event)
    {
        case EV_MORSE_DIT_TMR:
        case EV_MORSE_DLY_TMR:

            if((pChan->morseByte = *pChan->pBuf++) != '\0')
            {
                // Is it an inter-word space?
                if(pChan->morseByte == 0x80)
                    StartMDitTmr(pChan, MCWORD_SPACE_T);

                // Here, not inter-word space

                else
                {
                    pChan->bitCount = pChan->morseByte & 0x07;
                    pChan->morseByte = (pChan->morseByte >> 3) & 0x1F;

                    if(pChan->bitCount == 0x06) // 6 bit char w/ dit?
                    {
                        // Send initial dit
                        SendMorseBit(pChan, MCDIT_T);
                        pChan->bitCount = 5;        // 5 bits to go.
                    }

                    // 6 bit char w/ dah?
                    else if(pChan->bitCount == 0x07)
                    {
                        // Send initial dah
                        SendMorseBit(pChan, MCDAH_T);
                        pChan->bitCount = 5;        // 5 bits to go.
                    }

                    else
                    {
                        pChan->bitCount--;
                        SendNextMorseBit(pChan);
                    }

                    Transition_BitDelay(pChan);
                }
            }

            // Here we are all done sending.

            else
            {
                // Stop the dit timer.
                SSM_StopTmr(&pChan->ditTimer);
                StartMDlyTmr(pChan, TIME_POSTMORSE);
                Transition_EndString(pChan);
            }

            break;

        default:
            break;
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//  Morse State E: End of String Delay

//  Post Morse Delay.  This state delays some amount of time
//  after sending a morse string before signaling end of
//  morse output.


static void Transition_EndString(struct TAG_MORSE_CHAN *pChan)
{
    // Select next state = E
    SET_STATE(State_EndString,'E');
}


static void State_EndString(
    enum TAG_MORSE_EVIDS event, struct TAG_MORSE_CHAN *pChan)
{
    // Wait for a timer timeout

    switch(event)
    {
        case EV_MORSE_DLY_TMR:
            EMIT_SIGNAL(MORSE_Complete, pChan->chanID, 0);
            Transition_Idle(pChan);
            break;

        default:
            break;
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//  Here, send either a dit or a dah.  This function sets
//  the morse_tmr so the state machine will execute again
//  when the timer times out.

static void SendNextMorseBit(struct TAG_MORSE_CHAN *pChan)
{
    if(pChan->morseByte & 0x01)
    {
        SendMorseBit(pChan, MCDAH_T);       // Send a Dah
#ifdef MORSE_DEBUG
        IO_DBPrintChar('_');
#endif
    }

    else
    {
        SendMorseBit(pChan, MCDIT_T);       // Send a Dit
#ifdef MORSE_DEBUG
        IO_DBPrintChar('.');
#endif
    }

    pChan->morseByte = pChan->morseByte >> 1;
}

static void SendMorseBit(
                        struct TAG_MORSE_CHAN *pChan, uint8_t time)
{
    StartMDitTmr(pChan, time);
    FGen_SetChannelFreq(pChan->fgenChan,
                (uint16_t)(EEP_GetMorseFreq(pChan->chanID)*10));
    FGen_SetChannelGain(pChan->fgenChan, pChan->level);
    FGen_ChannelOn(pChan->fgenChan);
}

/****************************************************************
                    End of Morse Code Module
****************************************************************/
