/*********************************************************************/
/**
    @file ssm_satd.c

    Subaudible Tone Decoder for PSE-702 using simple state machine
    host.

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

*/
//*********************************************************************
//      *** INCLUDE FILES ***

#include <stdint.h>
#include <stdbool.h>        //lint -esym(766, stdbool.h)
#include <limits.h>

#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"

#include "simplesm.h"
#include "app.h"
#include "aplog2.h"
#include "astream.h"
#include "eep.h"
#include "ssm_satd.h"

    // Headers used during debug:

#include "apprint.h"        //lint -esym(766, apprint.h)
#include "io_debug.h"       //lint -esym(766, io_debug.h)
#include "io_main.h"        //lint -esym(766, io_main.h)

#ifdef UNIT_TEST
#include "utest.h"
#endif

//***************************************************************
//                     *** MACROS ****
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//lint -esym(750, PRINT_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, SET_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, STATE_CHAN) (Macro not always referenced)

#define PRINT_STATE_CHAR(x)  if(STATE_CHAN == pChan->chanID) \
        IO_DBPrintChar(x)

#if ((TP_SATD_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
  || (TP_SATD_STATE_FLAGS == STATE_FLAGS_LOWER_CH1))
#   define STATE_CHAN  RX_CHAN_1
#else
#   define STATE_CHAN  RX_CHAN_2
#endif

#if (TP_SATD_STATE_FLAGS == STATE_FLAGS_OFF)
#   define SET_STATE(X,Y)   pChan->pState = (X)

#elif ((TP_SATD_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
    || (TP_SATD_STATE_FLAGS == STATE_FLAGS_UPPER_CH2))
#   define SET_STATE(X,Y)   pChan->pState = (X); PRINT_STATE_CHAR(Y)

#elif ((TP_SATD_STATE_FLAGS == STATE_FLAGS_LOWER_CH1) \
    || (TP_SATD_STATE_FLAGS == STATE_FLAGS_LOWER_CH2))
#   define SET_STATE(X,Y)   pChan->pState = (X); PRINT_STATE_CHAR((Y)|0x20)
#else
#error "TP_SATD_STATE_FLAGS improperly defined."
#endif


//***************************************************************
//              *** MODULE ID DEFINITIONS ****
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // State Machine Event IDs:

enum TAG_SATD_EVIDS
{
    EV_SATD_ON = SATD_ID(0),    // Turn on detectors
    EV_SATD_OFF,                // Turn off detectors.
    EV_SATD_ZXING,              // A zero crossing was detected.
    EV_SATD_NOZXING,            // An expected zxing didn't occur
    EV_SATD_LOCK_TO             // Lock timeout
};

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // Timeout Values:

#define SATD_TIME_LOCK      (1000*TU_1MS)   ///< Lock debounce timer

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // Tone Qualification Constants

#define MIN_ZXING_FACTOR    230 ///< Min samples to qualify zxing.
                                ///< Represents approx. 90% ideal.

#define MAX_ZXING_FACTOR    282 ///< Max samples to qualify zxing
                                ///< Represents approx. 110% ideal.

    /* Capture and hold factors.
        The following constants determine the zero crossing period
        error tolerance that is allowed for capturing lock and for
        holding lock.

        The factors are multiplied by the ideal number of samples
        from the filter parameter table to obtain the absolute
        tolerance for zero crossing period error.
    */

#define ERR_CAPTURE_FACTOR  3       ///< Freq error capture factor

#define ERR_HOLD_FACTOR     4       ///< Freq error hold factor

#define RMS_CAPTURE DBP_2_LOG2(37.0) ///< Min RMS^2 to allow decoding.
                                    ///< Represents 300 Hz deviation.

#define RMS_RBURST  DBP_2_LOG2(10.0) ///< Min RMS diff to detect rburst

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

#define E2_BP_FILTER    14      ///< Bandpass filter upscale value

#define BP_CONST_A2 16356       ///< This constant doesn't change w/ freq.

#define BP_CONST_B0     14      ///< This constant doesn't change w/ freq.

    // The following table contains the constants for the
    // CTCSS decder bandpass filters, the ideal number of
    // zero crossing counts per cycle for each frequency, and
    // first order time constants for the smoothing filters used
    // with zero crossings and RMS readings.

    // (Note that the zero crossing counts are equal to the
    // ideal number of ASTREAM samples per cycle of the frequency,
    // multiplied by 256).

static const struct TAG_SATD_BP_TBL
{
    uint16_t bp_a1;
    uint16_t zxingTime;
    uint16_t zxing_a1;
    uint16_t rms_a1;
    uint16_t freq;

} satdFilterParmTbl[SATD_FILTER_TABLE_SIZE] =

{
    // Designed for sample rate of 11,718.75

    {32719, 44776, 101, 200, 670},    // 67.0
    {32716, 41725, 107, 203, 719},    // 71.9
    {32714, 40323, 111, 205, 744},    // 74.4
    {32712, 38961, 114, 206, 770},    // 77.0
    {32710, 37641, 117, 208, 797},    // 79.7
    {32708, 36364, 120, 209, 825},    // 82.5
    {32706, 35129, 123, 211, 854},    // 85.4
    {32703, 33898, 126, 212, 885},    // 88.5
    {32701, 32787, 129, 213, 915},    // 91.5
    {32698, 31646, 132, 215, 948},    // 94.8
    {32693, 30000, 137, 217, 1000},   // 100.0
    {32690, 28986, 140, 218, 1035},   // 103.5
    {32686, 27985, 143, 219, 1072},   // 107.2
    {32682, 27051, 146, 220, 1109},   // 110.9
    {32678, 26132, 149, 221, 1148},   // 114.8
    {32674, 25253, 151, 222, 1188},   // 118.8
    {32669, 24390, 154, 224, 1230},   // 123.0
    {32664, 23566, 157, 225, 1273},   // 127.3
    {32658, 22762, 159, 226, 1318},   // 131.8
    {32652, 21978, 162, 227, 1365},   // 136.5
    {32646, 21231, 164, 228, 1413},   // 141.3
    {32640, 20520, 167, 228, 1462},   // 146.2
    {32632, 19815, 169, 229, 1514},   // 151.4
    {32625, 19145, 172, 230, 1567},   // 156.7
    {32616, 18496, 174, 231, 1622},   // 162.2
    {32607, 17868, 176, 232, 1679},   // 167.9
    {32598, 17261, 179, 233, 1738},   // 173.8
    {32588, 16676, 181, 233, 1799},   // 179.9
    {32577, 16112, 183, 234, 1862},   // 186.2
    {32565, 15560, 185, 235, 1928},   // 192.8
    {32545, 14742, 188, 236, 2035},   // 203.5
    {32531, 14238, 190, 237, 2107},   // 210.7
    {32516, 13755, 192, 237, 2187},   // 218.1
    {32501, 13292, 194, 238, 2257},   // 225.7
    {32484, 12842, 196, 238, 2336},   // 233.6
    {32465, 12407, 198, 239, 2418},   // 241.8
    {32446, 11986, 199, 240, 2503}    // 250.3
};


//***************************************************************
//                    *** VARIABLES ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // A structure to represent each receiver channel for the ISR.

static struct TAG_SATD_ISR_RXCHAN
{
    enum TAG_RX_CHAN chanID;        // Must be initialized at reset.
    bool enable;            // True if channel is enabled.
    bool locked;            // True if channel is locked on a tone.

    // Digital Filter Parameters:

    uint16_t a1;                    // Filter constant
    int64_t z_1, z_2;     // Delay elements

    // Tone Qualification Parameters:

    int16_t y_sub_n_minus_1; // Used to detect zero xings.
    uint16_t sampleCount;       // Counts samples betw. zero xings.
    uint16_t minSampleCount;    // Used to qualify a valid zxing.
    uint16_t maxSampleCount;    // Used to qualify a valid zxing.
    uint32_t rmsSquared;        // Signal magnitude, squared.

} isrRxChans[MAX_RXCHAN];


    // A structure to represent each receiver channel for the task.

static struct TAG_SATD_TASK_RXCHAN
{

    enum TAG_RX_CHAN chanID;        // Must be initialized at reset.

    TMR lockTimer;          // Lock timeout timer.

    // Receiver Channel's Current State:
    void (*pState)(enum TAG_SATD_EVIDS event,
                    enum TAG_RX_CHAN rxChan, uintptr_t data);

    uint32_t zxing_z_1;       // Zero crossing LP filter delay element.
    int64_t rms_z_1;          // Log2 RMS LP filter delay element.

    uint16_t index;
    uint16_t r_sub_n_minus_1;   // Residual from zxing interpolation
    uint16_t zxingTimeError;

    uint16_t errCaptureTolerance;
    uint16_t errHoldTolerance;

    int32_t logRMSsquared;    // Log2 of rmsSquared.
    int32_t lpLog2RMSsquared; // Log2 of rmsSquared, low-pass filtered.

} taskRxChans[MAX_RXCHAN];


    // Message sent from interrupt to task after a
    // zero crossing.  Data in this structure is copied
    // from isr data, and is used by the task
    // to process zero crossings.

static struct TAG_SATD_XMSG
{
    int16_t y_sub_n, y_sub_n_minus_1;
    uint16_t sampleCount;
    uint32_t rmsSquared;

}xMsgs[MAX_RXCHAN];

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                           SIGNALS
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

DEFINE_SIGNAL(SATD_Locked)
DEFINE_SIGNAL(SATD_Unlocked)
DEFINE_SIGNAL(SATD_SigLevel)
DEFINE_SIGNAL(SATD_ZXing)

    // Debugging Table for Variable Display

const struct TAG_VARIABLE_TABLE satdVarTab[] =
{
//    {"stack", satdStack, VAR_TYPE_NONE, sizeof(satdStack)},
//    {"tcb", &satdTCB, VAR_TYPE_NONE, sizeof(satdTCB)},

    {NULL, NULL, VAR_TYPE_NONE, 0}
};

//***************************************************************
//                  *** STATIC FUNCTION PROTOTYPES ***
//***************************************************************

static void SATD_Task(
            unsigned event, uintptr_t data1, uintptr_t data2);
static void DecoderSampleSink_ISR(
    struct TAG_SATD_ISR_RXCHAN *pRxChan, int16_t x_sub_n);

static void ProcessZeroCrossings(enum TAG_RX_CHAN rxChan);

static void StartLockTimer(struct TAG_SATD_TASK_RXCHAN *pRxChan);
static void StopLockTimer(struct TAG_SATD_TASK_RXCHAN *pRxChan);

static void Transition_A(struct TAG_SATD_TASK_RXCHAN *pRxChan);
static void State_A(
    enum TAG_SATD_EVIDS event, enum TAG_RX_CHAN rxChan, uintptr_t data);

static void Transition_B(struct TAG_SATD_TASK_RXCHAN *pRxChan);
static void State_B(
    enum TAG_SATD_EVIDS event, enum TAG_RX_CHAN rxChan, uintptr_t data);

static void Transition_C(struct TAG_SATD_TASK_RXCHAN *pRxChan);
static void State_C(
    enum TAG_SATD_EVIDS event, enum TAG_RX_CHAN rxChan, uintptr_t data);

//***************************************************************
//                    *** API ***
//***************************************************************

    /** Initialize Subaudible Tone Decoder module:
    */
void SATD_Reset(void)
{
    isrRxChans[RX_CHAN_1].enable = false;
    isrRxChans[RX_CHAN_2].enable = false;
    isrRxChans[RX_CHAN_1].locked = false;
    isrRxChans[RX_CHAN_2].locked = false;

    isrRxChans[RX_CHAN_1].chanID = RX_CHAN_1;
    isrRxChans[RX_CHAN_2].chanID = RX_CHAN_2;
    taskRxChans[RX_CHAN_1].chanID = RX_CHAN_1;
    taskRxChans[RX_CHAN_2].chanID = RX_CHAN_2;

    SSM_InitTimer(&taskRxChans[RX_CHAN_1].lockTimer,
        EV_SATD_LOCK_TO, RX_CHAN_1, 0);
    SSM_InitTimer(&taskRxChans[RX_CHAN_2].lockTimer,
        EV_SATD_LOCK_TO, RX_CHAN_2, 0);

    SATD_UnsubscribeAll();
}

void SATD_Start(void)
{
    SSM_CreateTask(SATD_Task, MID_SATD);
    // Set the initial state.

    Transition_A(&taskRxChans[RX_CHAN_1]);
    Transition_A(&taskRxChans[RX_CHAN_2]);

    // Read the configuration from the User Configuration
    // Module (EEP).

    SATD_LoadDecoderParms(RX_CHAN_1);
    SATD_LoadDecoderParms(RX_CHAN_2);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */ /**
    * Unsubscribe all SATD signals.
    */

void SATD_UnsubscribeAll(void)
{
    Unsubscribe_SATD_Locked();
    Unsubscribe_SATD_Unlocked();
    Unsubscribe_SATD_SigLevel();
    Unsubscribe_SATD_ZXing();
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Return a frequency in 1/10 Hz given an index

unsigned SATD_TranslateIndex(unsigned index)
{
    return(satdFilterParmTbl[index].freq);
}
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Disable CTCSS Decoder:

void SATD_DisableDecoder(enum TAG_RX_CHAN rxChan)
{
    SSM_SendMsgExt(EV_SATD_OFF, (uintptr_t)rxChan, 0);
}

    /// Enable CTCSS Decoder:

void SATD_EnableDecoder(enum TAG_RX_CHAN rxChan)
{
    SSM_SendMsgExt(EV_SATD_ON, (uintptr_t)rxChan, 0);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void SATD_SampleSinkChan1_ISR(int16_t x_sub_n)
{
    DecoderSampleSink_ISR(&isrRxChans[RX_CHAN_1], x_sub_n);
}

void SATD_SampleSinkChan2_ISR(int16_t x_sub_n)
{
    DecoderSampleSink_ISR(&isrRxChans[RX_CHAN_2], x_sub_n);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Load a new set of parameters into a CTCSS decoder.

        This function is called on reset and whenever the parameters
        are thought to have changed (such as via EEP_SetPathRx()).

        @param rxChan: An enum identifying the receiver channel
            to which parameters should be loaded.

        @returns: Nothing.
    */

void SATD_LoadDecoderParms(enum TAG_RX_CHAN rxChan)
{
    struct TAG_SATD_ISR_RXCHAN *pIsrRxChan;
    struct TAG_SATD_TASK_RXCHAN *pTaskRxChan;
    uint8_t index;

    // Read detector parameters from EEP.

    index = EEP_GetCTCSSDecodeIndex(rxChan);

    pIsrRxChan = &isrRxChans[rxChan];
    pTaskRxChan = &taskRxChans[rxChan];

    // Disable the path so we don't get interrupts while
    // we are changing the parameters.

    DI();
    pIsrRxChan->enable = false;
    EI();

        // Init the bandpass filter parameters:

    pIsrRxChan->a1 = satdFilterParmTbl[index].bp_a1;

        // Init the zero crossing qualification parameters:
    pIsrRxChan->minSampleCount = (uint16_t)
        (((uint32_t)satdFilterParmTbl[index].zxingTime
        * MIN_ZXING_FACTOR) >> 16);

    pIsrRxChan->maxSampleCount = (uint16_t)
        (((uint32_t)satdFilterParmTbl[index].zxingTime
        * MAX_ZXING_FACTOR) >> 16);

        // Init task parameters:

    pTaskRxChan->index = index;

    pTaskRxChan->errCaptureTolerance = (uint16_t)
        (((uint32_t)satdFilterParmTbl[index].zxingTime
        * ERR_CAPTURE_FACTOR) >> 8);

    pTaskRxChan->errHoldTolerance = (uint16_t)
        (((uint32_t)satdFilterParmTbl[index].zxingTime
        * ERR_HOLD_FACTOR) >> 8);

#   ifdef UNIT_TEST_ZXING_LIMITS_REPORT
    UT_ZXingLimitsReport(
        pIsrRxChan->chanID, index,
        satdFilterParmTbl[index].zxingTime,
        pIsrRxChan->minSampleCount,
        pIsrRxChan->maxSampleCount,
        pTaskRxChan->errCaptureTolerance,
        pTaskRxChan->errHoldTolerance
    );

#   endif
}

//***************************************************************
//                 *** STATIC FUNCTIONS ***
//***************************************************************

static void StartLockTimer(struct TAG_SATD_TASK_RXCHAN *pRxChan)
{
    SSM_StopTmr(&pRxChan->lockTimer);
    SSM_StartTmr(&pRxChan->lockTimer, SATD_TIME_LOCK);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void StopLockTimer(struct TAG_SATD_TASK_RXCHAN *pRxChan)
{
    SSM_StopTmr(&pRxChan->lockTimer);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** This function receives the current sample from the
     * codec ISR.
     */
static void DecoderSampleSink_ISR(
    struct TAG_SATD_ISR_RXCHAN *pIsrRxChan, int16_t x_sub_n)
{
    int64_t e_sub_n;      // Intermediate result
    int16_t y_sub_n;      // Output value

    struct TAG_SATD_XMSG *pXMsg = &xMsgs[pIsrRxChan->chanID];

    if(pIsrRxChan->enable)
    {
#       ifdef UNIT_TEST
        // Unit test sample count trap
        if(pIsrRxChan->sampleCount >= 20)
            PutBreakpointHere();
#       endif

        // Standard biquad bandpass
        // filter with gain of k:
        // b0 = k, b1 = 0, b2 = -k.
        // k, a1 and a2 are from the CTCSS bandpass table.
        // Sample is down-shifted 4 because we don't need all the bits.

        e_sub_n = (x_sub_n >> 4) + (
            ((pIsrRxChan->a1 * pIsrRxChan->z_1) -
            (BP_CONST_A2 * pIsrRxChan->z_2)) >> E2_BP_FILTER);

        y_sub_n = (int16_t)(
            (BP_CONST_B0 * (e_sub_n - pIsrRxChan->z_2)) >> E2_BP_FILTER);

        pIsrRxChan->z_2 = pIsrRxChan->z_1; pIsrRxChan->z_1 = e_sub_n;

        // Accumulate the squared RMS:

        pIsrRxChan->rmsSquared += (uint32_t)(y_sub_n * y_sub_n);

//IO_DBTraceIn(x_sub_n, y_sub_n);

#       ifdef UNIT_TEST_FILTER_REPORT
        UT_FilterReport(pIsrRxChan->chanID, pIsrRxChan->sampleCount,
            sample, pIsrRxChan->z_1, pIsrRxChan->z_2,
            pIsrRxChan->y_sub_n_minus_1, pIsrRxChan->rmsSquared);
#       endif

        // Now find the zero crossing information.

        if(y_sub_n >= 0 && pIsrRxChan->y_sub_n_minus_1 < 0)
        {
#           ifdef UNIT_TEST_NEWLINE_PER_ZXING
            UT_NewLine();
#           endif

#           ifdef UNIT_TEST_ZXING_PARMS_REPORT
            UT_ZXingParmsReport(pIsrRxChan->sampleCount,
                pIsrRxChan->y_sub_n_minus_1, y_sub_n,
                pIsrRxChan->rmsSquared);
#           endif

            if(
                (pIsrRxChan->sampleCount >= pIsrRxChan->minSampleCount)
//              && (pIsrRxChan->rmsSquared >= RMS_CAPTURE))
                || pIsrRxChan->locked)
            {
                // Here, there has been a valid zero crossing.
                // Prepare a message and send it to the task.

                // Test statically allocated message for reuse:
//                ASSERT(pXMsg->msg.state == MSS_INACTIVE);

                pXMsg->y_sub_n = y_sub_n;
                pXMsg->y_sub_n_minus_1 = pIsrRxChan->y_sub_n_minus_1;
                pXMsg->sampleCount = pIsrRxChan->sampleCount;
                pXMsg->rmsSquared = pIsrRxChan->rmsSquared;

                SSM_SendMsgExt(
                            EV_SATD_ZXING, pIsrRxChan->chanID, 0);
            }

            pIsrRxChan->sampleCount = 0;
            pIsrRxChan->rmsSquared = 0;
        }

        // Here, no valid zero crossings occured.

        else if(pIsrRxChan->sampleCount > pIsrRxChan->maxSampleCount)
        {
            // Here, it has been too long since the last
            // zero crossing, so send a "missing zero crossing"
            // message.
            pXMsg->rmsSquared = pIsrRxChan->rmsSquared;
            SSM_SendMsgExt(
                        EV_SATD_NOZXING, pIsrRxChan->chanID, 0);

            pIsrRxChan->sampleCount = 0;
            pIsrRxChan->rmsSquared = 0;
        }

        // Here, not rejected, but not a zero crossing.  Keep
        // counting.

        else
            pIsrRxChan->sampleCount++;  // Samples since last zxing.

        pIsrRxChan->y_sub_n_minus_1 = y_sub_n;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Turn on all CTCSS detectors for a given Rx channel.

        @param rxChan: A enum identifying the receiver channel to
            be enabled.

        This function enables a receiver channel and causes it to
        begin decoding CTCSS tones.
    */

static void EnablePathDetectors(enum TAG_RX_CHAN rxChan)
{
    struct TAG_SATD_ISR_RXCHAN *pIsrRxChan = &isrRxChans[rxChan];

    // Disable the path until it is set up, so we
    // can write to its parameters.

    DI();
    pIsrRxChan->enable = false;
    EI();

    // Here, we are enabling the path.  Init the
    // variables.

    pIsrRxChan->z_1 = 0;
    pIsrRxChan->z_2 = 0;

    pIsrRxChan->sampleCount = 0;
    pIsrRxChan->rmsSquared = 0;
    pIsrRxChan->y_sub_n_minus_1 = 0;
    taskRxChans[rxChan].r_sub_n_minus_1 = 0;

    DI();
    pIsrRxChan->enable = true;
    EI();
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Subaudible Tone Detector Task

        This function performs the background operations that
        are required to decode a subaudible tone.  The task
        receives messages from the ASTREAM interrupts, as well
        as timer events, and processes them through a finite
        state machine to generate a "lock" or "unlock" condition.
    */

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//      ranksep=0.2;
//      ranksep=equally;
//      rankdir=LR;
//      ratio=0.5;
//      ratio=auto;
//      State_C -> State_C [ label = "EV_SATD_OFF\n(StartSquelchTimer)" ];
//      State_C -> State_C [ label = "EV_SATD_ON\n(StopSquelchTimer)" ];
//      State_C -> State_C [ label = "EV_SATD_ZXING\n&& Qualified\n(StartLockTimer)" ];
//      State_C -> State_C [ label = "EV_SATD_ZXING\n&& Unqualified"];

//      CTCSS Decoder State Machine Graph
/**
    @dot
    digraph G {

        size="5,8"
        ratio=0.7;
        rankdir=LR;

        node [shape = point]; Reset

        node [shape = rect];
        node [style=rounded];

        State_A State_B State_C;

        State_A [label="State_A:\nIdle"];
        State_B [label="State_B:\nUnlocked"];
        State_C [label="State_C:\nLocked"];

        edge [fontsize=10];

        Reset -> State_A [label = "Reset"];

        State_A -> State_B [ label = "EV_SATD_ON" ];

        State_B -> State_A [ label = "EV_SATD_OFF" ];
        State_B -> State_C [ label = "EV_SATD_ZXING\n&& Qualified" ];

        State_C -> State_B [ label = "EV_SATD_ZXING\n&& RevBurstDetected"];
        State_C -> State_B [ label = "EV_SATD_LOCK_TO"];
        State_C -> State_A [ label = "EV_SATD_SQUELCH_TO" ];
    }
    @enddot
*/

static void SATD_Task(
            unsigned event, uintptr_t data1, uintptr_t data2)
{
    ASSERT(data1 < MAX_RXCHAN);
    ASSERT(taskRxChans[data1].pState != NULL);

    (*taskRxChans[data1].pState)(
        (enum TAG_SATD_EVIDS)event, (enum TAG_RX_CHAN)data1, data2);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/**
    Determine the time of a zero crossing in counts.  Each
    sample time takes 256 counts.  The filter interrupt's zero
    crossing detector triggers when one negative sample
    (sample_n_minus_1) is followed by a zero or positive sample
    (sample_n).  When such an event occurs, a message is sent
    to the task.

    The function defined below is called by the task to calculate
    the number of counts since the last zero crossing until this
    most recent one.  The calculation involves linear interpolation
    of the time between samples sample_n_minus_1 and sample_n,
    to obtain the time of the zero crossing to greater precision
    than would be possible by simply considering the number of
    intervening sample times.

    @param rxChan: A receiver channel number indicating the
        receiver for which zero crossing time is to be calculated.

    @returns Nothing.
*/

static void ProcessZeroCrossings(enum TAG_RX_CHAN rxChan)
{
    uint16_t deltaSample;
    uint32_t r_sub_n;
    int zxingTimeError;
    uint32_t zxingTime; // Time in zxing units of last zero crossing.
    struct TAG_SATD_XMSG *pXMsg = &xMsgs[rxChan];
    struct TAG_SATD_TASK_RXCHAN *pTaskRxChan = &taskRxChans[rxChan];

    deltaSample = (uint16_t)(pXMsg->y_sub_n - pXMsg->y_sub_n_minus_1);

    r_sub_n =
        ((uint32_t)((int)-pXMsg->y_sub_n_minus_1) << 8) / deltaSample;

    zxingTime = (pXMsg->sampleCount << 8) + r_sub_n +
        (256 - pTaskRxChan->r_sub_n_minus_1);

    pTaskRxChan->r_sub_n_minus_1 = (uint16_t)r_sub_n;

    // Low-pass filter the zero crossing time.

    pTaskRxChan->zxing_z_1 = zxingTime + (
        (pTaskRxChan->zxing_z_1
        * satdFilterParmTbl[pTaskRxChan->index].zxing_a1)
        >> 8);

    EMIT_SIGNAL(SATD_ZXing, rxChan, zxingTime);

    zxingTimeError = (
        (pTaskRxChan->zxing_z_1
        * (256-satdFilterParmTbl[pTaskRxChan->index].zxing_a1))
        >> 8) - satdFilterParmTbl[pTaskRxChan->index].zxingTime;

    pTaskRxChan->zxingTimeError = (uint16_t)ABS(zxingTimeError);

    // Calculate the RMS in logrithmic units.

    pTaskRxChan->logRMSsquared =
                APLog2(pXMsg->rmsSquared/pXMsg->sampleCount);

    // Clip result to not be less than zero.  If zero is an argument
    // to the log function above, it will return the largest
    // possible negative integer, which will overflow the following
    // filter.

    pTaskRxChan->logRMSsquared =
                    pTaskRxChan->logRMSsquared < 0?
                                    0 : pTaskRxChan->logRMSsquared;

    pTaskRxChan->rms_z_1 = pTaskRxChan->logRMSsquared + (
        (pTaskRxChan->rms_z_1
        * satdFilterParmTbl[pTaskRxChan->index].rms_a1)
        >> 8);

    pTaskRxChan->lpLog2RMSsquared = (int32_t)((pTaskRxChan->rms_z_1
        * (256-satdFilterParmTbl[pTaskRxChan->index].rms_a1)) >> 8);

    // Emit the low-pass filtered Log RMS value as a
    // signal.  Note that the lpLog2RMS value was obtained from
    // samples that were down-shifted 4 before the bandpass filter.
    // This means that the range of the original samples is
    // +/- 2048 counts (since the original signal is +/- 32768
    // counts). Note also that we divide lpLog2RMSsquared by 2
    // to effectively take the square root.

    EMIT_SIGNAL(SATD_SigLevel, rxChan,
                (uintptr_t)(pTaskRxChan->lpLog2RMSsquared / 2));

#   ifdef UNIT_TEST_ZXING_TIME_REPORT
    UT_ZXingTimeReport(
        pTaskRxChan->chanID, zxingTime,
        zxingTime - satdFilterParmTbl[pTaskRxChan->index].zxingTime,
        pTaskRxChan->zxingTimeError, pTaskRxChan->logRMS,
        pTaskRxChan->lpLog2RMS);
#   endif

}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//             *** CTCSS Decoder State Machine ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** State A: Idle
    */

static void Transition_A(struct TAG_SATD_TASK_RXCHAN *pChan)
{

#ifdef UNIT_TEST_STATE_REPORT
    UT_StateReport("State_A");
#endif

    SET_STATE(State_A,'A');
}


static void State_A(enum TAG_SATD_EVIDS event,
                    enum TAG_RX_CHAN rxChan, uintptr_t unused)
{
    switch(event)
    {
        // Here, got a command to turn the detectors on.

        case EV_SATD_ON:
            EnablePathDetectors(rxChan);
            Transition_B(&taskRxChans[rxChan]);
            break;

        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** State B: Detector On, Unlocked.

        Come here when the detector is turned on.
    */

static void Transition_B(struct TAG_SATD_TASK_RXCHAN *pChan)
{
#if 0
    if(pChan->chanID == RX_CHAN_1)
        CDX_Subscribe(RXCHAN_2_CDXCHAN(RX_CHAN_1),
                            CDX_SLOT_SATD, DecoderSampleSinkChan1_ISR);
    else
        CDX_Subscribe(RXCHAN_2_CDXCHAN(RX_CHAN_2),
                            CDX_SLOT_SATD, DecoderSampleSinkChan2_ISR);
#endif
    DI();
    isrRxChans[pChan->chanID].locked = false;
    EI();

#ifdef UNIT_TEST_STATE_REPORT
    UT_StateReport("State_B");
#endif

    SET_STATE(State_B,'B');
}


static void State_B(enum TAG_SATD_EVIDS event,
                    enum TAG_RX_CHAN rxChan, uintptr_t unused)
{
    struct TAG_SATD_TASK_RXCHAN *pTaskRxChan = &taskRxChans[rxChan];

    switch(event)
    {
        // Turn off detectors.  This is a result of the receiver
        // going squelched.  If we are unlocked we don't debounce
        // the squelch.

        case EV_SATD_OFF:
            Transition_A(pTaskRxChan);
            break;

        // A zero crossing was detected.
        case EV_SATD_ZXING:
            ProcessZeroCrossings(rxChan);

            // Qualify the signal and go into lock:

            if(pTaskRxChan->zxingTimeError
                < pTaskRxChan->errCaptureTolerance &&
                pTaskRxChan->logRMSsquared >= RMS_CAPTURE &&
                (pTaskRxChan->lpLog2RMSsquared
                            - pTaskRxChan->logRMSsquared)
                                                    < RMS_RBURST)
            {
                // Here we are locked.
                ASSERT(SATD_Locked != NULL);
                EMIT_SIGNAL(SATD_Locked, rxChan, 0);
                Transition_C(pTaskRxChan);
            }

            break;

        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** State C: Detector On, Locked.

        Come here when the detector is locked on a frequency.
    */

static void Transition_C(struct TAG_SATD_TASK_RXCHAN *pChan)
{
    DI();
    isrRxChans[pChan->chanID].locked = true;
    EI();

#ifdef UNIT_TEST_STATE_REPORT
    UT_StateReport("State_C");
#endif

    StartLockTimer(pChan);
    SET_STATE(State_C,'C');
}


static void State_C(
    enum TAG_SATD_EVIDS event, enum TAG_RX_CHAN rxChan, uintptr_t data)
{
    struct TAG_SATD_TASK_RXCHAN *pTaskRxChan = &taskRxChans[rxChan];

    switch(event)
    {
        // Turn off detectors.  This is a result of the receiver
        // going squelched, and squelch may be bouncing, so if we
        // are currently locked, debounce it by starting a timer.

        case EV_SATD_OFF:
            Transition_A(pTaskRxChan);
            break;

        // A zero crossing was detected.
        case EV_SATD_ZXING:

            ProcessZeroCrossings(rxChan);

            if(pTaskRxChan->zxingTimeError
                < pTaskRxChan->errHoldTolerance)
            {
                StartLockTimer(pTaskRxChan);
            }

            // Here, look for a reverse burst:

            else
            {
                if((pTaskRxChan->lpLog2RMSsquared
                            - pTaskRxChan->logRMSsquared)
                                                >= RMS_RBURST)
                {
                    StopLockTimer(pTaskRxChan);
                    ASSERT(SATD_Unlocked != NULL);
                    EMIT_SIGNAL(SATD_Unlocked, rxChan, 0);
                    Transition_B(pTaskRxChan);
                }
            }
            break;

        // An expected zxing didn't occur
        case EV_SATD_NOZXING:
            break;

        // Lock Timeout: Tone is now unlocked.

        case EV_SATD_LOCK_TO:
            ASSERT(SATD_Unlocked != NULL);
            EMIT_SIGNAL(SATD_Unlocked, rxChan, 0);
            Transition_B(pTaskRxChan);
            break;

        default:
            break;
    }
} //lint !e715 data not used.

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                        END
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

