/****************************************************************//**

    @file ssm_sate.c
    Subaudible Tone Encoder (SATE) State Machine
    for PSE-702.  This state machine determines when to encode
    CTCSS tones.

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

********************************************************************/
//      *** INCLUDE FILES ***

#include <stdint.h>
#include <stdbool.h>        //lint -esym(766, stdbool.h)

#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"

#include "simplesm.h"
#include "app.h"
#include "astream.h"
#include "eep.h"
#include "fgen.h"

#include "ssm_satd.h"
#include "ssm_sate.h"

#include "io_debug.h"       //lint -esym(766, io_debug.h)
#include "apprint.h"        //lint -esym(766, apprint.h)

//***************************************************************
//      *** MACROS ****

//lint -esym(750, PRINT_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, SET_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, STATE_CHAN) (Macro not always referenced)


#define PRINT_STATE_CHAR(x)  if(STATE_CHAN == pChan->chanID) \
        IO_DBPrintChar(x)

#if ((TP_SATE_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
  || (TP_SATE_STATE_FLAGS == STATE_FLAGS_LOWER_CH1))
#   define STATE_CHAN  TX_CHAN_1
#else
#   define STATE_CHAN  TX_CHAN_2
#endif

#if (TP_SATE_STATE_FLAGS == STATE_FLAGS_OFF)
#   define SET_STATE(X,Y)   pChan->pState = (X)

#elif ((TP_SATE_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
    || (TP_SATE_STATE_FLAGS == STATE_FLAGS_UPPER_CH2))
#   define SET_STATE(X,Y)   pChan->pState = (X); PRINT_STATE_CHAR(Y)

#elif ((TP_SATE_STATE_FLAGS == STATE_FLAGS_LOWER_CH1) \
    || (TP_SATE_STATE_FLAGS == STATE_FLAGS_LOWER_CH2))
#   define SET_STATE(X,Y)   pChan->pState = (X); PRINT_STATE_CHAR((Y)|0x20)
#else
#error "TP_SATE_STATE_FLAGS improperly defined."
#endif


//***************************************************************
//             *** MODULE ID DEFINITIONS ****
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // State Machine Event IDs:

enum TAG_SATE_EVIDS
{
    EV_SATE_NULL = SATE_ID(0),  // User audio on
    EV_SATE_ON,                 // Start CTCSS encoder output.
    EV_SATE_OFF,                // Stop CTCSS encoder output
    EV_SATE_RBTO                // Reverse burst timeout
};

    // Reverse Burst Constants:

#define SATE_TIME_REV_BURST 180 // Reverse burst duration
#define SATE_INCR_REV_BURST ((235*65536)/360)   // 235 degrees

//***************************************************************
//             *** DATA STRUCTURE DEFINITIONS ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//***************************************************************
//                  *** VARIABLES ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static struct TAG_SATE_CHAN
{
    enum TAG_TX_CHAN chanID;    // Must be initialized at reset.
    uint8_t fgenChan;           // Function generator channel

    // Task state:
    void (*pState)(
        enum TAG_SATE_EVIDS event, struct TAG_SATE_CHAN *pChan);

    TMR revBurstTmr;            // Reverse burst timer

} chan[MAX_TXCHAN];

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//              Signals
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

DEFINE_SIGNAL(SATE_TurnedOn)
DEFINE_SIGNAL(SATE_TurnedOff)
DEFINE_SIGNAL(SATE_RevBurst)

//***************************************************************
//                 *** CONSTANTS ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Debugging Table for Variable Display

const struct TAG_VARIABLE_TABLE sateVarTab[] =
{
    {"state1", &chan[TX_CHAN_1].pState,
                                VAR_TYPE_HEXINT, sizeof(void *)},
    {"state2", &chan[TX_CHAN_2].pState,
                                VAR_TYPE_HEXINT, sizeof(void *)},
    {NULL, NULL, VAR_TYPE_NONE, 0}
};


//***************************************************************
//          *** STATIC FUNCTION PROTOTYPES ****
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void SATETask(
            unsigned event, uintptr_t data1, uintptr_t data2);

static void StartReverseBurstTimer(struct TAG_SATE_CHAN *pChan);
static void CancelReverseBurstTimer(struct TAG_SATE_CHAN *pChan);

static void Transition_A(struct TAG_SATE_CHAN *pChan);
static void State_A(
    enum TAG_SATE_EVIDS event, struct TAG_SATE_CHAN *pChan);

static void Transition_B(struct TAG_SATE_CHAN *pChan);
static void State_B(
    enum TAG_SATE_EVIDS event, struct TAG_SATE_CHAN *pChan);

static void Transition_C(struct TAG_SATE_CHAN *pChan);
static void State_C(
    enum TAG_SATE_EVIDS event, struct TAG_SATE_CHAN *pChan);

//***************************************************************
//                     *** API ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Reset initialization:

void SATE_Reset(void)
{
    chan[TX_CHAN_1].chanID = TX_CHAN_1;
    chan[TX_CHAN_2].chanID = TX_CHAN_2;

    chan[TX_CHAN_1].fgenChan = FGEN_TX1_ENCODER_CHAN;
    chan[TX_CHAN_2].fgenChan = FGEN_TX2_ENCODER_CHAN;

    SSM_InitTimer(&chan[TX_CHAN_1].revBurstTmr,
                    (int)EV_SATE_RBTO, (int)TX_CHAN_1 ,0);
    SSM_InitTimer(&chan[TX_CHAN_2].revBurstTmr,
                    (int)EV_SATE_RBTO, (int)TX_CHAN_2, 0);

    SATE_UnsubscribeAll();
}

void SATE_Start(void)
{
    SSM_CreateTask(SATETask, MID_SATE);

    SATE_UpdateFrequency(TX_CHAN_1);
    SATE_UpdateFrequency(TX_CHAN_2);

    Transition_A(&chan[TX_CHAN_1]);
    Transition_A(&chan[TX_CHAN_2]);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void SATE_UnsubscribeAll(void)
{
    Unsubscribe_SATE_TurnedOn();
    Unsubscribe_SATE_TurnedOff();
    Unsubscribe_SATE_RevBurst();
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell the CTCSS encoder to turn on the tone.

void SATE_On(enum TAG_TX_CHAN txChan)
{
    SSM_SendMsgExt((int)EV_SATE_ON, (uintptr_t)txChan, 0);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell the CTCSS encoder to turn off the tone.

void SATE_Off(enum TAG_TX_CHAN txChan)
{
    SSM_SendMsgExt((int)EV_SATE_OFF, (uintptr_t)txChan, 0);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell CTCSS encoder to update it's encode frequency.

void SATE_UpdateFrequency(enum TAG_TX_CHAN txChan)
{
    uint16_t freq;
    uint8_t index;

    index = EEP_GetCTCSSEncodeIndex(txChan);
    freq = (uint16_t)SATD_TranslateIndex(index);
    FGen_SetChannelFreq(chan[txChan].fgenChan, freq);

    // The gain below gives approx. 1.0 volts RMS at
    // J10/J11.  Chan. 1 output at card edge connector
    // is down from that because of RC divider following
    // J10 pin header.

    FGen_SetChannelGain(chan[txChan].fgenChan, 255);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                   *** STATIC FUNCTIONS ***
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void StartReverseBurstTimer(struct TAG_SATE_CHAN *pChan)
{
    SSM_StartTmr(&pChan->revBurstTmr, SATE_TIME_REV_BURST);
}

static void CancelReverseBurstTimer(struct TAG_SATE_CHAN *pChan)
{
    SSM_StopTmr(&pChan->revBurstTmr);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Subaudible Tone Encoder Task:

    // Define a single word to use to mean all ID task
    // events:

static void SATETask(
            unsigned event, uintptr_t txChan, uintptr_t unused)
{
    // Call the current state for the channel
    // addressed by the message with the message ID as
    // the state's event.

    ASSERT(VALID_TX_CHAN(txChan));
    ASSERT(chan[txChan].pState != NULL);

    // Do the FSM.
    (*chan[txChan].pState)(
        (enum TAG_SATE_EVIDS)event,
        (struct TAG_SATE_CHAN *)&chan[txChan]);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                      STATE MACHINE
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// State A: Idle

static void Transition_A(struct TAG_SATE_CHAN *pChan)
{
    FGen_ChannelOff(pChan->fgenChan);

    // Notify central that subaudible tone output sequence is complete.

    EMIT_SIGNAL(SATE_TurnedOff, pChan->chanID, 0);

    SET_STATE(State_A,'A');
}

static void State_A(
    enum TAG_SATE_EVIDS event, struct TAG_SATE_CHAN *pChan)
{
    switch(event)
    {
        // We've been keyed.  Turn on tone if EEPROM says
        // we are in RUS mode.

        case EV_SATE_ON:
            Transition_B(pChan);
            break;

        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // CTCSS encoder output is on.

static void Transition_B(struct TAG_SATE_CHAN *pChan)
{
    // Turn on encoder for indicated channel.

    FGen_ChannelOn(pChan->fgenChan);

    // Tell central that tone is on so it can keep the
    // transmitter keyed during reverse burst.

    EMIT_SIGNAL(SATE_TurnedOn, pChan->chanID, 0);

    SET_STATE(State_B,'B');
}

static void State_B(
    enum TAG_SATE_EVIDS event, struct TAG_SATE_CHAN *pChan)
{
    switch(event)
    {
        // Here, we've been told to stop output.  If there
        // is a reverse
        // burst to do, complete it first.

        case EV_SATE_OFF:
            // If reverse burst is enabled, do it here.

            if(EEP_GetTxFlags(
                        pChan->chanID, EEP_TX_FLAG_REVBURST))
                Transition_C(pChan);

            else
                Transition_A(pChan);
            break;

        default:
            break;
    }
}
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Reverse-burst.  Wait for timer to time out.

static void Transition_C(struct TAG_SATE_CHAN *pChan)
{
    StartReverseBurstTimer(pChan);
    EMIT_SIGNAL(SATE_RevBurst, pChan->chanID, 0);

    // Reverse phase of tone (or dead-air?)

    FGen_IncrementPhase(
                pChan->fgenChan, SATE_INCR_REV_BURST);

    SET_STATE(State_C,'C');
}

static void State_C(
    enum TAG_SATE_EVIDS event, struct TAG_SATE_CHAN *pChan)
{
    switch(event)
    {
        // Turn off tone (Transition_A() does this).

        case EV_SATE_RBTO:
            Transition_A(pChan);
            break;

        // We've been keyed again.

        case EV_SATE_ON:
            CancelReverseBurstTimer(pChan);
            Transition_B(pChan);
            break;

        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                          END
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
