/****************************************************************//**

    @file ssm_squelch.c

    Detects carrier squelch changes on two channels and reports
    changes to central.

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

********************************************************************/
//      *** INCLUDE FILES ***

#include <stdint.h>
#include <stdbool.h>        //lint -esym(766, stdbool.h)

#include "std.h"
#include "lpc213x.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"
#include "simplesm.h"

#include "app.h"
#include "io_map.h"
#include "astream.h"
#include "eep.h"

#include "ssm_squelch.h"

    // Headers used during debug:

#include "io_debug.h"   //lint -esym(766, io_debug.h)
#include "io_main.h"        //lint -esym(766, io_main.h)
#include "apprint.h"        //lint -esym(766, apprint.h)

//***************************************************************
//      *** MACROS ****

//lint -esym(750, PRINT_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, SET_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, STATE_CHAN) (Macro not always referenced)

#define PRINT_STATE_CHAR(x)  if(STATE_CHAN == pChan->chanID) \
        IO_DBPrintChar(x)

#if ((TP_SQU_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
  || (TP_SQU_STATE_FLAGS == STATE_FLAGS_LOWER_CH1))
#   define STATE_CHAN  RX_CHAN_1
#else
#   define STATE_CHAN  RX_CHAN_2
#endif

#if (TP_SQU_STATE_FLAGS == STATE_FLAGS_OFF)
#   define SET_STATE(X,Y)   pChan->pState = (X)

#elif ((TP_SQU_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
    || (TP_SQU_STATE_FLAGS == STATE_FLAGS_UPPER_CH2))
#   define SET_STATE(X,Y)   pChan->pState = (X); PRINT_STATE_CHAR(Y)

#elif ((TP_SQU_STATE_FLAGS == STATE_FLAGS_LOWER_CH1) \
    || (TP_SQU_STATE_FLAGS == STATE_FLAGS_LOWER_CH2))
#   define SET_STATE(X,Y)   pChan->pState = (X); PRINT_STATE_CHAR((Y)|0x20)
#else
#error "TP_SQU_STATE_FLAGS improperly defined."
#endif


//***************************************************************

    // Timeout Values:

#define SQU_TIME_POLL           10      ///< ms between RUS polls


enum TAG_SQU_EVIDS
{
    EV_SQU_NULL = SQU_ID(0),    // Used to return a null event.

    // The following events are associated with a receiver channel
    // number:

    EV_SQU_POLL,                // Polling timer timed out.

    EV_SQU_SQU_CLOSED,          // A channel's carrier squelch closed.
    EV_SQU_SQU_OPENED,          // A channel's carrier squelch opened.

    EV_SQU_MAX_EV               // Max event number.
};


//***************************************************************
//                    *** VARIABLES ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static TMR pollTimer;   // When timed out, check for RUS change.

    // Receiver channels: Each channel is represented by
    // a structure:

static struct TAG_SQU_RX_CHAN
{
    enum TAG_RX_CHAN chanID;        // Must be initialized at reset.

    // Receiver state:
    void (*pState)(enum TAG_SQU_EVIDS event,
        struct TAG_SQU_RX_CHAN *pRxChan);

} rxChans[MAX_RXCHAN];


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//              Signals
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */ /**

DEFINE_SIGNAL(SQU_SquelchClosed)
DEFINE_SIGNAL(SQU_SquelchOpened)

//***************************************************************
//                     *** CONSTANTS ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Debugging Table for Variable Display

const struct TAG_VARIABLE_TABLE squelchVarTab[] =
{
    {NULL, NULL, VAR_TYPE_NONE, 0}
};

//***************************************************************
//            *** STATIC FUNCTION PROTOTYPES ****
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void SquelchTask(
                unsigned event, uintptr_t data1, uintptr_t data2);

static bool IsRUSActive(enum TAG_RX_CHAN rxChan);

static void
    Transition_SquelchClosed(struct TAG_SQU_RX_CHAN *pChan);
static void State_SquelchClosed(
        enum TAG_SQU_EVIDS event, struct TAG_SQU_RX_CHAN *pChan);
static void
    Transition_SquelchOpened(struct TAG_SQU_RX_CHAN *pChan);
static void State_SquelchOpened(
        enum TAG_SQU_EVIDS event, struct TAG_SQU_RX_CHAN *pChan);

//********************************************************************
//                      *** API ***
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
*/

void SQU_Reset(void)
{
    SSM_InitTimer(&pollTimer, EV_SQU_POLL, 0, 0);

    rxChans[RX_CHAN_1].chanID = RX_CHAN_1;
    rxChans[RX_CHAN_2].chanID = RX_CHAN_2;

    SQU_UnsubscribeAll();
}

void SQU_Start(void)
{
    SSM_CreateTask(SquelchTask, MID_SQU);

    // Init the state machine.  This must be done here in the
    // task instead of in Reset because it calls functions
    // that are only available after Apex_Start() has been
    // called.

    Transition_SquelchClosed(&rxChans[RX_CHAN_1]);
    Transition_SquelchClosed(&rxChans[RX_CHAN_2]);

    // Start the polling timer.
    SSM_StartTmr(&pollTimer, SQU_TIME_POLL);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    * Unsubscribe all squelch signals.
    */

void SQU_UnsubscribeAll(void)
{
    Unsubscribe_SQU_SquelchClosed();
    Unsubscribe_SQU_SquelchOpened();
}


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void SquelchTask(
                unsigned event, uintptr_t data1, uintptr_t unused)
{
    struct TAG_SQU_RX_CHAN *pChan;

    ASSERT(VALID_RX_CHAN(data1));
    pChan = &rxChans[data1];

    if(event == EV_SQU_POLL)
    {
        // Restart the timer.
        SSM_StartTmr(&pollTimer, SQU_TIME_POLL);

        (*rxChans[RX_CHAN_1].pState)(
            EV_SQU_POLL, &rxChans[RX_CHAN_1]);

        (*rxChans[RX_CHAN_2].pState)(
            EV_SQU_POLL, &rxChans[RX_CHAN_2]);
    }

    else
    {
        (*(pChan->pState))(
                    (enum TAG_SQU_EVIDS)event, pChan); // Do FSM.
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Determine if a particular channel's squelch is active
    // (open).


static bool IsRUSActive(enum TAG_RX_CHAN rxChan)
{
    ASSERT(VALID_RX_CHAN(rxChan));

    if(rxChan == RX_CHAN_1)
    {
        if(EEP_GetRxFlags(RX_CHAN_1, EEP_RX_FLAG_RUSLEVEL))
        {
            return(!(I_RUS1()));
        }
        else    // Here, must be active low.
        {
            return(I_RUS1());
        }
    }

    else    // Must be channel 2.
    {
        if(EEP_GetRxFlags(RX_CHAN_2, EEP_RX_FLAG_RUSLEVEL))
            return(!(I_RUS2()));

        else    // Here, must be active low.
            return(I_RUS2());
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                     Squelch State Machine
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Carrier Squelched State:

static void
    Transition_SquelchClosed(struct TAG_SQU_RX_CHAN *pChan)
{
    ASSERT(SQU_SquelchClosed != NULL);
    EMIT_SIGNAL(SQU_SquelchClosed, pChan->chanID, 0);
    SET_STATE(State_SquelchClosed,'A');
}

static void State_SquelchClosed(
        enum TAG_SQU_EVIDS event, struct TAG_SQU_RX_CHAN *pChan)
{
    switch(event)
    {
        // Did polling timer time out?  If so, check for state
        // change in RUS on this channel.

        case EV_SQU_POLL:

            if(IsRUSActive(pChan->chanID))
            {
                Transition_SquelchOpened(pChan);
            }
            break;

        // Not processing other events in this state.
        default:
            break;
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Unsquelched State:  Channel is fully open.

static void
    Transition_SquelchOpened(struct TAG_SQU_RX_CHAN *pChan)
{
    ASSERT(SQU_SquelchOpened != NULL);
    EMIT_SIGNAL(SQU_SquelchOpened, pChan->chanID, 0);
    SET_STATE(State_SquelchOpened,'C');
}

static void State_SquelchOpened(
        enum TAG_SQU_EVIDS event, struct TAG_SQU_RX_CHAN *pChan)
{
    switch(event)
    {
        // Did polling timer time out?  If so, check for state
        // change in RUS on this channel.
        case EV_SQU_POLL:

            if(!IsRUSActive(pChan->chanID))
            {
                // Here, channel is squelched again.
                Transition_SquelchClosed(pChan);
            }
            break;

        // Not processing other events in this state.
        default:
            break;
    }
}


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                        END
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *



