/****************************************************************//**

    @file ssm_txsm.c
    Transmitter State Machine using Simple State Machine support.

    @defgroup ssm_txsm Transmitter State Machine

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

    <P><A href="txsm.svg"> Transmitter State Diagram (svg)</A></P>

********************************************************************/

//      *** INCLUDE FILES ***

#include <stdint.h>
#include <stdbool.h>

#include "lpc213x.h"
#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"
//#include "apex.h"
#include "simplesm.h"
#include "io_map.h"
#include "io_misc.h"
#include "app.h"
#include "astream.h"
#include "eep.h"

#include "ssm_txsm.h"

#include "io_debug.h"   //lint -esym(766, io_debug.h)
#include "apprint.h"        //lint -esym(766, apprint.h)
#include "io_main.h"        //lint -esym(766, io_main.h)

//***************************************************************
//      *** MACROS ****

//lint -esym(750, PRINT_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, SET_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, STATE_CHAN) (Macro not always referenced)

#define PRINT_STATE_CHAR(x)  if(STATE_CHAN == pChan->chanID) \
        IO_DBPrintChar(x)

#if ((TP_TXSM_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
  || (TP_TXSM_STATE_FLAGS == STATE_FLAGS_LOWER_CH1))
#   define STATE_CHAN  TX_CHAN_1
#else
#   define STATE_CHAN  TX_CHAN_2
#endif

#if (TP_TXSM_STATE_FLAGS == STATE_FLAGS_OFF)
#   define SET_STATE(X,Y)   pChan->pState = (X)

#elif ((TP_TXSM_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
    || (TP_TXSM_STATE_FLAGS == STATE_FLAGS_UPPER_CH2))
#   define SET_STATE(X,Y)   pChan->pState = (X); PRINT_STATE_CHAR(Y)

#elif ((TP_TXSM_STATE_FLAGS == STATE_FLAGS_LOWER_CH1) \
    || (TP_TXSM_STATE_FLAGS == STATE_FLAGS_LOWER_CH2))
#   define SET_STATE(X,Y)   pChan->pState = (X); PRINT_STATE_CHAR((Y)|0x20)
#else
#error "TP_TXSM_STATE_FLAGS improperly defined."
#endif

//***************************************************************
    // State Machine Event IDs:

enum TAG_TXSM_EVIDS
{
    EV_TXSM_UAON = TXSM_ID(0),  ///< Receiver audio on
    EV_TXSM_UAOFF,              ///< Receiver audio off
    EV_TXSM_GENAUDIO,           ///< Play generated audio
    EV_TXSM_DTMFM,              ///< DTMF Mute
    EV_TXSM_AUDIOCP,            ///< Audio complete
    EV_TXSM_HTCP,               ///< Hang time complete.
    EV_TXSM_TIMEOUT,            ///< Transmitter timeout
    EV_TXSM_SATE_OFF,           ///< CTCSS encoder turned off.
    EV_TXSM_SATE_ON,            ///< CTCSS encoder turned on.
    EV_TXSM_TIMEOUT_CLEARED,    ///< The cause of the timeout is cleared
    EV_TXSM_DISABLE,            ///< Disable transmitter.
    EV_TXSM_ENABLE,             ///< Re-enable transmitter.
    EV_TXSM_MAX                 ///< End of event list
};

//***************************************************************
//           *** DATA STRUCTURE DEFINITIONS ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

//***************************************************************
//                   *** VARIABLES ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static struct TAG_TXSM_CHAN
{
    enum TAG_TX_CHAN chanID;        // Must be initialized at reset.

    // Task state:
    bool (*pState)(
        enum TAG_TXSM_EVIDS event, struct TAG_TXSM_CHAN *pChan);

    TMR alligator;              // Alligator timer
    TMR hangTimer;              // Post-courtesy beep timer

    bool sateOn;                // true if CTCSS encoder is on.

} chan[MAX_TXCHAN];

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//              Signals
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

DEFINE_SIGNAL(TXSM_TxOff)
DEFINE_SIGNAL(TXSM_TxOn)
DEFINE_SIGNAL(TXSM_StartGeneratedAudio)
//DEFINE_SIGNAL(TXSM_ScheduleCB)
DEFINE_SIGNAL(TXSM_TxTimeout)
DEFINE_SIGNAL(TXSM_TxComplete)

//***************************************************************
//                    *** CONSTANTS ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Debugging Table for Variable Display

const struct TAG_VARIABLE_TABLE txsmVarTab[] =
{
    {"state1", &chan[TX_CHAN_1].pState,
                                VAR_TYPE_HEXINT, sizeof(void *)},
    {"state2", &chan[TX_CHAN_2].pState,
                                VAR_TYPE_HEXINT, sizeof(void *)},
    {NULL, NULL, VAR_TYPE_NONE, 0}
};


//***************************************************************
//              *** STATIC FUNCTION PROTOTYPES ****
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void TransmitTask(
                unsigned event, uintptr_t data1, uintptr_t unused);

static void StartTransmitterTimeout(struct TAG_TXSM_CHAN *pChan);
static void StopTransmitterTimeout(struct TAG_TXSM_CHAN *pChan);

static void StartHangTimer(struct TAG_TXSM_CHAN *pChan);
static void StopHangTimer(struct TAG_TXSM_CHAN *pChan);
static void EndHangTime(struct TAG_TXSM_CHAN *pChan);

static void Transition_Idle(struct TAG_TXSM_CHAN *pChan);      // A
static bool State_Idle(
    enum TAG_TXSM_EVIDS event, struct TAG_TXSM_CHAN *pChan);

static void Transition_Repeating(struct TAG_TXSM_CHAN *pChan); // B
static bool State_Repeating(
    enum TAG_TXSM_EVIDS event, struct TAG_TXSM_CHAN *pChan);

static void Transition_Disable(struct TAG_TXSM_CHAN *pChan);   // C
static bool State_Disable(
    enum TAG_TXSM_EVIDS event, struct TAG_TXSM_CHAN *pChan);

static void Transition_IsDisabled(
                                struct TAG_TXSM_CHAN *pChan);  // D
static bool State_IsDisabled(
    enum TAG_TXSM_EVIDS event, struct TAG_TXSM_CHAN *pChan);

static void
    Transition_Wait4HangTime(struct TAG_TXSM_CHAN *pChan);     // E
static bool State_Wait4HangTime(
    enum TAG_TXSM_EVIDS event, struct TAG_TXSM_CHAN *pChan);

static void
    Transition_Wait4AudioSm(struct TAG_TXSM_CHAN *pChan);      // F
static bool State_Wait4AudioSm(
    enum TAG_TXSM_EVIDS event, struct TAG_TXSM_CHAN *pChan);

static void Transition_Timeout(struct TAG_TXSM_CHAN *pChan);   // G
static bool State_Timeout(
    enum TAG_TXSM_EVIDS event, struct TAG_TXSM_CHAN *pChan);

static void Transition_Wait4Sate(struct TAG_TXSM_CHAN *pChan); // H
static bool State_Wait4Sate(
    enum TAG_TXSM_EVIDS event, struct TAG_TXSM_CHAN *pChan);

//***************************************************************
//                       *** API ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void TXSM_Reset(void)
{
    chan[TX_CHAN_1].chanID = TX_CHAN_1;
    chan[TX_CHAN_2].chanID = TX_CHAN_2;

    chan[TX_CHAN_1].sateOn = false;
    chan[TX_CHAN_2].sateOn = false;

    SSM_InitTimer(
        &chan[TX_CHAN_1].alligator, EV_TXSM_TIMEOUT, TX_CHAN_1, 0);
    SSM_InitTimer(
        &chan[TX_CHAN_2].alligator, EV_TXSM_TIMEOUT, TX_CHAN_2, 0);

    SSM_InitTimer(
        &chan[TX_CHAN_1].hangTimer, EV_TXSM_HTCP, TX_CHAN_1, 0);
    SSM_InitTimer(
        &chan[TX_CHAN_2].hangTimer, EV_TXSM_HTCP, TX_CHAN_2, 0);

    TXSM_UnsubscribeAll();
}

void TXSM_Start(void)
{
    SSM_CreateTask(TransmitTask, MID_TXSM);

    if(EEP_GetTxFlags(TX_CHAN_1, EEP_TX_FLAG_TXENA))
        Transition_Idle(&chan[TX_CHAN_1]);
    else
        Transition_IsDisabled(&chan[TX_CHAN_1]);

    if(EEP_GetTxFlags(TX_CHAN_2, EEP_TX_FLAG_TXENA))
        Transition_Idle(&chan[TX_CHAN_2]);
    else
        Transition_IsDisabled(&chan[TX_CHAN_2]);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Unsubscribe From All TXSM Signals

void TXSM_UnsubscribeAll(void)
{
    Unsubscribe_TXSM_TxOff();
    Unsubscribe_TXSM_TxOn();
    Unsubscribe_TXSM_StartGeneratedAudio();
    Unsubscribe_TXSM_TxTimeout();
    Unsubscribe_TXSM_TxComplete();
}
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell TxSM to transmit the audio source.

void TXSM_NotifyUserAudioOn(enum TAG_TX_CHAN txChan)
{
    SSM_SendMsgExt(EV_TXSM_UAON, (uintptr_t)txChan, 0);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Tell TxSM to stop transmitting the audio source.

void TXSM_NotifyUserAudioOff(enum TAG_TX_CHAN txChan)
{
    SSM_SendMsgExt(EV_TXSM_UAOFF, (uintptr_t)txChan, 0);
}

    // Tell TxSM that audio is being generated.
void TXSM_NotifyGeneratedAudioOn(enum TAG_TX_CHAN txChan)
{
    SSM_SendMsgExt(EV_TXSM_GENAUDIO, (uintptr_t)txChan, 0);
}

    // Tell TxSM that generated audio has completed.
void TXSM_NotifyGeneratedAudioOff(enum TAG_TX_CHAN txChan)
{
    SSM_SendMsgExt(EV_TXSM_AUDIOCP, (uintptr_t)txChan, 0);
}

    // Tell TxSM to mute for DTMF reception.
void TXSM_DTMFMute(enum TAG_TX_CHAN txChan)
{
    SSM_SendMsgExt(EV_TXSM_DTMFM, (uintptr_t)txChan, 0);
}

    // Tell TxSM that CTCSS encoder (SATE) has started.
void TXSM_NotifySATE_TurnedOn(enum TAG_TX_CHAN txChan)
{
    chan[txChan].sateOn = true;
    SSM_SendMsgExt(EV_TXSM_SATE_ON, (uintptr_t)txChan, 0);
}

    // Tell TxSM that CTCSS encoder (SATE) is finished.
void TXSM_NotifySATE_TurnedOff(enum TAG_TX_CHAN txChan)
{
    chan[txChan].sateOn = false;
    SSM_SendMsgExt(EV_TXSM_SATE_OFF, (uintptr_t)txChan, 0);
}

    // Tell TxSM that the paths that caused the transmitter
    // to time out have been cleared..
void TXSM_NotifyTxTimeoutCleared(enum TAG_TX_CHAN txChan)
{
    SSM_SendMsgExt(EV_TXSM_TIMEOUT_CLEARED, (uintptr_t)txChan, 0);
}

    // Tell TxSM that the transmitter should be disabled.
void TXSM_NotifyTxDisable(enum TAG_TX_CHAN txChan)
{
    SSM_SendMsgExt(EV_TXSM_DISABLE, (uintptr_t)txChan, 0);
}

    // Tell TxSM that the transmitter should be re-enabled.
void TXSM_NotifyTxEnable(enum TAG_TX_CHAN txChan)
{
    SSM_SendMsgExt(EV_TXSM_ENABLE, (uintptr_t)txChan, 0);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Turn PTT on and off

void TXSM_TurnPTTOn(enum TAG_TX_CHAN txChan)
{
    if(txChan == TX_CHAN_1)
    {
        if(EEP_GetTxFlags(TX_CHAN_1, EEP_TX_FLAG_PTTLEVEL))
        {
            O_PTT1_H();
        }
        else    // Here, must be active low.
        {
            O_PTT1_L();
        }

        IO_SetPTT1LED();    // Turn on the PTT LED for chan. 1
    }

    else    // Must be channel 2.
    {
        if(EEP_GetTxFlags(TX_CHAN_2, EEP_TX_FLAG_PTTLEVEL))
        {
            O_PTT2_H();
        }
        else    // Here, must be active low.
        {
            O_PTT2_L();
        }

        IO_SetPTT2LED();    // Turn on the PTT LED for chan. 2
    }
}

void TXSM_TurnPTTOff(enum TAG_TX_CHAN txChan)
{
    if(txChan == TX_CHAN_1)
    {
        if(EEP_GetTxFlags(TX_CHAN_1, EEP_TX_FLAG_PTTLEVEL))
        {
            O_PTT1_L();
        }
        else    // Here, must be active low. Set high to turn off.
        {
            O_PTT1_H();
        }

        IO_ClearPTT1LED();  // Turn off the PTT LED for chan. 1
    }

    else    // Must be channel 2.
    {
        if(EEP_GetTxFlags(TX_CHAN_2, EEP_TX_FLAG_PTTLEVEL))
        {
            O_PTT2_L();
        }
        else    // Here, must be active low. Set high to turn off.
        {
            O_PTT2_H();
        }

        IO_ClearPTT2LED();  // Turn off the PTT LED for chan. 1
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                *** STATIC FUNCTIONS ***
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void StartTransmitterTimeout(struct TAG_TXSM_CHAN *pChan)
{
    if(EEP_GetTxTimeout(pChan->chanID) != 0)
    {
        SSM_StartTmr(
            &pChan->alligator, EEP_GetTxTimeout(pChan->chanID));
    }
}

static void StopTransmitterTimeout(struct TAG_TXSM_CHAN *pChan)
{
    SSM_StopTmr(&pChan->alligator);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void StartHangTimer(struct TAG_TXSM_CHAN *pChan)
{
    SSM_StartTmr(
            &pChan->hangTimer, EEP_GetHangTime(pChan->chanID));
}

static void StopHangTimer(struct TAG_TXSM_CHAN *pChan)
{
    SSM_StopTmr(&pChan->hangTimer);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/**
    @image html txsm.svg "Transmitter State Diagram"
*/

static void TransmitTask(
                unsigned event, uintptr_t data1, uintptr_t unused)
{
    struct TAG_TXSM_CHAN *pChan;

    ASSERT(VALID_TX_CHAN(data1));
    pChan = &chan[data1];

    // Call the function associated with the current state, and
    // if it returns true, it means that no further processing is
    // required for the event.  If false is returned, it means
    // that we should do the default processing for the event.

    if(!(*(pChan->pState))((enum TAG_TXSM_EVIDS)event, pChan))
    {
        switch((enum TAG_TXSM_EVIDS)event)
        {
            case EV_TXSM_DISABLE:
                Transition_Disable(pChan);
                break;

            default:
                break;
        }
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // State A: Idle

static void Transition_Idle(struct TAG_TXSM_CHAN *pChan)
{
    TXSM_TurnPTTOff(pChan->chanID);
    EMIT_SIGNAL(TXSM_TxOff, pChan->chanID, 0);
    StopTransmitterTimeout(pChan);
    SET_STATE(State_Idle,'A');
}

static bool State_Idle(
    enum TAG_TXSM_EVIDS event, struct TAG_TXSM_CHAN *pChan)
{
    switch(event)
    {
        // This event means a receiver is unsquelched and
        // is feeding user audio to the transmitter.
        case EV_TXSM_UAON:
            StartTransmitterTimeout(pChan);
            Transition_Repeating(pChan);
            return(true);   // Event was processed.

        // This event means that the audio module has something
        // to transmit (like an ID or announcement).
        case EV_TXSM_GENAUDIO:
            TXSM_TurnPTTOn(pChan->chanID);
            EMIT_SIGNAL(TXSM_TxOn, pChan->chanID, 0);
            EMIT_SIGNAL(TXSM_StartGeneratedAudio, pChan->chanID, 0);
            StartTransmitterTimeout(pChan);
            Transition_Wait4AudioSm(pChan);
            return(true);   // Event was processed.

        default:
            return(false);   // Do default processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // User Audio On

static void Transition_Repeating(struct TAG_TXSM_CHAN *pChan)
{
    TXSM_TurnPTTOn(pChan->chanID);
    EMIT_SIGNAL(TXSM_TxOn, pChan->chanID, 0);

#if 0
    // Turn CTCSS encoder on if necessary.
    if(EEP_GetTxFlags(pChan->chanID, EEP_TX_FLAG_SATE))
    {
        EMIT_SIGNAL(TXSM_StartSATE, pChan->chanID, 0);
        pChan->sateOn = true;
    }
#endif

    SET_STATE(State_Repeating,'B');
}

static bool State_Repeating(
    enum TAG_TXSM_EVIDS event, struct TAG_TXSM_CHAN *pChan)
{
    switch(event)
    {
        // User audio disappeared.
        case EV_TXSM_UAOFF:
            Transition_Wait4AudioSm(pChan);
            return(true);   // Event was processed.

        // Transmitter timed out:
        case EV_TXSM_TIMEOUT:
            Transition_Timeout(pChan);
            return(true);   // Event was processed.

        default:
            return(false);   // Do default processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Hang Time

static void Transition_Wait4HangTime(struct TAG_TXSM_CHAN *pChan)
{
    if(EEP_GetHangTime(pChan->chanID) != 0)
    {
        StartHangTimer(pChan);
        SET_STATE(State_Wait4HangTime,'E');
    }
    else
    {
        EndHangTime(pChan);
    }
}

static bool State_Wait4HangTime(
    enum TAG_TXSM_EVIDS event, struct TAG_TXSM_CHAN *pChan)
{
    switch(event)
    {
        // Here, a user keyed up during hang time.  Go
        // back to repeating user audio.
        case EV_TXSM_UAON:
            StopHangTimer(pChan);
            Transition_Repeating(pChan);
            return(true);   // Event was processed.

        // Here, the audio system decided to do an ID or
        // something right in the middle of hang time,
        // so go back and wait for it to complete.
        case EV_TXSM_GENAUDIO:
            StopHangTimer(pChan);
            Transition_Wait4AudioSm(pChan);
            return(true);   // Event was processed.

        // Transmitter should turn off.  End of hang time.
        // Request permission to turn off from central, which
        // will coordinate with CTCSS encoder in case of
        // reverse-burst operation.

        case EV_TXSM_HTCP:
            EndHangTime(pChan);
            return(true);   // Event was processed.

        // Transmitter timed out:
        case EV_TXSM_TIMEOUT:
            Transition_Timeout(pChan);
            return(true);   // Event was processed.

        default:
            return(false);   // Do default processing.
    }
}


//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Wait for generated audio to complete

static void Transition_Wait4AudioSm(struct TAG_TXSM_CHAN *pChan)
{
    EMIT_SIGNAL(TXSM_StartGeneratedAudio, pChan->chanID, 0);
    SET_STATE(State_Wait4AudioSm,'F');
}

static bool State_Wait4AudioSm(
    enum TAG_TXSM_EVIDS event, struct TAG_TXSM_CHAN *pChan)
{
    switch(event)
    {
        // If user audio comes on we should cancel a courtesy
        // beep but not an ID.

        case EV_TXSM_UAON:
            Transition_Repeating(pChan);
            return(true);   // Event was processed.

        case EV_TXSM_AUDIOCP:
            StopTransmitterTimeout(pChan);

            // Check to see if hang time is enabled:
            if(EEP_GetHangTime(pChan->chanID) != 0)
                Transition_Wait4HangTime(pChan);
            else
                EndHangTime(pChan);
            return(true);   // Event was processed.

        // Transmitter timed out:
        case EV_TXSM_TIMEOUT:
            Transition_Timeout(pChan);
            return(true);   // Event was processed.

        default:
            return(false);   // Do default processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // Come here at end of hang time.

static void EndHangTime(struct TAG_TXSM_CHAN *pChan)
{
    EMIT_SIGNAL(TXSM_TxComplete, pChan->chanID, 0);

    // If the subaudible tone encoder is still on,
    // tell Central to turn it off, and wait.
    if(pChan->sateOn)
    {
        Transition_Wait4Sate(pChan);
    }
    else
        Transition_Idle(pChan);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Transmitter Timeout: Transmitter is turning off.  Waiting for
    // final ID to complete.

static void Transition_Timeout(struct TAG_TXSM_CHAN *pChan)
{
    // Tell central we timed out.
    EMIT_SIGNAL(TXSM_TxTimeout, pChan->chanID, 0);
    SET_STATE(State_Timeout,'G');
}

static bool State_Timeout(
    enum TAG_TXSM_EVIDS event, struct TAG_TXSM_CHAN *pChan)
{
    switch(event)
    {
        case EV_TXSM_AUDIOCP:
            EMIT_SIGNAL(TXSM_TxComplete, pChan->chanID, 0);
            Transition_Idle(pChan);
            return(true);   // Event was processed.

        default:
            return(false);   // Do default processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Transmitter Disable: Transmitter is turning off.  Waiting for
    // final ID to complete.

static void Transition_Disable(struct TAG_TXSM_CHAN *pChan)
{
    TXSM_TurnPTTOn(pChan->chanID);
    EMIT_SIGNAL(TXSM_TxOn, pChan->chanID, 0);
    EMIT_SIGNAL(TXSM_StartGeneratedAudio, pChan->chanID, 0);

    StopTransmitterTimeout(pChan);
    SET_STATE(State_Disable,'C');
}

static bool State_Disable(
    enum TAG_TXSM_EVIDS event, struct TAG_TXSM_CHAN *pChan)
{
    switch(event)
    {
        case EV_TXSM_AUDIOCP:
            EMIT_SIGNAL(TXSM_TxComplete, pChan->chanID, 0);
            Transition_IsDisabled(pChan);
            return(true);   // Event was processed.

        case EV_TXSM_ENABLE:
            EMIT_SIGNAL(TXSM_TxComplete, pChan->chanID, 0);
            Transition_Idle(pChan);
            return(true);   // Event was processed.

        default:
            return(false);   // Do default processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Transmitter Is Disabled: Transmitter is off.  Waiting for
    // command to re-enable.

static void Transition_IsDisabled(struct TAG_TXSM_CHAN *pChan)
{
    TXSM_TurnPTTOff(pChan->chanID);
    EMIT_SIGNAL(TXSM_TxOff, pChan->chanID, 0);
    SET_STATE(State_IsDisabled,'D');
}

static bool State_IsDisabled(
    enum TAG_TXSM_EVIDS event, struct TAG_TXSM_CHAN *pChan)
{
    switch(event)
    {
        case EV_TXSM_ENABLE:
            Transition_Idle(pChan);
            return(true);   // Event was processed.

        default:
            return(false);   // Do default processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Wait for CTCSS encoder to turn off. (Reverse burst)

static void Transition_Wait4Sate(struct TAG_TXSM_CHAN *pChan)
{
    SET_STATE(State_Wait4Sate,'H');
}

static bool State_Wait4Sate(
    enum TAG_TXSM_EVIDS event, struct TAG_TXSM_CHAN *pChan)
{
    switch(event)
    {
        case EV_TXSM_SATE_OFF:
            Transition_Idle(pChan);
            return(true);   // Event was processed.

        // Here, a user keyed up during reverse burst.  Go
        // back to repeating user audio.
        case EV_TXSM_UAON:
            Transition_Repeating(pChan);
            return(true);   // Event was processed.

        // Here, the audio system decided to do an ID or
        // something right in the middle of a reverse
        // burst so go back and wait for it to complete.
        case EV_TXSM_GENAUDIO:
            Transition_Wait4AudioSm(pChan);
            return(true);   // Event was processed.

        // Transmitter timed out:
        case EV_TXSM_TIMEOUT:
            Transition_Timeout(pChan);
            return(true);   // Event was processed.

        default:
            return(false);   // Do default processing.
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                          END
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
