/************************************************************************
*									*
*  STARTUP.S:  Startup Code for LPC213x Microcontrollers		*
*									*
    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

************************************************************************/


/* *** Startup Code (executed after Reset) ***
*/

/*	Memory Limits
*/

	.equ	SRAM_START,	0x40000000
	.equ	SRAM_END,	0x40003FFF
	.equ	SRAM_LENGTH,	0x00004000

/*	Standard definitions of Mode bits and Interrupt 
	(I & F) flags in PSRs
*/
	.equ	Mode_USR,	0x10
	.equ	Mode_FIQ,	0x11
	.equ	Mode_IRQ,	0x12
	.equ	Mode_SVC,	0x13
	.equ	Mode_ABT,	0x17
	.equ	Mode_UND,	0x1B
	.equ	Mode_SYS,	0x1F

        .equ    I_Bit,	0x80    /* when I bit is set, IRQ is disabled */
        .equ    F_Bit,	0x40    /* when F bit is set, FIQ is disabled */


/*	Stack Configuration
*/
	.equ	Top_Stack,	((SRAM_START+SRAM_LENGTH)-32)

	.equ	FIQ_Stack_Size, 256
	.equ	IRQ_Stack_Size, 512

/*	Phase Locked Loop (PLL) definitions

	Fosc = 12 MHz
	M = 5, P=2.  See LPC213X Users Manual Ch.4 Sec. 8.9:
	Fcco = Fosc * M * 2 * P = 240 MHz
	CCLK= M * Fosc = 60 MHz
*/
        .equ	PLL_BASE,       0xE01FC080  /* PLL Base Address */
        .equ	PLLCON_OFS,     0x00        /* PLL Control Offset*/
        .equ	PLLCFG_OFS,     0x04        /* PLL Configuration Offset */
        .equ	PLLSTAT_OFS,    0x08        /* PLL Status Offset */
        .equ	PLLFEED_OFS,    0x0C        /* PLL Feed Offset */
        .equ	PLLCON_PLLE,    (1<<0)      /* PLL Enable */
        .equ	PLLCON_PLLC,    (1<<1)      /* PLL Connect */
        .equ	PLLCFG_MSEL,    (0x04<<0)   /* PLL Multiplier */
        .equ	PLLCFG_PSEL,    (0x01<<5)   /* PLL Divider */
        .equ	PLLSTAT_PLOCK,  (1<<10)     /* PLL Lock Status */

/*	PLL Setup
*/
        .equ	PLL_SETUP,	1
        .equ	PLLCFG_Val,	0x00000024


/* Memory Accelerator Module (MAM) definitions
*/
        .equ    MAM_BASE,	0xE01FC000  /* MAM Base Address */
        .equ    MAMCR_OFS,	0x00        /* MAM Control Offset*/
        .equ    MAMTIM_OFS,	0x04        /* MAM Timing Offset */

/*	MAM Setup
*/
        .equ    MAM_SETUP,	1
/*	.equ    MAMCR_Val,	0x00000002	/ * MAM full enable */
        .equ    MAMCR_Val,	0x00000001	/* MAM partial enable */

        .equ    MAMTIM_Val,	0x00000004


/* Starupt Code must be linked first at Address at which it expects to run.
*/
        .text
        .arm

        .global _startup
        .func   _startup
_startup:


/* Exception Vectors
  Mapped to Address 0.
  Absolute addressing mode must be used.
  Dummy Handlers are implemented as infinite loops which can be modified.
*/

Vectors:
	LDR	PC, Reset_Addr         
	LDR	PC, Undef_Addr
	LDR	PC, SWI_Addr
	LDR	PC, PAbt_Addr
	LDR	PC, DAbt_Addr
	NOP                            /* Reserved Vector */
#	LDR	PC, IRQ_Addr
	LDR	PC, [PC, #-0x0FF0]     /* Vector from VicVectAddr */
	LDR	PC, FIQ_Addr

Reset_Addr:	.word	Reset_Handler
Undef_Addr:	.word	IO_UNDEF_Handler
SWI_Addr:	.word	IO_SWI_Handler
PAbt_Addr:	.word	IO_ABORT_Handler
DAbt_Addr:	.word	IO_ABORT_Handler
		.word	0                      /* Reserved Address */
IRQ_Addr:	.word	IO_IRQ_Handler
FIQ_Addr:	.word	IO_FIQ_Handler


/* Reset Handler
*/
Reset_Handler:  

	/* Configure and Enable PLL */

	LDR     R0, =PLL_BASE
	MOV     R1, #0xAA
	MOV     R2, #0x55

	MOV     R3, #PLLCFG_Val
	STR     R3, [R0, #PLLCFG_OFS] 
	MOV     R3, #PLLCON_PLLE
	STR     R3, [R0, #PLLCON_OFS]
	STR     R1, [R0, #PLLFEED_OFS]
	STR     R2, [R0, #PLLFEED_OFS]

	/* Wait until PLL Locked */
	
PLL_Loop:
	LDR     R3, [R0, #PLLSTAT_OFS]
	ANDS    R3, R3, #PLLSTAT_PLOCK
	BEQ     PLL_Loop

	/* Switch to PLL Clock */ 

	MOV     R3, #(PLLCON_PLLE | PLLCON_PLLC)
	STR     R3, [R0, #PLLCON_OFS]
	STR     R1, [R0, #PLLFEED_OFS]
	STR     R2, [R0, #PLLFEED_OFS]

	/* Memory Mapping (when Interrupt Vectors are in RAM) */ 

	LDR     R0, =MAM_BASE
	MOV     R1, #MAMTIM_Val
	STR     R1, [R0, #MAMTIM_OFS] 
	MOV     R1, #MAMCR_Val
	STR     R1, [R0, #MAMCR_OFS] 

	.equ    MEMMAP, 0xE01FC040  /* Memory Mapping Control */


	/* Initialise Interrupt System */

	/* Setup Stack for each mode.  Set the exception mode
	   stack pointers to the same locaton as the IRQ stack.
	   This saves space because once an exception occurs,
	   the system halts with interrupts disabled anyway.

	   (The primary disadvantage to this is that once an
	   exception occurs, it is not possible to examine
	   the base addresses of the interrupt stack).
	*/
	
	LDR     R0, =Top_Stack

	/*  Enter IRQ Mode and set its Stack Pointer */

	MSR     CPSR_c, #Mode_IRQ|I_Bit|F_Bit
	MOV     SP, R0

	/*  Enter Undefined Instruction Mode 
	and set its Stack Pointer */

	MSR     CPSR_c, #Mode_UND|I_Bit|F_Bit
	MOV     SP, R0

	/*  Enter Abort Mode and set its Stack Pointer */

	MSR     CPSR_c, #Mode_ABT|I_Bit|F_Bit
	MOV     SP, R0

	/* Allocate room for IRQ stack: */
	
	SUB     R0, R0, #IRQ_Stack_Size

	/*  System Mode is not used so its stack is not set. */
	/*  User Mode is not used so its stack is not set */

	/* MSR     CPSR_c, #Mode_SYS|I_Bit|F_Bit
	   MOV     SP, R0
	*/

	/*  Enter FIQ Mode and set its Stack Pointer */

	MSR     CPSR_c, #Mode_FIQ|I_Bit|F_Bit
	MOV     SP, R0

	/* Allocate room for FIQ stack: */
	
	SUB     R0, R0, #FIQ_Stack_Size

	/* Enter Supervisor Mode and set its stack pointer.
	   This will be the Apex Idle Task stack. */

	MSR     CPSR_c, #Mode_SVC|I_Bit|F_Bit
	MOV     SP, R0

	/* Clear all of RAM (Zero init) */

	MOV     R0, #0
	LDR     R1, =SRAM_START
	LDR     R2, =SRAM_END
LoopZI:
	CMP     R1, R2
	STRLO   R0, [R1], #4
	BLO     LoopZI

	/* Relocate .data section (Copy from ROM to RAM) */

	LDR     R1, =_etext
	LDR     R2, =_data
	LDR     R3, =_edata
LoopRel:
	CMP     R2, R3
	LDRLO   R0, [R1], #4
	STRLO   R0, [R2], #4
	BLO     LoopRel


	/* Clear .bss section.  (Not necessary since everything
	   was cleared above). 
	*/
/*
	MOV     R0, #0
	LDR     R1, =__bss_start__
	LDR     R2, =__bss_end__
LoopZI:
	CMP     R1, R2
	STRLO   R0, [R1], #4
	BLO     LoopZI
*/

	/* Enter the C code */

	LDR     LR, =main
	BX	LR

/*	B	main
*/

__main_exit:
	B       __main_exit


        .global _EI, _DI

        .size   _startup, . - _startup
        .endfunc

        .end
