//***************************************************************
/**
    @file susi.c

    Serial User Interface with X-Modem Support

    @author nlh/alp

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

*/
//***************************************************************

//    *** INCLUDE FILES ***

#include <stdint.h>         //lint -esym(766, stdint.h)
#include <stdbool.h>

#include "lpc213x.h"        //lint -esym(766, lpc213x.h)
#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)

#include "aplib.h"
#include "target.h"         //lint -esym(766, target.h)
#include "apex.h"

#include "app.h"
#include "apparse.h"
#include "apprint.h"
#include "io_main.h"
#include "io_flash.h"
#include "io_codec.h"
#include "astream.h"

#include "console.h"
#include "central.h"
#include "io_debug.h"       //lint -esym(766, io_debug.h)
//#include "ttcon.h"
//#include "io_misc.h"
#include "filesys.h"
//#include "ssm_morse.h"
//#include "ssm_satd.h"
#include "ssm_sate.h"
#include "eep.h"
//#include "audiosm.h"
#include "susi.h"
#include "susi_cmds.h"
#include "xmodem.h"
#include "signon.h"
#include "ansi_term.h"
#include "fgen.h"
#include "ssm_txsm.h"

/***************************************************************/
//    *** MACROS ****

//lint -esym(750, PRINT_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, SET_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, STATE_CHAN) (Macro not always referenced)

#define PRINT_STATE_CHAR(x)   IO_DBPrintChar(x)

#if (TP_SUSI_STATE_FLAGS == STATE_FLAGS_OFF)
#   define SET_STATE(X,Y)   pState = (X)

#elif ((TP_SUSI_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
    || (TP_SUSI_STATE_FLAGS == STATE_FLAGS_UPPER_CH2))
#   define SET_STATE(X,Y)   pState = (X); PRINT_STATE_CHAR(Y)

#elif ((TP_SUSI_STATE_FLAGS == STATE_FLAGS_LOWER_CH1) \
    || (TP_SUSI_STATE_FLAGS == STATE_FLAGS_LOWER_CH2))
#   define SET_STATE(X,Y)   pState = (X); PRINT_STATE_CHAR((Y)|0x20)
#else
#error "TP_SUSI_STATE_FLAGS improperly defined."
#endif

/***************************************************************/
//    *** GENERAL DEFINITIONS ****

    // Timer Values
#define SUSI_TIME_WRITE_DELAY   (200*TU_1MS)    ///< Flash write delay

/// 3000 Hz Deviation Gain Reference for Test Tone
///
/// This gain value causes the function generator to produce
/// a tone with peak counts corresponding to 3000 Hz deviation.

#define GAIN_3000HZ_DEVIATION   ((uint8_t)( \
                                 (FGEN_MAX_GAIN*3.0*0.2417)/5.0))

/***************************************************************/

//    *** STATIC FUNCTION PROTOTYPES ****

static void Slot_RxCharSink_ISR(char c);
static void SetUserConsoleContext(void);
static void SignOn(void);
static char GetRxBuffer(void);
static bool RxBufIsEmpty(void);
static void InitRxBuffer(void);
static bool PrintScreenfulOfStatus(bool first);
static bool PrintScreenfulOfHelp(bool first);

static uint8_t CmdShowHelp(const char *pLine);
static uint8_t CmdShowCommandNames(const char *pLine);
static uint8_t CmdTestTone(const char *pLine);


static uint8_t CmdStatus(const char *pLine);
static uint8_t CmdDump(const char *pLine);

    // General Static Functions

static unsigned FileTime(unsigned fileNum);
static unsigned FileSpaceLeft(void);

    // State Machine Functions

static void StateMachine(void);

static void State_Interactive(
        enum TAG_SUSI_EVIDS id, uintptr_t data1, uintptr_t data2);

static void State_XModemReceive(
        enum TAG_SUSI_EVIDS id, uintptr_t data1, uintptr_t data2);

static void State_Start(
        enum TAG_SUSI_EVIDS id, uintptr_t data1, uintptr_t data2);

static void State_TestToneRequest(
    enum TAG_SUSI_EVIDS id, uintptr_t data1, uintptr_t unused);

static void State_VoiceRecord(
    enum TAG_SUSI_EVIDS id, uintptr_t data1, uintptr_t unused);

static void State_StatusList(
    enum TAG_SUSI_EVIDS id, uintptr_t data1, uintptr_t unused);

static void State_HelpList(
    enum TAG_SUSI_EVIDS id, uintptr_t data1, uintptr_t unused2);

/***************************************************************/

//    *** DATA STRUCTURE Definitions ***

/***************************************************************/
//    *** CONSTANTS ***

//    Command Table

//    The following table identifies the command words and
//    functions available on the user console when issuing commands
//    from the serial port.

static const console_cmd_tbl_t susiCmdTable[] =
{
    {   // Set courtesy beep Morse character
        "cbc", SUSI_CmdSetBeepCharacter,
        "Set courtesy beep Morse character.",
        "cbc {chan} {char}",
        "\n\rWhere:\n\r\tchan\tReceiver channel number (1 or 2)"
        "\n\r\tchar\tA single character from the Morse character set."
    },

    {   // Set pre-courtesy beep delay
        "cbd", SUSI_CmdSetBeepDelay,
        "Set pre-courtesy beep delay time.",
        "cbd {chan} {delay}",
        "\n\rWhere:\n\r\tchan\tReceiver channel number (1 or 2)"
        "\n\r\tdelay\tDelay time in milliseconds (100 to 10000)"
    },

    {   // Set courtesy beep audio level.
        "cbl", SUSI_CmdSetBeepLevel,
        "Set courtesy beep audio level.",
        "cbl {chan} {level}",
        "\n\rWhere:\n\r\tchan\tTransmitter channel number (1 or 2)"
        "\n\r\tlevel\tAudio level in percent (0 to 100)."
    },

    {   // Set courtesy beep mode
        "cbm", SUSI_CmdSetBeepMode,
        "Set courtesy beep mode.",
        "cbm {chan} {mode}",
        "\n\rWhere:\n\r\tchan\tReceiver channel number (1 or 2)"
        "\n\r\tmode\t0 to disable courtesy beep,"
        "\n\r\t\t1 for single tone beep,"
        "\n\r\t\t2 for double tone beep,"
        "\n\r\t\t3 for Morse character beep."
    },

    {   // Set courtesy beep tone parameters
        "cbt", SUSI_CmdSetBeepTone,
        "Set courtesy beep tone parameters.",
        "cbt {chan} {toneNum} {freq} {length}",
        "\n\rWhere:\n\r\tchan\tReceiver channel number (1 or 2)"
        "\n\r\ttoneNum\tTone number 1 or 2"
        "\n\r\tfreq\tBeep frequency in Hz (400 to 2000)"
        "\n\r\tlength\tLength of beep in milliseconds (50 to 1000)"
    },

    {   // Set Morse rate and frequency.
        "cwr", SUSI_CmdSetMorseRate,
        "Set Morse rate and frequency.",
        "cwr {chan} {rate} {freq}",
        "\n\rWhere:\n\r\tchan\tTransmitter channel number (1 or 2)"
        "\n\r\trate\tCW rate in words per minute (10 to 20)"
        "\n\r\tfreq\tCW tone frequency in Hz (400 to 2000)"
    },

    {   // Set Morse audio level.
        "cwl", SUSI_CmdSetMorseLevel,
        "Set Morse audio level.",
        "cwl {chan} {level}",
        "\n\rWhere:\n\r\tchan\tTransmitter channel number (1 or 2)"
        "\n\r\tlevel\tAudio level in percent (0 to 100)"
    },

    {   // Display command list
        "help", CmdShowHelp,
        "Display summary or detailed help for console commands.",
        "help [command]",
        "\n\rWhere:\n\r\tcommand\tOptional command name."
    },

    {   // Display command names
        "?", CmdShowCommandNames,
        "Display command names only.",
        "?",
        ""
    },

    {   // Set ID string
        "id", SUSI_CmdSetIDString,
        "Set CW ID string.",
        "id {chan} {\"string\"}",
        "\n\rWhere:\n\r\tchan\tChannel number (1 or 2)"
        "\n\r\tstring\tCW ID string in quotes; max 19 characters"
    },

    {   // Set a Receiver's CTCSS decoder frequency index
        "sdi", SUSI_CmdSetCTCSSDecodeIndex,
        "Set a receiver's CTCSS decoder frequency index.",
        "sdi {rx_chan} {index}",
        "\n\rWhere: \n\trx_chan\tReceiver channel number (1 or 2)"
        "\n\r\tindex\t0 to 36, to select CTCSS frequency."
        "\n\r\t\tType \"help tonetab\" for a table of tone indices."
    },

    {   // Set a Transmitter's CTCSS encoder frequency
        "sei", SUSI_CmdSetCTCSSEncodeIndex,
        "Set a transmitter's CTCSS encoder frequency index.",
        "sei {chan} {index}",
        "\n\rWhere: \n\r\tchan\tTransmitter channel number (1 or 2)"
        "\n\r\tindex\t0 to 36, to select CTCSS frequency."
        "\n\r\t\tType \"help tonetab\" for a table of tone indices."
    },

    {   // Set a Transmitter's CTCSS encoder mode
        "sem", SUSI_CmdSetCTCSSEncodeMode,
        "Set a transmitter's CTCSS encoder mode.",
        "sem {chan} {mode}",
        "\n\rWhere: \n\r\tchan\tTransmitter channel number (1 or 2)"
        "\n\r\tmode\t1: Encode only while receiver is unsquelched, "
        "\n\r\t\t2: Encode whenever PTT is asserted."
    },

    {   // Set Rx-Tx Path Parameters
        "path", SUSI_CmdSetRxTxPath,
        "Set Rx-Tx path parameters.",
        "path {rx_chan} {tx_chan} {priority}",
        "\n\rWhere:"
        "\n\r\trx_chan:\tReceiver number (1 or 2)"
        "\n\r\ttx_chan:\tTransmitter number (1 or 2)"
        "\n\r\tpriority: 'L' (Low), 'M' (Medium/Mixer), 'H' (High)"
    },

    {   // Delete a Path
        "dpath", SUSI_CmdDeletePath,
        "Delete an Rx-Tx path.",
        "dpath {rx_chan} {tx_chan}",
        "\n\rWhere:"
        "\n\r\trx_chan:\tReceiver number (1 or 2)"
        "\n\r\ttx_chan:\tTransmitter number (1 or 2)"
    },

    {   // Schedule an Announcement
        "anno", SUSI_CmdScheduleAnno,
        "Schedule an announcement.",
        "anno {tx_chan} {fileNum} {period} {priority}",
        "\n\rWhere:"
        "\n\r\ttx_chan:\tTransmitter number (1 or 2)"
        "\n\r\tfileNum:\tFile number (1 through 8)"
        "\n\r\tperiod:\tTime, in minutes, between successive playbacks "
        "\n\r\t\tof this announcement"
        "\n\r\tpriority: 'L' (Low), 'M' (Medium/Mixer), 'H' (High)"
        "\n\n\rNote: scheduling an announcement automatically enables"
        "\n\rit.  To disable the announcement see 'eanno' command."
    },

    {   // Delete an Announcement
        "danno", SUSI_CmdDeleteAnno,
        "Delete an announcement.",
        "danno {tx_chan} {fileNum}",
        "\n\rWhere:"
        "\n\r\ttx_chan:\tTransmitter number (1 or 2)"
        "\n\r\tfileNum:\tFile number (1 through 8)"
        "\n\n\rNote that deleting the announcement does not delete"
        "\n\rthe voice file containing the announcement.  To delete"
        "\n\rthe file, see 'fdel'."
    },

    {   // Enable or Disable an Announcement
        "eanno", SUSI_CmdEnableAnno,
        "Enable or disable an announcement.",
        "eanno {fileNum} {enable}",
        "\n\rWhere:"
        "\n\r\tfileNum:\tFile number (1 through 8)"
        "\n\r\tenable: 1 to enable announcement, 0 to disable it."
        "\n\rEnables or disables all announcements associated with"
        "\n\rthe selected file number."
    },

    {   // Enable or disable DTMF (Touch-Tone) password
        "ttpena", SUSI_CmdEnableDTMFPassword,
        "Enable or disable DTMF (Touch Tone) access password.",
        "ttpena {flag}",
        "\n\rWhere:\n\r\tflag: 1 to enable, or 0 to disable, the "
        "\n\r\tDTMF password entry requirement."
    },

    {   // Set DTMF (Touch-Tone) password string
        "ttpwd", SUSI_CmdSetDTMFPwd,
        "Set DTMF (Touch Tone) access password code number.",
        "ttpwd {\"string\"}",
        "\n\rWhere:\n\r\tstring\tPassword in quotes; max 8 characters"
        "\n\r\tCharacters may be any valid DTMF character except * "
            "or #"
    },

    {   // Set DTMF (Touch-Tone) command character
        "ttchar", SUSI_CmdSetDTMFCommandChar,
        "Set DTMF (Touch Tone) command character.",
        "ttchar {char}",
        "\n\rWhere:\n\r\tchar: A DTMF character #, *, A, B, C, or D,"
        "\n\r\tto be used as the first character of all DTMF commands."
    },

    {   // Set receiver flags
        "rxf", SUSI_CmdSetRxFlags,
        "Set receiver active level and CTCSS decoder enable flag.",
        "rxf {chan} [Lx] [Px] [Rx]",
        "\n\rWhere:\n\r\tchan\tReceiver channel number (1 or 2)"
        "\n\r\tx\t0 (off or low) or 1 (on or high)"
        "\n\r\tL\tReceiver unsquelched (RUS) active high if 1,"
        "\n\r\tP\tCTCSS decoder enabled if 1."
        "\n\r\tR\t1 to enable the receiver, 0 to disable it."
    },

    {   // Set filter index
        "filt", SUSI_CmdSetFilterIndex,
        "Select filters.",
        "filt {chan} {type} {index}",
        "\n\rWhere:\n\r\tchan\tReceiver channel number (1 or 2)"
        "\n\r\tType\tFilter to which command pertains:"
        "\n\r\t\tD = Deemphasis Tone Filter,"
        "\n\r\t\tR = CTCSS Reject Filter."
        "\n\r\tIndex\tID number of filter to use:"
        "\n\r\t\tDeemphasis Tone: 0 = high, 1 = med, 2 = low,"
        "\n\r\t\tCTCSS Reject: 0 = None, 1 = 3rd order, 2 = 4th order."
    },

    {   // Set receiver level manually
        "rxl", SUSI_CmdSetRxLevel,
        "Set receiver level.",
        "rxl {chan} {level}",
        "\n\rWhere:\n\r\tchan\tReceiver channel number (1 or 2)"
        "\n\r\tlevel\tVolume in percent (0 to 100)"
    },

    {   // Set receiver level automatically
        "rxla", SUSI_CmdSetRxLevelAuto,
        "Set receiver level automatically.",
        "rxla {chan}",
        "\n\rWhere:\n\r\tchan\tReceiver channel number (1 or 2)"
        "\n\n\rFeed receiver channel with 1000 Hz tone at 3000 Hz"
        "\n\rmodulation level before entering this command."
    },


    {   // Output a test tone.
        "tone", CmdTestTone,
        "Output a test tone to a transmitter channel.",
        "tone {chan}",
        "\n\rWhere:\n\r\tchan\tChannel number (1 or 2)"
    },

    {   // Print parameter status.
        "status", CmdStatus,
        "Print parameter values on the console with pagination.",
        "status",
        ""
    },

    {   // Dump parameter values.
        "dump", CmdDump,
        "Print parameter values on the console without pagination.",
        "dump",
        ""
    },

    {   // Print firmware version info.
        "ver", SUSI_CmdVersion,
        "Output a line containing the firmware version and/or date.",
        "ver",
        ""
    },

    {   // Set tx channel timers
        "tmr",  SUSI_CmdSetTimers,
        "Set transmitter timers.",
        "tmr {chan} {id} {timeout} {hang}",
        "\n\rWhere:\n\r\tchan\tTransmitter channel number (1 or 2)"
        "\n\r\tid\tID timer in sec (60 to 1920). 0 = no ID"
        "\n\r\ttimeout\tTx timeout in sec (10 to 600). 0 = no timeout"
        "\n\r\thang\tHang time duration in mS (0 to 10000)"
    },

    {   // Set transmitter flags
        "txf",  SUSI_CmdSetTxFlags,
        "Set transmitter flags.",
        "txf {chan} [Bx] [Dx] [Lx] [Px] [Rx] [Tx] [Vx]",
        "\n\rWhere:\n\r\tchan\tTransmitter channel number (1 or 2)"
        "\n\r"
        "\n\r\tB\tCourtesy beep enabled if 1, disabled if 0"
        "\n\r\tD\tDTMF mute if 1, unmute if 0"
        "\n\r\tL\tPTT active high"
        "\n\r\tP\tCTCSS decoder on if 1, off if 0"
        "\n\r\tR\tCTCSS reverse burst on if 1, off if 0"
        "\n\r\tT\tTransmitter enable"
        "\n\r\tV\tVoice ID if 1, CW ID if 0"
    },

    {   // Set transmitter level
        "txl",  SUSI_CmdSetTxLevel,
        "Set transmitter level.",
        "txl {chan} {level}",
        "\n\rWhere:"
        "\n\r\tchan\tTransmitter channel number (1 or 2)"
        "\n\r\tlevel\tVolume in dB (-96 to 48)"
    },

    {   // Force reload of defaults
        "def", SUSI_CmdDefaults,
        "Load default parameter values into non-volatile memory.",
        "def 8958",
        "\n\rWhere:\n\r\t8958 is the \"Are you sure?\" code."
    },

    {   // Format the flash system
        "format", SUSI_CmdFileFormat,
        "Format the flash file system.",
        "format 8958",
        "\n\rWhere:\n\r\t8958 is the \"Are you sure?\" code."
    },

    {   // Configure Voice ID
        "vid", SUSI_CmdVoiceID,
        "Configure voice ID for a transmitter.",
        "vid {chan} {file}",
        "\n\rWhere:"
        "\n\r\tchan\tTransmitter channel number (1 or 2),"
        "\n\r\tfile\tFile number (1 to 8)."
    },

    {   // File System Status
        "files", SUSI_CmdFileSystemStatus,
        "Print file system status.",
        "fsys",
        ""
    },

#if EMULATOR == 0
    {   // Reset the microcontroller
        "reset", SUSI_CmdReset,
        "Reset the microcontroller.",
        "reset",
        NULL
    },
#endif

    {   // Display Tone Table Help
        "tonetab", NULL,
        "Type 'help tonetab' to display the CTCSS tone table.", "",

        "\n\r\tCTCSS Tone Table\n"
        "\n\r\t 0:  67.0\t13: 110.9\t25: 167.9"
        "\n\r\t 1:  71.9\t14: 114.8\t26: 173.8"
        "\n\r\t 2:  74.4\t15: 118.8\t27: 179.9"
        "\n\r\t 3:  77.0\t16: 123.0\t28: 186.2"
        "\n\r\t 4:  79.7\t17: 127.3\t29: 192.8"
        "\n\r\t 5:  82.5\t18: 131.8\t30: 203.5"
        "\n\r\t 6:  85.4\t19: 136.5\t31: 210.7"
        "\n\r\t 7:  88.5\t20: 141.3\t32: 218.1"
        "\n\r\t 8:  91.5\t21: 146.2\t33: 225.7"
        "\n\r\t 9:  94.8\t22: 151.4\t34: 233.6"
        "\n\r\t10: 100.0\t23: 156.7\t35: 241.8"
        "\n\r\t11: 103.5\t24: 162.2\t36: 250.3"
        "\n\r\t12: 107.2"
    },

    {   // Record a voice file
        "frec",  SUSI_CmdFileRecord,
        "Record a voice file.",
        "frec {chan} {file}",
        "\n\rWhere:\n\r\tchan\tReceiver channel number (1 or 2)"
        "\n\r\tfile\tFile number (1 through 8)"
    },

    {   // Delete a voice file
        "fdel",  SUSI_CmdFileDelete,
        "Delete a voice file.",
        "fdel {filenum}",
        "\n\rWhere:\n\r\tfilenum\tFile number (1 through 8)"
    },


    {   // Play a voice file
        "fplay",  SUSI_CmdPlayFile,
        "Play a voice file.",
        "fplay {chan} {filenum}",
        "\n\rWhere:\n\r\tchan\tReceiver channel number (1 or 2)"
        "\n\r\tfilenum\tFile number (1 through 8)"
    },


    {   // Play the ID
        "playid",  SUSI_CmdPlayID,
        "Play the current ID.",
        "playid {chan}",
        "\n\rWhere:\n\r\tchan\tTransmitter channel number (1 or 2)"
    },

    // * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

        // Debugging Commands

    {"apath", SUSI_CmdDisplayActivePaths, NULL, NULL, NULL},

        // Display a page from SRAM.
    {"dm", SUSI_CmdDisplaySRAM, NULL, NULL, NULL},

        // Display a variable in SRAM.
    {"dv", SUSI_CmdDisplayVariable, NULL, NULL, NULL},

        // Select a module for display.
    {"sm", SUSI_CmdSelectModule, NULL, NULL, NULL},

        // Trace Commands
#if TRACE != 0

        // Reset trace system
    {"trst", CmdResetTrace, NULL, NULL, NULL},

        // Enable tracing
    {"ton", CmdEnaTrace, NULL, NULL, NULL},

        // Disable tracing
    {"toff", CmdDisTrace, NULL, NULL, NULL},

        // Display trace buff to console
    {"tdb", CmdPrintTraceBuf, NULL, NULL, NULL},

        // Display trace count to console
    {"tdc", CmdPrintTraceCount, NULL, NULL, NULL},
#endif // TRACE

        // Breakpoint commands

    {"abp", SUSI_CmdArmBreakpoint, NULL, NULL, NULL},

        // Invoke Downloader

        // Download new program to flash.
    {"dl", SUSI_CmdDownload, NULL, NULL, NULL},

        // Flash Memory Commands

        // ff pattern inc
    {"ff", SUSI_CmdFillFlashBuf, NULL, NULL, NULL},

        // fwrp addr
    {"fwrp", SUSI_CmdWriteFlashPage, NULL, NULL, NULL},

        // frdp addr
    {"frdp", SUSI_CmdReadFlashPage, NULL, NULL, NULL},

        // fwrb addr data
    {"fwrb", SUSI_CmdWriteFlashByte, NULL, NULL, NULL},

        // frdb addr
    {"frdb", SUSI_CmdReadFlashByte, NULL, NULL, NULL},

    // frdb addr
    {"frsn", SUSI_CmdReadFlashSerialNumber, NULL, NULL, NULL},

        // fera addr
    {"fera", SUSI_CmdEraseFlash, NULL, NULL, NULL},

        // frs (no parameters)
    {"frs", SUSI_CmdReadFlashStatus, NULL, NULL, NULL},

        // fws status
    {"fws", SUSI_CmdWriteFlashStatus, NULL, NULL, NULL},

        // fd no parameters
    {"fd", SUSI_CmdDisplayFlash, NULL, NULL, NULL},

        // fup no parameters
    {"fup", SUSI_CmdUnprotectFlash, NULL, NULL, NULL},

        // File System Commands

        // Format File System
    {"format", SUSI_CmdFileFormat, NULL, NULL, NULL},

        // Read from file to buffer
    {"fread", SUSI_CmdFileRead, NULL, NULL, NULL},

        // Append buffer to file
    {"fappend", SUSI_CmdFileAppend, NULL, NULL, NULL},

        // Delete (erase) a file
    {"fdel", SUSI_CmdFileDelete, NULL, NULL, NULL},

        // RTC Commands

        // Set the real-time clock
    {"rtcset", SUSI_CmdSetRTC, NULL, NULL, NULL},

        // Read the real-time clock
    {"rtcread", SUSI_CmdReadRTC, NULL, NULL, NULL},


    {NULL, NULL, NULL, NULL, NULL}      // Marks end of table.
};

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


//****************************************************************
//                 *** VARIABLES ***
//****************************************************************

    // State to call when a character arrives.

static void (*pState)(
        enum TAG_SUSI_EVIDS id, uintptr_t data1, uintptr_t data2);

static MBOX mailbox;

static enum TAG_TX_CHAN testToneChan;   ///< Channel on which to
                                         ///< output a test tone.
static bool testToneOn;                 ///< True if tone is on.

static uint8_t recordFile;     ///< File number we are recording.

static TMR writeTimer;         ///< Times writes to flash.

    /// SUSI's variable table for debugger access

const struct TAG_VARIABLE_TABLE susiVarTab[] =
{
    {NULL, NULL, VAR_TYPE_NONE, 0}
};

//***************************************************************
//                     Public API Functions
//***************************************************************

void SUSI_Reset(void)
{
    Apex_InitMailbox(&mailbox);
    testToneOn = false;
    Apex_InitMsgTimer(
        &writeTimer, &mailbox, EV_SUSI_EEP_WRITE_TIMEOUT, 0, 0);
    SUSI_DebugReset();
    SUSI_RxCommandsReset();
}

void SUSI_Start(void)
{
    SetUserConsoleContext();
    SUSI_Transition_State_Start();
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void SUSI_SetUserConsoleContext(void)
{
    SetUserConsoleContext();
    SignOn();
    SUSI_Transition_State_Interactive();
//    Apex_SendMsg(&mailbox, EV_SUSI_SIGNON);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // Set new SUSI state.  This function allows the external
    // SUSI modules set the current state for SUSI's state
    // machine.

void SUSI_SetState(void (*pNewState)(
        enum TAG_SUSI_EVIDS id, uintptr_t data1, uintptr_t data2),
        char c)
{
    SET_STATE(pNewState,'C');
} //lint !e715 char c not used if not printing state characters.


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Init a timer for use by external states.

void SUSI_InitTimer(TMR *pTmr, unsigned id)
{
    Apex_InitMsgTimer(pTmr, &mailbox, id, 0, 0);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    // Central calls this after SUSI requests permission to
    // output a test tone, if the test tone cannot be allowed
    // right now.

    // Test tones can only be allowed if no transmitters or
    // receivers are active.

void SUSI_DenyTestTone(void)
{
    Apex_SendMsg(&mailbox, EV_SUSI_TEST_TONE_DENIED);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Central calls this after SUSI requests permission to
    // output a test tone, if the test tone can be allowed
    // right now.  No transmitters or receivers are currently
    // active.

void SUSI_GrantTestTone(void)
{
    Apex_SendMsg(&mailbox, EV_SUSI_TEST_TONE_GRANTED);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Central calls this to notify SUSI that a voice recording
 * has ended.
 */
void SUSI_NotifyRecordComplete(enum TAG_SUSI_RECORD_STATUS status)
{
    Apex_SendMsgExt(&mailbox, EV_SUSI_RECORD_COMPLETE, status, 0);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Central calls this to notify SUSI that a voice recording
 * has started.
 */
void SUSI_NotifyRecordStarted(void)
{
    Apex_SendMsgExt(&mailbox, EV_SUSI_RECORD_STARTED, 0, 0);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /// Restart the timer that causes the EEPROM to be written
    /// when it times out.

void SUSI_RestartWriteTimer(void)
{
    Apex_StopTmr(&writeTimer);
    Apex_StartTmr(&writeTimer, SUSI_TIME_WRITE_DELAY);
}

//***************************************************************
//                       Static Functions
//***************************************************************

static void SetUserConsoleContext(void)
{
    Con_SetContext(susiCmdTable, '>', STDOUT, StateMachine);
    InitRxBuffer();
    Subscribe_IO_RxUart0(Slot_RxCharSink_ISR); // Capture UART0
    CEN_SubscribeAll();     // Subscribe to other module's signals.
}

//***************************************************************
//               Character Input Buffer Object
//***************************************************************

#define Z_CHAR_BUF      192     ///< Size of input buffer.
static char buffer[Z_CHAR_BUF]; ///< Character input buffer.
static uint16_t bufCount;    ///< Number of chars in buffer.
static uint16_t inputIndex;     ///< Index to next empty slot.
static uint16_t outputIndex;    ///< Index to next full slot.

static void Slot_RxCharSink_ISR(char c)
{
    if(bufCount < Z_CHAR_BUF)
    {
        if(bufCount++ == 0)
            Apex_SendMsg(&mailbox, EV_SUSI_RXDATA);

        buffer[inputIndex] = c;

        // Wrap input index.
        if(inputIndex >= (Z_CHAR_BUF-1))
            inputIndex = 0;
        else
            inputIndex++;
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Get a character from the input buffer.
     *
     * @return Returns the next character from the input buffer.
     * Returns '\0' if the buffer is empty.
     * @see IsRxBufEmpty() to determine if there are characters
     * in the buffer to receive.
     */

static char GetRxBuffer(void)
{
    char c;

    DI();
    if(bufCount == 0)
        c = '\0';
    else
    {
        bufCount--;
        c = buffer[outputIndex];

        // Wrap output index.
        if(outputIndex >= (Z_CHAR_BUF-1))
            outputIndex = 0;
        else
            outputIndex++;
    }
    EI();
    return(c);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Test to see if there are characters in the buffer.
     *
     * @return Returns true if there are characters in the buffer,
     * false otherwise.
     */
static bool RxBufIsEmpty(void)
{
    bool result;

    DI();
    result = bufCount == 0;
    EI();
    return(result);
}

static void InitRxBuffer(void)
{
    DI();
    // Character input buffer initialization
    bufCount = inputIndex = outputIndex = 0;
    EI();
}

//***************************************************************
//                   Console State Machine
//***************************************************************

static void StateMachine(void)
{
    unsigned id;
    uintptr_t data1, data2;

    Apex_PendMsg(&mailbox, &id, &data1, &data2, 0);

    if((enum TAG_SUSI_EVIDS)id == EV_SUSI_RXDATA)
    {
        while(!RxBufIsEmpty())
        {
            data1 = (uintptr_t)GetRxBuffer();
            ASSERT(pState != NULL);
            (*pState)((enum TAG_SUSI_EVIDS)id, data1, data2);
            Apex_Delay(0);  // Give other tasks a shot.
        }
    }
    else if((enum TAG_SUSI_EVIDS)id == EV_SUSI_EEP_WRITE_TIMEOUT)
        EEP_WriteOperatorParameters();

    else
    {
        ASSERT(pState != NULL);
        (*pState)((enum TAG_SUSI_EVIDS)id, data1, data2);
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Wait in this state for an initial CR from the user console
    // before signing on.

void SUSI_Transition_State_Start(void)
{
    SET_STATE(State_Start,'C');
}

static void State_Start(
    enum TAG_SUSI_EVIDS id, uintptr_t data1, uintptr_t unused)
{
    char c = (char)data1;

    switch(id)
    {
        case EV_SUSI_RXDATA:
            if(c == ASCII_CR)
            {
                SignOn();
                SUSI_Transition_State_Interactive();
            }
            break;

        default:
            break;
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Stay in this state to receive console characters and
    // process commands.

void SUSI_Transition_State_Interactive(void)
{
    SET_STATE(State_Interactive,'A');
}

static void State_Interactive(
    enum TAG_SUSI_EVIDS id, uintptr_t data1, uintptr_t unused)
{
    char c = (char)data1;

    switch(id)
    {
        case EV_SUSI_RXDATA:
            if(c == ASCII_EOT)
            {
                TxUserConsole(ASCII_ACK);
                Con_Prompt();
            }
            else
            {
                Con_CharacterSink(c);
            }
            break;

        case EV_SUSI_SIGNON:
            SignOn();
            break;

        default:
            break;
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /// Send sign-on message to console.

static void SignOn(void)
{
    ANSI_ClearScreen(STDOUT);
    ANSI_MoveCursor(STDOUT, 0, 0);
    APPrint(STDOUT, "PSE-702 Operator Console: %s\n\r", signon);
    if(EEP_GetPwrUpChecksumWasOK())
        APPrint(STDOUT,
            "EEPROM Checksum was OK on power-up.\n\n\r");
    else
        APPrint(STDOUT,
            "EEPROM Checksum was bad on power-up. "
            "Defaults were loaded.\n\n\r");
    Con_Prompt();
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

/// @todo Write susi xmodem functions.  Unitl they are written,
/// make a lint exception for the following functions which appear
/// to be unreferenced.

//lint -esym(528, Slot_XMOD_Complete)
//lint -esym(528, Transition_State_XModemReceive)

void SUSI_Transition_State_XModemReceive(void)
{
    SET_STATE(State_XModemReceive,'B');
}

static void State_XModemReceive(
    enum TAG_SUSI_EVIDS id, uintptr_t data1, uintptr_t unused)
{
    XMODM_StateMachine(id, data1);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void Slot_XMOD_Complete(
        uintptr_t unused1, uintptr_t unused2)
{
    Unsubscribe_XMODM_Signal_Done();
    SUSI_Transition_State_Interactive();
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Set up and change state to Voice Record.
    // Could be incorporated into the transition function below
    // but requires a parameter.  Called by susi_sys_cmds.c.

void SUSI_StartVoiceRecord(uint8_t file)
{
    recordFile = file;
    SUSI_Transition_State_VoiceRecord();
}


    // Wait in this state for an initial CR from the user console
    // before signing on.

void SUSI_Transition_State_VoiceRecord(void)
{
    APPrint(STDOUT, " - Recording armed...");
    SET_STATE(State_VoiceRecord,'E');
}

static void State_VoiceRecord(
    enum TAG_SUSI_EVIDS id, uintptr_t data1, uintptr_t unused)
{
    switch(id)
    {
        case EV_SUSI_RXDATA:
            // If the user presses ESC, cancel the recording.
            if((char)data1 == ASCII_ESC)
            {
                CEN_CancelRecording();
                APPrint(STDOUT, "canceled by operator.");
                SUSI_Transition_State_Interactive();
                APPrint_CRLF(STDOUT);
                Con_Prompt();
            }
            break;

        case EV_SUSI_RECORD_STARTED:
            APPrint(STDOUT, "started...");
            break;

        // We get this event when central ends the recording.
        case EV_SUSI_RECORD_COMPLETE:
            switch(data1)
            {
                case SUSI_REC_STS_OK:
                    APPrint(STDOUT, "recording complete.");

                    APPrint(STDOUT,
                        "\n\rFile duration: %.1f sec.",
                         FileTime(recordFile));

                    APPrint(STDOUT,
                        "\n\rTime left: %.1f sec.", FileSpaceLeft());
                    break;

                case SUSI_REC_STS_OVF:
                    APPrint(STDOUT,
                        "stopped, out of memory.");
                    break;

                case SUSI_REC_STS_BUSY:
                    APPrint(STDOUT,
                        "cannot record, channel busy.");
                    break;

                case SUSI_REC_STS_TIMEOUT:
                    APPrint(STDOUT, "timeout.");
                    break;

                default:
                    FATAL();        // Should never get here.
                    break;

            }
            SUSI_Transition_State_Interactive();
            APPrint_CRLF(STDOUT);
            Con_Prompt();
            break;

        default:
            break;
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//      Test Tone
//      This command requires a state change.
//
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Output a test tone to a transmitter channel.

        @clsyntax > tone {chan}

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) to which the
            transmitter level setting applies.

        @description This command outputs a 1000 Hz sine wave at a
            level corresponding to 3/5 full scale on the selected
            channel.  The command operates as a toggle.  When it
            is first issued, the tone is turned on.  When it is
            issued again, it is turned off.
    */

static uint8_t CmdTestTone(const char *pLine)
{
    uint32_t chan;
    unsigned charCount;
    unsigned parmCount;

    if((parmCount = APParse(pLine, &charCount, " %d|", &chan)) == 0)
    {
        // If the tone is on, turn it off.
        if(testToneOn)
        {
            // Turn all test tones off.
            FGen_ChannelOff(FGEN_TX1_TONE_CHAN);
            FGen_ChannelOff(FGEN_TX1_ENCODER_CHAN);
            TXSM_TurnPTTOff(TX_CHAN_1);
            FGen_ChannelOff(FGEN_TX2_TONE_CHAN);
            FGen_ChannelOff(FGEN_TX2_ENCODER_CHAN);
            TXSM_TurnPTTOff(TX_CHAN_2);
            CEN_TestToneOff();
            APPrint(STDOUT, " - Test tones off.");
            testToneOn = false;
            return(CON_STS_OK);
        }
        else
            return(CON_STS_SYNTAX);
    }

    else if(parmCount == 1 && VALID_TX_CHAN(ORDINAL_2_TX_CHAN(chan)))
    {
        if(!testToneOn)
        {
            // Convert user channel to enum.
            testToneChan = ORDINAL_2_TX_CHAN(chan);
            SUSI_Transition_State_TestToneRequest();
            return(CON_STS_STATECHG);
        }

        else
        {
            APPrint(STDOUT, " - A tone is already on. "
                                "Enter 'tone' to turn tone off.");
            return(CON_STS_OK);
        }
    }

    else
    {
        return(CON_STS_SYNTAX);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Stay in this state while the test tone is sounding.

void SUSI_Transition_State_TestToneRequest(void)
{
    CEN_RequestTestTone();
    SET_STATE(State_TestToneRequest,'D');
}

static void State_TestToneRequest(
    enum TAG_SUSI_EVIDS id, uintptr_t unused1, uintptr_t unused2)
{
    switch(id)
    {
        case EV_SUSI_TEST_TONE_DENIED:
            APPrint(STDOUT,
                " - Cannot outupt test tone while Tx or Rx active.");
            SUSI_Transition_State_Interactive();
            APPrint_CRLF(STDOUT);
            Con_Prompt();
            break;

        case EV_SUSI_TEST_TONE_GRANTED:
            if(testToneChan == TX_CHAN_1)
            {
                FGen_SetChannelFreq(
                    FGEN_TX1_TONE_CHAN, 10000);  // 1000 Hz in 10ths
                FGen_SetChannelGain(
                    FGEN_TX1_TONE_CHAN, GAIN_3000HZ_DEVIATION);
                FGen_ChannelOn(FGEN_TX1_TONE_CHAN);

                if(EEP_GetTxFlags(TX_CHAN_1, EEP_TX_FLAG_SATE))
                                FGen_ChannelOn(FGEN_TX1_ENCODER_CHAN);

                TXSM_TurnPTTOn(TX_CHAN_1);
                APPrint(STDOUT, " - Tone 1 on. OK");
                testToneOn = true;
            }

            else if(testToneChan == TX_CHAN_2)
            {
                FGen_SetChannelFreq(
                    FGEN_TX2_TONE_CHAN, 10000);  // 1000 Hz in 10ths
                FGen_SetChannelGain(
                    FGEN_TX2_TONE_CHAN, GAIN_3000HZ_DEVIATION);
                FGen_ChannelOn(FGEN_TX2_TONE_CHAN);

                if(EEP_GetTxFlags(TX_CHAN_2, EEP_TX_FLAG_SATE))
                                FGen_ChannelOn(FGEN_TX2_ENCODER_CHAN);

                TXSM_TurnPTTOn(TX_CHAN_2);
                APPrint(STDOUT, " - Tone 2 on. OK");
                testToneOn = true;
            }

            else
                FATAL();

            SUSI_Transition_State_Interactive();
            APPrint_CRLF(STDOUT);
            Con_Prompt();
            break;

        default:
            break;
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Print parameter status to the console with pagination.

void Transition_State_StatusList(void)
{
    APPrint_CRLF(STDOUT);
    PrintScreenfulOfStatus(true);   //lint !e534 ignoring return
    SET_STATE(State_StatusList,'F');
}

static void State_StatusList(
    enum TAG_SUSI_EVIDS id, uintptr_t data1, uintptr_t unused2)
{
    char c = (char)data1;

    switch(id)
    {
        case EV_SUSI_RXDATA:
            // If the user presses ESC, cancel the recording.
            if(c == ASCII_ESC)
            {
                SUSI_Transition_State_Interactive();
                APPrint_CRLF(STDOUT);
                APPrint_CRLF(STDOUT);
                Con_Prompt();
            }

            else if(c == '\r' || c == ' ')
            {
                if(!PrintScreenfulOfStatus(false))
                {
                    SUSI_Transition_State_Interactive();
                    APPrint_CRLF(STDOUT);
                    APPrint_CRLF(STDOUT);
                    Con_Prompt();
                }
            }
            break;


        default:
            break;
    }
}

    // Print one screenful of status.  Argument 'first' should
    // be true to send the first screenful.
    // Returns true if there are more screens to print, or
    // false after the last screen prints.

static bool PrintScreenfulOfStatus(bool first)
{
    static int section = 0;

    if(first)
    {
        APPrint_String(STDOUT,
            "\n\rTimers, ID String, Courtesy Beeps:\n\n\r");
        section = 0;
        SUSI_StatusSetTimers();
        SUSI_StatusSetIDString();
        SUSI_StatusSetBeepTone();
        SUSI_StatusSetBeepDelay();
        SUSI_StatusSetBeepMode();
        SUSI_StatusSetBeepCharacter();
        SUSI_StatusSetBeepLevel();
    }
    else
    {
        switch(section)
        {
            case 1:
                APPrint_String(STDOUT,
                    "\n\rPaths, Announcements, and Voice ID:\n\n\r");
                SUSI_StatusSetRxTxPath();
                SUSI_StatusScheduleAnno();
                SUSI_StatusEnableAnno();
                SUSI_StatusVoiceID();
                break;

            case 2:
                APPrint_String(STDOUT,
                    "\n\rMorse and Touch-Tone Parameters, "
                    "Rx and Tx Flags:\n\n\r");
                SUSI_StatusSetMorseRate();
                SUSI_StatusSetMorseLevel();
                SUSI_StatusSetDTMFPwd();
                SUSI_StatusEnableDTMFPassword();
                SUSI_StatusSetDTMFCommandChar();
                SUSI_StatusSetTxFlags();
                SUSI_StatusSetRxFlags();
                break;

            case 3:
                APPrint_String(STDOUT,
                    "\n\rRx and Tx Levels, CTCSS Parameters:\n\n\r");
                SUSI_StatusSetRxLevel();
                SUSI_StatusSetTxLevel();
                SUSI_StatusCTCSSDecodeIndex();
                SUSI_StatusCTCSSEncodeIndex();
                SUSI_StatusCTCSSEncodeMode();
                return(false);

            default:
                FATAL();
        }
    }

    section++;
    APPrint_String(STDOUT,
         "Press 'Enter' for next screenful, or 'Esc' to quit.\n\r");
    return(true);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Print parameter status to the console with pagination.

void Transition_State_HelpList(void)
{
    PrintScreenfulOfHelp(true);       //lint !e534 ignoring return
    SET_STATE(State_HelpList,'G');
}

static void State_HelpList(
    enum TAG_SUSI_EVIDS id, uintptr_t data1, uintptr_t unused2)
{
    char c = (char)data1;

    switch(id)
    {
        case EV_SUSI_RXDATA:
            // If the user presses ESC, cancel the recording.
            if(c == ASCII_ESC)
            {
                SUSI_Transition_State_Interactive();
                APPrint_CRLF(STDOUT);
                APPrint_CRLF(STDOUT);
                Con_Prompt();
            }

            else if(c == '\r' || c == ' ')
            {
                if(!PrintScreenfulOfHelp(false))
                {
                    SUSI_Transition_State_Interactive();
                    APPrint_CRLF(STDOUT);
                    APPrint_CRLF(STDOUT);
                    Con_Prompt();
                }
            }
            break;


        default:
            break;
    }
}

    // Print one screenful of help.  Argument 'first' should
    // be true to send the first screenful.
    // Returns true if there are more screens to print, or
    // false after the last screen prints.

static bool PrintScreenfulOfHelp(bool first)
{
    static uint8_t cmdIndex = 0;
    static uint8_t page = 0;
    unsigned lineCount = 0;
    char *pName, *pDescription;

    if(first)
    {
        cmdIndex = 0;
        page = 1;
    }

    APPrint(STDOUT,
        "\n\rPSE 702 terminal interface commands, page %d.\n\r",
        page++);
    APPrint_String(STDOUT, "\n\r Name\tDescription");
    APPrint_String(STDOUT, "\n\r ====\t===========\n\r");

    while(lineCount<15)
    {
        if(susiCmdTable[cmdIndex].pName != NULL)
        {
            if(susiCmdTable[cmdIndex].pDescription != NULL)
            {
                pName = susiCmdTable[cmdIndex].pName;
                pDescription = susiCmdTable[cmdIndex].pDescription;
                APPrint(STDOUT, "\n\r %s\t%s", pName, pDescription);
                lineCount++;
            }
            cmdIndex++;
        }
        else
            return(false);  // Last page full.
    }

    APPrint_String(STDOUT,
        "\n\n\rPress 'Enter' for next screenful, "
        "or 'Esc' to quit.\n\r");
    return(true);
}

//****************************************************************
//                    COMMAND LINE FUNCTIONS:
/// @defgroup susi Console Operator Commands
/// @{

uint16_t SUSI_AssembleFlagWord(
                            uint16_t flags, bool value, uint16_t bit)
{
    if(value)
        flags |= bit;
    else
        flags &= ~bit;

    return(flags);
}


    // Calculate how much time a recording took,

static unsigned FileTime(unsigned fileNum)
{
    unsigned numPages;  // Count of pages allocated to file etc.
    unsigned time;      // Seconds duration of file etc.

    // Calculate how much time the recording took,
    // and how much time is left.
    numPages = FSys_FileSize(fileNum);

    // File time in 10ths of seconds
    time = (numPages * Z_FLASH_PAGE * 10)
        /((unsigned)(CDX_SAMPLE_RATE_HZ + 0.5));

    return(time);
}


    // and how much time is left.
static unsigned FileSpaceLeft(void)
{
    unsigned numFree;   // Count of unallocated pages.
    unsigned time;      // Seconds duration of file etc.

    numFree = FSys_FreePages();
    time = (numFree * Z_FLASH_PAGE * 10)
        /((unsigned)(CDX_SAMPLE_RATE_HZ + 0.5));
    return(time);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Print Status

        @clsyntax \> status

        @par Command Line Parameters: None.

        @description This command prints the current settings for
            all parameters to the operator console, and breaks
            for each page making it easier to examine the status
            than the dump command, which is best used to capture
            parameter status for storage and playback.
    */

static uint8_t CmdStatus(const char *pLine)
{

    // Here, if the line is empty, print current parameter
    // status.

    if(*pLine != '\0')
        return(CON_STS_SYNTAX);
    else
    {
        Transition_State_StatusList();
        return(CON_STS_STATECHG);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Dump Parameters

        @clsyntax \> dump

        @par Command Line Parameters: None.

        @description This command prints the current settings for
            all parameters to the operator console without
            stopping for page breaks.  The printout
            is formatted so that if it were played back into the
            PSE-702, it would re-enter all the parameters.
            This command is primarilly used for capturing
            parameter values for later playback.
    */

static uint8_t CmdDump(const char *pLine)
{
    // Here, if the line is empty, print current parameter
    // status.

    if(*pLine != '\0')
        return(CON_STS_SYNTAX);
    else
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetTimers();
        SUSI_StatusSetIDString();

        SUSI_StatusSetBeepTone();
        SUSI_StatusSetBeepDelay();
        SUSI_StatusSetBeepMode();
        SUSI_StatusSetBeepCharacter();
        SUSI_StatusSetBeepLevel();
        SUSI_StatusSetMorseRate();
        SUSI_StatusSetMorseLevel();

        SUSI_StatusSetDTMFPwd();
        SUSI_StatusEnableDTMFPassword();
        SUSI_StatusSetDTMFCommandChar();

        SUSI_StatusSetTxFlags();
        SUSI_StatusSetRxFlags();
        SUSI_StatusSetRxLevel();
        SUSI_StatusSetTxLevel();

        SUSI_StatusCTCSSDecodeIndex();
        SUSI_StatusCTCSSEncodeIndex();
        SUSI_StatusCTCSSEncodeMode();

        SUSI_StatusSetRxTxPath();
        SUSI_StatusVoiceID();
        SUSI_StatusScheduleAnno();
        SUSI_StatusEnableAnno();
        APPrint_CRLF(STDOUT);
        return(CON_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Show system help

        @clsyntax > help [command]

        @par Command Line Parameters:
            @oo [command]: The name of a 702 serial user interface
            command. (Optional parameter.)

        @description This command displays a list of the commands that
            can be used to configure the 702.
*/

static uint8_t CmdShowHelp(const char *pLine)
{
    char *pSyntax, *pLongSyntax, *pDescription;
    const console_cmd_tbl_t *pCmdTbl;
    unsigned charCount;
    bool found;
    char commandName[Z_WORD];

    if(*pLine == '\0')  // Show help for all 702 user commands
    {
        Transition_State_HelpList();
        return(CON_STS_STATECHG);
    }

    // Check parameters
    else if(APParse(pLine, &charCount,
                      " " GEN_FORMAT(w|, Z_WORD), commandName) != 1)
    {
        return(CON_STS_SYNTAX);
    }

    else                    // Look up help for the specified command
    {
        pCmdTbl = susiCmdTable;

        found = false;

        while((pCmdTbl->pName != NULL) && !found)
        {
            if(strcmp(commandName, pCmdTbl->pName) == 0)
                found = true;
            else
                pCmdTbl++;
        }

        if(found)
        {
            // Command not available to users:
            if(pCmdTbl->pDescription == NULL)
            {
                APPrint(STDOUT, "\r\n - Command not found.");
                return(CON_STS_NULL);
            }

            // This is what we do for non-command help, like the
            // tone table:
            else if(pCmdTbl->pCommand == NULL)
            {
                APPrint(STDOUT, "%s", pCmdTbl->pLongSyntax);
            }

            else
            {
                pDescription = pCmdTbl->pDescription;
                pSyntax = pCmdTbl->pSyntax;
                pLongSyntax = pCmdTbl->pLongSyntax;
                APPrint(STDOUT, "\n\r%s\n\rSyntax:\n\r\t> %s",
                                                pDescription, pSyntax);
                APPrint(STDOUT, "%s",pLongSyntax);
                APPrint(STDOUT, "\n\rParameters in {} are required; "
                    "parameters in [] are optional.\n\r");
            }
        }

        else
            APPrint(STDOUT, "\r\n - Command not found.");

        return(CON_STS_NULL);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Display Command Names

        @clsyntax > ?

        @par Command Line Parameters: None.

        @description This command displays a list of all command
        names on the console.
*/

static uint8_t CmdShowCommandNames(const char *pLine)
{
    const console_cmd_tbl_t *pCmdTbl;
    unsigned i;

    if(*pLine == '\0')  // Show names for all commands
    {
        APPrint(STDOUT,
                "\n\rPSE 702 terminal interface command names:");
        APPrint_CRLF(STDOUT);

        pCmdTbl = susiCmdTable;
        i = 0;

        while(pCmdTbl->pName != NULL)
        {
            // If the description is null, hide the name.
            if(pCmdTbl->pDescription != NULL)
            {
                if((i++ & 0x07) == 0)
                    APPrint_CRLF(STDOUT);
                else
                    APPrint_String(STDOUT, "\t");

                APPrint(STDOUT, "%s", pCmdTbl->pName);
            }
            pCmdTbl ++;
        }
        APPrint_CRLF(STDOUT); APPrint_CRLF(STDOUT);
        return(CON_STS_OK);
    }
    else
        return(CON_STS_SYNTAX);

}

#if 0
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Receive a File Via XModem.

        @clsyntax xmodem {fileNum}

        @par Command Line Parameters:
            @oo {fileNum} A file number (integer from 0 to 7
                inclusive) identifying the flash file into which
                the user's file will be stored.

        @description Receives a user file via XModem transfer
            protocol, and saves it in the flash.

        @ingroup ttcon_file
    */

static uint8_t SUSI_CmdXModem(const char *pLine)
{
    uint32_t fileNum;
    unsigned charCount;

    if(APParse(pLine, &charCount, " %d|", &fileNum) != 1)
        return(CON_STS_SYNTAX);

    else if(fileNum >= NUM_FILES)
        return(CON_STS_RANGE);

    else
    {
        APPrint(STDOUT, " - Start XModem file transfer...");
        Transition_State_Xmodem();
        return(CON_STS_STATECHG); // Console can skip the prompt.
    }
}

static void ReceiveXModemBuffer(const char *pBuf)
{

}   //lint !e715

static void Transition_State_Xmodem(void)
{
    Subscribe_XMODM_Signal_Done(
                        (slot_t *)Slot_XMODM_Done);
    XMODM_StartSession(ReceiveXModemBuffer);

    SET_STATE(State_Xmodem,'A');
}

static void State_Xmodem(
    unsigned event, uintptr_t data1, uintptr_t unused)
{
    XMODM_StateMachine(event, data1);
}

static void Slot_XMODM_Done(uintptr_t unused1, uintptr_t unused2)
{
    XMODM_UnsubscribeAllSignals();
    Transition_State_Interactive();
}
#endif

/// @}
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                          END
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
