//***************************************************************
/**
    @file susi_dbg_cmds.c

    Serial User Interface Debug Commands

    @author nlh/alp

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.
*/
//***************************************************************

//    *** INCLUDE FILES ***

#include <stdint.h>         //lint -esym(766, stdint.h)
#include <stdbool.h>

#include "lpc213x.h"        //lint -esym(766, lpc213x.h)
#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)

#include "aplib.h"
#include "target.h"         //lint -esym(766, target.h)
#include "apex.h"

#include "app.h"
#include "apparse.h"
#include "apprint.h"
#include "io_main.h"
#include "io_flash.h"
//#include "io_codec.h"
#include "astream.h"

#include "console.h"
#include "central.h"
#include "io_debug.h"       //lint -esym(766, io_debug.h)
//#include "ttcon.h"
//#include "io_misc.h"
#include "filesys.h"
#include "dtmfdet.h"
#include "io_spi.h"
#include "ssm_idsm.h"
#include "voice.h"
#include "filter.h"
#include "ssm_squelch.h"
#include "new_dusi.h"
#include "ssm_morse.h"
#include "ssm_satd.h"
#include "ssm_sate.h"
#include "eep.h"
#include "audiosm.h"
//#include "susi.h"
//#include "xmodem.h"
//#include "signon.h"
//#include "ansi_term.h"
//#include "fgen.h"
#include "ssm_txsm.h"

/***************************************************************/
//    *** MACROS ****

/***************************************************************/
//    *** GENERAL DEFINITIONS ****

/***************************************************************/

//    *** STATIC FUNCTION PROTOTYPES ****

static void Bootloader(void);

const struct TAG_VARIABLE_TABLE
                            *LookupModule(const char *pString);
const struct TAG_VARIABLE_TABLE
                            *LookupVariable(const char *pString);

/***************************************************************/

//    *** DATA STRUCTURE Definitions ***

/***************************************************************/
//    *** CONSTANTS ***


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Module Table

        This table contains pointers to the variable tables for
        each module that can be monitored in real-time.
    */
static const struct TAG_MODULE_TABLE
{
    const char *pName;                       // Name of module
    const struct TAG_VARIABLE_TABLE *pTable; // Module's var table

} moduleTable[] =
{
    {"apex", apexVarTab},       ///< Apex RTOS
    {"dtmf", dtmfVarTab},       ///< DTMF module: dtmf
    {"filesys", filesysVarTab}, ///< File System module: filesys
    {"con", consoleVarTab},     ///< Console module: con
    {"flash", flashVarTab},     ///< Flash I/O module: flash
    {"spi", spiVarTab},         ///< SPI I/O module: spi
    {"txsm", txsmVarTab},       ///< Tx State Machine: txsm
    {"sate", sateVarTab},       ///< CTCSS Encoder module: sate
    {"satd", satdVarTab},       ///< CTCSS Decoder module: satd
    {"morse", morseVarTab},     ///< Morse Encoder module: morse
    {"idsm", idsmVarTab},       ///< ID State Machine: idsm
    {"cen", cenVarTab},         ///< Central Control module: cen
    {"aud", audVarTab},         ///< Audio State Machine: aud
    {"voice", voiceVarTab},     ///< Voice Store & Forward: voice
    {"filt", filtVarTab},       ///< Digital Filter module: filt
    {"squ", squelchVarTab},     ///< Squelch State Machine: squ
    {"dusi", dusiVarTab},       ///< DTMF User Interface: dusi

    {NULL, NULL}                        // End of table
};

static const struct TAG_VARIABLE_TABLE nullVarTable[] =
{
    {NULL, NULL, VAR_TYPE_NONE, 0}
};


//****************************************************************
//                 *** VARIABLES ***
//****************************************************************

/// Buffer for flash reads/writes
static uint8_t flashBuf[Z_FLASH_PAGE];

/// pVarTable points to the module's variable table.

static const struct TAG_VARIABLE_TABLE *pVarTable;


//***************************************************************
//                     Public API Functions
//***************************************************************

//****************************************************************

void SUSI_DebugReset(void)
{
    pVarTable = nullVarTable;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/// @defgroup ttcon_flash Flash Memory Commands
/// @ingroup ttcon
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    Display all currently active paths in real time
    (for debugging).

    @clsyntax \> apath

    @par Command Line Parameters: None.

    @description This command displays all active paths.
*/

uint8_t SUSI_CmdDisplayActivePaths(const char *pLine)
{
    LINK *pNext;     // Pointer to next path in list.
    astream_path_t *pPath;
    astream_destsig_t i;

    // The rest of the line should be empty.

    if(*pLine != '\0')
        return(CON_STS_SYNTAX);

    else
    {
        // Go through all destinations and clear their path lists.
        for(i=ASTREAM_DSIG_TX1; i<ASTREAM_MAX_DSIG; i++)
        {
            pNext = ASTREAM_GetDestPathList(i)->pFLink;

            // Loop through the paths and remove inactive ones.
            while((pPath = pNext->pObject) != NULL)
            {
                APPrint(STDOUT, "\n\r\t%d %d %d",
                        pPath->id, pPath->srcSig, pPath->destSig);
                // Now look at the next path on the list:
                pNext = pPath->link.pFLink;
            }
        }

        APPrint_CRLF(STDOUT);
        return(CON_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Fill Flash Buffer.

        @clsyntax > ff {pattern} {inc}

        @par Command Line Parameters:
            @oo {pattern}: A hex integer from 0x00 to 0xFF
            containing the pattern to be written to the
            flash buffer.
            @oo {inc}: An optional decimal integer from 0 to 255
            containing an amount by which the pattern will be
            incremented.

        @description This command writes consecutive locations of the
            flash buffer with the indicated pattern, incrementing it
            each time with the {inc} value. If {inc} is not specified,
            the pattern is repeated identically in each flash buffer
            location

        @ingroup ttcon_flash
    */
uint8_t SUSI_CmdFillFlashBuf(const char *pLine)
{
    uint32_t parmCount;
    unsigned charCount;
    uint32_t pattern, inc;
    uint32_t i;

    parmCount = APParse(
        pLine, &charCount, " 0x%2x %d|", &pattern, &inc);

    if(parmCount == 1)
    {
        if(pattern > 255)
            return(CON_STS_RANGE);
        else
        {
            for(i=0; i<256; i++)
                flashBuf[i] = (uint8_t)pattern;

            return(CON_STS_OK);
        }
    }

    else if(parmCount == 2)
    {
        if(pattern > 255 || inc > 255)
            return(CON_STS_RANGE);
        else
        {
            for(i=0; i<256; i++, pattern+=inc)
                flashBuf[i] = (uint8_t)pattern;

            return(CON_STS_OK);
        }
    }

    else
        return(CON_STS_SYNTAX);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Write Flash Page

        @clsyntax > fwrp {page}

        @par Command Line Parameters:
            @oo {page}: A hex integer from 0x0000 to 0xFFFF
            containing the physical page number to be written.

        @description This command writes the contents of the flash
            buffer to the flash memory at the page number indicated
            by page.

        @note: In the case of the Atmel AT26F004 flash device, only
            2048 (800 Hex) pages are available. Page addresses
            greater than 7FF hex will wrap around.

        @ingroup ttcon_flash
    */
uint8_t SUSI_CmdWriteFlashPage(const char *pLine)
{
    unsigned charCount;
    uint32_t addr;

    if(APParse(
        pLine, &charCount, " 0x%4x|", &addr) != 1)
        return(CON_STS_SYNTAX);

    else
    {
        if(addr >= NUM_PAGES)
            return(CON_STS_RANGE);

        else
        {
            Flash_WritePage(addr, flashBuf);
            return(CON_STS_OK);
        }
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Read Flash Page

        @clsyntax > frdp {page}

        @par Command Line Parameters:
            @oo {page}: A hex integer from 0x0000 to 0xFFFF
            containing the physical page number to be read.

        @description This command reads the flash page addressed by
            page and places the contents into the flash buffer.

        @note In the case of the Atmel AT26F004 flash device, only
            2048 (800 Hex) pages are available. Page addresses
            greater than 7FF hex will wrap around.

        @ingroup ttcon_flash
    */
uint8_t SUSI_CmdReadFlashPage(const char *pLine)
{
    unsigned charCount;
    uint32_t addr;

    if(APParse(pLine, &charCount, " 0x%4x|", &addr) != 1)
        return(CON_STS_SYNTAX);

    else
    {
        if(addr > 65535)
            return(CON_STS_RANGE);

        else
        {
            // Read flash into buffer.

            Flash_ReadPage((uint16_t)addr, flashBuf);
            return(CON_STS_OK);
        }
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Write Flash Byte

        @clsyntax > fwrb {addr}, {data}

        @par Command Line Parameters:
            @oo {addr}: A 6 digit (24 bit) hex integer containing
            the physical address of the uint8_t to be written.
            @oo {data}: A 2 digit (8 bit) hex integer containing
            the data to be written at that uint8_t address.

        @description This command writes the data contained in data
            to the flash at the address specified by addr.

        @note In the case of the Atmel AT26F004 flash device, only
            524,288 (80000 Hex) bytes are available.    Byte addresses
            greater than 7FFFF hex will wrap around.

        @ingroup ttcon_flash
    */
uint8_t SUSI_CmdWriteFlashByte(const char *pLine)
{
    unsigned charCount;
    uint32_t addr, data;
    uint8_t c;

    if(APParse(
        pLine, &charCount, " %6x 0x%2x|", &addr, &data) != 2)
        return(CON_STS_SYNTAX);

    else
    {
        if(addr >= Z_FLASH)
            return(CON_STS_RANGE);

        else
        {
            c = (uint8_t)data;
            Flash_WriteBytes(addr, &c, 1);
            return(CON_STS_OK);
        }
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Read flash byte.

        @clsyntax > frdb {addr}

        @par Command Line Parameters:
            @oo {addr}: A 6 digit (24 bit) hex integer containing
            the physical uint8_t number to be read.

        @description This command reads the uint8_t in flash addressed by
            addr and prints the contents to the monitor.

        @note In the case of the Atmel AT26F004 flash device, only
            524,288 (80000 Hex) bytes are available.    Byte addresses
            greater than 7FFFF hex will wrap around.

        @ingroup ttcon_flash
    */
uint8_t SUSI_CmdReadFlashByte(const char *pLine)
{
    unsigned charCount;
    uint32_t addr;
    uint8_t data;

    if(APParse(pLine, &charCount, " 0x%6x|", &addr) != 1)
        return(CON_STS_SYNTAX);

    else
    {
        if(addr >= Z_FLASH)
            return(CON_STS_RANGE);

        else
        {
            // Read flash into buffer.

            Flash_ReadBytes(addr, &data, 1);
            APPrint(STDOUT, " %xb", data);
            return(CON_STS_OK);
        }
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Read flash serial number.

        @clsyntax > frsn

        @par Command Line Parameters:
            @oo None.

        @description This command reads the serial number string
        in the flash security register and puts it in the flash
        buffer.

        @ingroup ttcon_flash
    */
uint8_t SUSI_CmdReadFlashSerialNumber(const char *pLine)
{
    if(*pLine != '\0')
        return(CON_STS_SYNTAX);

    else
    {
        // Read flash serial number into buffer.

        Flash_ReadSerialString(flashBuf);
        return(CON_STS_OK);
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Display contents of flash buffer on console.

        @clsyntax > fd

        @par Command Line Parameters: None.

        @description This command displays the contents of the flash
            buffer on the console.  The flash buffer is a 256 byte
            buffer that is used to contain data read from or written
            to the flash.

        @ingroup ttcon_flash
    */
uint8_t SUSI_CmdDisplayFlash(const char *pLine)
{
    if(*pLine != '\0')
        return(CON_STS_SYNTAX);

    IO_DisplayMemoryBytes(STDOUT, flashBuf, 0, 256);

    return(CON_STS_OK);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Read the flash status register.

        @clsyntax > frs

        @par Command Line Parameters: None.

        @description This command reads and displays the 16
            bit contents of the flash status register.

        @ingroup ttcon_flash
    */
uint8_t SUSI_CmdReadFlashStatus(const char *pLine)
{
    uint8_t status[Z_FLASH_STATUS];
    unsigned i;

    if(*pLine != '\0')
        return(CON_STS_SYNTAX);

    else
    {
        Flash_ReadStatusRegister(status, Z_FLASH_STATUS);
        APPrint(STDOUT, " 0x");
        for(i=0; i<Z_FLASH_STATUS; i++)
          APPrint(STDOUT, "%xb", status[i]);

        return(CON_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Write Flash Status Register.

        @clsyntax > fws {data}

        @par Command Line Parameters:
            @oo {data}: A 16 bit hexadecimal value to write to
            the flash status register.  The low order 8 bits
            are written to SR1, the next 8 are written to
            SR2, etc.

        @description This command writes the value
            contained in the data to the flash status register.
            The data must contain exactly the number of bytes
            (expressed in hexadecimal) needed to write the
            status register(s).

        @ingroup ttcon_flash
    */

uint8_t SUSI_CmdWriteFlashStatus(const char *pLine)
{
    uint32_t data;
    unsigned charCount;
    flash_status_t status;

    if(APParse(pLine, &charCount, " 0x%x|", &data) != 1)
        return(CON_STS_SYNTAX);

    else if(charCount != (Z_FLASH_STATUS*2))
        return(CON_STS_SYNTAX);

    else
    {
          status = (flash_status_t)data;
        Flash_WriteStatusRegister(&status);
        return(CON_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Erase a flash block.

        @clsyntax > fera {block_addr}

        @par Command Line Parameters:
            @oo {block_addr}: A hexadecimal block address indicating
            which block to erase.

        @description Erases the addressed block.    An erased block
            will contain FF in every uint8_t position. Blocks are 4096
            (1000 Hex) bytes long.  Block addresses start with 0
            and increment by one for each block from the beginning
            of flash to the end.

        @ingroup ttcon_flash
    */

uint8_t SUSI_CmdEraseFlash(const char *pLine)
{
    unsigned charCount;
    uint32_t addr;

    if(APParse(
        pLine, &charCount, " 0x%3x|", &addr) != 1)
        return(CON_STS_SYNTAX);

    else
    {
        if(addr >= NUM_BLOCKS)
            return(CON_STS_RANGE);

        else
        {
            // Erase flash sector.

            Flash_EraseBlock(addr);
            return(CON_STS_OK);
        }
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Unprotect the flash.

        @clsyntax > fup

        @par Command Line Parameters: None.

        @description The flash contains internal write-protection
            bits that may be used to prevent writes to various flash
            locations.  When the flash is powered up, all write-
            protect bits are set and no flash locations may be written.
            This function clears all write-protect bits to enable
            writing to all flash locations.

        @note The flash write-protect bits are automatically cleared by
            the PSE-702 on power-up so calling this function seperately
            is not normally necessary.  It is included for test and
            troubleshooting purposes only.

            The AT25DF321A flash contains 64 sectors of 65536 bytes
            each.

        @ingroup ttcon_flash
    */

uint8_t SUSI_CmdUnprotectFlash(const char *pLine)
{
    if(*pLine != '\0')
        return(CON_STS_SYNTAX);

    else
    {
        Flash_UnprotectChip();
        return(CON_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/// @defgroup ttcon_fsys File System Commands
/// @ingroup ttcon
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Read a page from the file into the flash buffer.

        @clsyntax fread {fileNum} {pageNum}

        @par Command Line Parameters:
            @oo {fileNum}: A decimal integer specifying the file
            number of the file to read.

            @oo {pageNum}: A hexadecimal integer logical page
            number, beginning at 0 for the first page in the file,
            and incrementing by 1 for successive pages.

        @description Reads the addressed logical page from the addressed
        file, into the flash buffer.    Pages may be read in any order.

        @par
        An error message will be given if an attempt is made to read
        beyond the end of the file.

        @see CmdFileAppend, CmdFileClose, CmdFileDelete,
        CmdFileFormat, CmdFileOpen

        @ingroup ttcon_fsys
    */

uint8_t SUSI_CmdFileRead(const char *pLine)
{
    uint32_t fileNum;           // Number of file to append to.
    uint32_t pageNum;
    unsigned charCount;

    if(APParse(
        pLine, &charCount, " %d 0x%4x|", &fileNum, &pageNum) != 2)
        return(CON_STS_SYNTAX);

    else if(fileNum >= NUM_FILES)
        return(CON_STS_RANGE);

    else if(pageNum >= NUM_PAGES)
        return(CON_STS_RANGE);

    else
    {
        if(FSys_ReadPage(fileNum, pageNum, flashBuf) == false)
            APPrint(STDOUT,
                " - Unable to read. Page not allocated to file.");

        return(CON_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Append a page to a file.

        @clsyntax fappend {fileNum}, {count}

        @par Command Line Parameters:
            @oo {fileNum}: The number of the file to append to.
            @oo {count}: An optional integer number of pages to append.

        @description Appends the contents of the flash buffer to
        the currently open file.    The buffer is appended the
        number of times given by the optional count parameter. If
        count is omitted, the buffer shall be appended once.

        @par
        An error message is given if the write was not successful
        because the flash was full.

        @see CmdFileClose, CmdFileDelete,
        CmdFileFormat, CmdFileOpen, CmdFileRead

        @ingroup ttcon_fsys
    */

uint8_t SUSI_CmdFileAppend(const char *pLine)
{
    uint32_t fileNum;           // Number of file to append to.
    uint32_t itterCount;        // Counts itterations of append().
    unsigned charCount;
    uint32_t parmCount;

    parmCount = APParse(
            pLine, &charCount, " %d %d|", &fileNum, &itterCount);

    if(parmCount < 1 || parmCount > 2)
        return(CON_STS_SYNTAX);

    if(fileNum >= NUM_FILES)
        return(CON_STS_RANGE);

    else
    {
        if(parmCount == 1)
            itterCount = 1;

        while(itterCount > 0)
        {
            if(FSys_AppendPage(fileNum, flashBuf) == false)
            {
                APPrint(STDOUT, "Unable to append. Memory full.");
                break;
            }

            itterCount--;
        }

        return(CON_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/// @defgroup ttcon_rtc Real-Time Clock Commands
/// @ingroup ttcon
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Set the real-time clock.

        @clsyntax rtcset {min} {hr} {dom} {month} {year}

        @par Command Line Parameters:
            @oo {min}: An integer from 0 to 59 indicating the number of
            minutes to set in the clock.

            @oo {hr}: An integer from 0 to 59 indicating the number of
            hours to set in the clock.

            @oo {dom}: An integer from 0 to 31 indicating the day of
            the month to set in the clock.

            @oo {month}: An integer from 1 to 12 indicating the month
            to set in the clock.

            @oo {year}: An integer from 2000 to 2099 indicating the
            year to set in the clock.

        @description Sets the real-time clock to the date and time
        indicated in the parameters.

        @see CmdReadRTC

        @ingroup ttcon_rtc
    */

uint8_t SUSI_CmdSetRTC(const char *pLine)
{
    uint32_t min, hr, dom, month, year;
    unsigned charCount;

    if(APParse(pLine, &charCount, " %d %d %d %d %d|",
        &min, &hr, &dom, &month, &year) != 5)
        return(CON_STS_SYNTAX);

    else if((min >= 60) || (hr >= 60) || (dom > 31) ||
        (month == 0) || (month > 11) || (year < 2000) ||
        (year > 2099))
        return(CON_STS_RANGE);

    else
    {
        IO_RTC_SetTime((uint8_t)min, (uint8_t)hr,
            (uint8_t) dom, (uint8_t)month, (uint16_t)year);

        return(CON_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Read the contents of the real-time clock.

        @clsyntax rtcread

        @par Command Line Parameters:
            @oo None.

        @description Reads the contents of the real-time clock and
        prints the date and time to the console.

        @see CmdSetRTC

        @ingroup ttcon_rtc
    */

uint8_t SUSI_CmdReadRTC(const char *pLine)
{
    uint8_t min, hr, dom, month;
    uint16_t year;

    if(*pLine != '\0')
        return(CON_STS_SYNTAX);

    else
    {
        IO_RTC_ReadTime(&min, &hr, &dom, &month, &year);
        APPrint(
            STDOUT, " %d %d %d %d %d", min, hr, dom, month, year);

        return(CON_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//          DEBUG COMMANDS
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** @defgroup susi_debug Debug Commands

    These commands support debugging.  In order to use these
    commands, the variable DEBUG must be defined in susi and
    in all module called by these functions.  Usually this
    is done by placing the file in the DEBUG_OBJ line in the
    makefile.

    @ingroup ttcon
*/

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Prepare PSE-702 for downloading.

        @clsyntax dl

        @par Command Line Parameters:
            @oo None.

        @description Prepares the CPU to accept the download of new
        firmware to flash.  The download should occur on UART0, the
        9 pin D connector on the PSE-702 card edge.

        @see CmdReset

        @note After download is complete the PSE-702 does not seem
        to be able to reset itself, so a hard reset must be
        performed.

        @note Also, this function is available externally and may be
        called by xconsole.c in response to having received a DC2
        character.  This integrates the download command with Flash
        Magic software for Windows.

        @ingroup ttcon_admin
    */
uint8_t SUSI_CmdDownload(const char *pLine)
{
    if(*pLine != '\0')
        return(CON_STS_SYNTAX);

    else
    {
        Bootloader();
        return(CON_STS_OK); // There should not be a return from
                                    // the bootloader, so this line
                                    // shouldn't be executed.
    }
}


static void Bootloader(void)
{
    void (*bootloader_entry)(void);
    uint32_t temp;

    // Output the OK to the console so the user knows the
    // command was accepted, and wait for the OK to clear
    // the transmitter before continuing.

    APPrint(STDOUT, " OK\n\r");
    LOOP_TIL_true(U0LSR & 0x40);        // Wait until transmitter empty.

    // Invoke the bootloader: The following code was pasted
    // from NXP App Note # 10356.

    temp = PINSEL0;

    // Connect RXD0 & TXD0 pins to GPIO

    PINSEL0 = temp & 0xFFFFFFF3;

    // Select P0.14 as an output and P0.1 as an input

    temp = IODIR0;
    temp = temp | 0x4000;
    temp = temp & 0xFFFFFFFD;
    IODIR0 = temp;

    // Clear P0.14

    IOCLR0 = 0x4000;

    // Disable Interrupts in the VIC

    VICIntEnClr =
        BV(VIC_SPI0) | BV(VIC_SSP) |
        BV(VIC_TIMER0) | BV(VIC_UART0);

    // Set the VPB divider to 1/4 if your application
    // changes the VPBDIV value.    The bootloader is
    // hard-coded to use the reset value of VPBDIV
    // register

    VPBDIV = 0x0;

    // Restore reset state of Timer1

    T1PR = 0x0;
    T1MCR = 0x0;
    T1CCR = 0x0;

    // Init UART0

    U0FDR = 0x10;

    // Disconnect PLL.

    PLLCON = 0x0;
    PLLFEED = 0xAA;
    PLLFEED = 0x55;

    // Map bootloader vectors

    MEMMAP = 0x0;

    // Point to bootloader entry point i.e. reset vector 0x0

    bootloader_entry = (void (*)(void))(0x0);
    (*bootloader_entry)();  //lint !e413 // Yes, it's a NULL pointer.
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Display a variable from the currently selected module.

        @clsyntax dv {moduleName} {variableName}

        @par Command Line Parameters:
            @oo {moduleName}: The name of the module containing the
                variable to be displayed.  This parameter is optional
                and if omitted, the module last selected by the "sm"
                command shall be used.
            @oo {variableName}: The name of the selected variable.

        @description Displays the contents of the named variable
        in the currently selected module on the console.

        @see CmdSelectModule

        @note A TAG_VARIABLE_TABLE struct containing an entry for
        the variable must have been set up in the selected module.

        @ingroup ttcon_debug
    */

uint8_t SUSI_CmdDisplayVariable(const char *pLine)
{
    char wordBuf1[20], wordBuf2[20], *pVarName;
    const struct TAG_VARIABLE_TABLE *pTable;
    unsigned charCount, parmCount;

    if((parmCount = APParse(
        pLine, &charCount, " %20w %20w|",
        wordBuf1, wordBuf2)) == 2)
    {
        // Here we need to look up the module name first.

        if((pTable = LookupModule(wordBuf1)) == NULL)
        {
            APPrint(STDOUT, "\n\r**Module not found.");
            return(CON_STS_OK);
        }
        else
        {
            pVarTable = pTable;
            pVarName = wordBuf2;
        }
    }

    else if(parmCount == 1)
    {
        // Here, use selected module.

        pVarName = wordBuf1;
    }

    else
        return(CON_STS_SYNTAX);

    // Now pVarTable points at the module's variable table.  Look
    // up the name in that table.

    if((pTable = LookupVariable(pVarName)) == NULL)
    {
        APPrint(STDOUT, "\n\r**Variable not found.");
        return(CON_STS_OK);
    }
    else
    {
        switch(pTable->type)
        {
            case VAR_TYPE_NONE:
                IO_DisplayMemoryBytes(STDOUT, pTable->pAddr,
                    (uintptr_t)pTable->pAddr, pTable->length);
                break;

            case VAR_TYPE_CHAR:
                APPrint(STDOUT, ": %c", *((char *)pTable->pAddr));
                break;

            case VAR_TYPE_STRING:
                APPrint(STDOUT, ": %s", (char *)pTable->pAddr);
                break;

            case VAR_TYPE_HEXBYTE:
                APPrint(STDOUT, ": %xb", *((uint8_t *)pTable->pAddr));
                break;

            case VAR_TYPE_HEXWORD:
                APPrint(STDOUT, ": %xw", *((uint16_t *)pTable->pAddr));
                break;

            case VAR_TYPE_HEXINT:
                APPrint(STDOUT, ": %xd", *((uint32_t *)pTable->pAddr));
                break;

            default:
                APPrint(STDOUT,
                    " - Variable type not properly defined");
                break;
        }
    }

    return(CON_STS_OK);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Select a module for variable display.

        @clsyntax sm {moduleName}

        @par Command Line Parameters:
            @oo {moduleName}: The name of the module to be selected.

        @description Selects the named module.  After selection, the
        module's variables may be displayed using the
        CmdDisplayVariable function.

        @see CmdDisplayVariable

        @note An entry in the TAG_MODULE_TABLE struct in ttcon.c
        containing the module name must have been set up prior
        to execution of this command.

        @ingroup ttcon_debug
    */

uint8_t SUSI_CmdSelectModule(const char *pLine)
{
    char wordBuf[21];
    const struct TAG_VARIABLE_TABLE *pTable;
    unsigned charCount;

    if(APParse(pLine, &charCount, " %20w|", wordBuf) != 1)
        return(CON_STS_SYNTAX);

    else
    {
        if((pTable = LookupModule(wordBuf)) == NULL)
            APPrint(STDOUT, "\n\r**Module not found.");
        else
            pVarTable = pTable; // Set global table

        return(CON_STS_OK);
    }
}

const struct TAG_VARIABLE_TABLE *LookupModule(const char *pString)
{
    struct TAG_MODULE_TABLE *pTable;
    bool found;

    pTable = (struct TAG_MODULE_TABLE *)moduleTable;
    found = false;

    while((pTable->pName != NULL) && !found)
    {
        if(strcmp(pString, pTable->pName) == 0)
            return(pTable->pTable);
        else
            pTable++;
    }
    return(NULL);
}


    // Looks up string in table selected by pVarTable.

const struct TAG_VARIABLE_TABLE *LookupVariable(const char *pString)
{
    const struct TAG_VARIABLE_TABLE *pTable;
    bool found;

    pTable = pVarTable;
    found = false;

    while((pTable->pName != NULL) && !found)
    {
        if(strcmp(pString, pTable->pName) == 0)
            return(pTable);
        else
            pTable++;
    }
    return(NULL);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Display SRAM.

        @clsyntax dm {pageNum}

        @par Command Line Parameters:
            @oo {pageNum}: A hexadecimal integer from 00 to 3F
            representing the logical page address of the page
            to be displayed.    Logical page 00 is the first
            page of LPC-213X SRAM.

        @description Displays the contents of the addressed SRAM
        page on the console.

        @ingroup ttcon_debug
    */

uint8_t SUSI_CmdDisplaySRAM(const char *pLine)
{
    uint32_t pageNum;
    unsigned charCount;

    if(APParse(pLine, &charCount, " %2x|", &pageNum) != 1)
        return(CON_STS_SYNTAX);

    else if(pageNum >= NUM_PAGES)
        return(CON_STS_RANGE);

    else
    {
        pageNum = (pageNum << 8) + SRAM_START;
        IO_DisplayMemoryBytes(STDOUT, (uint8_t *)pageNum, pageNum, 256);
        return(CON_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#if TRACE != 0

    /** Reset trace system

        @clsyntax trst

        @par Command Line Parameters: None.

        @description Initializes the trace system.

        @see CmdEnaTrace, CmdDisTrace,
        CmdPrintTraceCount CmdPrintTraceBuf

        @test Needs testing.

        @ingroup ttcon_debug
    */

uint8_t SUSI_CmdResetTrace(const char *pLine)
{
    if(*pLine != '\0')
        return(CON_STS_SYNTAX);
    else
    {
        IO_DBTraceReset();
        return(CON_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Enable tracing

        @clsyntax ton

        @par Command Line Parameters: None.

        @description Allows the code to begin capturing trace buffer
        entries.

        @see CmdResetTrace, CmdDisTrace,
        CmdPrintTraceCount CmdPrintTraceBuf

        @test Needs testing.

        @ingroup ttcon_debug
    */

uint8_t SUSI_CmdEnaTrace(const char *pLine)
{
    if(*pLine != '\0')
        return(CON_STS_SYNTAX);
    {
        IO_DBTraceEnable();
        return(CON_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Disable tracing

        @clsyntax toff

        @par Command Line Parameters: None.

        @description Stops the trace system from capturing more
        trace entries in the buffer.

        @see CmdResetTrace, CmdEnaTrace,
        CmdPrintTraceCount CmdPrintTraceBuf

        @test Needs testing.

        @ingroup ttcon_debug
    */

uint8_t SUSI_CmdDisTrace(const char *pLine)
{
    if(*pLine != '\0')
        return(CON_STS_SYNTAX);
    else
    {
        IO_DBTraceDisable();
        return(CON_STS_OK);
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Print trace count.

        @clsyntax tdc

        @par Command Line Parameters: None.

        @description Prints the contents of the trace counter to
        the user console.

        @see CmdResetTrace, CmdEnaTrace, CmdDisTrace,
        CmdPrintTraceBuf

        @test Needs testing.

        @ingroup ttcon_debug
    */

uint8_t SUSI_CmdPrintTraceCount(const char *pLine)
{
    if(*pLine != '\0')
        return(CON_STS_SYNTAX);
    {
        APPrint(STDOUT, " %d", IO_DBTraceGetCount());
        return(CON_STS_OK);
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Print trace buffer.

        @clsyntax tdb

        @par Command Line Parameters: None

        @description Prints the contents of the trace buffer to
        the user console.

        @see CmdResetTrace, CmdEnaTrace, CmdDisTrace,
        CmdPrintTraceCount

        @test Needs testing.

        @ingroup ttcon_debug
    */

uint8_t SUSI_CmdPrintTraceBuf(const char *pLine)
{
    uint32_t count;
    uint32_t d, t;

    if(*pLine != '\0')
        return(CON_STS_SYNTAX);
    {
        count = IO_DBTraceGetCount();

        APPrint(STDOUT, " %d Records:", count);

        if(count > 0)
        {
            APPrint_String(STDOUT, "\n\n\rTag\tData");

            while(count-- > 0)
            {
                IO_DBTraceOut(&d, &t);  //lint !e64
                APPrint_HexDWord(STDOUT, t);
                APPrint_Char(STDOUT, '\t');
                APPrint_HexDWord(STDOUT, d);
                APPrint_CRLF(STDOUT);
            }

            APPrint_CRLF(STDOUT);
        }
        return(CON_STS_OK);
    }
}
#endif // TRACE


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Arm the breakpoint

        @clsyntax abp

        @par Command Line Parameters: None.

        @description Arms the breakpoint so that the next time
        it is called, it will call PutBreakpointHere();

        @ingroup ttcon_debug
    */

uint8_t SUSI_CmdArmBreakpoint(const char *pLine)
{
    if(*pLine != '\0')
        return(CON_STS_SYNTAX);
    {
        IO_DB_BreakpointArm();
        return(CON_STS_OK);
    }
}

/// @}
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                          END
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
