//***************************************************************
/**
    @file susi_rx_cmds.c

    Serial User Interface Receiver Commands

    @author nlh/alp

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.
*/
//***************************************************************

//    *** INCLUDE FILES ***

#include <stdint.h>         //lint -esym(766, stdint.h)
#include <stdbool.h>

#include "lpc213x.h"        //lint -esym(766, lpc213x.h)
#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)

#include "target.h"         //lint -esym(766, target.h)
#include "apex.h"

#include "app.h"
#include "ansi_term.h"

#include "apparse.h"
#include "apprint.h"
#include "central.h"
#include "io_main.h"
#include "io_misc.h"
#include "astream.h"

#include "console.h"
#include "io_debug.h"       //lint -esym(766, io_debug.h)
#include "io_codec.h"
#include "ssm_morse.h"
#include "ssm_satd.h"
#include "ssm_sate.h"
#include "eep.h"
#include "susi.h"
#include "susi_cmds.h"
#include "apmath.h"

/***************************************************************/
//    *** MACROS ****

/***************************************************************/
/***************************************************************/

//    *** STATIC FUNCTION PROTOTYPES ****

    // General Static Functions

static void DisassembleRxFlagWord(enum TAG_RX_CHAN chan);

static void State_RxLevelAuto(
        enum TAG_SUSI_EVIDS id, uintptr_t data1, uintptr_t data2);

//****************************************************************
//                 *** VARIABLES ***
//****************************************************************

    // Variables used by the Rx Level Servo:
static enum TAG_RX_CHAN rxLevelServoChan;   ///< Channel number
static TMR rxLevelServoSampleTimer;         ///< Time to adjust level.
static int rxLevelServo_InitialPotLevel;   ///< Pot level
static int rxLevelServo_InitialRms;
static unsigned rxLevelServoTrys;

//***************************************************************
//                     Public API Functions
//***************************************************************

void SUSI_RxCommandsReset(void)
{
    SUSI_InitTimer(&rxLevelServoSampleTimer, EV_SUSI_LEVEL_TIMEOUT);
}

//***************************************************************
//                   State Functions
//***************************************************************

    // The value below is the target RMS value in codec counts
    // for a receiver channel.  It takes into account the
    // following:
    //      1) The input during the adjustment is a tone with
    //          3KHz deviation, and 5 KHz deviation is absolute
    //          max.
    //      2) The target level should be about 10% lower than
    //          absolute max to avoid saturation due to
    //          variations in the user's input tone level etc.
    //      3) The codec 0dB reference is a peak value, and we
    //          will be targeting an RMS value.
    //      4) If a deemphasis filter is employed, the level
    //          when excited by a 300 Hz signal will be higher
    //          than that of the 1000 Hz signal that is being
    //          used to take the measurements.  So we must take
    //          that into account so a 300 Hz component does not
    //          saturate the receiver.

#define RX_LEVEL_SERVO_TARGET        ((CDX_0DB_REFERENCE*2)/5)
#define RX_LEVEL_SERVO_SAMPLE_TIME   (500*TU_1MS)

    /* Gain Analysis:

        Using the LTSpice model of the input circuit, I varied
        the resistance ratio of the Rx input pot and observed
        the effect on the codec input voltage.  Here are the
        values I observed:

            Pot             Codec Input
            Setting         Voltage
            =======         ===========

              90%               599 mVRMS
              50                333
              45                300
              10                 67

        From the above values I determined that the pot gain
        was approximately 6.65 mV per pot-count (Kpot).

        Using the Mathcad spreadsheet in which the codec's
        input voltages were analyzed, and using the worst-case
        values for codec reference and input levels, I found
        the codec gain was approximately 26.48 uV per codec-
        count (Kcodec).

        Combining the two gains, I get the overall gain of
        251 codec-counts per pot-count.

        The servo will correct a fraction of the total
        error on each itteration.  That fraction will be
        determined by the number of bits of downshift that
        will be applied to the error to get the correction
        amount.
    */

#define RX_LEVEL_SERVO_GAIN_ESTIMATE       30  // Use exponent below.
#define RX_LEVEL_SERVO_GAIN_ESTIMATE_EXP   13  // Binary exponent

void SUSI_Transition_State_RxLevelAuto(
                                            enum TAG_RX_CHAN chan)
{
    uint32_t rmsSquared;
    int deltaRms;
    int deltaPotLevel;
    int newPotLevel;

    rxLevelServoChan = chan;

    EEP_InqRxLevel(
            RX_CHAN_1, (uint32_t *)&rxLevelServo_InitialPotLevel);
    rmsSquared = ASTREAM_GetRmsSquared(
                ASTREAM_RX_CHAN_2_SOURCE(rxLevelServoChan));
    rxLevelServo_InitialRms = (int)APML_Sqrt(rmsSquared);

    // Use approximate gain to guess at the correct result.
    deltaRms = RX_LEVEL_SERVO_TARGET - rxLevelServo_InitialRms;
    deltaPotLevel = (deltaRms * RX_LEVEL_SERVO_GAIN_ESTIMATE)
                                  >> RX_LEVEL_SERVO_GAIN_ESTIMATE_EXP;

    newPotLevel = rxLevelServo_InitialPotLevel + deltaPotLevel;
    newPotLevel = CLIP(newPotLevel, MAX_RX_LEVEL, MIN_RX_LEVEL);

    // Set the level in the user parameters (but
    // don't write to flash yet):
    EEP_SetRxLevel((rxLevelServoChan), (uint8_t)newPotLevel);

    // Set the pot itself.
    if(rxLevelServoChan == RX_CHAN_1)
    {
        IO_SetRx1Level((uint8_t)newPotLevel);
    }
    else
    {
        IO_SetRx2Level((uint8_t)newPotLevel);
    }

    // Set up level timer to allow RMS to stabilize.
    Apex_StartTmr(
            &rxLevelServoSampleTimer, RX_LEVEL_SERVO_SAMPLE_TIME);
    rxLevelServoTrys = 2;

    SUSI_SetState(State_RxLevelAuto,'H');
}

static void State_RxLevelAuto(
        enum TAG_SUSI_EVIDS id, uintptr_t data1, uintptr_t unused)
{
    uint32_t rmsSquared;
    int currentRms;
    int currentPotLevel;
    int deltaRms;
    int deltaPotLevel;
    int gain;
    int newPotLevel;

    switch(id)
    {
        case EV_SUSI_RXDATA:
            // If the user presses ESC, cancel the level setting.
            if((char)data1 == ASCII_ESC)
            {
                SUSI_Transition_State_Interactive();
                APPrint_CRLF(STDOUT);
                Con_Prompt();
            }
            break;

        case EV_SUSI_LEVEL_TIMEOUT:
            EEP_InqRxLevel(RX_CHAN_1, (uint32_t *)&currentPotLevel);
            rmsSquared = ASTREAM_GetRmsSquared(
                        ASTREAM_RX_CHAN_2_SOURCE(rxLevelServoChan));
            currentRms = (int)APML_Sqrt(rmsSquared);

            // Compute the actual gain:
            deltaRms = currentRms - rxLevelServo_InitialRms;
            deltaPotLevel =
                    currentPotLevel - rxLevelServo_InitialPotLevel;
            gain = (deltaPotLevel
                << RX_LEVEL_SERVO_GAIN_ESTIMATE_EXP) / deltaRms;

            deltaRms = RX_LEVEL_SERVO_TARGET - currentRms;

            deltaPotLevel = (deltaRms * gain)
                                  >> RX_LEVEL_SERVO_GAIN_ESTIMATE_EXP;

            newPotLevel = currentPotLevel + deltaPotLevel;
            newPotLevel = CLIP(newPotLevel, MAX_RX_LEVEL, MIN_RX_LEVEL);

            // Set the level in the user parameters (but
            // don't write to flash yet):
            EEP_SetRxLevel((rxLevelServoChan), (uint8_t)newPotLevel);

            // Set the pot itself.
            if(rxLevelServoChan == RX_CHAN_1)
            {
                IO_SetRx1Level((uint8_t)newPotLevel);
            }
            else
            {
                IO_SetRx2Level((uint8_t)newPotLevel);
            }

            // Iterate a few times to get really close.
            if(--rxLevelServoTrys == 0)
            {
                // Make sure we write the result to flash.
                SUSI_RestartWriteTimer();
                SUSI_Transition_State_Interactive();
                APPrint_CRLF(STDOUT);
                Con_Prompt();
            }
            else
            {
                // Get ready for next iteration:
                rxLevelServo_InitialRms = currentRms;
                rxLevelServo_InitialPotLevel = currentPotLevel;

                // Set up level timer to allow RMS to stabilize.
                Apex_StartTmr(&rxLevelServoSampleTimer,
                                    RX_LEVEL_SERVO_SAMPLE_TIME);
            }
            break;

        default:
            break;
    }
}

//****************************************************************
//                    COMMAND LINE FUNCTIONS:
/// @defgroup susi Console Operator Commands
/// @{
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set courtesy beep frequency and duration

        @clsyntax \> cbt {rx_chan} {tone} {freq} {duration}

        @par Command Line Parameters:
            @oo {rx_chan}: Receiver channel number (1 or 2) to which
                the courtesy beep settings apply.
            @oo {tone}: The digit 1 or 2 corresponding to the
                portion of the double courtesy beep tone being
                changed.  If using only a single tone beep, use
                tone 1 only.
            @oo {freq}: Tone frequency in Hz.  FG_MIN_FREQ is the
                minimum frequency, FG_MAX_FREQ is the maximum.
            @oo {duration}: Duration of beep tone in milliseconds.

        @description This command sets courtesy beep frequency and
        duration for one of the two possible tones.
    */

uint8_t SUSI_CmdSetBeepTone(const char *pLine)
{
    uint32_t chan, toneNum, freq, duration;
    unsigned charCount;
    eep_courtsey_beep_t *pBeep;

    // Here, if the line is empty, print current contents
    // of the courtesy beep parameters.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetBeepTone();
        return(CON_STS_NULL);
    }

    else if(APParse(pLine, &charCount, " %d %d %d %d|",
        &chan, &toneNum, &freq, &duration) != 4)
        return(CON_STS_SYNTAX);

    else if(!VALID_RX_CHAN(ORDINAL_2_RX_CHAN(chan)))
        return(CON_STS_RANGE);

    else if(freq < FG_MIN_FREQ || freq > FG_MAX_FREQ)
        return(CON_STS_RANGE);

    else if(duration < CB_MIN_DURATION || duration > CB_MAX_DURATION)
        return(CON_STS_RANGE);

    else
    {
        pBeep = EEP_GetCourtesyBeepParms(ORDINAL_2_RX_CHAN(chan));

        if(toneNum == 1)
        {
            pBeep->freq1 = (uint16_t)freq;
            pBeep->duration1 = (uint16_t)duration;
            SUSI_RestartWriteTimer();
            return(CON_STS_OK);
        }

        else if(toneNum == 2)
        {
            pBeep->freq2 = (uint16_t)freq;
            pBeep->duration2 = (uint16_t)duration;
            SUSI_RestartWriteTimer();
            return(CON_STS_OK);
        }

        else
            return(CON_STS_RANGE);
    }
}


void SUSI_StatusSetBeepTone(void)
{
    eep_courtsey_beep_t *pBeep;

    pBeep = EEP_GetCourtesyBeepParms(RX_CHAN_1);

    APPrint(STDOUT,
        "cbt 1 1 %d %d\n\r", pBeep->freq1, pBeep->duration1);
    APPrint(STDOUT,
        "cbt 1 2 %d %d\n\r", pBeep->freq2, pBeep->duration2);

    pBeep = EEP_GetCourtesyBeepParms(RX_CHAN_2);

    APPrint(STDOUT,
        "cbt 2 1 %d %d\n\r", pBeep->freq1, pBeep->duration1);
    APPrint(STDOUT,
        "cbt 2 2 %d %d\n\r", pBeep->freq2, pBeep->duration2);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set pre-courtesy beep delay

        @clsyntax \> cbd {chan} {delay}

        @par Command Line Parameters:
            @oo {chan}: Receiver channel number (1 or 2) to which
                the courtesy beep settings apply.
            @oo {delay}: The time delay from squelch to start
                of courtesy beep in milliseconds.

        @description This command sets the time delay from the
        receiver squelch closing until the start of the courtesy
        beep.
    */

uint8_t SUSI_CmdSetBeepDelay(const char *pLine)
{
    uint32_t chan, delay;
    unsigned charCount;
    eep_courtsey_beep_t *pBeep;

    // Here, if the line is empty, print current parameter
    // status.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetBeepDelay();
        return(CON_STS_NULL);
    }

    else if(APParse(
            pLine, &charCount, " %d %d|", &chan, &delay) != 2)
        return(CON_STS_SYNTAX);

    else if(!VALID_RX_CHAN(ORDINAL_2_RX_CHAN(chan)))
        return(CON_STS_RANGE);

    else if(delay < CB_MIN_DELAY || delay > CB_MAX_DELAY)
        return(CON_STS_RANGE);

    else
    {
        pBeep = EEP_GetCourtesyBeepParms(ORDINAL_2_RX_CHAN(chan));
        pBeep->delay = (uint16_t)delay;
        SUSI_RestartWriteTimer();
        return(CON_STS_OK);
    }
}


void SUSI_StatusSetBeepDelay(void)
{
    eep_courtsey_beep_t *pBeep;

    pBeep = EEP_GetCourtesyBeepParms(RX_CHAN_1);
    APPrint(STDOUT, "cbd 1 %d\n\r", pBeep->delay);

    pBeep = EEP_GetCourtesyBeepParms(RX_CHAN_2);
    APPrint(STDOUT, "cbd 2 %d\n\r", pBeep->delay);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set courtesy beep mode

        @clsyntax \> cbm {chan} {mode}

        @par Command Line Parameters:
            @oo {chan}: Receiver channel number (1 or 2) to which
                the courtesy beep settings apply.
            @oo {mode}: A code from the following table identifying
                the desired courtesy beep mode.

        @verbatim
            Code
            Number  Mode Description
            ======  ================
              0     Courtesy beep disabled
              1     Single tone beep
              2     Double tone beep
              3     Morse code character
        @endverbatim

        @description This command sets the type of courtesy beep
        desired for the given receiver channel.
    */

uint8_t SUSI_CmdSetBeepMode(const char *pLine)
{
    uint32_t chan, mode;
    unsigned charCount;
    eep_courtsey_beep_t *pBeep;

    // Here, if the line is empty, print current parameter
    // status.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetBeepMode();
        return(CON_STS_NULL);
    }

    else if(APParse(
            pLine, &charCount, " %d %d|", &chan, &mode) != 2)
        return(CON_STS_SYNTAX);

    else if(!VALID_RX_CHAN(ORDINAL_2_RX_CHAN(chan)))
        return(CON_STS_RANGE);

    else if(mode > CB_MODE_MAX)
        return(CON_STS_RANGE);

    else
    {
        pBeep = EEP_GetCourtesyBeepParms(ORDINAL_2_RX_CHAN(chan));
        pBeep->mode = (enum TAG_CB_MODE)mode;
        SUSI_RestartWriteTimer();
        return(CON_STS_OK);
    }
}


void SUSI_StatusSetBeepMode(void)
{
    eep_courtsey_beep_t *pBeep;

    pBeep = EEP_GetCourtesyBeepParms(RX_CHAN_1);
    APPrint(STDOUT, "cbm 1 %d\n\r", pBeep->mode);

    pBeep = EEP_GetCourtesyBeepParms(RX_CHAN_2);
    APPrint(STDOUT, "cbm 2 %d\n\r", pBeep->mode);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set courtesy beep character

        @clsyntax \> cbc {chan} {char}

        @par Command Line Parameters:
            @oo {chan}: Receiver channel number (1 or 2) to which
                the courtesy beep settings apply.
            @oo {char}: A character from the following list
                that shall be used as a courtesy beep.

        @verbatim
        ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789/-.?
        @endverbatim

        @description This command sets the character that shall
        be used for a courtesy beep when 'Morse Code Character'
        mode is selected.
    */

uint8_t SUSI_CmdSetBeepCharacter(const char *pLine)
{
    uint32_t chan;
    unsigned charCount;
    eep_courtsey_beep_t *pBeep;
    char c;

    // Here, if the line is empty, print current parameter
    // status.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetBeepCharacter();
        return(CON_STS_NULL);
    }

    else if(APParse(
            pLine, &charCount, " %d %c|", &chan, &c) != 2)
        return(CON_STS_SYNTAX);

    else if(!VALID_RX_CHAN(ORDINAL_2_RX_CHAN(chan)))
        return(CON_STS_RANGE);

    else if(!Morse_IsMorse(c))
        return(CON_STS_RANGE);

    else
    {
        pBeep = EEP_GetCourtesyBeepParms(ORDINAL_2_RX_CHAN(chan));
        pBeep->morse[0] = c;
        pBeep->morse[1] = '\0';
        SUSI_RestartWriteTimer();
        return(CON_STS_OK);
    }
}


void SUSI_StatusSetBeepCharacter(void)
{
    eep_courtsey_beep_t *pBeep;

    pBeep = EEP_GetCourtesyBeepParms(RX_CHAN_1);
    APPrint(STDOUT, "cbc 1 %c\n\r", pBeep->morse[0]);

    pBeep = EEP_GetCourtesyBeepParms(RX_CHAN_2);
    APPrint(STDOUT, "cbc 2 %c\n\r", pBeep->morse[0]);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set courtesy beep level

        @clsyntax \> cbl {chan} {level}

        @par Command Line Parameters:
            @oo {chan}: Transmitter channel number (1 or 2) to which
                the setting applies.
            @oo {level}: A number in percent indicating the desired
                audio level.

        @description This command sets the audio level
        with which the courtesy beep shall be output on the
        selected transmitter channel.
    */

uint8_t SUSI_CmdSetBeepLevel(const char *pLine)
{
    uint32_t chan, level;
    unsigned charCount;

    // Here, if the line is empty, print current parameter
    // status.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetBeepLevel();
        return(CON_STS_NULL);
    }

    else if(APParse(
            pLine, &charCount, " %d %d|", &chan, &level) != 2)
        return(CON_STS_SYNTAX);

    else if(!VALID_TX_CHAN(ORDINAL_2_TX_CHAN(chan)))
        return(CON_STS_RANGE);

    else if(level > 100)
        return(CON_STS_RANGE);

    else
    {
        EEP_SetCourtesyBeepLevel(
                ORDINAL_2_TX_CHAN(chan), (uint8_t)level);
        SUSI_RestartWriteTimer();
        return(CON_STS_OK);
    }
}


void SUSI_StatusSetBeepLevel(void)
{
    unsigned level;

    level = EEP_GetCourtesyBeepLevel(TX_CHAN_1);
    APPrint(STDOUT, "cbl 1 %d\n\r", level);

    level = EEP_GetCourtesyBeepLevel(TX_CHAN_2);
    APPrint(STDOUT, "cbl 2 %d\n\r", level);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set receiver flags

        @clsyntax \> rxf {chan} [Lx] [Px] [Rx]

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) to which the receiver
                flag settings apply.
            @oo {Lx}: The character L/l followed immediately by 0 or
                1 to set RUS level to active low or active high
                (respectively) on the selected channel.
            @oo {Px}: The character P/p followed immediately by 0 or
                1 to disable or enable (respectively) CTCSS input
                (decode) tone on the selected channel.
            @oo {Rx}: The character R/r followed immediately by 0 or
                1 to disable or enable (respectively), the selected
                channel.

        @description This command sets the various flags governing the
            operation of the selected receiver.

        @note At least one flag must b specified. The defaults are
            L1, P0, D1.

Deprecated: Cx and Dx
            @oo {Cx}: 1 to enable CTCSS reject filter, 0 to disable it.
            @oo {Dx}: The character D/d followed immediately by 0 or
                1 to set de-emphasis on the selected channel's
                received audio.

    */

uint8_t SUSI_CmdSetRxFlags(const char *pLine)
{
    uint32_t chan;
    unsigned charCount;
    char flag;
    bool value;
    uint16_t flags;
    enum TAG_RX_CHAN rxChan;

    // If the line is empty, print current flag settings.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetRxFlags();
        return(CON_STS_NULL);
    }

    else if(APParse(pLine, &charCount, " %d", &chan) != 1)
        return(CON_STS_SYNTAX);

    else
    {
        if(chan != 1 && chan != 2)
            return(CON_STS_RANGE);

        else
        {
            rxChan = ORDINAL_2_RX_CHAN(chan);

            // Get the flags so we can change them.
            EEP_InqRxFlags(rxChan, &flags);
            pLine += charCount;
            while(*pLine != '\0')
            {
                if(APParse(
                    pLine, &charCount, " %c%b", &flag, &value) != 2)
                    return(CON_STS_SYNTAX);

                switch(flag)
                {
#if 0 // Flags C and D are deprecated.
                    // Enable/Disable CTCSS Reject Filter
                    case 'c': case 'C':
                        flags = SUSI_AssembleFlagWord(
                            flags, value, EEP_RX_FLAG_SATR);
                        ASTREAM_SelectCtcssRejectionFilter(
                                                        rxChan, value);
                        break;

                    // Evaluate deemphasis setting.
                    case 'd': case 'D':
                        flags = SUSI_AssembleFlagWord(
                            flags, value, EEP_RX_FLAG_DEEMPH);
                        ASTREAM_SelectDeemphasis(rxChan, value);
                        break;
#endif
                    // Evaluate RUS active level setting.
                    case 'l': case 'L':
                        flags = SUSI_AssembleFlagWord(
                            flags, value, EEP_RX_FLAG_RUSLEVEL);
                        break;

                    // Enable/Disable a receiver.
                    case 'r': case 'R':
                        flags = SUSI_AssembleFlagWord(
                            flags, value, EEP_RX_FLAG_ENABLE);
                        CEN_RxEnable(rxChan, value);
                        break;

                    // Evaluate subaudible tone decode enable/disable.
                    case 'p': case 'P':
                        flags = SUSI_AssembleFlagWord(
                            flags, value, EEP_RX_FLAG_SATD);
                        break;

                    default:
                        return(CON_STS_SYNTAX);
                }
                pLine += charCount;

            }   // while

            EEP_SetRxFlags(rxChan, flags);
            SUSI_RestartWriteTimer();
        }
    }

    return(CON_STS_OK);
}

void SUSI_StatusSetRxFlags(void)
{
    APPrint(STDOUT, "rxf 1 ");
    DisassembleRxFlagWord(RX_CHAN_1);
    APPrint_CRLF(STDOUT);

    APPrint(STDOUT, "rxf 2 ");
    DisassembleRxFlagWord(RX_CHAN_2);
    APPrint_CRLF(STDOUT);
}


static void DisassembleRxFlagWord(enum TAG_RX_CHAN chan)
{
    uint16_t flags;

    EEP_InqRxFlags(chan, &flags);
#if 0   // Flags C and D are deprecated.
    APPrint(STDOUT, " C%b",
        (uint32_t)((flags & EEP_RX_FLAG_SATR) != 0));
    APPrint(STDOUT, " D%b",
        (uint32_t)((flags & EEP_RX_FLAG_DEEMPH) != 0));
#endif

    APPrint(STDOUT, " L%b",
        (uint32_t)((flags & EEP_RX_FLAG_RUSLEVEL) != 0));
    APPrint(STDOUT, " R%b",
        (uint32_t)((flags & EEP_RX_FLAG_ENABLE) != 0));
    APPrint(STDOUT, " P%b",
        (uint32_t)((flags & EEP_RX_FLAG_SATD) != 0));
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set a Receiver's Filter Index

        @clsyntax > filt {rx_chan} {type} {index}

        @par Command Line Parameters:
            @oo {rx_chan}: 1 or 2, identifying the
                receiver channel to which the command applies.
            @oo {type}: D = Deemphasis, R = CTCSS Reject
            @oo {index}: Deemphasis: 0 to 2, for low, med, high,
                Reject: 0 = None, 1 = 3rd order, 2 = 4th order.

        @description This command selects the filter to be used
        for the function indicated by the 'type' parameter.
*/

uint8_t SUSI_CmdSetFilterIndex(const char *pLine)
{
    uint32_t rxChan, index;
    unsigned charCount;
    char c;

    // Here, if the line is empty, print current parameter
    // status.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetFilterIndex();
        return(CON_STS_NULL);
    }

    // Parse the line.

    if(APParse(
        pLine, &charCount, " %d %c %d|", &rxChan, &c, &index) != 3)
        return(CON_STS_SYNTAX);

    else if(!VALID_RX_CHAN(ORDINAL_2_RX_CHAN(rxChan)))
        return(CON_STS_RANGE);

    else
    {
        switch(c)
        {
            case 'D': case 'd':
                if(index >= MAX_DEEMPH_FILTERS)
                    return(CON_STS_RANGE);
                else
                {
                    // Store the new index and inform the decoder.
                    EEP_SetDeemphasisFilterIndex(
                            ORDINAL_2_RX_CHAN(rxChan), index);
                    ASTREAM_SetDeemphasisFilterIndex(
                            ORDINAL_2_RX_CHAN(rxChan), index);
                }
                break;

            case 'R': case 'r':
                if(index >= MAX_CTCSS_REJECT_FILTERS)
                    return(CON_STS_RANGE);
                else
                {
                    // Store the new index and inform the decoder.
                    EEP_SetCtcssRejectionFilterIndex(
                            ORDINAL_2_RX_CHAN(rxChan), index);
                    ASTREAM_SetCtcssRejectionFilterIndex(
                            ORDINAL_2_RX_CHAN(rxChan), index);
                }
                break;

            default:
                return(CON_STS_SYNTAX);
        }

        SUSI_RestartWriteTimer();
        return(CON_STS_OK);
    }
}

    // Print status for filt command:
    // {rx_chan} {index}

void SUSI_StatusSetFilterIndex(void)
{
    unsigned index;

    index = EEP_GetDeemphasisFilterIndex(RX_CHAN_1);
    APPrint(STDOUT, "filt 1 D %d\n\r", index);
    index = EEP_GetDeemphasisFilterIndex(RX_CHAN_2);
    APPrint(STDOUT, "filt 2 D %d\n\r", index);
    index = EEP_GetCtcssRejectionFilterIndex(RX_CHAN_1);
    APPrint(STDOUT, "filt 1 R %d\n\r", index);
    index = EEP_GetCtcssRejectionFilterIndex(RX_CHAN_2);
    APPrint(STDOUT, "filt 2 R %d\n\r", index);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set a Receiver's CTCSS Decoder Frequency Index

        @clsyntax > sdi {rx_chan} {index}

        @par Command Line Parameters:
            @oo {rx_chan}: 1 or 2, identifying the
                receiver channel to which the command applies.
            @oo {index}: 0 to 36, to select tone decoder frequency.

        @description This command sets the CTCSS frequency index
            for the addressed receiver's decoder.
*/

uint8_t SUSI_CmdSetCTCSSDecodeIndex(const char *pLine)
{
    uint32_t parmCount, rxChan, index;
    unsigned charCount;

    // Here, if the line is empty, print current parameter
    // status.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusCTCSSDecodeIndex();
        return(CON_STS_NULL);
    }

    // Parse the line.

    parmCount = APParse(
        pLine, &charCount, " %d %d|", &rxChan, &index);

    if(parmCount != 2)
        return(CON_STS_SYNTAX);

    else if(rxChan != 1 && rxChan != 2)
        return(CON_STS_RANGE);

    else if(index >= SATD_FILTER_TABLE_SIZE)
        return(CON_STS_RANGE);

    else
    {
        // Store the new index and inform the decoder.
        EEP_SetCTCSSDecodeIndex(
                        ORDINAL_2_RX_CHAN(rxChan), (uint8_t)index);
        SUSI_RestartWriteTimer();
        SATD_LoadDecoderParms(ORDINAL_2_RX_CHAN(rxChan));

        return(CON_STS_OK);
    }
}

    // Print status for sdi command:
    // {rx_chan} {index}

void SUSI_StatusCTCSSDecodeIndex(void)
{
    uint8_t index;

    index = EEP_GetCTCSSDecodeIndex(RX_CHAN_1);
    APPrint(STDOUT, "sdi 1 %d\n\r", index);
    index = EEP_GetCTCSSDecodeIndex(RX_CHAN_2);
    APPrint(STDOUT, "sdi 2 %d\n\r", index);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set receiver channel digital level

        @clsyntax \> rxl {chan} {rxLevel}

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) to which the receiver
                level setting applies.
            @oo {rxLevel}: Volume level in percent (0 to 100).

        @description This command sets the digital potentiometer
            associated with the specified receiver channel.

    */

uint8_t SUSI_CmdSetRxLevel(const char *pLine)
{
    uint32_t chan, rxLevel;
    unsigned charCount;

    // Here, if the line is empty, print current level set for the
    // respective receiver channels.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetRxLevel();
        return(CON_STS_NULL);
    }

    else if(APParse(
            pLine, &charCount, " %d %d|", &chan, &rxLevel) != 2)
        return(CON_STS_SYNTAX);

    else if(!VALID_RX_CHAN(ORDINAL_2_RX_CHAN(chan)))
        return(CON_STS_RANGE);

    else if(rxLevel > MAX_RX_LEVEL)
        return(CON_STS_RANGE);

    else
    {
        // Set the level in the user parameters:
        EEP_SetRxLevel((ORDINAL_2_RX_CHAN(chan)),
            (uint8_t)rxLevel);
        SUSI_RestartWriteTimer();

        if(chan == 1)
        {
            IO_SetRx1Level((uint8_t)rxLevel);
        }
        else
        {
            IO_SetRx2Level((uint8_t)rxLevel);
        }

        return(CON_STS_OK);
    }
}

void SUSI_StatusSetRxLevel(void)
{
    uint32_t level;

    EEP_InqRxLevel(RX_CHAN_1, &level);
    APPrint(STDOUT, "rxl 1 %d\n\r", level);

    EEP_InqRxLevel(RX_CHAN_2, &level);
    APPrint(STDOUT, "rxl 2 %d\n\r", level);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set receiver channel digital level automatically

        @clsyntax \> rxla {chan}

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) to which the receiver
                level setting applies.

        @description This command sets the digital potentiometer
            associated with the specified receiver channel
            automatically in the presence of a 1000 Hz tone
            transmitted at a 3000 Hz deviation level.
    */

uint8_t SUSI_CmdSetRxLevelAuto(const char *pLine)
{
    uint32_t chan;
    unsigned charCount;

    // Here, if the line is empty, print current level set for the
    // respective receiver channels.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetRxLevel();
        return(CON_STS_NULL);
    }

    else if(APParse(
            pLine, &charCount, " %d|", &chan) != 1)
        return(CON_STS_SYNTAX);

    else if(!VALID_RX_CHAN(ORDINAL_2_RX_CHAN(chan)))
            return(CON_STS_RANGE);

    else
    {
        SUSI_Transition_State_RxLevelAuto(ORDINAL_2_RX_CHAN(chan));
        return(CON_STS_STATECHG);
    }
}


/// @}
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                          END
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
