//***************************************************************
/**
    @file susi_sys_cmds.c

    Serial User Interface System Commands

    @author nlh/alp

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.
*/
//***************************************************************

//    *** INCLUDE FILES ***

#include <stdint.h>         //lint -esym(766, stdint.h)
#include <stdbool.h>

#include "lpc213x.h"        //lint -esym(766, lpc213x.h)
#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)

#include "aplib.h"
#include "target.h"         //lint -esym(766, target.h)
#include "apex.h"

#include "app.h"
#include "apparse.h"
#include "apprint.h"
#include "io_main.h"
#include "io_flash.h"
#include "io_codec.h"
#include "astream.h"

#include "console.h"
#include "central.h"
#include "io_debug.h"       //lint -esym(766, io_debug.h)
//#include "ttcon.h"
#include "io_misc.h"
#include "filesys.h"
//#include "ssm_morse.h"
//#include "ssm_satd.h"
#include "ssm_sate.h"
#include "eep.h"
#include "audiosm.h"
#include "susi.h"
#include "susi_cmds.h"
//#include "xmodem.h"
#include "signon.h"
//#include "ansi_term.h"
//#include "fgen.h"
//#include "ssm_txsm.h"

/***************************************************************/
//    *** MACROS ****

/***************************************************************/
//    *** GENERAL DEFINITIONS ****

/***************************************************************/

//    *** STATIC FUNCTION PROTOTYPES ****


    // General Static Functions

static unsigned FileTime(unsigned fileNum);
static unsigned FileSpaceLeft(void);

/***************************************************************/
//                 *** VARIABLES ***
//****************************************************************

//***************************************************************
//                    COMMAND LINE FUNCTIONS:
/// @defgroup susi Console Operator Commands
/// @{
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    Print file system status.

    @clsyntax > files

    @par Command Line Parameters: None.

    @description This command prints the status of the operator
    files to the console.
    */
uint8_t SUSI_CmdFileSystemStatus(const char *pLine)
{
    unsigned time, totalTime, i;

    // The rest of the line should be empty.

    if(*pLine != '\0')
        return(CON_STS_SYNTAX);

    else
    {
        totalTime = 0;

        for(i=0; i<NUM_USER_FILES; i++)
        {
            time = FileTime(i);
            totalTime += time;
            if(time != 0)
            {
                APPrint(STDOUT, "\n\rFile %d: %.1f seconds.",
                    FILENUM_2_ORDINAL(i), time);
            }
        }

        APPrint(STDOUT, "\n\rTotal recorded time: %.1f seconds, "
            "time left: %.1f seconds.\n\n\r",
            totalTime, FileSpaceLeft());

        return(CON_STS_OK);
    }

}

    // Calculate how much time a recording took,

static unsigned FileTime(unsigned fileNum)
{
    unsigned numPages;  // Count of pages allocated to file etc.
    unsigned time;      // Seconds duration of file etc.

    // Calculate how much time the recording took,
    // and how much time is left.
    numPages = FSys_FileSize(fileNum);

    // File time in 10ths of seconds
    time = (numPages * Z_FLASH_PAGE * 10)
        /((unsigned)(CDX_SAMPLE_RATE_HZ + 0.5));

    return(time);
}


    // and how much time is left.
static unsigned FileSpaceLeft(void)
{
    unsigned numFree;   // Count of unallocated pages.
    unsigned time;      // Seconds duration of file etc.

    numFree = FSys_FreePages();
    time = (numFree * Z_FLASH_PAGE * 10)
        /((unsigned)(CDX_SAMPLE_RATE_HZ + 0.5));
    return(time);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Start recording a voice file

        @clsyntax > frec {chan} {filenum}

        @par Command Line Parameters:
            @oo {chan}: Receiver number (1 or 2) indicating from
                which receiver to record,
            @oo {filenum}: File number (1 through 8) indicating
                which file to record.

        @description This command sets a timer for 10 seconds and
        waits for the user to key the selected receiver.  When
        the receiver is keyed, the audio is captured and stored in
        the selected file.

        @par
        The audio is recorded from the 'flat' source corresponding
        to the receiver indicated.  If de-emphasis is required on
        the recording it should be applied during playback.
    */

uint8_t SUSI_CmdFileRecord(const char *pLine)
{
    uint32_t chan, fileNum;
    unsigned charCount;
    uint8_t recordFile;

    if(APParse(
        pLine, &charCount, " %d %d|", &chan, &fileNum) != 2)
        return(CON_STS_SYNTAX);

    else if(!VALID_RX_CHAN(ORDINAL_2_RX_CHAN(chan)))
        return(CON_STS_RANGE);

    else if(!VALID_USER_FILENUM(ORDINAL_2_FILENUM(fileNum)))
        return(CON_STS_RANGE);

    else
    {
        recordFile = ORDINAL_2_FILENUM(fileNum);

        // Test to see if the file is empty.  If not, give an
        // error.  You must explicitly delete a file.
        if(FSys_FileSize(recordFile) != 0)
        {
            APPrint(STDOUT, "\n\rFile not empty! Delete it first.");
            return(CON_STS_NULL);
        }
        else
        {
            CEN_RecordAudioFile(ORDINAL_2_RX_CHAN(chan), recordFile);
            SUSI_StartVoiceRecord(recordFile);
            return(CON_STS_STATECHG);
        }
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Delete a voice file

        @clsyntax > fdel {filenum}

        @par Command Line Parameters:
            @oo {filenum}: File number (1 through 8) indicating
            which file to delete.

        @description This command deletes all data from the file
            addressed by the filenum parameter.
    */

uint8_t SUSI_CmdFileDelete(const char *pLine)
{
    uint32_t fileNum;
    unsigned charCount;

    if(APParse(
        pLine, &charCount, " %d|", &fileNum) != 1)
        return(CON_STS_SYNTAX);

    else if(!VALID_USER_FILENUM(ORDINAL_2_FILENUM(fileNum)))
        return(CON_STS_RANGE);

    else
    {
        FSys_Delete(ORDINAL_2_FILENUM(fileNum));
        return(CON_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Play a file

        @clsyntax > fplay {chan} {fileNum}

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) indicating on
                which channel to play the ID,
            @oo {fileNum}: A number from 1 to 8 indicating the
                file to be played.

        @description This command plays the indicated file on
        the indicated transmitter.
    */
uint8_t SUSI_CmdPlayFile(const char *pLine)
{
    unsigned charCount, chan, fileNum;

    if(APParse(
        pLine, &charCount, " %d %d|", &chan, &fileNum) != 2)
        return(CON_STS_SYNTAX);

    else if(!VALID_TX_CHAN(ORDINAL_2_TX_CHAN(chan)))
        return(CON_STS_RANGE);

    // If file number is out of range, return an error.
    else if(!VALID_USER_FILENUM(ORDINAL_2_FILENUM(fileNum)))
        return(CON_STS_RANGE);

    else
    {
        CEN_PlayFile(
            ORDINAL_2_TX_CHAN(chan), ORDINAL_2_FILENUM(fileNum));

        return(CON_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Format file system for first use.

        @clsyntax format {code}

        @par Command Line Parameters:
            @oo {code}: The number 8958. This is required to
            insure that the user is executing this command
            deliberately.

        @description Formats the file system.   All data contained
        in files prior to the execution of this command are lost.
        The flash is erased and directory information is written.
        All files are cleared, initialized, and are ready to be
        re-written after execution of this command.

        @see CmdFileAppend, CmdFileClose, CmdFileDelete,
        CmdFileOpen, CmdFileRead

        @ingroup ttcon_fsys
    */

uint8_t SUSI_CmdFileFormat(const char *pLine)
{
    unsigned charCount, code;

    if(APParse(pLine, &charCount, " %d|", &code) != 1)
        return(CON_STS_SYNTAX);

    else if(code != 8958)
        return(CON_STS_RANGE);

    else
    {
        APPrint(STDOUT,
            " - Please wait. Operation may take up to 2 minutes.");

        FSys_Format();
        return(CON_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#if EMULATOR == 0

    /** Reset the microcontroller

        @clsyntax \> reset

        @description This command resets the microcontroller by
        disabling interrupts and hanging the CPU until the watchdog
        timer times out.
     */
uint8_t SUSI_CmdReset(const char *unused)
{
    APPrint(STDOUT, " - Resetting. Press 'Enter'.");
    DI();
    FOREVER
        ;
}
#endif

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Force reload of default parameters.

        @clsyntax \> def {code}

        @par Command Line Parameters:
            @oo {code}: The number 8958. This is required to insure
            that the user is executing this command deliberately.

        @description This command forces the reload of default
        parameters into the EEPROM.
*/
uint8_t SUSI_CmdDefaults(const char *pLine)
{
    uint32_t code, potLevel;
    unsigned charCount;

    if(APParse(pLine, &charCount, " %d|", &code) != 1)
        return(CON_STS_SYNTAX);

    else if(code != 8958)
        return(CON_STS_RANGE);

    else
    {
        EEP_SetDefaults();
        SUSI_RestartWriteTimer();
        AUD_DeleteAllAnno();

        // Need to set Rx pot level here.
        EEP_InqRxLevel(RX_CHAN_1, &potLevel);
        IO_SetRx1Level((uint8_t)potLevel);

        EEP_InqRxLevel(RX_CHAN_2, &potLevel);
        IO_SetRx2Level((uint8_t)potLevel);

        APPrint(STDOUT,"\n\rDefaults have been set.");
        return(CON_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set DTMF password

        @clsyntax \> ttpwd {word}

        @par Command Line Parameters:
            @oo {word}: A word containing the password characters
                in quotes.  The password must contain only valid
                DTMF characters 0-9, A-D, a-d.

        @description This command sets the DTMF password..
        @note Defaults: string defaults to "8958"
        @note Max length of the string is Z_DTMF_PASSWORD, minimum
            length is 4.
    */
uint8_t SUSI_CmdSetDTMFPwd(const char *pLine)
{
    unsigned charCount;
    char string[9];
    char *pString;

    // Here, if the line is empty, print current contents of
    // DTMF password.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetDTMFPwd();
        return(CON_STS_NULL);
    }

    if(APParse(pLine, &charCount,
        " " GEN_FORMAT(w|, Z_DTMF_PASSWORD), string) != 1)
        return(CON_STS_SYNTAX);

    else if(strlen(string) < MIN_DTMF_PASSWORD)
    {
        APPrint(STDOUT,
            " DTMF password must be at least 4 characters.\n\r");
        return(CON_STS_OK);
    }

    else
    {
        // do test for valid chars here
        pString = string; //get start of string
        while(*pString != '\0')
        {
            if((*pString >= '0' && *pString <= '9') ||
               (*pString >= 'A' && *pString <= 'D') ||
               (*pString >= 'a' && *pString <= 'd'))
                pString++;
            else
                return(CON_STS_RANGE);
        }

        EEP_SetDTMFPwd(string);
        SUSI_RestartWriteTimer();
        return(CON_STS_OK);
    }
}


void SUSI_StatusSetDTMFPwd(void)
{
    char *pString;

    EEP_InqDTMFPwd(&pString);
    APPrint(STDOUT, "ttpwd %s\n\r", pString);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    Enable/Disable Touch Tone (DTMF) Password

    @clsyntax > ttpena {flag}

    @par Command Line Parameters:
        @oo {flag}: 0 to disable or 1 to enable the password.

    @description This command enables or disables the requirement
    that a DTMF password must be entered before DTMF commands
    will be accepted.
    */

uint8_t SUSI_CmdEnableDTMFPassword(const char *pLine)
{
    bool enable;
    uint16_t flags;
    unsigned charCount;

    // If the line is empty, print current settings.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusEnableDTMFPassword();
        return(CON_STS_NULL);
    }

    else if(APParse(
            pLine, &charCount, " %b|", &enable) != 1)
        return(CON_STS_SYNTAX);

    else
    {
        EEP_InqSysFlags(&flags);
        if(enable)
            flags |= EEP_SYS_FLAG_DTMF_PASS;
        else
            flags &= ~EEP_SYS_FLAG_DTMF_PASS;

        EEP_SetSysFlags(flags);
        SUSI_RestartWriteTimer();
        return(CON_STS_OK);
    }
}

void SUSI_StatusEnableDTMFPassword(void)
{
    uint16_t flags;

    EEP_InqSysFlags(&flags);
    if(EEP_GetSysFlags(EEP_SYS_FLAG_DTMF_PASS))
        APPrint_String(STDOUT, "ttpena 1\n\r");
    else
        APPrint_String(STDOUT, "ttpena 0\n\r");
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    Set Touch Tone (DTMF) Command Character

    @clsyntax > ttchar {char}

    @par Command Line Parameters:
        @oo {char}: DTMF character that will signify the start
            of a command.

    @description This command sets the DTMF character that will
    henceforth be used to signify the start of a command.  The
    character may be any of #,  *, A, B, C, or D.
    */

uint8_t SUSI_CmdSetDTMFCommandChar(const char *pLine)
{
    unsigned charCount;
    char c;

    // Here, if the line is empty, print current parameter
    // status.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetDTMFCommandChar();
        return(CON_STS_NULL);
    }

    else if(APParse(
            pLine, &charCount, " %c|", &c) != 1)
        return(CON_STS_SYNTAX);

    else if(c != '#' && c != '*' &&  c != 'A'
            && c != 'B' && c != 'C' && c != 'D')
        return(CON_STS_RANGE);

    else
    {
        EEP_SetDTMFCommandChar(c);
        SUSI_RestartWriteTimer();
        return(CON_STS_OK);
    }
}

void SUSI_StatusSetDTMFCommandChar(void)
{
    char c = EEP_GetDTMFCommandChar();
    APPrint(STDOUT, "ttchar %c\n\r", c);
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    Set Path Parameters

    @clsyntax \> path {source} {dest} {priority}

        @par Command Line Parameters:
        @oo {source}: 1 for receiver 1, 2 for receiver 2.
        @oo {dest}: 1 for transmitter 1, 2 for transmitter 2.
        @oo {priority} A character, 'L', 'M', or 'H' to set the path's
            priority to low, medium (mixer), or high.

        @description This command selects which transmitters are
        to be activated when a particular receiver is unsquelched.
    */

uint8_t SUSI_CmdSetRxTxPath(const char *pLine)
{
    uint32_t source, dest;
    eep_path_t *pPath;
    eep_path_t path;
    unsigned charCount;
    char priority;

    // Here, if the line is empty, print current parameter
    // status.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetRxTxPath();
        return(CON_STS_OK);
    }

    // Parse the parameters.

    if(APParse(pLine, &charCount, " %d %d %c|",
                            &source, &dest, &priority) != 3)
        return(CON_STS_SYNTAX);

    // Here, add a new path or edit an existing one.
    else
    {
        switch(source)
        {
            case 1: path.src = ASTREAM_SSIG_RX1; break;
            case 2: path.src = ASTREAM_SSIG_RX2; break;
            default: return(CON_STS_RANGE);
        }

        switch(dest)
        {
            case 1: path.dest = ASTREAM_DSIG_TX1; break;
            case 2: path.dest = ASTREAM_DSIG_TX2; break;
            default: return(CON_STS_RANGE);
        }

        // Decode the priority parameter:
        switch(priority)
        {
            case 'H': case 'h': path.priority = 'H'; break;
            case 'M': case 'm': path.priority = 'M'; break;
            case 'L': case 'l': path.priority = 'L'; break;
            default: return(CON_STS_SYNTAX);
        }

        // Make sure the path is unique:

        // Use an existing path with the same source and destination
        // if one exists:
        if((pPath = EEP_FindPath(path.src, path.dest)) != NULL)
        {
            EEP_SetPath(pPath, &path);
            SUSI_RestartWriteTimer();
            return(CON_STS_OK);
        }

        // If no existing path with same src and dest exists, find
        // an empty path slot to use.
        else if((pPath = EEP_FindPath(
                     ASTREAM_NULL_SOURCE, ASTREAM_NULL_DEST)) != NULL)
        {
            EEP_SetPath(pPath, &path);
            SUSI_RestartWriteTimer();
            return(CON_STS_OK);
        }

        // If we can't find any empty path slots, give a
        // message and return.
        else
        {
            APPrint(STDOUT,"\n\rError: No room in path table.");
            return(CON_STS_NULL);
        }
    }
}

    // Print status for path command:

void SUSI_StatusSetRxTxPath(void)
{
    unsigned source, dest, i;
    struct TAG_EEP_PATH *pPath;

    // Loop through all paths and print them.
    for(i=0; i<MAX_PATHS; i++)
    {
        pPath = EEP_GetPath(i);

        if(pPath->src != ASTREAM_NULL_SOURCE)
        {
            switch(pPath->src)
            {
                case ASTREAM_SSIG_RX1: source = 1; break;
                case ASTREAM_SSIG_RX2: source = 2; break;
                default: FATAL();
            }

            switch(pPath->dest)
            {
                case ASTREAM_DSIG_TX1: dest = 1; break;
                case ASTREAM_DSIG_TX2: dest = 2; break;
                default: FATAL();
            }

            APPrint(STDOUT,"path %u %u %c\n\r",
                                source, dest, pPath->priority);
        }
    }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    Delete Path

    @clsyntax \> dpath {source} {dest}

    @par Command Line Parameters:
    @oo {source}: 1 for receiver 1, 2 for receiver 2.
    @oo {dest}: 1 for transmitter 1, 2 for transmitter 2.

    @description This command deletes any path matching
    the given source and destination parameters.
*/

uint8_t SUSI_CmdDeletePath(const char *pLine)
{
    uint32_t source, dest;
    unsigned charCount;
    astream_srcsig_t srcSig;
    astream_destsig_t destSig;
    eep_path_t *pPath;

    // Parse the parameters.
    if(APParse(
        pLine, &charCount, " %d %d|", &source, &dest) != 2)
        return(CON_STS_SYNTAX);

    // Here, add a new path or edit an existing one.
    else
    {
        switch(source)
        {
            case 1: srcSig = ASTREAM_SSIG_RX1; break;
            case 2: srcSig = ASTREAM_SSIG_RX2; break;
            default: return(CON_STS_RANGE);
        }

        switch(dest)
        {
            case 1: destSig = ASTREAM_DSIG_TX1; break;
            case 2: destSig = ASTREAM_DSIG_TX2; break;
            default: return(CON_STS_RANGE);
        }

        if((pPath = EEP_FindPath(srcSig, destSig)) != NULL)
        {
            EEP_DeletePath(pPath);
            SUSI_RestartWriteTimer();
        }

        return(CON_STS_OK);
    }
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    Schedule an Announcement

    @clsyntax \> anno {txNum} {fileNum} {period} {priority}

        @par Command Line Parameters:
        @oo {txNum}: The transmitter channel to which the
            announcement is to be assigned.
        @oo {fileNum}: A number from 1 to 8 indicating the file
            that contains the announcement audio.
        @oo {period}: The number of minutes between successive
            playbacks of this announcement.
        @oo {priority} A character, 'L', 'M', or 'H' to set the
            path's priority to low, medium (mixer), or high.

        @description This command registers an announcement to
        be played back at the interval given in the time parameter.
        The priority parameter determines if the announcement may
        be preempted by user audio, or may preempt user audio, or
        mixes with user audio.

        @par
        If an announcement using the same file number is already
        scheduled, the new announcement will be given the same
        enable status as the existing announcement.  Otherwise
        the new announcement will be created with 'disabled'
        status.  To enable it, see the 'eanno' command.
 */

uint8_t SUSI_CmdScheduleAnno(const char *pLine)
{
    eep_anno_t *pAnno;
    eep_anno_t anno;
    unsigned i, charCount, chan, fileNum, period;
    char priority;

    // Here, if the line is empty, print current parameter
    // status.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusScheduleAnno();
        return(CON_STS_NULL);
    }

    // Parse the parameters.

    if(APParse(pLine, &charCount, " %d %d %d %c|",
                &chan, &fileNum, &period, &priority) != 4)
        return(CON_STS_SYNTAX);

    // If file number is out of range, return an error.
    else if(!VALID_USER_FILENUM(ORDINAL_2_FILENUM(fileNum)))
        return(CON_STS_RANGE);

    else if(period == 0 || period > MAX_ANNO_PERIOD)
        return(CON_STS_RANGE);

    // Here, add a new path or edit an existing one.
    else
    {
        anno.fileNum = (uint8_t)ORDINAL_2_FILENUM(fileNum);
        anno.period = period;
        anno.enabled = false;

        switch(chan)
        {
            case 1: anno.chan = TX_CHAN_1; break;
            case 2: anno.chan = TX_CHAN_2; break;
            default: return(CON_STS_RANGE);
        }

        // Decode the priority parameter:
        switch(priority)
        {
            case 'H': case 'h': anno.priority = 'H'; break;
            case 'M': case 'm': anno.priority = 'M'; break;
            case 'L': case 'l': anno.priority = 'L'; break;
            default: return(CON_STS_SYNTAX);
        }

        // Make sure the announcement is unique:

        // If this announcement is the same as an existing one,
        // replace the existing one.
        for(i=0; i<MAX_ANNO; i++)
        {
            pAnno = EEP_GetAnno(i);
            if(pAnno->fileNum == anno.fileNum)
            {
                // If another announcement exists with the same file
                // number, set the enable parameter for the new
                // announcement to be the same as the other one.
                anno.enabled = pAnno->enabled;

                // If this one is a duplicate, use the same slot.
                if(pAnno->chan == anno.chan)
                    break;
            }
        }

        // If we didn't find a duplicate, go through the list
        // and find an empty slot.
        if(i == MAX_ANNO)
        {
            for(i=0; i<MAX_ANNO; i++)
            {
                pAnno = EEP_GetAnno(i);
                if(pAnno->chan == TXCHAN_INVALID)
                    break;
            }
        }

        // Give an error if we didn't find an open slot.
        if(i == MAX_ANNO)
        {
            APPrint_String(STDOUT,
                            "\n\rError: Announcement memory full.");
            return(CON_STS_NULL);
        }
        else
        {
            // Add the announcement in EEP before adding it in
            // audio task because audio task looks in EEP to
            // get the announcement parameters.
            EEP_SetAnno(pAnno, &anno);
            AUD_AddAnno(i);
            SUSI_RestartWriteTimer();
            return(CON_STS_OK);
        }
    }
}

void SUSI_StatusScheduleAnno(void)
{
    eep_anno_t *pAnno;
    unsigned i;

    // Loop through all announcements and print them.
    for(i=0; i<MAX_ANNO; i++)
    {
        pAnno = EEP_GetAnno(i);

        if(pAnno->chan != TXCHAN_INVALID)
        {
            APPrint(STDOUT,"anno %u %u %u %c\n\r",
                TX_CHAN_2_ORDINAL(pAnno->chan),
                FILENUM_2_ORDINAL(pAnno->fileNum),
                pAnno->period, pAnno->priority);
        }
    }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    Delete an Announcement

    @clsyntax \> danno {txChan} {fileNum}

        @par Command Line Parameters:
        @oo {txChan}: The transmitter channel to which the
            announcement is assigned.
        @oo {fileNum}: A number from 1 to 8 indicating the file
            that contains the announcement audio.
 */
uint8_t SUSI_CmdDeleteAnno(const char *pLine)
{
    unsigned charCount, chan, fileNum;
    short i;

    // Parse the parameters.
    if(APParse(
        pLine, &charCount, " %d %d|", &chan, &fileNum) != 2)
        return(CON_STS_SYNTAX);

    else if(!VALID_TX_CHAN(ORDINAL_2_TX_CHAN(chan)))
        return(CON_STS_RANGE);

    // If file number is out of range, return an error.
    else if(!VALID_USER_FILENUM(ORDINAL_2_FILENUM(fileNum)))
        return(CON_STS_RANGE);

    // Here, find the announcement and delete it if found.
    else
    {
        if((i = EEP_FindAnno(ORDINAL_2_TX_CHAN(chan),
                        ORDINAL_2_FILENUM(fileNum))) != -1)
        {
            // Delete in audio task before deleting in
            // EEP because audio task reads the record in
            // EEP to determine which announcement to delete.
            AUD_DeleteAnno((unsigned short)i);
            EEP_DeleteAnno(EEP_GetAnno((unsigned short)i));
            SUSI_RestartWriteTimer();
            return(CON_STS_OK);
        }
        else
            return(CON_STS_RANGE);
    }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    @clsyntax \> eanno {fileNum} {enable}

    @par Command Line Parameters:

    @oo {fileNum}: A number from 1 to 8 indicating the file
        that contains the announcement audio.
    @oo {enable}: 1 to enable playback, or 0 to disable
        playback of the announcement.

    @description This command enables or disables the playback
    of an announcement.  The announcement will be enabled
    or disabled for all channels on which the file number is
    assigned.  Disabling an announcement does not delete the
    schedule or the audio file.  It only prevents playback
    until the announcement is re-enabled.
 */
uint8_t SUSI_CmdEnableAnno(const char *pLine)
{
    eep_anno_t *pAnno;
    unsigned i, charCount, fileNum;
    bool enable, found;

    // Here, if the line is empty, print current parameter
    // status.
    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusEnableAnno();
        return(CON_STS_NULL);
    }

    // Parse the parameters.
    if(APParse(
        pLine, &charCount, " %d %b|", &fileNum, &enable) != 2)
        return(CON_STS_SYNTAX);

    // If file number is out of range, return an error.
    else if(!VALID_USER_FILENUM(ORDINAL_2_FILENUM(fileNum)))
        return(CON_STS_RANGE);

    // Here, find the announcement and enable or disable it if found.
    else
    {
        found = false;

        for(i=0; i<MAX_ANNO; i++)
        {
            pAnno = EEP_GetAnno(i);
            if(pAnno->chan != TXCHAN_INVALID &&
                pAnno->fileNum == ORDINAL_2_FILENUM(fileNum))
            {
                found = true;
                EEP_EnableAnno(pAnno, enable);

                // Tell audio task to add the timer if it's being
                // enabled, or delete it if it's being disabled.
                if(enable)
                    AUD_AddAnno(i);

                else
                    AUD_DeleteAnno(i);

                SUSI_RestartWriteTimer();
            }
        }

        if(found)
            return(CON_STS_OK);
        else
        {
            APPrint_String(
                STDOUT, " Error - No such announcement.");
            return(CON_STS_NULL);
        }
    }
}

void SUSI_StatusEnableAnno(void)
{
    eep_anno_t *pAnno;
    unsigned i, enabled;
    uint16_t fileFlags = 0;

    // Loop through all announcements and print the enable status.
    for(i=0; i<MAX_ANNO; i++)
    {
        pAnno = EEP_GetAnno(i);

        // the file flags make sure we don't list a file number
        // more than once even though there may be more than one
        // announcement assigned to that file.
        if(pAnno->chan != TXCHAN_INVALID &&
            (fileFlags & BV(pAnno->fileNum)) == 0)
        {
            fileFlags |= (uint16_t)BV(pAnno->fileNum);
            enabled = pAnno->enabled? 1 : 0;
            APPrint(STDOUT,"eanno %u %u\n\r",
                FILENUM_2_ORDINAL(pAnno->fileNum), enabled);
        }
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Print Version Info

        @clsyntax \> ver

        @par Command Line Parameters: None.

        @description This command prints a line containing the
        firmware version number and/or date.
    */
uint8_t SUSI_CmdVersion(const char *pLine)
{
    // Here, if the line is empty, print current parameter
    // status.

    if(*pLine != '\0')
        return(CON_STS_SYNTAX);
    else
    {
        APPrint(STDOUT, "\n\r%s\n\r", signon);
        return(CON_STS_OK);
    }
}


/// @}
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                          END
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
