//***************************************************************
/**
    @file susi_tx_cmds.c

    Serial User Interface Transmitter Commands

    @author nlh/alp

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.
*/
//***************************************************************

//    *** INCLUDE FILES ***

#include <stdint.h>         //lint -esym(766, stdint.h)
#include <stdbool.h>

#include "lpc213x.h"        //lint -esym(766, lpc213x.h)
#include "std.h"
#include "tp_application.h" //lint -esym(766, tp_application.h)

#include "aplib.h"
#include "target.h"         //lint -esym(766, target.h)
#include "apex.h"

#include "app.h"
#include "apparse.h"
#include "apprint.h"
#include "io_main.h"
#include "astream.h"

#include "console.h"
#include "central.h"
#include "io_debug.h"       //lint -esym(766, io_debug.h)
//#include "io_misc.h"
#include "ssm_morse.h"
#include "ssm_satd.h"
#include "ssm_sate.h"
#include "eep.h"
#include "susi.h"
#include "susi_cmds.h"
//#include "fgen.h"
//#include "ssm_txsm.h"

/***************************************************************/
//    *** MACROS ****


/***************************************************************/

//    *** STATIC FUNCTION PROTOTYPES ****
static void DisassembleTxFlagWord(enum TAG_TX_CHAN chan);


/***************************************************************/

//    *** DATA STRUCTURE Definitions ***

/***************************************************************/
//****************************************************************
//                 *** VARIABLES ***
//****************************************************************

//***************************************************************
//                     Public API Functions
//                    COMMAND LINE FUNCTIONS:
/// @defgroup susi Console Operator Commands
/// @{
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Load timers.

        @clsyntax \> tmr {chan} {idTmr} {txTimeout} {hangTime}

        @par Command Line Parameters:
            @oo {chan}: Transmitter channel number (1 or 2) to which
                the timer settings apply.
            @oo {idTmr}: ID timer setting in seconds.  A setting
                of zero means no ID timer. MIN_IDTIME is min;
                MAX_IDTIME is max.
            @oo {txTimeout}: Transmitter timeout setting in seconds.
                A setting of zero means no timeout. MIN_TXTIME is min;
                MAX_TXTIME is max.
            @oo {hangTime}: Duration of hang-timer in milliseconds.
                MIN_HANGTIME is min; MAX_HANGTIME is max.

        @description This command sets the timers on the addressed
            channel to the values specified by the parameters.

        @note The limits on the above parameters are given in
            uilimits.h.
    */

uint8_t SUSI_CmdSetTimers(const char *pLine)
{
    unsigned charCount;
    uint32_t chan, idTmr, txTimeout, hangTime;

    // If the line is empty, print current contents of timers.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetTimers();
        return(CON_STS_NULL);
    }

    else if(APParse(
        pLine, &charCount, " %d %d %d %d|",
        &chan, &idTmr, &txTimeout, &hangTime) != 4)
        return(CON_STS_SYNTAX);

    else
    {
        if(!VALID_TX_CHAN(ORDINAL_2_TX_CHAN(chan)))
            return(CON_STS_RANGE);

        if(idTmr != 0)
        {
            if(idTmr < MIN_IDTIME || idTmr > MAX_IDTIME)
                return(CON_STS_RANGE);
        }

        if(txTimeout != 0)
        {
            if(txTimeout < MIN_TXTIME || txTimeout > MAX_TXTIME)
                return(CON_STS_RANGE);
        }

        if(hangTime != 0)
        {
            if((hangTime < MIN_HANGTIME) || (hangTime > MAX_HANGTIME))
                return(CON_STS_RANGE);
        }

        //subtract 1 to convert this to enum
        EEP_SetTimers(ORDINAL_2_TX_CHAN(chan),
            (uint16_t)idTmr, (uint16_t)txTimeout, (uint16_t)hangTime);
        SUSI_RestartWriteTimer();
//        IDSM_NotifyChangedTmrs(ORDINAL_2_TX_CHAN(chan));

        return(CON_STS_OK);
    }
}

void SUSI_StatusSetTimers(void)
{
    uint32_t idTmr, txTimeout, hangTime;

    EEP_InqTimers(TX_CHAN_1, &idTmr, &txTimeout, &hangTime);
    APPrint(
        STDOUT, "tmr 1 %d %d %d\n\r", idTmr, txTimeout, hangTime);

    EEP_InqTimers(TX_CHAN_2, &idTmr, &txTimeout, &hangTime);
    APPrint(
        STDOUT, "tmr 2 %d %d %d\n\r", idTmr, txTimeout, hangTime);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set ID string

        @clsyntax \> id {chan} {"string"}

        @par Command Line Parameters:
            @oo {chan}: Transmitter channel number (1 or 2) to which the ID
                string applies.
            @oo {string}: A quoted string containing the ID characters.

        @description This command sets the ID string for either transmitter
            channel 1 or 2.

        @note Defaults: string defaults to "NO ID" on both channels.
        @note Max length of the string is 20 characters (including
            terminating null).
    */

uint8_t SUSI_CmdSetIDString(const char *pLine)
{
    unsigned charCount;
    uint32_t chan;
    char string[MAX_ID_LENGTH];

    // Here, if the line is empty, print current contents of ID strings.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetIDString();
        return(CON_STS_NULL);
    }

    if(APParse(pLine, &charCount,
        " %d " GEN_FORMAT(s, MAX_ID_LENGTH), &chan, string) != 2)
        return(CON_STS_SYNTAX);

    else if(!VALID_TX_CHAN(ORDINAL_2_TX_CHAN(chan)))
        return(CON_STS_RANGE);

    else if(strlen(string) < MIN_ID_LENGTH)
    {
        APPrint(STDOUT,
            " Error: Minimum ID length is %d.\n\r", MIN_ID_LENGTH);
        return(CON_STS_OK);
    }

    else
    {
        EEP_SetIDString(ORDINAL_2_TX_CHAN(chan), string);
        SUSI_RestartWriteTimer();
        return(CON_STS_OK);
    }
}


void SUSI_StatusSetIDString(void)
{
    char *pString;

    EEP_InqIDString(TX_CHAN_1, &pString);
    APPrint(STDOUT, "id 1 \"%s\"\n\r", pString);

    EEP_InqIDString(TX_CHAN_2, &pString);
    APPrint(STDOUT, "id 2 \"%s\"\n\r", pString);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set Morse rate and frequency

        @clsyntax > cwr {tx_chan} {rate} {freq}

        @par Command Line Parameters:
            @oo {tx_chan}: Channel number (1 or 2) to which the CW
                (Morse code) ID parameter settings apply.
            @oo {rate}: CW rate in words per minute.
                MIN_MORSE_CODE_RATE is min, MAX_MORSE_CODE_RATE is max.
            @oo {freq}: CW tone frequency in Hz. FG_MIN_FREQ
                is min, FG_MAX_FREQ is max

        @description This command sets CW (Morse code) rate (in wpm),
            and frequency in Hz for the selected transmitter channel.
    */

uint8_t SUSI_CmdSetMorseRate(const char *pLine)
{
    uint32_t chan, rate, freq;
    unsigned charCount;

    // Here, if the line is empty, print current contents
    // of the CW (Morse code) ID parameters.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetMorseRate();
        return(CON_STS_NULL);
    }

    else if(APParse(
        pLine, &charCount, " %d %d %d|", &chan, &rate, &freq) != 3)
        return(CON_STS_SYNTAX);

    else if(!VALID_TX_CHAN(ORDINAL_2_TX_CHAN(chan)))
        return(CON_STS_RANGE);

    else if(rate < MIN_MORSE_CODE_RATE || rate > MAX_MORSE_CODE_RATE)
        return(CON_STS_RANGE);

    else if(freq < FG_MIN_FREQ || freq > FG_MAX_FREQ)
        return(CON_STS_RANGE);

    else
    {
        EEP_SetMorseFreq(ORDINAL_2_TX_CHAN(chan), (uint16_t)freq);
        EEP_SetMorseRate(ORDINAL_2_TX_CHAN(chan), (uint8_t)rate);
        SUSI_RestartWriteTimer();
        return(CON_STS_OK);
    }
}


void SUSI_StatusSetMorseRate(void)
{
    APPrint(STDOUT, "cwr 1 %d %d\n\r",
        EEP_GetMorseRate(TX_CHAN_1), EEP_GetMorseFreq(TX_CHAN_1));

    APPrint(STDOUT, "cwr 2 %d %d\n\r",
        EEP_GetMorseRate(TX_CHAN_2), EEP_GetMorseFreq(TX_CHAN_2));
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set Morse audio level

        @clsyntax \> cwl {chan} {level}

        @par Command Line Parameters:
            @oo {chan}: Transmitter channel number (1 or 2) to which
                the setting applies.
            @oo {level}: A number in percent indicating the desired
                audio level.

        @description This command sets the audio level
        with which the Morse CW tones shall be output on the
        selected transmitter channel.
    */

uint8_t SUSI_CmdSetMorseLevel(const char *pLine)
{
    uint32_t chan, level;
    unsigned charCount;

    // Here, if the line is empty, print current parameter
    // status.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetMorseLevel();
        return(CON_STS_NULL);
    }

    else if(APParse(
            pLine, &charCount, " %d %d|", &chan, &level) != 2)
        return(CON_STS_SYNTAX);

    else if(!VALID_TX_CHAN(ORDINAL_2_TX_CHAN(chan)))
        return(CON_STS_RANGE);

    else if(level > 100)
        return(CON_STS_RANGE);

    else
    {
        EEP_SetMorseLevel(ORDINAL_2_TX_CHAN(chan), (uint8_t)level);
        SUSI_RestartWriteTimer();
        return(CON_STS_OK);
    }
}


void SUSI_StatusSetMorseLevel(void)
{
    unsigned level;

    level = EEP_GetMorseLevel(TX_CHAN_1);
    APPrint(STDOUT, "cwl 1 %d\n\r", level);

    level = EEP_GetMorseLevel(TX_CHAN_2);
    APPrint(STDOUT, "cwl 2 %d\n\r", level);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   /** Set transmitter flags

   @clsyntax \> txf {chan} [Bx] [Dx] [Lx] [Px] [Rx] [Tx] [Vx]

    @par Command Line Parameters:
        @oo {chan}: Channel number (1 or 2) to which the transmitter
            flag settings apply.
        @oo {Bx}: The character B/b followed immediately by 0 or 1 to
            disable or enable (respectively), courtesy beeps on the
            selected transmitter.
        @oo {Dx}: The character D/d followed immediately by 0 or 1 to
            disable or enable (respectively) transmitted DTMF audio
            on the selected channel.
        @oo {Lx}: The character L/l followed immediately by 0 or 1 to
            set PTT level to active low or active high (respectively)
            on the selected channel.
        @oo {Px}: The character P/p followed immediately by 0 or 1 to
            disable or enable (respectively) CTCSS encoders
            (master enable/disable).
            on the selected channel.
        @oo {Rx}: The character R/r followed immediately by 0 or 1 to
            disable or enable (respectively) a CTCSS reverse burst on
            the selected channel.
        @oo {Tx}: The character T/t followed immediately by 0 or 1 to
            disable or enable (respectively) the selected transmitter
            itself.
        @oo {Vx}: The character V/v followed immediately by 0 or 1 to
            disable or enable (respectively) voice ID on the selected
            channel.  If V=0, CW ID will be performed, otherwise
            voice ID will be performed using the voice file selected
            by the 'vid' command.

    @description This command sets the various flags governing the
        operation of the selected transmitter.

    @note At least one flag must be specified. The defaults are
        B1, D0, H1, L1, P0, R0, T1 V0.

    */
uint8_t SUSI_CmdSetTxFlags(const char *pLine)
{
    uint32_t chan;
    unsigned charCount;
    char flag;
    bool value;
    uint16_t flags;

    // If the line is empty, print current flag settings.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetTxFlags();
        return(CON_STS_NULL);
    }

    else if(APParse(pLine, &charCount, " %d", &chan) != 1)
        return(CON_STS_SYNTAX);

    else
    {
        if(!VALID_TX_CHAN(ORDINAL_2_TX_CHAN(chan)))
            return(CON_STS_RANGE);

        else
        {
            EEP_InqTxFlags(ORDINAL_2_TX_CHAN(chan), &flags);
            pLine += charCount;
            while(*pLine != '\0')
            {
                if(APParse(
                    pLine, &charCount, " %c%b", &flag, &value) != 2)
                    return(CON_STS_SYNTAX);

                if(value != 0 && value != 1)
                    return(CON_STS_RANGE);

                switch(flag)    // Examine and evaluate each flag
                {
                    case 'b': case 'B': // Courtesy beep enable.
                        flags = SUSI_AssembleFlagWord(
                            flags, value, EEP_TX_CB_ENABLE);
                        break;

                    case 'd': case 'D': // DTMF mute if true.
                        flags = SUSI_AssembleFlagWord(
                            flags, value, EEP_TX_FLAG_DMUTE);
                        break;

                    case 'l': case 'L': // PTT active level setting.
                        flags = SUSI_AssembleFlagWord(
                            flags, value, EEP_TX_FLAG_PTTLEVEL);
                        break;

                    case 'p': case 'P': // CTCSS encoder ena/dis.
                        flags = SUSI_AssembleFlagWord(
                            flags, value, EEP_TX_FLAG_SATE);
                        break;

                    case 'r': case 'R': // Reverse burst ena/dis.
                        flags = SUSI_AssembleFlagWord(
                            flags, value, EEP_TX_FLAG_REVBURST);
                        break;

                    case 't': case 'T': // Transmitter enable/disable.

                        // If going from off to on, tell Central.
                        if(value)
                            CEN_TxEnable(ORDINAL_2_TX_CHAN(chan));

                        // If going from on to off, tell Central.
                        if(!value)
                            CEN_TxDisable(ORDINAL_2_TX_CHAN(chan));

                        flags = SUSI_AssembleFlagWord(
                            flags, value, EEP_TX_FLAG_TXENA);
                        break;

                    case 'v': case 'V': // Voice ID enable/disable.
                        flags = SUSI_AssembleFlagWord(
                            flags, value, EEP_TX_FLAG_VID);
                        break;

                    default:
                        return(CON_STS_SYNTAX);
                }
                pLine += charCount;

            }   // while

            EEP_SetTxFlags(ORDINAL_2_TX_CHAN(chan), flags);
            SUSI_RestartWriteTimer();
        }
    }

    return(CON_STS_OK);
}

void SUSI_StatusSetTxFlags(void)
{
    APPrint(STDOUT, "txf 1 ");
    DisassembleTxFlagWord(TX_CHAN_1);
    APPrint_CRLF(STDOUT);

    APPrint(STDOUT, "txf 2 ");
    DisassembleTxFlagWord(TX_CHAN_2);
    APPrint_CRLF(STDOUT);
}

static void DisassembleTxFlagWord(enum TAG_TX_CHAN chan)
{
    uint16_t flags;

    EEP_InqTxFlags(chan, &flags);

    APPrint(STDOUT,
            " B%b", (uint32_t)((flags & EEP_TX_CB_ENABLE) != 0));
    APPrint(STDOUT,
            " D%b", (uint32_t)((flags & EEP_TX_FLAG_DMUTE) != 0));
    APPrint(STDOUT,
            " L%b", (uint32_t)((flags & EEP_TX_FLAG_PTTLEVEL) != 0));
    APPrint(STDOUT,
            " P%b", (uint32_t)((flags & EEP_TX_FLAG_SATE) != 0));
    APPrint(STDOUT,
            " R%b", (uint32_t)((flags & EEP_TX_FLAG_REVBURST) != 0));
    APPrint(STDOUT,
            " T%b", (uint32_t)((flags & EEP_TX_FLAG_TXENA) != 0));
    APPrint(STDOUT,
            " V%b", (uint32_t)((flags & EEP_TX_FLAG_VID) != 0));
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *//**
    Enter Voice ID Parameters

    @clsyntax > vid {chan} {file}

    @par Command Line Parameters:
        @oo {chan}: Transmitter number (1 or 2) indicating which
            transmitter to which the ID pertains.
        @oo {file}: File number (1 through 8) indicating
            which file to play for the ID.

    @description This command sets the file number for a
    transmitter channel's voice ID.
    */

uint8_t SUSI_CmdVoiceID(const char *pLine)
{
    uint32_t chan, fileNum;
    unsigned charCount;

    // If the line is empty, print current settings.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusVoiceID();
        return(CON_STS_NULL);
    }

    else if(APParse(
            pLine, &charCount, " %d %d|", &chan, &fileNum) != 2)
        return(CON_STS_SYNTAX);

    else if(!VALID_TX_CHAN(ORDINAL_2_TX_CHAN(chan)))
        return(CON_STS_RANGE);

    else if(!VALID_USER_FILENUM(ORDINAL_2_FILENUM(fileNum)))
        return(CON_STS_RANGE);

    else
    {
        EEP_SetVidFileNum(ORDINAL_2_TX_CHAN(chan),
                            (uint8_t)ORDINAL_2_FILENUM(fileNum));
        SUSI_RestartWriteTimer();
        return(CON_STS_OK);
    }
}

void SUSI_StatusVoiceID(void)
{
    uint8_t fileNum;

    fileNum = EEP_GetVidFileNum(TX_CHAN_1);
    APPrint(STDOUT, "vid 1 %d\n\r", FILENUM_2_ORDINAL(fileNum));

    fileNum = EEP_GetVidFileNum(TX_CHAN_2);
    APPrint(STDOUT, "vid 2 %d\n\r", FILENUM_2_ORDINAL(fileNum));
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Play a transmitter's ID

        @clsyntax > playid {chan}

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) indicating on
                which channel to play the ID,

        @description This command plays the ID for the selected
        transmitter.
    */
uint8_t SUSI_CmdPlayID(const char *pLine)
{
    uint32_t chan;
    unsigned charCount;

    if(APParse(
        pLine, &charCount, " %d|", &chan) != 1)
        return(CON_STS_SYNTAX);

    else if(!VALID_TX_CHAN(ORDINAL_2_TX_CHAN(chan)))
        return(CON_STS_RANGE);

    else
    {
        CEN_PlayID(ORDINAL_2_TX_CHAN(chan));
        return(CON_STS_OK);
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set transmitter channel gain/attenuation level

        @clsyntax > txl {chan} {level}

        @par Command Line Parameters:
            @oo {chan}: Channel number (1 or 2) to which the
                transmitter level setting applies.
            @oo {txLevel}: Volume level index 0-100.

        @description This command sets the final gain or attenuation
        for the selected transmitter channel.
    */

uint8_t SUSI_CmdSetTxLevel(const char *pLine)
{
    uint32_t chan;
    uint32_t txLevel;
    unsigned charCount;

    // Here, if the line is empty, print current level set for the
    // respective transmitter channels.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusSetTxLevel();
        return(CON_STS_NULL);
    }

    else if(APParse(
            pLine, &charCount, " %d %d|", &chan, &txLevel) != 2)
        return(CON_STS_SYNTAX);

    else if(!VALID_TX_CHAN(ORDINAL_2_TX_CHAN(chan)))
        return(CON_STS_RANGE);

    else if(txLevel > MAX_TX_LEVEL)
        return(CON_STS_RANGE);

    else
    {
        // Convert channel number to enum and set level.
        EEP_SetTxLevel(ORDINAL_2_TX_CHAN(chan), (uint8_t)txLevel);
        SUSI_RestartWriteTimer();

        if(chan == 1)
        {
            ASTREAM_SetTx1Level((uint8_t)txLevel);
        }
        else
        {
            ASTREAM_SetTx2Level((uint8_t)txLevel);
        }

        return(CON_STS_OK);
    }
}


void SUSI_StatusSetTxLevel(void)
{
    uint32_t txLevel;

    EEP_InqTxLevel(TX_CHAN_1, &txLevel);
    APPrint(STDOUT, "txl 1 %d\n\r", txLevel);

    EEP_InqTxLevel(TX_CHAN_2, &txLevel);
    APPrint(STDOUT, "txl 2 %d\n\r", txLevel);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set a Transmitter's CTCSS Frequency Index

        @clsyntax > sei {chan} {index}

        @par Command Line Parameters:
            @oo {chan}: 1 or 2, identifying the transmitter
                channel.
            @oo {index}: 0 to 36, to select tone encoder frequency.

        @description This command sets the CTCSS frequency index
            for the addressed transmitter's encoder.
*/

uint8_t SUSI_CmdSetCTCSSEncodeIndex(const char *pLine)
{
    uint32_t parmCount, txChan, index;
    unsigned charCount;

    // Here, if the line is empty, print current parameter
    // status.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusCTCSSEncodeIndex();
        return(CON_STS_NULL);
    }

    // Parse the line.

    parmCount = APParse(
        pLine, &charCount, " %d %d|", &txChan, &index);

    if(parmCount != 2)
        return(CON_STS_SYNTAX);

    else if(txChan != 1 && txChan != 2)
        return(CON_STS_RANGE);

    else if(index >= SATD_FILTER_TABLE_SIZE)
        return(CON_STS_RANGE);

    else
    {
        //subtract 1 to convert chan to enum
        EEP_SetCTCSSEncodeIndex(
                        ORDINAL_2_TX_CHAN(txChan), (uint8_t)index);
        SUSI_RestartWriteTimer();
        SATE_UpdateFrequency(ORDINAL_2_TX_CHAN(txChan));

        return(CON_STS_OK);
    }
}

    // Print status for sei command:
    // {rx_chan} {index}

void SUSI_StatusCTCSSEncodeIndex(void)
{
    uint8_t index;

    index = EEP_GetCTCSSEncodeIndex(TX_CHAN_1);
    APPrint(STDOUT, "sei 1 %d\n\r", index);
    index = EEP_GetCTCSSEncodeIndex(TX_CHAN_2);
    APPrint(STDOUT, "sei 2 %d\n\r", index);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    /** Set a Transmitter's CTCSS Mode

        @clsyntax > sem {chan} {mode}

        @par Command Line Parameters:
            @oo {chan}: 1 or 2, identifying the
                transmitter channel to which the command applies.
            @oo {mode}: 1 or 2, depending on the desired mode:
                '1' if the transmitter should encode only while
                the receiver is unsquelched, or '2' if the
                transmitter should encode whenever PTT is asserted
        @description This command sets the CTCSS operation mode.
            The mode determines the circumstances under which the
            transmitter encodes tone.
*/

uint8_t SUSI_CmdSetCTCSSEncodeMode(const char *pLine)
{
    uint32_t parmCount, chan, mode;
    unsigned charCount;

    // Here, if the line is empty, print current parameter
    // status.

    if(*pLine == '\0')
    {
        APPrint_CRLF(STDOUT);
        SUSI_StatusCTCSSEncodeMode();
        return(CON_STS_NULL);
    }

    // Parse the line.

    parmCount = APParse(
        pLine, &charCount, " %d %d|", &chan, &mode);

    if(parmCount != 2)
        return(CON_STS_SYNTAX);

    else if(!VALID_TX_CHAN(ORDINAL_2_TX_CHAN(chan)))
        return(CON_STS_RANGE);

    else if(!VALID_SATE_MODE(ORDINAL_2_SATE_MODE(mode)))
        return(CON_STS_RANGE);

    else
    {
        EEP_SetCTCSSEncodeMode(
            ORDINAL_2_TX_CHAN(chan), ORDINAL_2_SATE_MODE(mode));
        SUSI_RestartWriteTimer();
        return(CON_STS_OK);
    }
}


    // Print status for sem command:
    // {rx_chan} {index}

void SUSI_StatusCTCSSEncodeMode(void)
{
    uint8_t mode;

    mode = (uint8_t)EEP_GetCTCSSEncodeMode(TX_CHAN_1);
    APPrint(STDOUT, "sem 1 %d\n\r", SATE_MODE_2_ORDINAL(mode));
    mode = (uint8_t)EEP_GetCTCSSEncodeMode(TX_CHAN_2);
    APPrint(STDOUT, "sem 2 %d\n\r", SATE_MODE_2_ORDINAL(mode));
}

/// @}
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                          END
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
