/******************************************************************//**

    @file voice.c
    Voice store and playback for PSE-702.

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.

**********************************************************************/

//      *** COMPILE SWITCHES ***

//***************************************************************

//      *** INCLUDE FILES ***

#include <stdint.h>
#include <stdbool.h>

#include "std.h"
//#include "tp_system.h"      //lint -esym(766, tp_system.h)
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"

#include "apex.h"
#include "app.h"
#include "esalaw.h"

#include "io_flash.h"
//#include "io_codec.h"
#include "filesys.h"
#include "voice.h"

    // Headers used during debug:

#include "io_debug.h"       //lint -esym(766, io_debug.h)
#include "lpc213x.h"        //lint -esym(766, lpc213x.h)
#include "io_map.h"         //lint -esym(766, io_map.h)
#include "apprint.h"        //lint -esym(766, apprint.h)
#include "io_main.h"        //lint -esym(766, io_main.h)

//***************************************************************
//                  *** MACROS ****

//lint -esym(750, PRINT_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, SET_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, STATE_CHAN) (Macro not always referenced)

#define PRINT_STATE_CHAR(x)  if(STATE_CHAN == pChan->chanID) \
        IO_DBPrintChar(x)

#if ((TP_VOI_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
  || (TP_VOI_STATE_FLAGS == STATE_FLAGS_LOWER_CH1))
#   define STATE_CHAN  VOI_CHAN_1
#else
#   define STATE_CHAN  VOI_CHAN_2
#endif

#if (TP_VOI_STATE_FLAGS == STATE_FLAGS_OFF)
#   define SET_STATE(X,Y)   pChan->pState = (X)

#elif ((TP_VOI_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
    || (TP_VOI_STATE_FLAGS == STATE_FLAGS_UPPER_CH2))
#   define SET_STATE(X,Y)   pChan->pState = (X); PRINT_STATE_CHAR(Y)

#elif ((TP_VOI_STATE_FLAGS == STATE_FLAGS_LOWER_CH1) \
    || (TP_VOI_STATE_FLAGS == STATE_FLAGS_LOWER_CH2))
#   define SET_STATE(X,Y)   pChan->pState = (X); \
                                            PRINT_STATE_CHAR((Y)|0x20)
#else
#error "TP_VOI_STATE_FLAGS improperly defined."
#endif


//***************************************************************
//              *** MODULE ID DEFINITIONS ****
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Stack Size: Measured use of 156 bytes on 3/13/12, but
    /// module was not exercised.

#define Z_STK_VOI       512     // Stack size in bytes

    /// Event IDs

enum TAG_VOI_EVIDS
{
    EV_VOI_RECORD = VOI_ID(0),  ///< Start recording
    EV_VOI_PLAYBACK,            ///< Play audio file
    EV_VOI_STOP,                ///< Stop capture/playback
    EV_VOI_PLAY_END,            ///< End of playback
    EV_VOI_REC_BUF,             ///< Record buffer full
    EV_VOI_PLAY_BUF,            ///< Playback buffer empty
    EV_VOI_MAX_EVENT            ///< Marks end of event list
};


//***************************************************************
//            *** DATA STRUCTURE DEFINITIONS ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


//***************************************************************
//      *** Signals ***

DEFINE_SIGNAL(VOI_Signal_RecordComplete)
DEFINE_SIGNAL(VOI_Signal_PlaybackComplete)

//***************************************************************
//                      *** VARIABLES ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static TCB voiceTCB;
static uint32_t voiceStack[Z_STK_VOI/sizeof(uint32_t)];
static MBOX mailbox;

    /// Typedef for each channel structure.

typedef struct TAG_VOI_CHAN_VARS
{
    /// Pointer to current state function.
    void (*pState)(enum TAG_VOI_EVIDS event,
        volatile struct TAG_VOI_CHAN_VARS *pChan, uint32_t data);

    int8_t pageBuf1[Z_FLASH_PAGE];  ///< Buffer 1 for capture/playback
    int8_t pageBuf2[Z_FLASH_PAGE];  ///< Buffer 2 for capture/playback

    volatile int8_t *pTaskBuf;   ///< Points to buffer used by task.
    volatile int8_t *pISRBuf;    ///< Points to buffer used by ISR

    unsigned bufCount;
    unsigned pageCount;   ///< Number of pages in file being played.
    unsigned pageNum;     ///< Current page being read during palyback.
    unsigned fileNum;     ///< File number being recorded or played.
    bool recordEnable;    ///< Enable capture/playback if true.
    bool playbackenable;  ///< Enable capture/playback if true.
    enum TAG_VOI_CHAN chanID; ///< Channel number of this channel

} channel_struct_t;

    /// Variables for each channel.

static volatile channel_struct_t channels[MAX_VOI_CHAN];

//***************************************************************
//                      *** CONSTANTS ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Debugging Table for Variable Display

const struct TAG_VARIABLE_TABLE voiceVarTab[] =
{
    {"stack", voiceStack, VAR_TYPE_NONE, sizeof(voiceStack)},
    {"tcb", &voiceTCB, VAR_TYPE_NONE, sizeof(voiceTCB)},
    {NULL, NULL, VAR_TYPE_NONE, 0}
};

//***************************************************************
//            *** STATIC FUNCTION PROTOTYPES ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void VoiceTask(void);

static void Transition_Idle(
                            volatile channel_struct_t *pChan); // A
static void State_Idle(enum TAG_VOI_EVIDS event,
            volatile channel_struct_t *pChan, uint32_t unused);

static void Transition_Record(
                            volatile channel_struct_t *pChan); // B
static void State_Record(enum TAG_VOI_EVIDS event,
            volatile channel_struct_t *pChan, uint32_t unused);

static void Transition_Playback(
                        volatile channel_struct_t *pChan);   // C
static void State_Playback(enum TAG_VOI_EVIDS event,
            volatile channel_struct_t *pChan, uint32_t unused);

//***************************************************************
//                          *** API ***
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void VOI_Reset(void)
{
    Apex_InitMailbox(&mailbox);
//    APEX_INIT_MUTEX_UNLOCKED(&voiMutex);

    channels[VOI_CHAN_1].chanID = VOI_CHAN_1;
    channels[VOI_CHAN_2].chanID = VOI_CHAN_2;

    channels[VOI_CHAN_1].recordEnable = false;
    channels[VOI_CHAN_2].recordEnable = false;
    channels[VOI_CHAN_1].playbackenable = false;
    channels[VOI_CHAN_2].playbackenable = false;

    // Initialize signals (Unsubscribe)
    VOI_UnsubscribeAllSignals();
}

void VOI_Start(void)
{
    Apex_CreateTask(&voiceTCB, TID_VOI, VoiceTask,
        voiceStack, (sizeof(voiceStack)), TPRI_VOI);
}


void VOI_UnsubscribeAllSignals(void)
{
    Unsubscribe_VOI_Signal_RecordComplete();
    Unsubscribe_VOI_Signal_PlaybackComplete();
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Inerface functions called by Central to start voice
    // capture and playback.

void VOI_RecordFile(enum TAG_VOI_CHAN chan, unsigned file)
{
    Apex_SendMsgExt(&mailbox, EV_VOI_RECORD,
                            (uintptr_t)chan, (uintptr_t)file);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void VOI_PlaybackFile(enum TAG_VOI_CHAN chan, unsigned file)
{
    Apex_SendMsgExt(&mailbox, EV_VOI_PLAYBACK,
                            (uintptr_t)chan, (uintptr_t)file);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

void VOI_StopRecording(enum TAG_VOI_CHAN chan)
{
    Apex_SendMsgExt(&mailbox, EV_VOI_STOP, chan, 0);
}

void VOI_StopPlayback(enum TAG_VOI_CHAN chan)
{
    Apex_SendMsgExt(&mailbox, EV_VOI_STOP, chan, 0);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Service routines to handle flash record interrupts
    /// from codec ADC:

    /// Record compresses sample using ALaw compression before
    /// saving

void VOI_PutSample_ISR(enum TAG_VOI_CHAN chan, int16_t sample)
{
    int8_t aLawSamp;
    volatile channel_struct_t *pChan = &channels[chan];

    if(pChan->recordEnable)
    {
        /// Compress the sample from 16 bits to 8 bits.
        aLawSamp = ALAW_Compress(sample);

        /// Put the sample in the buffer.
        *(pChan->pISRBuf + pChan->bufCount++) = aLawSamp;

        /// If the buffer is full, tell the task
        /// to write it to the file.
        if(pChan->bufCount == Z_FLASH_PAGE)
        {
            volatile int8_t *pTemp = pChan->pISRBuf;
            pChan->pISRBuf = pChan->pTaskBuf;
            pChan->pTaskBuf = pTemp;
            pChan->bufCount = 0;
            Apex_SendMsgExt(&mailbox, EV_VOI_REC_BUF, chan, 0);
        }
    }
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Service routine to handle flash playback sample interrupts
    // from ASTREAM:

    // Playback expands the 8 bit samples to 16 bits

int16_t VOI_GetSample_ISR(enum TAG_VOI_CHAN chan)
{
    int16_t sample = 0;
    volatile channel_struct_t *pChan = &channels[chan];

    if(pChan->playbackenable)
    {
        TP_VOICE_GETSAMPLE_STROBE

        /// Expand the sample from 8 bits to 16 bits.
        sample = ALAW_Expand(
            *(pChan->pISRBuf + pChan->bufCount++));

        /// If this was the last byte in the buffer, tell the
        /// task to either refill the buffer or quit.
        if(pChan->bufCount == Z_FLASH_PAGE)
        {
            /// If this is the last page in the file, tell the
            /// task that we are done.  Otherwise tell it to
            /// refill the buffer.
            if(--pChan->pageCount > 0)
            {
                volatile int8_t *pTemp = pChan->pISRBuf;
                pChan->pISRBuf = pChan->pTaskBuf;
                pChan->pTaskBuf = pTemp;
                pChan->bufCount = 0;
                Apex_SendMsgExt(
                    &mailbox, EV_VOI_PLAY_BUF, pChan->chanID, 0);
            }
            else
            {
                pChan->playbackenable = false;
                Apex_SendMsgExt(
                    &mailbox, EV_VOI_PLAY_END, pChan->chanID, 0);
            }
        }
    }

    return(sample);
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                      STATIC FUNCTIONS
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

static void VoiceTask(void)
{
    unsigned event;
    uintptr_t chan, data;
    volatile channel_struct_t *pChan;

    Transition_Idle(&channels[VOI_CHAN_1]);
    Transition_Idle(&channels[VOI_CHAN_2]);

    FOREVER
    {
        // Wait for a message.

        Apex_PendMsg(&mailbox, &event, &chan, &data, 0);

        ASSERT(VALID_VOI_CHAN(chan));
        pChan = &channels[(enum TAG_VOI_CHAN)chan];
        (*pChan->pState)((enum TAG_VOI_EVIDS)event, pChan, data);
    }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */ /**

*/

static void Transition_Idle(volatile channel_struct_t *pChan)
{
    SET_STATE(State_Idle, 'A');
}

static void State_Idle(enum TAG_VOI_EVIDS event,
                volatile channel_struct_t *pChan, uint32_t data)
{
    switch(event)
    {
        /// Start recording
        case EV_VOI_RECORD:
            pChan->fileNum = data;
            Transition_Record(pChan);
            break;

        /// Play audio file
        case EV_VOI_PLAYBACK:
            pChan->fileNum = data;
            Transition_Playback(pChan);
            break;

        default:
            break;

    }
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */ /**

*/

static void Transition_Record(volatile channel_struct_t *pChan)
{
     // Init page buffer pointers.

    pChan->pISRBuf = pChan->pageBuf1;
    pChan->pTaskBuf = pChan->pageBuf2;
    pChan->bufCount = 0;
    DI(); pChan->recordEnable = true; EI();

    SET_STATE(State_Record, 'B');
}

static void State_Record(enum TAG_VOI_EVIDS event,
                volatile channel_struct_t *pChan, uint32_t unused)
{
    bool status;

    switch(event)
    {
        /// Stop recording
        case EV_VOI_STOP:
            DI(); pChan->recordEnable = false; EI();

            // Zero-fill what is left of the buffer.
            while(pChan->bufCount < Z_FLASH_PAGE)
            {
                *(pChan->pISRBuf + pChan->bufCount++) = 0;
            }

            status = FSys_AppendPage(
                        pChan->fileNum, (uint8_t *)pChan->pISRBuf);
            EMIT_SIGNAL(VOI_Signal_RecordComplete,
                                   pChan->chanID, (uintptr_t)status);
            Transition_Idle(pChan);
            break;

        /// Start recording
        case EV_VOI_REC_BUF:
            status = FSys_AppendPage(
                        pChan->fileNum, (uint8_t *)pChan->pTaskBuf);
            if(!status)
            {
                // Here, we ran out of room.
                EMIT_SIGNAL(VOI_Signal_RecordComplete,
                                   pChan->chanID, (uintptr_t)status);
                Transition_Idle(pChan);
            }
            break;

        default:
            break;

    }
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */ /**

*/

static void Transition_Playback(volatile channel_struct_t *pChan)
{
    if((pChan->pageCount = FSys_FileSize(pChan->fileNum)) > 0)
    {
        pChan->pageNum = 0;
        pChan->bufCount = Z_FLASH_PAGE-1;
        pChan->pISRBuf = pChan->pageBuf1;

        // Make first sample = 0.
        pChan->pISRBuf[Z_FLASH_PAGE-1] = 0;
        pChan->pTaskBuf = pChan->pageBuf2;

        TP_VOICE_READPAGE_START
        if(!FSys_ReadPage(
            pChan->fileNum, pChan->pageNum++,
            (uint8_t *)pChan->pageBuf2))
            FATAL();
        TP_VOICE_READPAGE_END

        DI(); pChan->playbackenable = true; EI();
        SET_STATE(State_Playback, 'C');
    }
    else
    {
        // Here, zero length file.
        EMIT_SIGNAL(VOI_Signal_PlaybackComplete,
                                pChan->chanID, (uintptr_t)true);
        Transition_Idle(pChan);
    }
}

static void State_Playback(enum TAG_VOI_EVIDS event,
                volatile channel_struct_t *pChan, uint32_t unused)
{
    switch(event)
    {
        /// Stop Playback
        case EV_VOI_STOP:
            DI(); pChan->playbackenable = false; EI();
            EMIT_SIGNAL(VOI_Signal_PlaybackComplete,
                                    pChan->chanID, (uintptr_t)true);
            Transition_Idle(pChan);
            break;

        /// Interrupt finished last page and has turned itself
        /// off.
        case EV_VOI_PLAY_END:
            EMIT_SIGNAL(VOI_Signal_PlaybackComplete,
                                    pChan->chanID, (uintptr_t)true);
            Transition_Idle(pChan);
            break;

        /// Time to fill buffer with next page.
        case EV_VOI_PLAY_BUF:
            TP_VOICE_READPAGE_START
            if(!FSys_ReadPage(pChan->fileNum, pChan->pageNum++,
                                        (uint8_t *)pChan->pTaskBuf))
                FATAL();
            TP_VOICE_READPAGE_END
            break;

        default:
            break;
    }
}

//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                         END
//* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *



