//***************************************************************
/**
    @file xmodem.c

    State Machine for X-Modem Support

    @author alp

    @copyright
        Copyright 2018 by Pion and Simon Electronics.  All
        rights reserved.

    This file is part of the PSE-702 project

    The PSE-702 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSE-702 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file.  If not, see <http://www.gnu.org/licenses/>.
*/
//***************************************************************

//    *** INCLUDE FILES ***

#include <stdint.h>          //lint -esym(766, stdint.h)
#include <stdbool.h>

#include "std.h"
//#include "tp_system.h"      //lint -esym(766, tp_system.h)
#include "tp_application.h" //lint -esym(766, tp_application.h)
#include "target.h"         //lint -esym(766, target.h)
#include "apex.h"

#include "app.h"
#include "io_main.h"
#include "ansi_term.h"
#include "io_debug.h"       //lint -esym(766, io_debug.h)
#include "xmodem.h"

/***************************************************************/
//    *** MACROS ****

//lint -esym(750, PRINT_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, SET_STATE_CHAR) (Macro not always referenced)
//lint -esym(750, STATE_CHAN) (Macro not always referenced)

#define PRINT_STATE_CHAR(x)   IO_DBPrintChar(x)

#if (TP_XMODM_STATE_FLAGS == STATE_FLAGS_OFF)
#   define SET_STATE(X,Y)   pState = (X)

#elif ((TP_XMODM_STATE_FLAGS == STATE_FLAGS_UPPER_CH1) \
    || (TP_XMODM_STATE_FLAGS == STATE_FLAGS_UPPER_CH2))
#   define SET_STATE(X,Y)   pState = (X); PRINT_STATE_CHAR(Y)

#elif ((TP_XMODM_STATE_FLAGS == STATE_FLAGS_LOWER_CH1) \
    || (TP_XMODM_STATE_FLAGS == STATE_FLAGS_LOWER_CH2))
#   define SET_STATE(X,Y)   pState = (X); PRINT_STATE_CHAR((Y)|0x20)
#else
#error "TP_XMODM_STATE_FLAGS improperly defined."
#endif


/***************************************************************/
//    *** GENERAL DEFINITIONS ****

    /// X-Modem Module ID

#define XMODM_ID(x)   (MODULE(MID_XMODM) + (x))


enum TAG_XMODM_EVIDS
{
    EV_XMODM_TESTMSG = 0,                ///< Test message.
    EV_XMODM_RXDATA = XMODM_ID(0),       ///< Got a character.
    EV_XMODM_CHAR_TIMEOUT,               ///< Character timeout
    EV_XMODM_SESS_TIMEOUT                ///< Session timeout
};

    // Time Definitions

#define XMODM_TIME_CHAR   (TU_1SEC)       ///< XModem character timer.
#define XMODM_TIME_SESS   (3 * TU_10SEC)  ///< XModem session timer.


/***************************************************************/
//    *** STATIC FUNCTION PROTOTYPES ****

// State Machine Functions

static void Transition_State_Idle(void);
static void State_Idle(enum TAG_XMODM_EVIDS event, char c);

static void Transition_State_SyncHeader(void);
static void State_SyncHeader(enum TAG_XMODM_EVIDS event, char c);

static void Transition_State_RxHeader(void);
static void State_RxHeader(enum TAG_XMODM_EVIDS event, char c);

static void Transition_State_RxPayload(void);
static void State_RxPayload(
        enum TAG_XMODM_EVIDS event, char c);

static void Transition_State_WaitForTimeout(void);
static void State_WaitForTimeout(
        enum TAG_XMODM_EVIDS event, char c);

static void NullPayloadFunction(const char *pBuff);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                  SIGNALS
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Tell subscriber a digit was pressed
DEFINE_SIGNAL(XMODM_Signal_Done)

/***************************************************************/
//              *** DATA STRUCTURE Definitions ***

/***************************************************************/
//                     *** CONSTANTS ***


//****************************************************************
//                     *** VARIABLES ***
//****************************************************************

    // State to call when a character arrives.

static void (*pState)(enum TAG_XMODM_EVIDS event, char c);


static TMR charTimer, sessionTimer;

static MBOX mailbox;

static char header1;
static char buffer[XMODEM_PAYLOAD_Z];   // XModem Payload
static uint32_t bufIndex;
static uint8_t checksum;
static uint8_t blockNum;       // Serial block number.

static void (*pPayloadFunction)(char *pBuff);   // Save payload.

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /// Variable table for debugger access

const struct TAG_VARIABLE_TABLE xmodmVarTab[] =
{
    {NULL, NULL, VAR_TYPE_NONE, 0}
};

/***************************************************************/

void XMODM_Reset(void)
{
    pPayloadFunction = NullPayloadFunction;

    Apex_InitMsgTimer(
        &charTimer, &mailbox, EV_XMODM_CHAR_TIMEOUT, 0, 0);
    Apex_InitMsgTimer(
        &sessionTimer, &mailbox, EV_XMODM_SESS_TIMEOUT, 0, 0);

    Transition_State_Idle();
}

void XMODM_UnsubscribeAllSignals(void)
{
    Unsubscribe_XMODM_Signal_Done();
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    /** Start an X-Modem Session
     *
     * This function is called by the x-modem client module to
     * begin an x-modem session.  The client should feed messages
     * received at the console mailbox to the x-modem state
     * machine after calling this function.
     */

void XMODM_StartSession(void (*pFunc)(char *pBuff))
{
    pPayloadFunction = pFunc;
    blockNum = 1;       // Serial block number.
    TxUserConsole(ASCII_NAK);
    Transition_State_SyncHeader();
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
/** X-Modem State Machine Entry Point
 *
 * This function is called by the client module that is using
 * x-modem services.  This state machine is actually run by the
 * console task and uses the console task's stack.  The client
 * module calls the state machine each time a message arrives
 * in the console mailbox.
 *
*/

void XMODM_StateMachine(unsigned event, uintptr_t data1)
{
    ASSERT(pState != NULL);
    (*pState)((enum TAG_XMODM_EVIDS)event, (char)data1);
}

//***************************************************************

    // Null Payload Function

static void NullPayloadFunction(const char *pBuff){} //lint !e715

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // This is the timer that is used to determine how long
    // we will wait between characters before sending a NAK.

static void StartCharTimer(void)
{

    if(charTimer.state == TMS_COUNTING)
        Apex_StopTmr(&charTimer);

    Apex_StartTmr(&charTimer, XMODM_TIME_CHAR);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // If this timer times out, we end the session.

static void StartSessionTimer(void)
{
    Apex_StartTmr(&sessionTimer, XMODM_TIME_SESS);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//              X-Modem States
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Stay in this state while idle.

static void Transition_State_Idle(void)
{
    Apex_StopTmr(&charTimer);
    Apex_StopTmr(&sessionTimer);
    pPayloadFunction = NullPayloadFunction;

    SET_STATE(State_Idle,'A');
}

static void State_Idle(enum TAG_XMODM_EVIDS event, char c)
{
    FATAL();    // There are no events possible in this state.
}   //lint !e715    // parms not referenced.

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

    // Here, X-Modem transfer has been initiated.

static void Transition_State_SyncHeader(void)
{
    StartCharTimer();
    StartSessionTimer();
    SET_STATE(State_SyncHeader,'B');
}

static void State_SyncHeader(enum TAG_XMODM_EVIDS event, char c)
{
    switch(event)
    {
        case EV_XMODM_RXDATA:
            if(c == ASCII_EOT)
            {
                TxUserConsole(ASCII_ACK);
                ASSERT(XMODM_Signal_Done != NULL);
                EMIT_SIGNAL(XMODM_Signal_Done, 0, 0);
                Transition_State_Idle();
            }

            if(c == ASCII_SOH)
            {
                StartCharTimer();   // Refresh character timer
                Transition_State_RxHeader();     // Get header.
            }

            if(c == ASCII_CAN || c == ASCII_ESC)
            {
                TxUserConsole(ASCII_ACK);
                ASSERT(XMODM_Signal_Done != NULL);
                EMIT_SIGNAL(XMODM_Signal_Done, 0, 0);
                Transition_State_Idle();
            }
            break;

        // Took to long to receive the next character.

        case EV_XMODM_CHAR_TIMEOUT:
            TxUserConsole(ASCII_NAK);
            StartCharTimer();
            break;

        // The session timed out.
        case EV_XMODM_SESS_TIMEOUT:
                ASSERT(XMODM_Signal_Done != NULL);
                EMIT_SIGNAL(XMODM_Signal_Done, 0, 0);
            Transition_State_Idle();
            break;

        default:
            FATAL();        // There are no other events possible.
    }
}

//***************************************************************

    // Receive Header:

static void Transition_State_RxHeader(void)
{
    Apex_StopTmr(&sessionTimer);
    bufIndex = 0;
    SET_STATE(State_RxHeader,'C');
}


static void State_RxHeader(enum TAG_XMODM_EVIDS event, char c)
{
    switch(event)
    {
        case EV_XMODM_RXDATA:
            StartCharTimer();       // Refresh character timer.

            if(bufIndex == 1)
            {
                // Last header byte.
                if(
                    ((c & header1) == 0)
                        && ((uint8_t)(c | header1) == 0xFF)
                    )
                {
                    // Here, good header. Get payload
                    Transition_State_RxPayload();
                }

                else
                {
                    // Here, bad header.
                    Transition_State_WaitForTimeout();
                }
            }

            else
            {
                // First header byte.
                header1 = c;
                bufIndex++;
            }

            break;

        case EV_XMODM_CHAR_TIMEOUT:
            TxUserConsole(ASCII_NAK);
            Transition_State_SyncHeader();     // Send NAK.
            break;

        default:
            FATAL();        // There are no other events possible.
    }
}

//***************************************************************

    // Receive Payload.

static void Transition_State_RxPayload(void)
{
    bufIndex = 0;
    checksum = 0;
    SET_STATE(State_RxPayload,'D');
}

static void State_RxPayload(enum TAG_XMODM_EVIDS event, char c)
{
    switch(event)
    {
        case EV_XMODM_RXDATA:
/// @bug The following FTP0 set and clear instructions may be
    // used to test for the bug in apex.c that is described in
    // the comment above Apex_StartTmr().

            StartCharTimer();       // Refresh character timer.

            if(bufIndex == sizeof(buffer))
            {
                // c contains the checksum.
                if(c == checksum)
                {
                    // Here, good checksum.  Examine block number.

                    if(header1 == blockNum)
                    {
                        blockNum++;
                        TxUserConsole(ASCII_ACK);
                        (*pPayloadFunction)(buffer); // Save payload.
                        Transition_State_SyncHeader();
                    }

                    // Was it a retransmission of a previous block?
                    else if((header1+1) == blockNum)
                    {
                        TxUserConsole(ASCII_ACK);
                        Transition_State_SyncHeader();
                    }

                    else
                    {
                        Transition_State_WaitForTimeout();
                    }
                }

                else
                    Transition_State_WaitForTimeout();
            }

            else
            {
                buffer[bufIndex++] = c;
                checksum += (uint8_t)c;
            }
            break;

        case EV_XMODM_CHAR_TIMEOUT:
            TxUserConsole(ASCII_NAK);
            Transition_State_SyncHeader();
            break;

        default:
            FATAL();        // There are no other events possible.
    }

}

//***************************************************************

    // Error.  Wait for Timeout.

static void Transition_State_WaitForTimeout(void)
{
    SET_STATE(State_WaitForTimeout,'E');
}

static void State_WaitForTimeout(
        enum TAG_XMODM_EVIDS event, char c)
{
    switch(event)
    {
        case EV_XMODM_RXDATA:
            StartCharTimer();       // Refresh character timer.
            break;

        case EV_XMODM_CHAR_TIMEOUT:
            TxUserConsole(ASCII_NAK);
            Transition_State_SyncHeader();
            break;

        default:
            FATAL();        // There are no other events possible.
    }
}   //lint !e715    // c not referenced.

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
//                          END
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
